! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
MODULE sla
!
! this file contains SLALIB from Starlink, version 1998 Jan 22, 
! with the following changes to turn it into a Fortran90 module:
!
! 1. passed through the program to_f90 of Alan Miller to convert the
!    syntax to F90 style.  The most important feature is that
!    END SUBROUTINE /END FUNCTION statements have been added.
!
! 2. MODULE heading and tailings have been added.
!
! 3. A few duplicate declarations of the data types of functions had
!    to be removed, as they made the linker think that the were
!    references to external functions of the same name.
!
! This version produced 1998 April 28 by Clive Page, Leicester University.
!
CONTAINS
SUBROUTINE sla_addet (rm, dm, EQ, rc, dc)
!+
!     - - - - - -
!      A D D E T
!     - - - - - -

!  Add the E-terms (elliptic component of annual aberration)
!  to a pre IAU 1976 mean place to conform to the old
!  catalogue convention (double precision)

!  Given:
!     RM,DM     dp     RA,Dec (radians) without E-terms
!     EQ        dp     Besselian epoch of mean equator and equinox

!  Returned:
!     RC,DC     dp     RA,Dec (radians) with E-terms included

!  Called:
!     sla_ETRMS, sla_DCS2C, sla_DCC2S, sla_DRANRM, sla_DRANGE

!  Explanation:
!     Most star positions from pre-1984 optical catalogues (or
!     derived from astrometry using such stars) embody the
!     E-terms.  If it is necessary to convert a formal mean
!     place (for example a pulsar timing position) to one
!     consistent with such a star catalogue, then the RA,Dec
!     should be adjusted using this routine.

!  Reference:
!     Explanatory Supplement to the Astronomical Ephemeris,
!     section 2D, page 48.

!  P.T.Wallace   Starlink   July 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rm,dm,EQ,rc,dc

DOUBLE PRECISION :: a(3),v(3)

INTEGER :: i



!  E-terms vector
CALL sla_etrms(EQ,a)

!  Spherical to Cartesian
CALL sla_dcs2c(rm,dm,v)

!  Include the E-terms
DO i=1,3
  v(i)=v(i)+a(i)
END DO

!  Cartesian to spherical
CALL sla_dcc2s(v,rc,dc)

!  Bring RA into conventional range
rc=sla_dranrm(rc)

END SUBROUTINE sla_addet
SUBROUTINE sla_afin (string, iptr, a, j)
!+
!     - - - - -
!      A F I N
!     - - - - -

!  Sexagesimal character string to angle (single precision)

!  Given:
!     STRING  c*(*)   string containing deg, arcmin, arcsec fields
!     IPTR      i     pointer to start of decode (1st = 1)

!  Returned:
!     IPTR      i     advanced past the decoded angle
!     A         r     angle in radians
!     J         i     status:  0 = OK
!                             +1 = default, A unchanged
!                             -1 = bad degrees      )
!                             -2 = bad arcminutes   )  (note 3)
!                             -3 = bad arcseconds   )

!  Example:

!    argument    before                           after

!    STRING      '-57 17 44.806  12 34 56.7'      unchanged
!    IPTR        1                                16 (points to 12...)
!    A           ?                                -1.00000
!    J           ?                                0

!    A further call to sla_AFIN, without adjustment of IPTR, will
!    decode the second angle, 12deg 34min 56.7sec.

!  Notes:

!     1)  The first three "fields" in STRING are degrees, arcminutes,
!         arcseconds, separated by spaces or commas.  The degrees field
!         may be signed, but not the others.  The decoding is carried
!         out by the DFLTIN routine and is free-format.

!     2)  Successive fields may be absent, defaulting to zero.  For
!         zero status, the only combinations allowed are degrees alone,
!         degrees and arcminutes, and all three fields present.  If all
!         three fields are omitted, a status of +1 is returned and A is
!         unchanged.  In all other cases A is changed.

!     3)  Range checking:

!           The degrees field is not range checked.  However, it is
!           expected to be integral unless the other two fields are
!           absent.

!           The arcminutes field is expected to be 0-59, and integral if
!           the arcseconds field is present.  If the arcseconds field
!           is absent, the arcminutes is expected to be 0-59.9999...

!           The arcseconds field is expected to be 0-59.9999...

!     4)  Decoding continues even when a check has failed.  Under these
!         circumstances the field takes the supplied value, defaulting
!         to zero, and the result A is computed and returned.

!     5)  Further fields after the three expected ones are not treated
!         as an error.  The pointer IPTR is left in the correct state
!         for further decoding with the present routine or with DFLTIN
!         etc.  See the example, above.

!     6)  If STRING contains hours, minutes, seconds instead of degrees
!         etc, or if the required units are turns (or days) instead of
!         radians, the result A should be multiplied as follows:

!           for        to obtain    multiply
!           STRING     A in         A by

!           d ' "      radians      1       =  1.0
!           d ' "      turns        1/2pi   =  0.1591549430918953358
!           h m s      radians      15      =  15.0
!           h m s      days         15/2pi  =  2.3873241463784300365

!  Called:  sla_DAFIN

!  P.T.Wallace   Starlink   13 September 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: iptr
REAL :: a
INTEGER :: j

DOUBLE PRECISION :: ad



!  Call the double precision version
CALL sla_dafin(string,iptr,ad,j)
IF (j <= 0) a=REAL(ad)

END SUBROUTINE sla_afin
DOUBLE PRECISION FUNCTION sla_airmas (zd)
!+
!     - - - - - - -
!      A I R M A S
!     - - - - - - -

!  Air mass at given zenith distance (double precision)

!  Given:
!     ZD     d     Observed zenith distance (radians)

!  The result is an estimate of the air mass, in units of that
!  at the zenith.

!  Notes:

!  1)  The "observed" zenith distance referred to above means "as
!      affected by refraction".

!  2)  Uses Hardie's (1962) polynomial fit to Bemporad's data for
!      the relative air mass, X, in units of thickness at the zenith
!      as tabulated by Schoenberg (1929). This is adequate for all
!      normal needs as it is accurate to better than 0.1% up to X =
!      6.8 and better than 1% up to X = 10. Bemporad's tabulated
!      values are unlikely to be trustworthy to such accuracy
!      because of variations in density, pressure and other
!      conditions in the atmosphere from those assumed in his work.

!  3)  The sign of the ZD is ignored.

!  4)  At zenith distances greater than about ZD = 87 degrees the
!      air mass is held constant to avoid arithmetic overflows.

!  References:
!     Hardie, R.H., 1962, in "Astronomical Techniques"
!        ed. W.A. Hiltner, University of Chicago Press, p180.
!     Schoenberg, E., 1929, Hdb. d. Ap.,
!        Berlin, Julius Springer, 2, 268.

!  Original code by P.W.Hill, St Andrews
!  Adapted by P.T.Wallace, Starlink, 5 December 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: zd

DOUBLE PRECISION :: seczm1


seczm1 = 1D0/(COS(MIN(1.52D0,ABS(zd))))-1D0
sla_airmas = 1D0 + seczm1*(0.9981833D0  &
- seczm1*(0.002875D0 + 0.0008083D0*seczm1))

END FUNCTION sla_airmas
SUBROUTINE sla_altaz (ha, dec, phi,az, azd, azdd, el, eld, eldd, pa, pad, padd)
!+
!     - - - - - -
!      A L T A Z
!     - - - - - -

!  Positions, velocities and accelerations for an altazimuth
!  telescope mount.

!  (double precision)

!  Given:
!     HA      d     hour angle
!     DEC     d     declination
!     PHI     d     observatory latitude

!  Returned:
!     AZ      d     azimuth
!     AZD     d        "    velocity
!     AZDD    d        "    acceleration
!     EL      d     elevation
!     ELD     d         "     velocity
!     ELDD    d         "     acceleration
!     PA      d     parallactic angle
!     PAD     d         "      "   velocity
!     PADD    d         "      "   acceleration

!  Notes:

!  1)  Natural units are used throughout.  HA, DEC, PHI, AZ, EL
!      and ZD are in radians.  The velocities and accelerations
!      assume constant declination and constant rate of change of
!      hour angle (as for tracking a star);  the units of AZD, ELD
!      and PAD are radians per radian of HA, while the units of AZDD,
!      ELDD and PADD are radians per radian of HA squared.  To
!      convert into practical degree- and second-based units:

!        angles * 360/2pi -> degrees
!        velocities * (2pi/86400)*(360/2pi) -> degree/sec
!        accelerations * ((2pi/86400)**2)*(360/2pi) -> degree/sec/sec

!      Note that the seconds here are sidereal rather than SI.  One
!      sidereal second is about 0.99727 SI seconds.

!      The velocity and acceleration factors assume the sidereal
!      tracking case.  Their respective numerical values are (exactly)
!      1/240 and (approximately) 1/3300236.9.

!  2)  Azimuth is returned in the range 0-2pi;  north is zero,
!      and east is +pi/2.  Elevation and parallactic angle are
!      returned in the range +/-pi/2.  Position angle is +ve
!      for a star west of the meridian and is the angle NP-star-zenith.

!  3)  The latitude is geodetic as opposed to geocentric.  The
!      hour angle and declination are topocentric.  Refraction and
!      deficiencies in the telescope mounting are ignored.  The
!      purpose of the routine is to give the general form of the
!      quantities.  The details of a real telescope could profoundly
!      change the results, especially close to the zenith.

!  4)  No range checking of arguments is carried out.

!  5)  In applications which involve many such calculations, rather
!      than calling the present routine it will be more efficient to
!      use inline code, having previously computed fixed terms such
!      as sine and cosine of latitude, and (for tracking a star)
!      sine and cosine of declination.

!  P.T.Wallace   Starlink   14 March 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ha,dec,phi,az,azd,azdd,el,eld,eldd,pa,pad,padd

DOUBLE PRECISION :: dpi,d2pi,tiny
PARAMETER (dpi=3.1415926535897932384626433832795D0,  &
d2pi=6.283185307179586476925286766559D0,tiny=1D-30)

DOUBLE PRECISION :: sh,ch,sd,cd,sp,cp,chcd,sdcp,x,y,z,rsq,r,a,e,c,s,  &
q,qd,ad,ed,edr,add,edd,qdd


!  Useful functions
sh=SIN(ha)
ch=COS(ha)
sd=SIN(dec)
cd=COS(dec)
sp=SIN(phi)
cp=COS(phi)
chcd=ch*cd
sdcp=sd*cp
x=-chcd*sp+sdcp
y=-sh*cd
z=chcd*cp+sd*sp
rsq=x*x+y*y
r=SQRT(rsq)

!  Azimuth and elevation
IF (rsq == 0D0) THEN
  a=0D0
ELSE
  a=ATAN2(y,x)
END IF
IF (a < 0D0) a=a+d2pi
e=ATAN2(z,r)

!  Parallactic angle
c=cd*sp-ch*sdcp
s=sh*cp
IF (c*c+s*s > 0) THEN
  q=ATAN2(s,c)
ELSE
  q=dpi-ha
END IF

!  Velocities and accelerations (clamped at zenith/nadir)
IF (rsq < tiny) THEN
  rsq=tiny
  r=SQRT(rsq)
END IF
qd=-x*cp/rsq
ad=sp+z*qd
ed=cp*y/r
edr=ed/r
add=edr*(z*sp+(2D0-rsq)*qd)
edd=-r*qd*ad
qdd=edr*(sp+2D0*z*qd)

!  Results
az=a
azd=ad
azdd=add
el=e
eld=ed
eldd=edd
pa=q
pad=qd
padd=qdd

END SUBROUTINE sla_altaz
SUBROUTINE sla_amp (ra, da, date, EQ, rm, dm)
!+
!     - - - -
!      A M P
!     - - - -

!  Convert star RA,Dec from geocentric apparent to mean place

!  The mean coordinate system is the post IAU 1976 system,
!  loosely called FK5.

!  Given:
!     RA       d      apparent RA (radians)
!     DA       d      apparent Dec (radians)
!     DATE     d      TDB for apparent place (JD-2400000.5)
!     EQ       d      equinox:  Julian epoch of mean place

!  Returned:
!     RM       d      mean RA (radians)
!     DM       d      mean Dec (radians)

!  References:
!     1984 Astronomical Almanac, pp B39-B41.
!     (also Lederle & Schwan, Astron. Astrophys. 134,
!      1-6, 1984)

!  Notes:

!  1)  The distinction between the required TDB and TT is
!      always negligible.  Moreover, for all but the most
!      critical applications UTC is adequate.

!  2)  The accuracy is limited by the routine sla_EVP, called
!      by sla_MAPPA, which computes the Earth position and
!      velocity using the methods of Stumpff.  The maximum
!      error is about 0.3 milliarcsecond.

!  3)  Iterative techniques are used for the aberration and
!      light deflection corrections so that the routines
!      sla_AMP (or sla_AMPQK) and sla_MAP (or sla_MAPQK) are
!      accurate inverses;  even at the edge of the Sun's disc
!      the discrepancy is only about 1 nanoarcsecond.

!  4)  Where multiple apparent places are to be converted to
!      mean places, for a fixed date and equinox, it is more
!      efficient to use the sla_MAPPA routine to compute the
!      required parameters once, followed by one call to
!      sla_AMPQK per star.

!  Called:  sla_MAPPA, sla_AMPQK

!  P.T.Wallace   Starlink   19 January 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ra,da,date,EQ,rm,dm

DOUBLE PRECISION :: amprms(21)



CALL sla_mappa(EQ,date,amprms)
CALL sla_ampqk(ra,da,amprms,rm,dm)

END SUBROUTINE sla_amp
SUBROUTINE sla_ampqk (ra, da, amprms, rm, dm)
!+
!     - - - - - -
!      A M P Q K
!     - - - - - -

!  Convert star RA,Dec from geocentric apparent to mean place

!  The mean coordinate system is the post IAU 1976 system,
!  loosely called FK5.

!  Use of this routine is appropriate when efficiency is important
!  and where many star positions are all to be transformed for
!  one epoch and equinox.  The star-independent parameters can be
!  obtained by calling the sla_MAPPA routine.

!  Given:
!     RA       d      apparent RA (radians)
!     DA       d      apparent Dec (radians)

!     AMPRMS   d(21)  star-independent mean-to-apparent parameters:

!       (1)      time interval for proper motion (Julian years)
!       (2-4)    barycentric position of the Earth (AU)
!       (5-7)    heliocentric direction of the Earth (unit vector)
!       (8)      (grav rad Sun)*2/(Sun-Earth distance)
!       (9-11)   ABV: barycentric Earth velocity in units of c
!       (12)     sqrt(1-v**2) where v=modulus(ABV)
!       (13-21)  precession/nutation (3,3) matrix

!  Returned:
!     RM       d      mean RA (radians)
!     DM       d      mean Dec (radians)

!  References:
!     1984 Astronomical Almanac, pp B39-B41.
!     (also Lederle & Schwan, Astron. Astrophys. 134,
!      1-6, 1984)

!  Notes:

!  1)  The accuracy is limited by the routine sla_EVP, called
!      by sla_MAPPA, which computes the Earth position and
!      velocity using the methods of Stumpff.  The maximum
!      error is about 0.3 milliarcsecond.

!  2)  Iterative techniques are used for the aberration and
!      light deflection corrections so that the routines
!      sla_AMP (or sla_AMPQK) and sla_MAP (or sla_MAPQK) are
!      accurate inverses;  even at the edge of the Sun's disc
!      the discrepancy is only about 1 nanoarcsecond.

!  Called:  sla_DCS2C, sla_DIMXV, sla_DVDV, sla_DVN, sla_DCC2S,
!           sla_DRANRM

!  P.T.Wallace   Starlink   21 June 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ra,da,amprms(21),rm,dm

INTEGER :: i,j

DOUBLE PRECISION :: gr2e,ab1,ehn(3),abv(3),p3(3),p2(3),  &
ab1p1,p1dv,p1dvp1,p1(3),w,pde,pdep1,p(3)



!  Unpack scalar and vector parameters
gr2e = amprms(8)
ab1 = amprms(12)
DO i=1,3
  ehn(i) = amprms(i+4)
  abv(i) = amprms(i+8)
END DO

!  Apparent RA,Dec to Cartesian
CALL sla_dcs2c(ra,da,p3)

!  Precession and nutation
CALL sla_dimxv(amprms(13),p3,p2)

!  Aberration
ab1p1 = ab1+1D0
DO i=1,3
  p1(i) = p2(i)
END DO
DO j=1,2
  p1dv = sla_dvdv(p1,abv)
  p1dvp1 = 1D0+p1dv
  w = 1D0+p1dv/ab1p1
  DO i=1,3
    p1(i) = (p1dvp1*p2(i)-w*abv(i))/ab1
  END DO
  CALL sla_dvn(p1,p3,w)
  DO i=1,3
    p1(i) = p3(i)
  END DO
END DO

!  Light deflection
DO i=1,3
  p(i) = p1(i)
END DO
DO j=1,5
  pde = sla_dvdv(p,ehn)
  pdep1 = 1D0+pde
  w = pdep1-gr2e*pde
  DO i=1,3
    p(i) = (pdep1*p1(i)-gr2e*ehn(i))/w
  END DO
  CALL sla_dvn(p,p2,w)
  DO i=1,3
    p(i) = p2(i)
  END DO
END DO

!  Mean RA,Dec
CALL sla_dcc2s(p,rm,dm)
rm = sla_dranrm(rm)

END SUBROUTINE sla_ampqk
SUBROUTINE sla_aop (rap, dap, date, dut, elongm, phim, hm,  &
xp, yp, tdk, pmb, rh, wl, tlr,aob, zob, hob, dob, rob)
!+
!     - - - -
!      A O P
!     - - - -

!  Apparent to observed place, for optical sources distant from
!  the solar system.

!  Given:
!     RAP    d      geocentric apparent right ascension
!     DAP    d      geocentric apparent declination
!     DATE   d      UTC date/time (Modified Julian Date, JD-2400000.5)
!     DUT    d      delta UT:  UT1-UTC (UTC seconds)
!     ELONGM d      mean longitude of the observer (radians, east +ve)
!     PHIM   d      mean geodetic latitude of the observer (radians)
!     HM     d      observer's height above sea level (metres)
!     XP     d      polar motion x-coordinate (radians)
!     YP     d      polar motion y-coordinate (radians)
!     TDK    d      local ambient temperature (DegK; std=273.155D0)
!     PMB    d      local atmospheric pressure (mB; std=1013.25D0)
!     RH     d      local relative humidity (in the range 0D0-1D0)
!     WL     d      effective wavelength (micron, e.g. 0.55D0)
!     TLR    d      tropospheric lapse rate (DegK/metre, e.g. 0.0065D0)

!  Returned:
!     AOB    d      observed azimuth (radians: N=0,E=90)
!     ZOB    d      observed zenith distance (radians)
!     HOB    d      observed Hour Angle (radians)
!     DOB    d      observed Declination (radians)
!     ROB    d      observed Right Ascension (radians)

!  Notes:

!   1)  This routine returns zenith distance rather than elevation
!       in order to reflect the fact that no allowance is made for
!       depression of the horizon.

!   2)  The accuracy of the result is limited by the corrections for
!       refraction.  Providing the meteorological parameters are
!       known accurately and there are no gross local effects, the
!       predicted apparent RA,Dec should be within about 0.1 arcsec
!       for a zenith distance of less than 70 degrees.  Even at a
!       topocentric zenith distance of 90 degrees, the accuracy in
!       elevation should be better than 1 arcmin;  useful results
!       are available for a further 3 degrees, beyond which the
!       sla_REFRO routine returns a fixed value of the refraction.
!       The complementary routines sla_AOP (or sla_AOPQK) and sla_OAP
!       (or sla_OAPQK) are self-consistent to better than 1 micro-
!       arcsecond all over the celestial sphere.

!   3)  It is advisable to take great care with units, as even
!       unlikely values of the input parameters are accepted and
!       processed in accordance with the models used.

!   4)  "Apparent" place means the geocentric apparent right ascension
!       and declination, which is obtained from a catalogue mean place
!       by allowing for space motion, parallax, precession, nutation,
!       annual aberration, and the Sun's gravitational lens effect.  For
!       star positions in the FK5 system (i.e. J2000), these effects can
!       be applied by means of the sla_MAP etc routines.  Starting from
!       other mean place systems, additional transformations will be
!       needed;  for example, FK4 (i.e. B1950) mean places would first
!       have to be converted to FK5, which can be done with the
!       sla_FK425 etc routines.

!   5)  "Observed" Az,El means the position that would be seen by a
!       perfect theodolite located at the observer.  This is obtained
!       from the geocentric apparent RA,Dec by allowing for Earth
!       orientation and diurnal aberration, rotating from equator
!       to horizon coordinates, and then adjusting for refraction.
!       The HA,Dec is obtained by rotating back into equatorial
!       coordinates, using the geodetic latitude corrected for polar
!       motion, and is the position that would be seen by a perfect
!       equatorial located at the observer and with its polar axis
!       aligned to the Earth's axis of rotation (n.b. not to the
!       refracted pole).  Finally, the RA is obtained by subtracting
!       the HA from the local apparent ST.

!   6)  To predict the required setting of a real telescope, the
!       observed place produced by this routine would have to be
!       adjusted for the tilt of the azimuth or polar axis of the
!       mounting (with appropriate corrections for mount flexures),
!       for non-perpendicularity between the mounting axes, for the
!       position of the rotator axis and the pointing axis relative
!       to it, for tube flexure, for gear and encoder errors, and
!       finally for encoder zero points.  Some telescopes would, of
!       course, exhibit other properties which would need to be
!       accounted for at the appropriate point in the sequence.

!   7)  This routine takes time to execute, due mainly to the
!       rigorous integration used to evaluate the refraction.
!       For processing multiple stars for one location and time,
!       call sla_AOPPA once followed by one call per star to sla_AOPQK.
!       Where a range of times within a limited period of a few hours
!       is involved, and the highest precision is not required, call
!       sla_AOPPA once, followed by a call to sla_AOPPAT each time the
!       time changes, followed by one call per star to sla_AOPQK.

!   8)  The DATE argument is UTC expressed as an MJD.  This is,
!       strictly speaking, wrong, because of leap seconds.  However,
!       as long as the delta UT and the UTC are consistent there
!       are no difficulties, except during a leap second.  In this
!       case, the start of the 61st second of the final minute should
!       begin a new MJD day and the old pre-leap delta UT should
!       continue to be used.  As the 61st second completes, the MJD
!       should revert to the start of the day as, simultaneously,
!       the delta UTC changes by one second to its post-leap new value.

!   9)  The delta UT (UT1-UTC) is tabulated in IERS circulars and
!       elsewhere.  It increases by exactly one second at the end of
!       each UTC leap second, introduced in order to keep delta UT
!       within +/- 0.9 seconds.

!  10)  IMPORTANT -- TAKE CARE WITH THE LONGITUDE SIGN CONVENTION.
!       The longitude required by the present routine is east-positive,
!       in accordance with geographical convention (and right-handed).
!       In particular, note that the longitudes returned by the
!       sla_OBS routine are west-positive, following astronomical
!       usage, and must be reversed in sign before use in the present
!       routine.

!  11)  The polar coordinates XP,YP can be obtained from IERS
!       circulars and equivalent publications.  The maximum amplitude
!       is about 0.3 arcseconds.  If XP,YP values are unavailable,
!       use XP=YP=0D0.  See page B60 of the 1988 Astronomical Almanac
!       for a definition of the two angles.

!  12)  The height above sea level of the observing station, HM,
!       can be obtained from the Astronomical Almanac (Section J
!       in the 1988 edition), or via the routine sla_OBS.  If P,
!       the pressure in millibars, is available, an adequate
!       estimate of HM can be obtained from the expression

!             HM ~ -8149.9415D0*LOG(P/1013.25D0).

!       (See Astrophysical Quantities, C.W.Allen, 3rd edition,
!       section 52.)  Similarly, if the pressure P is not known,
!       it can be estimated from the height of the observing
!       station, HM as follows:

!             P ~ 1013.25D0*EXP(-HM/8149.9415D0).

!       Note, however, that the refraction is proportional to the
!       pressure and that an accurate P value is important for
!       precise work.

!  13)  The azimuths etc produced by the present routine are with
!       respect to the celestial pole.  Corrections to the terrestrial
!       pole can be computed using sla_POLMO.

!  Called:  sla_AOPPA, sla_AOPQK

!  P.T.Wallace   Starlink   22 February 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rap,dap,date,dut,elongm,phim,hm,  &
xp,yp,tdk,pmb,rh,wl,tlr,aob,zob,hob,dob,rob

DOUBLE PRECISION :: aoprms(14)


CALL sla_aoppa(date,dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr,aoprms)
CALL sla_aopqk(rap,dap,aoprms,aob,zob,hob,dob,rob)

END SUBROUTINE sla_aop
SUBROUTINE sla_aoppa (date, dut, elongm, phim, hm,  &
xp, yp, tdk, pmb, rh, wl, tlr, aoprms)
!+
!     - - - - - -
!      A O P P A
!     - - - - - -

!  Precompute apparent to observed place parameters required by
!  sla_AOPQK and sla_OAPQK.

!  Given:
!     DATE   d      UTC date/time (modified Julian Date, JD-2400000.5)
!     DUT    d      delta UT:  UT1-UTC (UTC seconds)
!     ELONGM d      mean longitude of the observer (radians, east +ve)
!     PHIM   d      mean geodetic latitude of the observer (radians)
!     HM     d      observer's height above sea level (metres)
!     XP     d      polar motion x-coordinate (radians)
!     YP     d      polar motion y-coordinate (radians)
!     TDK    d      local ambient temperature (DegK; std=273.155D0)
!     PMB    d      local atmospheric pressure (mB; std=1013.25D0)
!     RH     d      local relative humidity (in the range 0D0-1D0)
!     WL     d      effective wavelength (micron, e.g. 0.55D0)
!     TLR    d      tropospheric lapse rate (DegK/metre, e.g. 0.0065D0)

!  Returned:
!     AOPRMS d(14)  star-independent apparent-to-observed parameters:

!       (1)      geodetic latitude (radians)
!       (2,3)    sine and cosine of geodetic latitude
!       (4)      magnitude of diurnal aberration vector
!       (5)      height (HM)
!       (6)      ambient temperature (TDK)
!       (7)      pressure (PMB)
!       (8)      relative humidity (RH)
!       (9)      wavelength (WL)
!       (10)     lapse rate (TLR)
!       (11,12)  refraction constants A and B (radians)
!       (13)     longitude + eqn of equinoxes + sidereal DUT (radians)
!       (14)     local apparent sidereal time (radians)

!  Notes:

!   1)  It is advisable to take great care with units, as even
!       unlikely values of the input parameters are accepted and
!       processed in accordance with the models used.

!   2)  The DATE argument is UTC expressed as an MJD.  This is,
!       strictly speaking, improper, because of leap seconds.  However,
!       as long as the delta UT and the UTC are consistent there
!       are no difficulties, except during a leap second.  In this
!       case, the start of the 61st second of the final minute should
!       begin a new MJD day and the old pre-leap delta UT should
!       continue to be used.  As the 61st second completes, the MJD
!       should revert to the start of the day as, simultaneously,
!       the delta UTC changes by one second to its post-leap new value.

!   3)  The delta UT (UT1-UTC) is tabulated in IERS circulars and
!       elsewhere.  It increases by exactly one second at the end of
!       each UTC leap second, introduced in order to keep delta UT
!       within +/- 0.9 seconds.

!   4)  IMPORTANT -- TAKE CARE WITH THE LONGITUDE SIGN CONVENTION.
!       The longitude required by the present routine is east-positive,
!       in accordance with geographical convention (and right-handed).
!       In particular, note that the longitudes returned by the
!       sla_OBS routine are west-positive, following astronomical
!       usage, and must be reversed in sign before use in the present
!       routine.

!   5)  The polar coordinates XP,YP can be obtained from IERS
!       circulars and equivalent publications.  The maximum amplitude
!       is about 0.3 arcseconds.  If XP,YP values are unavailable,
!       use XP=YP=0D0.  See page B60 of the 1988 Astronomical Almanac
!       for a definition of the two angles.

!   6)  The height above sea level of the observing station, HM,
!       can be obtained from the Astronomical Almanac (Section J
!       in the 1988 edition), or via the routine sla_OBS.  If P,
!       the pressure in millibars, is available, an adequate
!       estimate of HM can be obtained from the expression

!             HM ~ -8149.9415D0*LOG(P/1013.25D0).

!       (See Astrophysical Quantities, C.W.Allen, 3rd edition,
!       section 52.)  Similarly, if the pressure P is not known,
!       it can be estimated from the height of the observing
!       station, HM as follows:

!             P ~ 1013.25D0*EXP(-HM/8149.9415D0).

!       Note, however, that the refraction is proportional to the
!       pressure and that an accurate P value is important for
!       precise work.

!   7)  Repeated, computationally-expensive, calls to sla_AOPPA for
!       times that are very close together can be avoided by calling
!       sla_AOPPA just once and then using sla_AOPPAT for the subsequent
!       times.  Fresh calls to sla_AOPPA will be needed only when changes
!       in the precession have grown to unacceptable levels or when
!       anything affecting the refraction has changed.

!  Called:  sla_GEOC, sla_REFCO, sla_EQEQX, sla_AOPPAT

!  P.T.Wallace   Starlink   27 December 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr,aoprms(14)


!  2Pi
DOUBLE PRECISION :: d2pi
PARAMETER (d2pi=6.283185307179586476925287D0)

!  Seconds of time to radians
DOUBLE PRECISION :: s2r
PARAMETER (s2r=7.272205216643039903848712D-5)

!  Speed of light (AU per day)
DOUBLE PRECISION :: c
PARAMETER (c=173.14463331D0)

!  Ratio between solar and sidereal time
DOUBLE PRECISION :: solsid
PARAMETER (solsid=1.00273790935D0)

DOUBLE PRECISION :: cphim,xt,yt,zt,xc,yc,zc,elong,phi,uau,vau



!  Observer's location corrected for polar motion
cphim = COS(phim)
xt = COS(elongm)*cphim
yt = SIN(elongm)*cphim
zt = SIN(phim)
xc = xt-xp*zt
yc = yt+yp*zt
zc = xp*xt-yp*yt+zt
IF (xc == 0D0.AND.yc == 0D0) THEN
  elong = 0D0
ELSE
  elong = ATAN2(yc,xc)
END IF
phi = ATAN2(zc,SQRT(xc*xc+yc*yc))
aoprms(1) = phi
aoprms(2) = SIN(phi)
aoprms(3) = COS(phi)

!  Magnitude of the diurnal aberration vector
CALL sla_geoc(phi,hm,uau,vau)
aoprms(4) = d2pi*uau*solsid/c

!  Copy the refraction parameters and compute the A & B constants
aoprms(5) = hm
aoprms(6) = tdk
aoprms(7) = pmb
aoprms(8) = rh
aoprms(9) = wl
aoprms(10) = tlr
CALL sla_refco(hm,tdk,pmb,rh,wl,phi,tlr,1D-10,aoprms(11),aoprms(12))

!  Longitude + equation of the equinoxes + sidereal equivalent of DUT
!  (ignoring change in equation of the equinoxes between UTC and TDB)
aoprms(13) = elong+sla_eqeqx(date)+dut*solsid*s2r

!  Sidereal time
CALL sla_aoppat(date,aoprms)

END SUBROUTINE sla_aoppa
SUBROUTINE sla_aoppat (date, aoprms)
!+
!     - - - - - - -
!      A O P P A T
!     - - - - - - -

!  Recompute the sidereal time in the apparent to observed place
!  star-independent parameter block.

!  Given:
!     DATE   d      UTC date/time (modified Julian Date, JD-2400000.5)
!                   (see AOPPA source for comments on leap seconds)

!     AOPRMS d(14)  star-independent apparent-to-observed parameters

!       (1-12)   not required
!       (13)     longitude + eqn of equinoxes + sidereal DUT
!       (14)     not required

!  Returned:
!     AOPRMS d(14)  star-independent apparent-to-observed parameters:

!       (1-13)   not changed
!       (14)     local apparent sidereal time (radians)

!  For more information, see sla_AOPPA.

!  Called:  sla_GMST

!  P.T.Wallace   Starlink   1 July 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,aoprms(14)




aoprms(14) = sla_gmst(date)+aoprms(13)

END SUBROUTINE sla_aoppat
SUBROUTINE sla_aopqk (rap, dap, aoprms, aob, zob, hob, dob, rob)
!+
!     - - - - - -
!      A O P Q K
!     - - - - - -

!  Quick apparent to observed place (but see note 8, below, for
!  remarks about speed).

!  Given:
!     RAP    d      geocentric apparent right ascension
!     DAP    d      geocentric apparent declination
!     AOPRMS d(14)  star-independent apparent-to-observed parameters:

!       (1)      geodetic latitude (radians)
!       (2,3)    sine and cosine of geodetic latitude
!       (4)      magnitude of diurnal aberration vector
!       (5)      height (HM)
!       (6)      ambient temperature (T)
!       (7)      pressure (P)
!       (8)      relative humidity (RH)
!       (9)      wavelength (WL)
!       (10)     lapse rate (TLR)
!       (11,12)  refraction constants A and B (radians)
!       (13)     longitude + eqn of equinoxes + sidereal DUT (radians)
!       (14)     local apparent sidereal time (radians)

!  Returned:
!     AOB    d      observed azimuth (radians: N=0,E=90)
!     ZOB    d      observed zenith distance (radians)
!     HOB    d      observed Hour Angle (radians)
!     DOB    d      observed Declination (radians)
!     ROB    d      observed Right Ascension (radians)

!  Notes:

!   1)  This routine returns zenith distance rather than elevation
!       in order to reflect the fact that no allowance is made for
!       depression of the horizon.

!   2)  The accuracy of the result is limited by the corrections for
!       refraction.  Providing the meteorological parameters are
!       known accurately and there are no gross local effects, the
!       observed RA,Dec predicted by this routine should be within
!       about 0.1 arcsec for a zenith distance of less than 70 degrees.
!       Even at a topocentric zenith distance of 90 degrees, the
!       accuracy in elevation should be better than 1 arcmin;  useful
!       results are available for a further 3 degrees, beyond which
!       the slaRefro routine returns a fixed value of the refraction.
!       The complementary routines slaAop (or slaAopqk) and slaOap
!       (or slaOapqk) are self-consistent to better than 1 micro-
!       arcsecond all over the celestial sphere.

!   3)  It is advisable to take great care with units, as even
!       unlikely values of the input parameters are accepted and
!       processed in accordance with the models used.

!   4)  "Apparent" place means the geocentric apparent right ascension
!       and declination, which is obtained from a catalogue mean place
!       by allowing for space motion, parallax, precession, nutation,
!       annual aberration, and the Sun's gravitational lens effect.  For
!       star positions in the FK5 system (i.e. J2000), these effects can
!       be applied by means of the sla_MAP etc routines.  Starting from
!       other mean place systems, additional transformations will be
!       needed;  for example, FK4 (i.e. B1950) mean places would first
!       have to be converted to FK5, which can be done with the
!       sla_FK425 etc routines.

!   5)  "Observed" Az,El means the position that would be seen by a
!       perfect theodolite located at the observer.  This is obtained
!       from the geocentric apparent RA,Dec by allowing for Earth
!       orientation and diurnal aberration, rotating from equator
!       to horizon coordinates, and then adjusting for refraction.
!       The HA,Dec is obtained by rotating back into equatorial
!       coordinates, using the geodetic latitude corrected for polar
!       motion, and is the position that would be seen by a perfect
!       equatorial located at the observer and with its polar axis
!       aligned to the Earth's axis of rotation (n.b. not to the
!       refracted pole).  Finally, the RA is obtained by subtracting
!       the HA from the local apparent ST.

!   6)  To predict the required setting of a real telescope, the
!       observed place produced by this routine would have to be
!       adjusted for the tilt of the azimuth or polar axis of the
!       mounting (with appropriate corrections for mount flexures),
!       for non-perpendicularity between the mounting axes, for the
!       position of the rotator axis and the pointing axis relative
!       to it, for tube flexure, for gear and encoder errors, and
!       finally for encoder zero points.  Some telescopes would, of
!       course, exhibit other properties which would need to be
!       accounted for at the appropriate point in the sequence.

!   7)  The star-independent apparent-to-observed-place parameters
!       in AOPRMS may be computed by means of the sla_AOPPA routine.
!       If nothing has changed significantly except the time, the
!       sla_AOPPAT routine may be used to perform the requisite
!       partial recomputation of AOPRMS.

!   8)  At zenith distances beyond about 76 degrees, the need for
!       special care with the corrections for refraction causes a
!       marked increase in execution time.  Moreover, the effect
!       gets worse with increasing zenith distance.  Adroit
!       programming in the calling application may allow the
!       problem to be reduced.  Prepare an alternative AOPRMS array,
!       computed for zero air-pressure;  this will disable the
!       refraction corrections and cause rapid execution.  Using
!       this AOPRMS array, a preliminary call to the present routine
!       will, depending on the application, produce a rough position
!       which may be enough to establish whether the full, slow
!       calculation (using the real AOPRMS array) is worthwhile.
!       For example, there would be no need for the full calculation
!       if the preliminary call had already established that the
!       source was well below the elevation limits for a particular
!       telescope.

!  9)   The azimuths etc produced by the present routine are with
!       respect to the celestial pole.  Corrections to the terrestrial
!       pole can be computed using sla_POLMO.

!  Called:  sla_DCS2C, sla_REFZ, sla_REFRO, sla_DCC2S, sla_DRANRM

!  P.T.Wallace   Starlink   22 February 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rap,dap,aoprms(14),aob,zob,hob,dob,rob

!  Breakpoint for fast/slow refraction algorithm:
!  ZD greater than arctan(4), (see sla_REFCO routine)
!  or vector Z less than cosine(arctan(Z)) = 1/sqrt(17)
DOUBLE PRECISION :: zbreak
PARAMETER (zbreak=0.242535625D0)

INTEGER :: i

DOUBLE PRECISION :: sphi,cphi,st,v(3),xhd,yhd,zhd,diurab,f,  &
xhdt,yhdt,zhdt,xaet,yaet,zaet,azobs,zdt,refa,refb,zdobs,dzd,dref,ce,  &
xaeo,yaeo,zaeo,hmobs,dcobs,raobs



!  Sin, cos of latitude
sphi = aoprms(2)
cphi = aoprms(3)

!  Local apparent sidereal time
st = aoprms(14)

!  Apparent RA,Dec to Cartesian -HA,Dec
CALL sla_dcs2c(rap-st,dap,v)
xhd = v(1)
yhd = v(2)
zhd = v(3)

!  Diurnal aberration
diurab = aoprms(4)
f = (1D0-diurab*yhd)
xhdt = f*xhd
yhdt = f*(yhd+diurab)
zhdt = f*zhd

!  Cartesian -HA,Dec to Cartesian Az,El (S=0,E=90)
xaet = sphi*xhdt-cphi*zhdt
yaet = yhdt
zaet = cphi*xhdt+sphi*zhdt

!  Azimuth (N=0,E=90)
IF (xaet == 0D0.AND.yaet == 0D0) THEN
  azobs = 0D0
ELSE
  azobs = ATAN2(yaet,-xaet)
END IF

!  Topocentric zenith distance
zdt = ATAN2(SQRT(xaet*xaet+yaet*yaet),zaet)


!  Refraction
!  ----------

!  Fast algorithm using two constant model
refa = aoprms(11)
refb = aoprms(12)
CALL sla_refz(zdt,refa,refb,zdobs)

!  Large zenith distance?
IF (COS(zdobs) < zbreak) THEN
  
!     Yes: use rigorous algorithm
  
!     Initialize loop (maximum of 10 iterations)
  i = 1
  dzd = 1D1
  DO WHILE (ABS(dzd) > 1D-10.AND.i <= 10)
    
!        Compute refraction using current estimate of observed ZD
    CALL sla_refro(zdobs,aoprms(5),aoprms(6),aoprms(7),  &
    aoprms(8),aoprms(9),aoprms(1),aoprms(10),1D-8,dref)
    
!        Remaining discrepancy
    dzd = zdobs+dref-zdt
    
!        Update the estimate
    zdobs = zdobs-dzd
    
!        Increment the iteration counter
    i = i+1
  END DO
END IF

!  To Cartesian Az/ZD
ce = SIN(zdobs)
xaeo = -COS(azobs)*ce
yaeo = SIN(azobs)*ce
zaeo = COS(zdobs)

!  Cartesian Az/ZD to Cartesian -HA,Dec
v(1) = sphi*xaeo+cphi*zaeo
v(2) = yaeo
v(3) = -cphi*xaeo+sphi*zaeo

!  To spherical -HA,Dec
CALL sla_dcc2s(v,hmobs,dcobs)

!  Right Ascension
raobs = sla_dranrm(st+hmobs)

!  Return the results
aob = azobs
zob = zdobs
hob = -hmobs
dob = dcobs
rob = raobs

END SUBROUTINE sla_aopqk
SUBROUTINE sla_atmdsp (tdk, pmb, rh, wl1, a1, b1, wl2, a2, b2)
!+
!     - - - - - - -
!      A T M D S P
!     - - - - - - -

!  Apply atmospheric-dispersion adjustments to refraction coefficients.

!  Given:
!     TDK       d       ambient temperature, degrees K
!     PMB       d       ambient pressure, millibars
!     RH        d       ambient relative humidity, 0-1
!     WL1       d       reference wavelength, micrometre (0.4D0 recommended)
!     A1        d       refraction coefficient A for wavelength WL1 (radians)
!     B1        d       refraction coefficient B for wavelength WL1 (radians)
!     WL2       d       wavelength for which adjusted A,B required

!  Returned:
!     A2        d       refraction coefficient A for wavelength WL2 (radians)
!     B2        d       refraction coefficient B for wavelength WL2 (radians)

!  Notes:

!  1  To use this routine, first call sla_REFCO specifying WL1 as the
!     wavelength.  This yields refraction coefficients A1,B1, correct
!     for that wavelength.  Subsequently, calls to sla_ATMDSP specifying
!     different wavelengths will produce new, slightly adjusted
!     refraction coefficients which apply to the specified wavelength.

!  2  Most of the atmospheric dispersion happens between 0.7 micrometre
!     and the UV atmospheric cutoff, and the effect increases strongly
!     towards the UV end.  For this reason a blue reference wavelength
!     is recommended, for example 0.4 micrometres.

!  3  The accuracy, for this set of conditions:

!        height above sea level    2000 m
!                      latitude    29 deg
!                      pressure    793 mB
!                   temperature    17 degC
!                      humidity    50%
!                    lapse rate    0.0065 degC/m
!          reference wavelength    0.4 micrometre
!                star elevation    15 deg

!     is about 2.5 mas RMS between 0.3 and 1.0 micrometres, and stays
!     within 4 mas for the whole range longward of 0.3 micrometres
!     (compared with a total dispersion from 0.3 to 20.0 micrometres
!     of about 11 arcsec).  These errors are typical for ordinary
!     conditions and the given elevation;  in extreme conditions values
!     a few times this size may occur, while at higher elevations the
!     errors become much smaller.

!  4  If either wavelength exceeds 100 micrometres, the radio case
!     is assumed and the returned refraction coefficients are the
!     same as the given ones.

!  5  The algorithm consists of calculation of the refractivity of the
!     air at the observer for the two wavelengths, using the methods
!     of the sla_REFRO routine, and then scaling of the two refraction
!     coefficients according to classical refraction theory.  This
!     amounts to scaling the A coefficient in proportion to (n-1) and
!     the B coefficient almost in the same ratio (see R.M.Green,
!     "Spherical Astronomy", Cambridge University Press, 1985).

!  P.T.Wallace   Starlink   6 October 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: tdk,pmb,rh,wl1,a1,b1,wl2,a2,b2

DOUBLE PRECISION :: f,tdkok,pmbok,rhok,psat,pwo,w1,wlok,wlsq,w2,dn1,dn2


!  Check for radio wavelengths
IF (wl1 > 100D0.OR.wl2 > 100D0) THEN
  
!     Radio: no dispersion
  a2 = a1
  b2 = b1
ELSE
  
!     Optical: keep arguments within safe bounds
  tdkok = MIN(MAX(tdk,100D0),500D0)
  pmbok = MIN(MAX(pmb,0D0),10000D0)
  rhok = MIN(MAX(rh,0D0),1D0)
  
!     Atmosphere parameters at the observer
  psat = 10D0**(-8.7115D0+0.03477D0*tdkok)
  pwo = rhok*psat
  w1 = 11.2684D-6*pwo
  
!     Refractivity at the observer for first wavelength
  wlok = MAX(wl1,0.1D0)
  wlsq = wlok*wlok
  w2 = 77.5317D-6+(0.43909D-6+0.00367D-6/wlsq)/wlsq
  dn1 = (w2*pmbok-w1)/tdkok
  
!     Refractivity at the observer for second wavelength
  wlok = MAX(wl2,0.1D0)
  wlsq = wlok*wlok
  w2 = 77.5317D-6+(0.43909D-6+0.00367D-6/wlsq)/wlsq
  dn2 = (w2*pmbok-w1)/tdkok
  
!     Scale the refraction coefficients (see Green 4.31, p93)
  IF (dn1 /= 0D0) THEN
    f = dn2/dn1
    a2 = a1*f
    b2 = b1*f
    IF (dn1 /= a1) b2=b2*(1D0+dn1*(dn1-dn2)/(2D0*(dn1-a1)))
  ELSE
    a2 = a1
    b2 = b1
  END IF
END IF

END SUBROUTINE sla_atmdsp
SUBROUTINE sla__atms (rt, tt, dnt, gamal, r, dn, rdndr)
!+
!     - - - - -
!      A T M S
!     - - - - -

!  Internal routine used by REFRO

!  Refractive index and derivative with respect to height for the
!  stratosphere.

!  Given:
!    RT      d    height of tropopause from centre of the Earth (metre)
!    TT      d    temperature at the tropopause (deg K)
!    DNT     d    refractive index at the tropopause
!    GAMAL   d    constant of the atmospheric model = G*MD/R
!    R       d    current distance from the centre of the Earth (metre)

!  Returned:
!    DN      d    refractive index at R
!    RDNDR   d    R * rate the refractive index is changing at R

!  P.T.Wallace   Starlink   14 July 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rt,tt,dnt,gamal,r,dn,rdndr

DOUBLE PRECISION :: b,w


b = gamal/tt
w = (dnt-1D0)*EXP(-b*(r-rt))
dn = 1D0+w
rdndr = -r*b*w

END SUBROUTINE sla__atms
SUBROUTINE sla__atmt (r0, t0, alpha, gamm2, delm2,  &
c1, c2, c3, c4, c5, c6, r, t, dn, rdndr)
!+
!     - - - - -
!      A T M T
!     - - - - -

!  Internal routine used by REFRO

!  Refractive index and derivative with respect to height for the
!  troposphere.

!  Given:
!    R0      d    height of observer from centre of the Earth (metre)
!    T0      d    temperature at the observer (deg K)
!    ALPHA   d    alpha          )
!    GAMM2   d    gamma minus 2  ) see HMNAO paper
!    DELM2   d    delta minus 2  )
!    C1      d    useful term  )
!    C2      d    useful term  )
!    C3      d    useful term  ) see source
!    C4      d    useful term  ) of sla_REFRO
!    C5      d    useful term  )
!    C6      d    useful term  )
!    R       d    current distance from the centre of the Earth (metre)

!  Returned:
!    T       d    temperature at R (deg K)
!    DN      d    refractive index at R
!    RDNDR   d    R * rate the refractive index is changing at R

!  Note that in the optical case C5 and C6 are zero.

!  P.T.Wallace   Starlink   30 MAY 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: r0,t0,alpha,gamm2,delm2,c1,c2,c3,c4,c5,c6,r,t,dn,rdndr

DOUBLE PRECISION :: tt0,tt0gm2,tt0dm2


t = MAX(MIN(t0-alpha*(r-r0),320D0),100D0)
tt0 = t/t0
tt0gm2 = tt0**gamm2
tt0dm2 = tt0**delm2
dn = 1D0+(c1*tt0gm2-(c2-c5/t)*tt0dm2)*tt0
rdndr = r*(-c3*tt0gm2+(c4-c6/tt0)*tt0dm2)

END SUBROUTINE sla__atmt
SUBROUTINE sla_av2m (axvec, rmat)
!+
!     - - - - -
!      A V 2 M
!     - - - - -

!  Form the rotation matrix corresponding to a given axial vector.

!  (single precision)

!  A rotation matrix describes a rotation about some arbitrary axis.
!  The axis is called the Euler axis, and the angle through which the
!  reference frame rotates is called the Euler angle.  The axial
!  vector supplied to this routine has the same direction as the
!  Euler axis, and its magnitude is the Euler angle in radians.

!  Given:
!    AXVEC  r(3)     axial vector (radians)

!  Returned:
!    RMAT   r(3,3)   rotation matrix

!  If AXVEC is null, the unit matrix is returned.

!  The reference frame rotates clockwise as seen looking along
!  the axial vector from the origin.

!  P.T.Wallace   Starlink   June 1989

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: axvec(3),rmat(3,3)

REAL :: x,y,z,phi,s,c,w



!  Euler angle - magnitude of axial vector - and functions
x = axvec(1)
y = axvec(2)
z = axvec(3)
phi = SQRT(x*x+y*y+z*z)
s = SIN(phi)
c = COS(phi)
w = 1.0-c

!  Euler axis - direction of axial vector (perhaps null)
IF (phi /= 0.0) THEN
  x = x/phi
  y = y/phi
  z = z/phi
END IF

!  Compute the rotation matrix
rmat(1,1) = x*x*w+c
rmat(1,2) = x*y*w+z*s
rmat(1,3) = x*z*w-y*s
rmat(2,1) = x*y*w-z*s
rmat(2,2) = y*y*w+c
rmat(2,3) = y*z*w+x*s
rmat(3,1) = x*z*w+y*s
rmat(3,2) = y*z*w-x*s
rmat(3,3) = z*z*w+c

END SUBROUTINE sla_av2m
REAL FUNCTION sla_bear (a1, b1, a2, b2)
!+
!     - - - - -
!      B E A R
!     - - - - -

!  Bearing (position angle) of one point on a sphere relative to another
!  (single precision)

!  Given:
!     A1,B1    r    spherical coordinates of one point
!     A2,B2    r    spherical coordinates of the other point

!  (The spherical coordinates are RA,Dec, Long,Lat etc, in radians.)

!  The result is the bearing (position angle), in radians, of point
!  A2,B2 as seen from point A1,B1.  It is in the range +/- pi.  If
!  A2,B2 is due east of A1,B1 the bearing is +pi/2.  Zero is returned
!  if the two points are coincident.

!  P.T.Wallace   Starlink   23 March 1991

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: a1,b1,a2,b2

REAL :: da,x,y


da=a2-a1
y=SIN(da)*COS(b2)
x=SIN(b2)*COS(b1)-COS(b2)*SIN(b1)*COS(da)
IF (x /= 0.0.OR.y /= 0.0) THEN
  sla_bear=ATAN2(y,x)
ELSE
  sla_bear=0.0
END IF

END FUNCTION sla_bear
SUBROUTINE sla_caf2r (ideg, iamin, asec, rad, j)
!+
!     - - - - - -
!      C A F 2 R
!     - - - - - -

!  Convert degrees, arcminutes, arcseconds to radians
!  (single precision)

!  Given:
!     IDEG        int       degrees
!     IAMIN       int       arcminutes
!     ASEC        real      arcseconds

!  Returned:
!     RAD         real      angle in radians
!     J           int       status:  0 = OK
!                                    1 = IDEG outside range 0-359
!                                    2 = IAMIN outside range 0-59
!                                    3 = ASEC outside range 0-59.999...

!  Notes:

!  1)  The result is computed even if any of the range checks
!      fail.

!  2)  The sign must be dealt with outside this routine.

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ideg,iamin
REAL :: asec,rad
INTEGER :: j

!  Arc seconds to radians
REAL :: as2r
PARAMETER (as2r=0.484813681109535994E-5)



!  Preset status
j=0

!  Validate arcsec, arcmin, deg
IF (asec < 0.0.OR.asec >= 60.0) j=3
IF (iamin < 0.OR.iamin > 59) j=2
IF (ideg < 0.OR.ideg > 359) j=1

!  Compute angle
rad=as2r*(60.0*(60.0*REAL(ideg)+REAL(iamin))+asec)

END SUBROUTINE sla_caf2r
SUBROUTINE sla_caldj (iy, im, id, djm, j)
!+
!     - - - - - -
!      C A L D J
!     - - - - - -

!  Gregorian Calendar to Modified Julian Date

!  (Includes century default feature:  use sla_CLDJ for years
!   before 100AD.)

!  Given:
!     IY,IM,ID     int    year, month, day in Gregorian calendar

!  Returned:
!     DJM          dp     modified Julian Date (JD-2400000.5) for 0 hrs
!     J            int    status:
!                           0 = OK
!                           1 = bad year   (MJD not computed)
!                           2 = bad month  (MJD not computed)
!                           3 = bad day    (MJD computed)

!  Acceptable years are 00-49, interpreted as 2000-2049,
!                       50-99,     "       "  1950-1999,
!                       100 upwards, interpreted literally.

!  Called:  sla_CLDJ

!  P.T.Wallace   Starlink   November 1985

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: iy,im,id
DOUBLE PRECISION :: djm
INTEGER :: j

INTEGER :: ny




!  Default century if appropriate
IF (iy >= 0.AND.iy <= 49) THEN
  ny=iy+2000
ELSE IF (iy >= 50.AND.iy <= 99) THEN
  ny=iy+1900
ELSE
  ny=iy
END IF

!  Modified Julian Date
CALL sla_cldj(ny,im,id,djm,j)

END SUBROUTINE sla_caldj
SUBROUTINE sla_calyd (iy, im, id, ny, nd, j)
!+
!     - - - - - -
!      C A L Y D
!     - - - - - -

!  Gregorian calendar date to year and day in year (in a Julian
!  calendar aligned to the 20th/21st century Gregorian calendar).

!  (Includes century default feature:  use sla_CLYD for years
!   before 100AD.)

!  Given:
!     IY,IM,ID   int    year, month, day in Gregorian calendar
!                       (year may optionally omit the century)
!  Returned:
!     NY         int    year (re-aligned Julian calendar)
!     ND         int    day in year (1 = January 1st)
!     J          int    status:
!                         0 = OK
!                         1 = bad year (before -4711)
!                         2 = bad month
!                         3 = bad day (but conversion performed)

!  Notes:

!  1  This routine exists to support the low-precision routines
!     sla_EARTH, sla_MOON and sla_ECOR.

!  2  Between 1900 March 1 and 2100 February 28 it returns answers
!     which are consistent with the ordinary Gregorian calendar.
!     Outside this range there will be a discrepancy which increases
!     by one day for every non-leap century year.

!  3  Years in the range 50-99 are interpreted as 1950-1999, and
!     years in the range 00-49 are interpreted as 2000-2049.

!  Called:  sla_CLYD

!  P.T.Wallace   Starlink   23 November 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: iy,im,id,ny,nd,j

INTEGER :: i



!  Default century if appropriate
IF (iy >= 0.AND.iy <= 49) THEN
  i=iy+2000
ELSE IF (iy >= 50.AND.iy <= 99) THEN
  i=iy+1900
ELSE
  i=iy
END IF

!  Perform the conversion
CALL sla_clyd(i,im,id,ny,nd,j)

END SUBROUTINE sla_calyd
SUBROUTINE sla_cc2s (v, a, b)
!+
!     - - - - -
!      C C 2 S
!     - - - - -

!  Direction cosines to spherical coordinates (single precision)

!  Given:
!     V     r(3)   x,y,z vector

!  Returned:
!     A,B   r      spherical coordinates in radians

!  The spherical coordinates are longitude (+ve anticlockwise
!  looking from the +ve latitude pole) and latitude.  The
!  Cartesian coordinates are right handed, with the x axis
!  at zero longitude and latitude, and the z axis at the
!  +ve latitude pole.

!  If V is null, zero A and B are returned.
!  At either pole, zero A is returned.

!  P.T.Wallace   Starlink   July 1989

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: v(3),a,b

REAL :: x,y,z,r


x = v(1)
y = v(2)
z = v(3)
r = SQRT(x*x+y*y)

IF (r == 0.0) THEN
  a = 0.0
ELSE
  a = ATAN2(y,x)
END IF

IF (z == 0.0) THEN
  b = 0.0
ELSE
  b = ATAN2(z,r)
END IF

END SUBROUTINE sla_cc2s
SUBROUTINE sla_cc62s (v, a, b, r, ad, bd, rd)
!+
!     - - - - - -
!      C C 6 2 S
!     - - - - - -

!  Conversion of position & velocity in Cartesian coordinates
!  to spherical coordinates (single precision)

!  Given:
!     V      r(6)   Cartesian position & velocity vector

!  Returned:
!     A      r      longitude (radians)
!     B      r      latitude (radians)
!     R      r      radial coordinate
!     AD     r      longitude derivative (radians per unit time)
!     BD     r      latitude derivative (radians per unit time)
!     RD     r      radial derivative

!  P.T.Wallace   Starlink   28 April 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: v(6),a,b,r,ad,bd,rd

REAL :: x,y,z,xd,yd,zd,rxy2,rxy,r2,xyp



!  Components of position/velocity vector
x=v(1)
y=v(2)
z=v(3)
xd=v(4)
yd=v(5)
zd=v(6)

!  Component of R in XY plane squared
rxy2=x*x+y*y

!  Modulus squared
r2=rxy2+z*z

!  Protection against null vector
IF (r2 == 0.0) THEN
  x=xd
  y=yd
  z=zd
  rxy2=x*x+y*y
  r2=rxy2+z*z
END IF

!  Position and velocity in spherical coordinates
rxy=SQRT(rxy2)
xyp=x*xd+y*yd
IF (rxy2 /= 0.0) THEN
  a=ATAN2(y,x)
  b=ATAN2(z,rxy)
  ad=(x*yd-y*xd)/rxy2
  bd=(zd*rxy2-z*xyp)/(r2*rxy)
ELSE
  a=0.0
  IF (z /= 0.0) THEN
    b=ATAN2(z,rxy)
  ELSE
    b=0.0
  END IF
  ad=0.0
  bd=0.0
END IF
r=SQRT(r2)
IF (r /= 0.0) THEN
  rd=(xyp+z*zd)/r
ELSE
  rd=0.0
END IF

END SUBROUTINE sla_cc62s
SUBROUTINE sla_cd2tf (ndp, days, SIGN, ihmsf)
!+
!     - - - - - -
!      C D 2 T F
!     - - - - - -

!  Convert an interval in days into hours, minutes, seconds

!  (single precision)

!  Given:
!     NDP       int      number of decimal places of seconds
!     DAYS      real     interval in days

!  Returned:
!     SIGN      char     '+' or '-'
!     IHMSF     int(4)   hours, minutes, seconds, fraction

!  Notes:

!     1)  NDP less than zero is interpreted as zero.

!     2)  The largest useful value for NDP is determined by the size of
!         DAYS, the format of REAL floating-point numbers on the target
!         machine, and the risk of overflowing IHMSF(4).  For example,
!         on the VAX, for DAYS up to 1.0, the available floating-point
!         precision corresponds roughly to NDP=3.  This is well below
!         the ultimate limit of NDP=9 set by the capacity of the 32-bit
!         integer IHMSF(4).

!     3)  The absolute value of DAYS may exceed 1.0.  In cases where it
!         does not, it is up to the caller to test for and handle the
!         case where DAYS is very nearly 1.0 and rounds up to 24 hours,
!         by testing for IHMSF(1)=24 and setting IHMSF(1-4) to zero.

!  Called:  sla_DD2TF

!  P.T.Wallace   Starlink   12 December 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ndp
REAL :: days
CHARACTER :: SIGN*(*)
INTEGER :: ihmsf(4)



!  Call double precision version
CALL sla_dd2tf(ndp,DBLE(days),SIGN,ihmsf)

END SUBROUTINE sla_cd2tf
SUBROUTINE sla_cldj (iy, im, id, djm, j)
!+
!     - - - - -
!      C L D J
!     - - - - -

!  Gregorian Calendar to Modified Julian Date

!  Given:
!     IY,IM,ID     int    year, month, day in Gregorian calendar

!  Returned:
!     DJM          dp     modified Julian Date (JD-2400000.5) for 0 hrs
!     J            int    status:
!                           0 = OK
!                           1 = bad year   (MJD not computed)
!                           2 = bad month  (MJD not computed)
!                           3 = bad day    (MJD computed)

!  The year must be -4699 (i.e. 4700BC) or later.

!  The algorithm is derived from that of Hatcher 1984
!  (QJRAS 25, 53-55).

!  P.T.Wallace   Starlink   December 1985

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: iy,im,id
DOUBLE PRECISION :: djm
INTEGER :: j

!  Month lengths in days
INTEGER :: mtab(12)
DATA mtab/31,28,31,30,31,30,31,31,30,31,30,31/



!  Preset status
j=0

!  Validate year
IF (iy < -4699) THEN
  j=1
ELSE
  
!     Validate month
  IF (im >= 1.AND.im <= 12) THEN
    
!        Allow for leap year
    IF (MOD(iy,4) == 0) THEN
      mtab(2)=29
    ELSE
      mtab(2)=28
    END IF
    IF (MOD(iy,100) == 0.AND.MOD(iy,400) /= 0) mtab(2)=28
    
!        Validate day
    IF (id < 1.OR.id > mtab(im)) j=3
    
!        Modified Julian Date
    djm=DBLE((1461*(iy-(12-im)/10+4712))/4+(306*MOD(im+9,12)+5)/10  &
    -(3*((iy-(12-im)/10+4900)/100))/4+id-2399904)
    
!        Bad month
  ELSE
    j=2
  END IF
  
END IF

END SUBROUTINE sla_cldj
SUBROUTINE sla_clyd (iy, im, id, ny, nd, jstat)
!+
!     - - - - -
!      C L Y D
!     - - - - -

!  Gregorian calendar to year and day in year (in a Julian calendar
!  aligned to the 20th/21st century Gregorian calendar).

!  Given:
!     IY,IM,ID   i    year, month, day in Gregorian calendar

!  Returned:
!     NY         i    year (re-aligned Julian calendar)
!     ND         i    day in year (1 = January 1st)
!     JSTAT      i    status:
!                       0 = OK
!                       1 = bad year (before -4711)
!                       2 = bad month
!                       3 = bad day (but conversion performed)

!  Notes:

!  1  This routine exists to support the low-precision routines
!     sla_EARTH, sla_MOON and sla_ECOR.

!  2  Between 1900 March 1 and 2100 February 28 it returns answers
!     which are consistent with the ordinary Gregorian calendar.
!     Outside this range there will be a discrepancy which increases
!     by one day for every non-leap century year.

!  3  The essence of the algorithm is first to express the Gregorian
!     date as a Julian Day Number and then to convert this back to
!     a Julian calendar date, with day-in-year instead of month and
!     day.  See 12.92-1 and 12.95-1 in the reference.

!  Reference:  Explanatory Supplement to the Astronomical Almanac,
!              ed P.K.Seidelmann, University Science Books (1992),
!              p604-606.

!  P.T.Wallace   Starlink   26 November 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: iy,im,id,ny,nd,jstat

INTEGER :: i,j,k,l,n

!  Month lengths in days
INTEGER :: mtab(12)
DATA mtab/31,28,31,30,31,30,31,31,30,31,30,31/



!  Preset status
jstat=0

!  Validate year
IF (iy >= -4711) THEN
  
!     Validate month
  IF (im >= 1.AND.im <= 12) THEN
    
!        Allow for (Gregorian) leap year
    IF (MOD(iy,4) == 0.AND. (MOD(iy,100) /= 0.OR.MOD(iy,400) == 0)) THEN
      mtab(2)=29
    ELSE
      mtab(2)=28
    END IF
    
!        Validate day
    IF (id < 1.OR.id > mtab(im)) jstat=3
    
!        Perform the conversion
    i=(14-im)/12
    k=iy-i
    j=(1461*(k+4800))/4+(367*(im-2+12*i))/12-(3*((k+4900)/100))/4+id-30660
    k=(j-1)/1461
    l=j-1461*k
    n=(l-1)/365-l/1461
    j=((80*(l-365*n+30))/2447)/11
    i=n+j
    nd=59+l-365*i+((4-n)/4)*(1-j)
    ny=4*k+i-4716
    
!        Bad month
  ELSE
    jstat=2
  END IF
ELSE
  
!     Bad year
  jstat=1
END IF

END SUBROUTINE sla_clyd
SUBROUTINE sla_cr2af (ndp, angle, SIGN, idmsf)
!+
!     - - - - - -
!      C R 2 A F
!     - - - - - -

!  Convert an angle in radians into degrees, arcminutes, arcseconds
!  (single precision)

!  Given:
!     NDP       int      number of decimal places of arcseconds
!     ANGLE     real     angle in radians

!  Returned:
!     SIGN      char     '+' or '-'
!     IDMSF     int(4)   degrees, arcminutes, arcseconds, fraction

!  Notes:

!     1)  NDP less than zero is interpreted as zero.

!     2)  The largest useful value for NDP is determined by the size of
!         ANGLE, the format of REAL floating-point numbers on the target
!         machine, and the risk of overflowing IDMSF(4).  For example,
!         on the VAX, for ANGLE up to 2pi, the available floating-point
!         precision corresponds roughly to NDP=3.  This is well below
!         the ultimate limit of NDP=9 set by the capacity of the 32-bit
!         integer IHMSF(4).

!     3)  The absolute value of ANGLE may exceed 2pi.  In cases where it
!         does not, it is up to the caller to test for and handle the
!         case where ANGLE is very nearly 2pi and rounds up to 360 deg,
!         by testing for IDMSF(1)=360 and setting IDMSF(1-4) to zero.

!  Called:
!     sla_CD2TF

!  P.T.Wallace   Starlink   11 September 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ndp
REAL :: angle
CHARACTER :: SIGN*(*)
INTEGER :: idmsf(4)

!  Hours to degrees * radians to turns
REAL :: f
PARAMETER (f=15.0/6.283185307179586476925287)



!  Scale then use days to h,m,s routine
CALL sla_cd2tf(ndp,angle*f,SIGN,idmsf)

END SUBROUTINE sla_cr2af
SUBROUTINE sla_cr2tf (ndp, angle, SIGN, ihmsf)
!+
!     - - - - - -
!      C R 2 T F
!     - - - - - -

!  Convert an angle in radians into hours, minutes, seconds
!  (single precision)

!  Given:
!     NDP       int      number of decimal places of seconds
!     ANGLE     real     angle in radians

!  Returned:
!     SIGN      char     '+' or '-'
!     IHMSF     int(4)   hours, minutes, seconds, fraction

!  Notes:

!  1)  NDP less than zero is interpreted as zero.

!  2)  The largest useful value for NDP is determined by the size of
!      ANGLE, the format of REAL floating-point numbers on the target
!      machine, and the risk of overflowing IHMSF(4).  For example,
!      on the VAX, for ANGLE up to 2pi, the available floating-point
!      precision corresponds roughly to NDP=3.  This is well below
!      the ultimate limit of NDP=9 set by the capacity of the 32-bit
!      integer IHMSF(4).

!  3)  The absolute value of ANGLE may exceed 2pi.  In cases where it
!      does not, it is up to the caller to test for and handle the
!      case where ANGLE is very nearly 2pi and rounds up to 24 hours,
!      by testing for IHMSF(1)=24 and setting IHMSF(1-4) to zero.

!  Called:
!     sla_CD2TF

!  P.T.Wallace   Starlink   11 September 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ndp
REAL :: angle
CHARACTER :: SIGN*(*)
INTEGER :: ihmsf(4)

!  Turns to radians
REAL :: t2r
PARAMETER (t2r=6.283185307179586476925287)



!  Scale then use days to h,m,s routine
CALL sla_cd2tf(ndp,angle/t2r,SIGN,ihmsf)

END SUBROUTINE sla_cr2tf
SUBROUTINE sla_cs2c (a, b, v)
!+
!     - - - - -
!      C S 2 C
!     - - - - -

!  Spherical coordinates to direction cosines (single precision)

!  Given:
!     A,B      real      spherical coordinates in radians
!                        (RA,Dec), (Long,Lat) etc

!  Returned:
!     V        real(3)   x,y,z unit vector

!  The spherical coordinates are longitude (+ve anticlockwise
!  looking from the +ve latitude pole) and latitude.  The
!  Cartesian coordinates are right handed, with the x axis
!  at zero longitude and latitude, and the z axis at the
!  +ve latitude pole.

!  P.T.Wallace   Starlink   October 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: a,b,v(3)

REAL :: cosb



cosb=COS(b)

v(1)=COS(a)*cosb
v(2)=SIN(a)*cosb
v(3)=SIN(b)

END SUBROUTINE sla_cs2c
SUBROUTINE sla_cs2c6 (a, b, r, ad, bd, rd, v)
!+
!     - - - - - -
!      C S 2 C 6
!     - - - - - -

!  Conversion of position & velocity in spherical coordinates
!  to Cartesian coordinates (single precision)

!  Given:
!     A     r      longitude (radians)
!     B     r      latitude (radians)
!     R     r      radial coordinate
!     AD    r      longitude derivative (radians per unit time)
!     BD    r      latitude derivative (radians per unit time)
!     RD    r      radial derivative

!  Returned:
!     V     r(6)   Cartesian position & velocity vector

!  P.T.Wallace   Starlink   November 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: a,b,r,ad,bd,rd,v(6)

REAL :: sa,ca,sb,cb,rcb,x,y,rbd,cbrd,w



!  Useful functions
sa=SIN(a)
ca=COS(a)
sb=SIN(b)
cb=COS(b)
rcb=r*cb
x=rcb*ca
y=rcb*sa
rbd=r*bd
cbrd=cb*rd
w=rbd*sb-cb*rd

!  Position
v(1)=x
v(2)=y
v(3)=r*sb

!  Velocity
v(4)=-y*ad-w*ca
v(5)=x*ad-w*sa
v(6)=rbd*cb+sb*rd

END SUBROUTINE sla_cs2c6
SUBROUTINE sla_ctf2d (ihour, imin, sec, days, j)
!+
!     - - - - - -
!      C T F 2 D
!     - - - - - -

!  Convert hours, minutes, seconds to days (single precision)

!  Given:
!     IHOUR       int       hours
!     IMIN        int       minutes
!     SEC         real      seconds

!  Returned:
!     DAYS        real      interval in days
!     J           int       status:  0 = OK
!                                    1 = IHOUR outside range 0-23
!                                    2 = IMIN outside range 0-59
!                                    3 = SEC outside range 0-59.999...

!  Notes:

!  1)  The result is computed even if any of the range checks
!      fail.

!  2)  The sign must be dealt with outside this routine.

!  P.T.Wallace   Starlink   November 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ihour,imin
REAL :: sec,days
INTEGER :: j

!  Seconds per day
REAL :: d2s
PARAMETER (d2s=86400.0)



!  Preset status
j=0

!  Validate sec, min, hour
IF (sec < 0.0.OR.sec >= 60.0) j=3
IF (imin < 0.OR.imin > 59) j=2
IF (ihour < 0.OR.ihour > 23) j=1

!  Compute interval
days=(60.0*(60.0*REAL(ihour)+REAL(imin))+sec)/d2s

END SUBROUTINE sla_ctf2d
SUBROUTINE sla_ctf2r (ihour, imin, sec, rad, j)
!+
!     - - - - - -
!      C T F 2 R
!     - - - - - -

!  Convert hours, minutes, seconds to radians (single precision)

!  Given:
!     IHOUR       int       hours
!     IMIN        int       minutes
!     SEC         real      seconds

!  Returned:
!     RAD         real      angle in radians
!     J           int       status:  0 = OK
!                                    1 = IHOUR outside range 0-23
!                                    2 = IMIN outside range 0-59
!                                    3 = SEC outside range 0-59.999...

!  Called:
!     sla_CTF2D

!  Notes:

!  1)  The result is computed even if any of the range checks
!      fail.

!  2)  The sign must be dealt with outside this routine.

!  P.T.Wallace   Starlink   November 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ihour,imin
REAL :: sec,rad
INTEGER :: j

REAL :: turns

!  Turns to radians
REAL :: t2r
PARAMETER (t2r=6.283185307179586476925287)



!  Convert to turns then radians
CALL sla_ctf2d(ihour,imin,sec,turns,j)
rad=t2r*turns

END SUBROUTINE sla_ctf2r
SUBROUTINE sla_daf2r (ideg, iamin, asec, rad, j)
!+
!     - - - - - -
!      D A F 2 R
!     - - - - - -

!  Convert degrees, arcminutes, arcseconds to radians
!  (double precision)

!  Given:
!     IDEG        int       degrees
!     IAMIN       int       arcminutes
!     ASEC        dp        arcseconds

!  Returned:
!     RAD         dp        angle in radians
!     J           int       status:  0 = OK
!                                    1 = IDEG outside range 0-359
!                                    2 = IAMIN outside range 0-59
!                                    3 = ASEC outside range 0-59.999...

!  Notes:
!     1)  The result is computed even if any of the range checks
!         fail.
!     2)  The sign must be dealt with outside this routine.

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ideg,iamin
DOUBLE PRECISION :: asec,rad
INTEGER :: j

!  Arc seconds to radians
DOUBLE PRECISION :: as2r
PARAMETER (as2r=0.484813681109535994D-5)



!  Preset status
j=0

!  Validate arcsec, arcmin, deg
IF (asec < 0D0.OR.asec >= 60D0) j=3
IF (iamin < 0.OR.iamin > 59) j=2
IF (ideg < 0.OR.ideg > 359) j=1

!  Compute angle
rad=as2r*(60D0*(60D0*DBLE(ideg)+DBLE(iamin))+asec)

END SUBROUTINE sla_daf2r
SUBROUTINE sla_dafin (string, iptr, a, j)
!+
!     - - - - - -
!      D A F I N
!     - - - - - -

!  Sexagesimal character string to angle (double precision)

!  Given:
!     STRING  c*(*)   string containing deg, arcmin, arcsec fields
!     IPTR      i     pointer to start of decode (1st = 1)

!  Returned:
!     IPTR      i     advanced past the decoded angle
!     A         d     angle in radians
!     J         i     status:  0 = OK
!                             +1 = default, A unchanged
!                             -1 = bad degrees      )
!                             -2 = bad arcminutes   )  (note 3)
!                             -3 = bad arcseconds   )

!  Example:

!    argument    before                           after

!    STRING      '-57 17 44.806  12 34 56.7'      unchanged
!    IPTR        1                                16 (points to 12...)
!    A           ?                                -1.00000D0
!    J           ?                                0

!    A further call to sla_DAFIN, without adjustment of IPTR, will
!    decode the second angle, 12deg 34min 56.7sec.

!  Notes:

!     1)  The first three "fields" in STRING are degrees, arcminutes,
!         arcseconds, separated by spaces or commas.  The degrees field
!         may be signed, but not the others.  The decoding is carried
!         out by the DFLTIN routine and is free-format.

!     2)  Successive fields may be absent, defaulting to zero.  For
!         zero status, the only combinations allowed are degrees alone,
!         degrees and arcminutes, and all three fields present.  If all
!         three fields are omitted, a status of +1 is returned and A is
!         unchanged.  In all other cases A is changed.

!     3)  Range checking:

!           The degrees field is not range checked.  However, it is
!           expected to be integral unless the other two fields are absent.

!           The arcminutes field is expected to be 0-59, and integral if
!           the arcseconds field is present.  If the arcseconds field
!           is absent, the arcminutes is expected to be 0-59.9999...

!           The arcseconds field is expected to be 0-59.9999...

!     4)  Decoding continues even when a check has failed.  Under these
!         circumstances the field takes the supplied value, defaulting
!         to zero, and the result A is computed and returned.

!     5)  Further fields after the three expected ones are not treated
!         as an error.  The pointer IPTR is left in the correct state
!         for further decoding with the present routine or with DFLTIN
!         etc. See the example, above.

!     6)  If STRING contains hours, minutes, seconds instead of degrees
!         etc, or if the required units are turns (or days) instead of
!         radians, the result A should be multiplied as follows:

!           for        to obtain    multiply
!           STRING     A in         A by

!           d ' "      radians      1       =  1D0
!           d ' "      turns        1/2pi   =  0.1591549430918953358D0
!           h m s      radians      15      =  15D0
!           h m s      days         15/2pi  =  2.3873241463784300365D0

!  Called:  sla_DFLTIN

!  P.T.Wallace   Starlink   1 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: iptr
DOUBLE PRECISION :: a
INTEGER :: j

DOUBLE PRECISION :: as2r
PARAMETER (as2r=4.84813681109535993589914102358D-6)
INTEGER :: jf,jd,jm,js
DOUBLE PRECISION :: deg,arcmin,arcsec



!  Preset the status to OK
jf=0

!  Defaults
deg=0D0
arcmin=0D0
arcsec=0D0

!  Decode degrees, arcminutes, arcseconds
CALL sla_dfltin(string,iptr,deg,jd)
IF (jd > 1) THEN
  jf=-1
ELSE
  CALL sla_dfltin(string,iptr,arcmin,jm)
  IF (jm < 0.OR.jm > 1) THEN
    jf=-2
  ELSE
    CALL sla_dfltin(string,iptr,arcsec,js)
    IF (js < 0.OR.js > 1) THEN
      jf=-3
      
!        See if the combination of fields is credible
    ELSE IF (jd > 0) THEN
!           No degrees:  arcmin, arcsec ought also to be absent
      IF (jm == 0) THEN
!              Suspect arcmin
        jf=-2
      ELSE IF (js == 0) THEN
!              Suspect arcsec
        jf=-3
      ELSE
!              All three fields absent
        jf=1
      END IF
!        Degrees present:  if arcsec present so ought arcmin to be
    ELSE IF (jm /= 0.AND.js == 0) THEN
      jf=-3
      
!        Tests for range and integrality
      
!        Degrees
    ELSE IF (jm == 0.AND.DINT(deg) /= deg) THEN
      jf=-1
!        Arcminutes
    ELSE IF ((js == 0.AND.DINT(arcmin) /= arcmin).OR.  &
      arcmin >= 60D0) THEN
      jf=-2
!        Arcseconds
    ELSE IF (arcsec >= 60D0) THEN
      jf=-3
    END IF
  END IF
END IF

!  Unless all three fields absent, compute angle value
IF (jf <= 0) THEN
  a=as2r*(60D0*(60D0*ABS(deg)+arcmin)+arcsec)
  IF (jd < 0) a=-a
END IF

!  Return the status
j=jf

END SUBROUTINE sla_dafin
DOUBLE PRECISION FUNCTION sla_dat (utc)
!+
!     - - - -
!      D A T
!     - - - -

!  Increment to be applied to Coordinated Universal Time UTC to give
!  International Atomic Time TAI (double precision)

!  Given:
!     UTC      d      UTC date as a modified JD (JD-2400000.5)

!  Result:  TAI-UTC in seconds

!  Notes:

!  1  The UTC is specified to be a date rather than a time to indicate
!     that care needs to be taken not to specify an instant which lies
!     within a leap second.  Though in most cases UTC can include the
!     fractional part, correct behaviour on the day of a leap second
!     can only be guaranteed up to the end of the second 23:59:59.

!  2  For epochs from 1961 January 1 onwards, the expressions from the
!     file ftp://maia.usno.navy.mil/ser7/tai-utc.dat are used.

!  3  The 5ms timestep at 1961 January 1 is taken from 2.58.1 (p87) of
!     the 1992 Explanatory Supplement.

!  4  UTC began at 1960 January 1.0 (JD 2436934.5) and it is improper
!     to call the routine with an earlier epoch.  However, if this
!     is attempted, the TAI-UTC expression for the year 1960 is used.


!     :-----------------------------------------:
!     :                                         :
!     :                IMPORTANT                :
!     :                                         :
!     :  This routine must be updated on each   :
!     :     occasion that a leap second is      :
!     :                announced                :
!     :                                         :
!     :  Latest leap second:  1997 July 1       :
!     :                                         :
!     :-----------------------------------------:

!  P.T.Wallace   Starlink   5 January 1998

!  Copyright (C) 1998 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: utc

DOUBLE PRECISION :: dt



IF (.false.) THEN
  
! - - - - - - - - - - - - - - - - - - - - - - *
!  Add new code here on each occasion that a  *
!  leap second is announced, and update the   *
!  preamble comments appropriately.           *
! - - - - - - - - - - - - - - - - - - - - - - *
  
!     1997 July 1
ELSE IF (utc >= 50630D0) THEN
  dt=31D0
  
!     1996 January 1
ELSE IF (utc >= 50083D0) THEN
  dt=30D0
  
!     1994 July 1
ELSE IF (utc >= 49534D0) THEN
  dt=29D0
  
!     1993 July 1
ELSE IF (utc >= 49169D0) THEN
  dt=28D0
  
!     1992 July 1
ELSE IF (utc >= 48804D0) THEN
  dt=27D0
  
!     1991 January 1
ELSE IF (utc >= 48257D0) THEN
  dt=26D0
  
!     1990 January 1
ELSE IF (utc >= 47892D0) THEN
  dt=25D0
  
!     1988 January 1
ELSE IF (utc >= 47161D0) THEN
  dt=24D0
  
!     1985 July 1
ELSE IF (utc >= 46247D0) THEN
  dt=23D0
  
!     1983 July 1
ELSE IF (utc >= 45516D0) THEN
  dt=22D0
  
!     1982 July 1
ELSE IF (utc >= 45151D0) THEN
  dt=21D0
  
!     1981 July 1
ELSE IF (utc >= 44786D0) THEN
  dt=20D0
  
!     1980 January 1
ELSE IF (utc >= 44239D0) THEN
  dt=19D0
  
!     1979 January 1
ELSE IF (utc >= 43874D0) THEN
  dt=18D0
  
!     1978 January 1
ELSE IF (utc >= 43509D0) THEN
  dt=17D0
  
!     1977 January 1
ELSE IF (utc >= 43144D0) THEN
  dt=16D0
  
!     1976 January 1
ELSE IF (utc >= 42778D0) THEN
  dt=15D0
  
!     1975 January 1
ELSE IF (utc >= 42413D0) THEN
  dt=14D0
  
!     1974 January 1
ELSE IF (utc >= 42048D0) THEN
  dt=13D0
  
!     1973 January 1
ELSE IF (utc >= 41683D0) THEN
  dt=12D0
  
!     1972 July 1
ELSE IF (utc >= 41499D0) THEN
  dt=11D0
  
!     1972 January 1
ELSE IF (utc >= 41371D0) THEN
  dt=10D0
  
!     1968 February 1
ELSE IF (utc >= 39887D0) THEN
  dt=4.2131700D0+(utc-39126D0)*0.002592D0
  
!     1966 January 1
ELSE IF (utc >= 39126D0) THEN
  dt=4.3131700D0+(utc-39126D0)*0.002592D0
  
!     1965 September 1
ELSE IF (utc >= 39004D0) THEN
  dt=3.8401300D0+(utc-38761D0)*0.001296D0
  
!     1965 July 1
ELSE IF (utc >= 38942D0) THEN
  dt=3.7401300D0+(utc-38761D0)*0.001296D0
  
!     1965 March 1
ELSE IF (utc >= 38820D0) THEN
  dt=3.6401300D0+(utc-38761D0)*0.001296D0
  
!     1965 January 1
ELSE IF (utc >= 38761D0) THEN
  dt=3.5401300D0+(utc-38761D0)*0.001296D0
  
!     1964 September 1
ELSE IF (utc >= 38639D0) THEN
  dt=3.4401300D0+(utc-38761D0)*0.001296D0
  
!     1964 April 1
ELSE IF (utc >= 38486D0) THEN
  dt=3.3401300D0+(utc-38761D0)*0.001296D0
  
!     1964 January 1
ELSE IF (utc >= 38395D0) THEN
  dt=3.2401300D0+(utc-38761D0)*0.001296D0
  
!     1963 November 1
ELSE IF (utc >= 38334D0) THEN
  dt=1.9458580D0+(utc-37665D0)*0.0011232D0
  
!     1962 January 1
ELSE IF (utc >= 37665D0) THEN
  dt=1.8458580D0+(utc-37665D0)*0.0011232D0
  
!     1961 August 1
ELSE IF (utc >= 37512D0) THEN
  dt=1.3728180D0+(utc-37300D0)*0.001296D0
  
!     1961 January 1
ELSE IF (utc >= 37300D0) THEN
  dt=1.4228180D0+(utc-37300D0)*0.001296D0
  
!     Before that
ELSE
  dt=1.4178180D0+(utc-37300D0)*0.001296D0
  
END IF

sla_dat=dt

END FUNCTION sla_dat
SUBROUTINE sla_dav2m (axvec, rmat)
!+
!     - - - - - -
!      D A V 2 M
!     - - - - - -

!  Form the rotation matrix corresponding to a given axial vector.
!  (double precision)

!  A rotation matrix describes a rotation about some arbitrary axis.
!  The axis is called the Euler axis, and the angle through which the
!  reference frame rotates is called the Euler angle.  The axial
!  vector supplied to this routine has the same direction as the
!  Euler axis, and its magnitude is the Euler angle in radians.

!  Given:
!    AXVEC  d(3)     axial vector (radians)

!  Returned:
!    RMAT   d(3,3)   rotation matrix

!  If AXVEC is null, the unit matrix is returned.

!  The reference frame rotates clockwise as seen looking along
!  the axial vector from the origin.

!  P.T.Wallace   Starlink   June 1989

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: axvec(3),rmat(3,3)

DOUBLE PRECISION :: x,y,z,phi,s,c,w



!  Euler angle - magnitude of axial vector - and functions
x = axvec(1)
y = axvec(2)
z = axvec(3)
phi = SQRT(x*x+y*y+z*z)
s = SIN(phi)
c = COS(phi)
w = 1D0-c

!  Euler axis - direction of axial vector (perhaps null)
IF (phi /= 0D0) THEN
  x = x/phi
  y = y/phi
  z = z/phi
END IF

!  Compute the rotation matrix
rmat(1,1) = x*x*w+c
rmat(1,2) = x*y*w+z*s
rmat(1,3) = x*z*w-y*s
rmat(2,1) = x*y*w-z*s
rmat(2,2) = y*y*w+c
rmat(2,3) = y*z*w+x*s
rmat(3,1) = x*z*w+y*s
rmat(3,2) = y*z*w-x*s
rmat(3,3) = z*z*w+c

END SUBROUTINE sla_dav2m
DOUBLE PRECISION FUNCTION sla_dbear (a1, b1, a2, b2)
!+
!     - - - - - -
!      D B E A R
!     - - - - - -

!  Bearing (position angle) of one point on a sphere relative to another
!  (double precision)

!  Given:
!     A1,B1    d    spherical coordinates of one point
!     A2,B2    d    spherical coordinates of the other point

!  (The spherical coordinates are RA,Dec, Long,Lat etc, in radians.)

!  The result is the bearing (position angle), in radians, of point
!  A2,B2 as seen from point A1,B1.  It is in the range +/- pi.  If
!  A2,B2 is due east of A1,B1 the bearing is +pi/2.  Zero is returned
!  if the two points are coincident.

!  P.T.Wallace   Starlink   23 March 1991

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: a1,b1,a2,b2

DOUBLE PRECISION :: da,x,y


da=a2-a1
y=SIN(da)*COS(b2)
x=SIN(b2)*COS(b1)-COS(b2)*SIN(b1)*COS(da)
IF (x /= 0D0.OR.y /= 0D0) THEN
  sla_dbear=ATAN2(y,x)
ELSE
  sla_dbear=0D0
END IF

END FUNCTION sla_dbear
SUBROUTINE sla_dbjin (string, nstrt, dreslt, j1, j2)
!+
!     - - - - - -
!      D B J I N
!     - - - - - -

!  Convert free-format input into double precision floating point,
!  using DFLTIN but with special syntax extensions.

!  The purpose of the syntax extensions is to help cope with mixed
!  FK4 and FK5 data.  In addition to the syntax accepted by DFLTIN,
!  the following two extensions are recognized by DBJIN:

!     1)  A valid non-null field preceded by the character 'B'
!         (or 'b') is accepted.

!     2)  A valid non-null field preceded by the character 'J'
!         (or 'j') is accepted.

!  The calling program is notified of the incidence of either of these
!  extensions through an supplementary status argument.  The rest of
!  the arguments are as for DFLTIN.

!  Given:
!     STRING      char       string containing field to be decoded
!     NSTRT       int        pointer to 1st character of field in string

!  Returned:
!     NSTRT       int        incremented
!     DRESLT      double     result
!     J1          int        DFLTIN status: -1 = -OK
!                                            0 = +OK
!                                           +1 = null field
!                                           +2 = error
!     J2          int        syntax flag:  0 = normal DFLTIN syntax
!                                         +1 = 'B' or 'b'
!                                         +2 = 'J' or 'j'

!  Called:  sla_DFLTIN

!  For details of the basic syntax, see sla_DFLTIN.

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: nstrt
DOUBLE PRECISION :: dreslt
INTEGER :: j1,j2

INTEGER :: j2a,lenstr,na,j1a,nb,j1b
CHARACTER :: c



!   Preset syntax flag
j2a=0

!   Length of string
lenstr=LEN(string)

!   Pointer to current character
na=nstrt

!   Attempt normal decode
CALL sla_dfltin(string,na,dreslt,j1a)

!   Proceed only if pointer still within string
IF (na >= 1.AND.na <= lenstr) THEN
  
!      See if DFLTIN reported a null field
  IF (j1a == 1) THEN
    
!         It did: examine character it stuck on
    c=string(na:na)
    IF (c == 'B'.OR.c == 'b') THEN
!            'B' - provisionally note
      j2a=1
    ELSE IF (c == 'J'.OR.c == 'j') THEN
!            'J' - provisionally note
      j2a=2
    END IF
    
!         Following B or J, attempt to decode a number
    IF (j2a == 1.OR.j2a == 2) THEN
      nb=na+1
      CALL sla_dfltin(string,nb,dreslt,j1b)
      
!            If successful, copy pointer and status
      IF (j1b <= 0) THEN
        na=nb
        j1a=j1b
!            If not, forget about the B or J
      ELSE
        j2a=0
      END IF
      
    END IF
    
  END IF
  
END IF

!   Return argument values and exit
nstrt=na
j1=j1a
j2=j2a

END SUBROUTINE sla_dbjin
SUBROUTINE sla_dc62s (v, a, b, r, ad, bd, rd)
!+
!     - - - - - -
!      D C 6 2 S
!     - - - - - -

!  Conversion of position & velocity in Cartesian coordinates
!  to spherical coordinates (double precision)

!  Given:
!     V      d(6)   Cartesian position & velocity vector

!  Returned:
!     A      d      longitude (radians)
!     B      d      latitude (radians)
!     R      d      radial coordinate
!     AD     d      longitude derivative (radians per unit time)
!     BD     d      latitude derivative (radians per unit time)
!     RD     d      radial derivative

!  P.T.Wallace   Starlink   28 April 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: v(6),a,b,r,ad,bd,rd

DOUBLE PRECISION :: x,y,z,xd,yd,zd,rxy2,rxy,r2,xyp



!  Components of position/velocity vector
x=v(1)
y=v(2)
z=v(3)
xd=v(4)
yd=v(5)
zd=v(6)

!  Component of R in XY plane squared
rxy2=x*x+y*y

!  Modulus squared
r2=rxy2+z*z

!  Protection against null vector
IF (r2 == 0D0) THEN
  x=xd
  y=yd
  z=zd
  rxy2=x*x+y*y
  r2=rxy2+z*z
END IF

!  Position and velocity in spherical coordinates
rxy=SQRT(rxy2)
xyp=x*xd+y*yd
IF (rxy2 /= 0D0) THEN
  a=ATAN2(y,x)
  b=ATAN2(z,rxy)
  ad=(x*yd-y*xd)/rxy2
  bd=(zd*rxy2-z*xyp)/(r2*rxy)
ELSE
  a=0D0
  IF (z /= 0D0) THEN
    b=ATAN2(z,rxy)
  ELSE
    b=0D0
  END IF
  ad=0D0
  bd=0D0
END IF
r=SQRT(r2)
IF (r /= 0D0) THEN
  rd=(xyp+z*zd)/r
ELSE
  rd=0D0
END IF

END SUBROUTINE sla_dc62s
SUBROUTINE sla_dcc2s (v, a, b)
!+
!     - - - - - -
!      D C C 2 S
!     - - - - - -

!  Direction cosines to spherical coordinates (double precision)

!  Given:
!     V     d(3)   x,y,z vector

!  Returned:
!     A,B   d      spherical coordinates in radians

!  The spherical coordinates are longitude (+ve anticlockwise
!  looking from the +ve latitude pole) and latitude.  The
!  Cartesian coordinates are right handed, with the x axis
!  at zero longitude and latitude, and the z axis at the
!  +ve latitude pole.

!  If V is null, zero A and B are returned.
!  At either pole, zero A is returned.

!  P.T.Wallace   Starlink   July 1989

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: v(3),a,b

DOUBLE PRECISION :: x,y,z,r


x = v(1)
y = v(2)
z = v(3)
r = SQRT(x*x+y*y)

IF (r == 0D0) THEN
  a = 0D0
ELSE
  a = ATAN2(y,x)
END IF

IF (z == 0D0) THEN
  b = 0D0
ELSE
  b = ATAN2(z,r)
END IF

END SUBROUTINE sla_dcc2s
SUBROUTINE sla_dcmpf (coeffs,xz,yz,xs,ys,perp,orient)
!+
!     - - - - - -
!      D C M P F
!     - - - - - -

!  Decompose an [X,Y] linear fit into its constituent parameters:
!  zero points, scales, nonperpendicularity and orientation.

!  Given:
!     COEFFS  d(6)      transformation coefficients (see note)

!  Returned:
!     XZ       d        x zero point
!     YZ       d        y zero point
!     XS       d        x scale
!     YS       d        y scale
!     PERP     d        nonperpendicularity (radians)
!     ORIENT   d        orientation (radians)

!  The model relates two sets of [X,Y] coordinates as follows.
!  Naming the elements of COEFFS:

!     COEFFS(1) = A
!     COEFFS(2) = B
!     COEFFS(3) = C
!     COEFFS(4) = D
!     COEFFS(5) = E
!     COEFFS(6) = F

!  the model transforms coordinates [X1,Y1] into coordinates
!  [X2,Y2] as follows:

!     X2 = A + B*X1 + C*Y1
!     Y2 = D + E*X1 + F*Y1

!  The transformation can be decomposed into four steps:

!     1)  Zero points:

!             x' = XZ + X1
!             y' = YZ + Y1

!     2)  Scales:

!             x'' = XS*x'
!             y'' = YS*y'

!     3)  Nonperpendicularity:

!             x''' = cos(PERP/2)*x'' + sin(PERP/2)*y''
!             y''' = sin(PERP/2)*x'' + cos(PERP/2)*y''

!     4)  Orientation:

!             X2 = cos(ORIENT)*x''' + sin(ORIENT)*y'''
!             Y2 =-sin(ORIENT)*y''' + cos(ORIENT)*y'''

!  See also sla_FITXY, sla_PXY, sla_INVF, sla_XY2XY

!  P.T.Wallace   Starlink   14 August 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: coeffs(6),xz,yz,xs,ys,perp,orient

DOUBLE PRECISION :: a,b,c,d,e,f,rb2e2,rc2f2,xsc,ysc,p1,p2,p,ws,wc,  &
OR,hp,shp,chp,sor,cor,det,x0,y0



!  Copy the six coefficients
a=coeffs(1)
b=coeffs(2)
c=coeffs(3)
d=coeffs(4)
e=coeffs(5)
f=coeffs(6)

!  Scales
rb2e2=SQRT(b*b+e*e)
rc2f2=SQRT(c*c+f*f)
IF (b*f-c*e >= 0D0) THEN
  xsc=rb2e2
ELSE
  b=-b
  c=-c
  xsc=-rb2e2
END IF
ysc=rc2f2

!  Non-perpendicularity
IF (c /= 0D0.OR.f /= 0D0) THEN
  p1=ATAN2(c,f)
ELSE
  p1=0D0
END IF
IF (e /= 0D0.OR.b /= 0D0) THEN
  p2=ATAN2(e,b)
ELSE
  p2=0D0
END IF
p=p1+p2

!  Orientation
ws=c*rb2e2-e*rc2f2
wc=b*rc2f2+f*rb2e2
IF (ws /= 0D0.OR.wc /= 0D0) THEN
  OR=ATAN2(ws,wc)
ELSE
  OR=0D0
END IF

!  Zero corrections
hp=p/2D0
shp=SIN(hp)
chp=COS(hp)
sor=SIN(OR)
cor=COS(OR)
det=xsc*ysc*(chp+shp)*(chp-shp)
IF (ABS(det) > 0D0) THEN
  x0=ysc*(a*(chp*cor-shp*sor)-d*(chp*sor+shp*cor))/det
  y0=xsc*(a*(chp*sor-shp*cor)+d*(chp*cor+shp*sor))/det
ELSE
  x0=0D0
  y0=0D0
END IF

!  Results
xz=x0
yz=y0
xs=xsc
ys=ysc
perp=p
orient=OR

END SUBROUTINE sla_dcmpf
SUBROUTINE sla_dcs2c (a, b, v)
!+
!     - - - - - -
!      D C S 2 C
!     - - - - - -

!  Spherical coordinates to direction cosines (double precision)

!  Given:
!     A,B       dp      spherical coordinates in radians
!                        (RA,Dec), (Long,Lat) etc

!  Returned:
!     V         dp(3)   x,y,z unit vector

!  The spherical coordinates are longitude (+ve anticlockwise
!  looking from the +ve latitude pole) and latitude.  The
!  Cartesian coordinates are right handed, with the x axis
!  at zero longitude and latitude, and the z axis at the
!  +ve latitude pole.

!  P.T.Wallace   Starlink   October 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: a,b,v(3)

DOUBLE PRECISION :: cosb



cosb=COS(b)

v(1)=COS(a)*cosb
v(2)=SIN(a)*cosb
v(3)=SIN(b)

END SUBROUTINE sla_dcs2c
SUBROUTINE sla_dd2tf (ndp, days, SIGN, ihmsf)
!+
!     - - - - - -
!      D D 2 T F
!     - - - - - -

!  Convert an interval in days into hours, minutes, seconds
!  (double precision)

!  Given:
!     NDP       int      number of decimal places of seconds
!     DAYS      dp       interval in days

!  Returned:
!     SIGN      char     '+' or '-'
!     IHMSF     int(4)   hours, minutes, seconds, fraction

!  Notes:

!     1)  NDP less than zero is interpreted as zero.

!     2)  The largest useful value for NDP is determined by the size
!         of DAYS, the format of DOUBLE PRECISION floating-point numbers
!         on the target machine, and the risk of overflowing IHMSF(4).
!         For example, on the VAX, for DAYS up to 1D0, the available
!         floating-point precision corresponds roughly to NDP=12.
!         However, the practical limit is NDP=9, set by the capacity of
!         the 32-bit integer IHMSF(4).

!     3)  The absolute value of DAYS may exceed 1D0.  In cases where it
!         does not, it is up to the caller to test for and handle the
!         case where DAYS is very nearly 1D0 and rounds up to 24 hours,
!         by testing for IHMSF(1)=24 and setting IHMSF(1-4) to zero.

!  P.T.Wallace   Starlink   11 September 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ndp
DOUBLE PRECISION :: days
CHARACTER :: SIGN*(*)
INTEGER :: ihmsf(4)

!  Days to seconds
DOUBLE PRECISION :: d2s
PARAMETER (d2s=86400D0)

INTEGER :: nrs,n
DOUBLE PRECISION :: rs,rm,rh,a,ah,am,as,af



!  Handle sign
IF (days >= 0D0) THEN
  SIGN='+'
ELSE
  SIGN='-'
END IF

!  Field units in terms of least significant figure
nrs=1
DO n=1,ndp
  nrs=nrs*10
END DO
rs=DBLE(nrs)
rm=rs*60D0
rh=rm*60D0

!  Round interval and express in smallest units required
a=ANINT(rs*d2s*ABS(days))

!  Separate into fields
ah=AINT(a/rh)
a=a-ah*rh
am=AINT(a/rm)
a=a-am*rm
as=AINT(a/rs)
af=a-as*rs

!  Return results
ihmsf(1)=MAX(nint(ah),0)
ihmsf(2)=MAX(MIN(nint(am),59),0)
ihmsf(3)=MAX(MIN(nint(as),59),0)
ihmsf(4)=MAX(nint(MIN(af,rs-1D0)),0)

END SUBROUTINE sla_dd2tf
SUBROUTINE sla_de2h (ha, dec, phi, az, el)
!+
!     - - - - -
!      D E 2 H
!     - - - - -

!  Equatorial to horizon coordinates:  HA,Dec to Az,El

!  (double precision)

!  Given:
!     HA      d     hour angle
!     DEC     d     declination
!     PHI     d     observatory latitude

!  Returned:
!     AZ      d     azimuth
!     EL      d     elevation

!  Notes:

!  1)  All the arguments are angles in radians.

!  2)  Azimuth is returned in the range 0-2pi;  north is zero,
!      and east is +pi/2.  Elevation is returned in the range
!      +/-pi/2.

!  3)  The latitude must be geodetic.  In critical applications,
!      corrections for polar motion should be applied.

!  4)  In some applications it will be important to specify the
!      correct type of hour angle and declination in order to
!      produce the required type of azimuth and elevation.  In
!      particular, it may be important to distinguish between
!      elevation as affected by refraction, which would
!      require the "observed" HA,Dec, and the elevation
!      in vacuo, which would require the "topocentric" HA,Dec.
!      If the effects of diurnal aberration can be neglected, the
!      "apparent" HA,Dec may be used instead of the topocentric
!      HA,Dec.

!  5)  No range checking of arguments is carried out.

!  6)  In applications which involve many such calculations, rather
!      than calling the present routine it will be more efficient to
!      use inline code, having previously computed fixed terms such
!      as sine and cosine of latitude, and (for tracking a star)
!      sine and cosine of declination.

!  P.T.Wallace   Starlink   9 July 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ha,dec,phi,az,el

DOUBLE PRECISION :: d2pi
PARAMETER (d2pi=6.283185307179586476925286766559D0)

DOUBLE PRECISION :: sh,ch,sd,cd,sp,cp,x,y,z,r,a


!  Useful trig functions
sh=SIN(ha)
ch=COS(ha)
sd=SIN(dec)
cd=COS(dec)
sp=SIN(phi)
cp=COS(phi)

!  Az,El as x,y,z
x=-ch*cd*sp+sd*cp
y=-sh*cd
z=ch*cd*cp+sd*sp

!  To spherical
r=SQRT(x*x+y*y)
IF (r == 0D0) THEN
  a=0D0
ELSE
  a=ATAN2(y,x)
END IF
IF (a < 0D0) a=a+d2pi
az=a
el=ATAN2(z,r)

END SUBROUTINE sla_de2h
SUBROUTINE sla_deuler (order, phi, theta, psi, rmat)
!+
!     - - - - - - -
!      D E U L E R
!     - - - - - - -

!  Form a rotation matrix from the Euler angles - three successive
!  rotations about specified Cartesian axes (double precision)

!  Given:
!    ORDER   c*(*)   specifies about which axes the rotations occur
!    PHI     d       1st rotation (radians)
!    THETA   d       2nd rotation (   "   )
!    PSI     d       3rd rotation (   "   )

!  Returned:
!    RMAT    d(3,3)  rotation matrix

!  A rotation is positive when the reference frame rotates
!  anticlockwise as seen looking towards the origin from the
!  positive region of the specified axis.

!  The characters of ORDER define which axes the three successive
!  rotations are about.  A typical value is 'ZXZ', indicating that
!  RMAT is to become the direction cosine matrix corresponding to
!  rotations of the reference frame through PHI radians about the
!  old Z-axis, followed by THETA radians about the resulting X-axis,
!  then PSI radians about the resulting Z-axis.

!  The axis names can be any of the following, in any order or
!  combination:  X, Y, Z, uppercase or lowercase, 1, 2, 3.  Normal
!  axis labelling/numbering conventions apply;  the xyz (=123)
!  triad is right-handed.  Thus, the 'ZXZ' example given above
!  could be written 'zxz' or '313' (or even 'ZxZ' or '3xZ').  ORDER
!  is terminated by length or by the first unrecognized character.

!  Fewer than three rotations are acceptable, in which case the later
!  angle arguments are ignored.  If all rotations are zero, the
!  identity matrix is produced.

!  P.T.Wallace   Starlink   23 May 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) order
DOUBLE PRECISION :: phi,theta,psi,rmat(3,3)

INTEGER :: j,i,l,n,k
DOUBLE PRECISION :: result(3,3),rotn(3,3),angle,s,c,w,wm(3,3)
CHARACTER :: axis



!  Initialize result matrix
DO j=1,3
  DO i=1,3
    IF (i /= j) THEN
      result(i,j) = 0D0
    ELSE
      result(i,j) = 1D0
    END IF
  END DO
END DO

!  Establish length of axis string
l = LEN(order)

!  Look at each character of axis string until finished
DO n=1,3
  IF (n <= l) THEN
    
!        Initialize rotation matrix for the current rotation
    DO j=1,3
      DO i=1,3
        IF (i /= j) THEN
          rotn(i,j) = 0D0
        ELSE
          rotn(i,j) = 1D0
        END IF
      END DO
    END DO
    
!        Pick up the appropriate Euler angle and take sine & cosine
    IF (n == 1) THEN
      angle = phi
    ELSE IF (n == 2) THEN
      angle = theta
    ELSE
      angle = psi
    END IF
    s = SIN(angle)
    c = COS(angle)
    
!        Identify the axis
    axis = order(n:n)
    IF (axis == 'X'.OR. axis == 'x'.OR.  &
      axis == '1') THEN
      
!           Matrix for x-rotation
      rotn(2,2) = c
      rotn(2,3) = s
      rotn(3,2) = -s
      rotn(3,3) = c
      
    ELSE IF (axis == 'Y'.OR.  &
      axis == 'y'.OR.axis == '2') THEN
      
!           Matrix for y-rotation
      rotn(1,1) = c
      rotn(1,3) = -s
      rotn(3,1) = s
      rotn(3,3) = c
      
    ELSE IF (axis == 'Z'.OR.  &
      axis == 'z'.OR.axis == '3') THEN
      
!           Matrix for z-rotation
      rotn(1,1) = c
      rotn(1,2) = s
      rotn(2,1) = -s
      rotn(2,2) = c
      
    ELSE
      
!           Unrecognized character - fake end of string
      l = 0
      
    END IF
    
!        Apply the current rotation (matrix ROTN x matrix RESULT)
    DO i=1,3
      DO j=1,3
        w = 0D0
        DO k=1,3
          w = w+rotn(i,k)*result(k,j)
        END DO
        wm(i,j) = w
      END DO
    END DO
    DO j=1,3
      DO i=1,3
        result(i,j) = wm(i,j)
      END DO
    END DO
    
  END IF
  
END DO

!  Copy the result
DO j=1,3
  DO i=1,3
    rmat(i,j) = result(i,j)
  END DO
END DO

END SUBROUTINE sla_deuler
SUBROUTINE sla_dfltin (string, nstrt, dreslt, jflag)
!+
!     - - - - - - -
!      D F L T I N
!     - - - - - - -

!  Convert free-format input into double precision floating point

!  Given:
!     STRING     c     string containing number to be decoded
!     NSTRT      i     pointer to where decoding is to start
!     DRESLT     d     current value of result

!  Returned:
!     NSTRT      i      advanced to next number
!     DRESLT     d      result
!     JFLAG      i      status: -1 = -OK, 0 = +OK, 1 = null, 2 = error

!  Called:  sla__IDCHF

!  Notes:

!     1     The reason DFLTIN has separate OK status values for +
!           and - is to enable minus zero to be detected.   This is
!           of crucial importance when decoding mixed-radix numbers.
!           For example, an angle expressed as deg, arcmin, arcsec
!           may have a leading minus sign but a zero degrees field.

!     2     A TAB is interpreted as a space, and lowercase characters
!           are interpreted as uppercase.

!     3     The basic format is the sequence of fields #^.^@#^, where
!           # is a sign character + or -, ^ means a string of decimal
!           digits, and @, which indicates an exponent, means D or E.
!           Various combinations of these fields can be omitted, and
!           embedded blanks are permissible in certain places.

!     4     Spaces:

!             .  Leading spaces are ignored.

!             .  Embedded spaces are allowed only after +, -, D or E,
!                and after the decomal point if the first sequence of
!                digits is absent.

!             .  Trailing spaces are ignored;  the first signifies
!                end of decoding and subsequent ones are skipped.

!     5     Delimiters:

!             .  Any character other than +,-,0-9,.,D,E or space may be
!                used to signal the end of the number and terminate
!                decoding.

!             .  Comma is recognized by DFLTIN as a special case;  it
!                is skipped, leaving the pointer on the next character.
!                See 13, below.

!     6     Both signs are optional.  The default is +.

!     7     The mantissa ^.^ defaults to 1.

!     8     The exponent @#^ defaults to D0.

!     9     The strings of decimal digits may be of any length.

!     10    The decimal point is optional for whole numbers.

!     11    A "null result" occurs when the string of characters being
!           decoded does not begin with +,-,0-9,.,D or E, or consists
!           entirely of spaces.  When this condition is detected, JFLAG
!           is set to 1 and DRESLT is left untouched.

!     12    NSTRT = 1 for the first character in the string.

!     13    On return from DFLTIN, NSTRT is set ready for the next
!           decode - following trailing blanks and any comma.  If a
!           delimiter other than comma is being used, NSTRT must be
!           incremented before the next call to DFLTIN, otherwise
!           all subsequent calls will return a null result.

!     14    Errors (JFLAG=2) occur when:

!             .  a +, -, D or E is left unsatisfied;  or

!             .  the decimal point is present without at least
!                one decimal digit before or after it;  or

!             .  an exponent more than 100 has been presented.

!     15    When an error has been detected, NSTRT is left
!           pointing to the character following the last
!           one used before the error came to light.  This
!           may be after the point at which a more sophisticated
!           program could have detected the error.  For example,
!           DFLTIN does not detect that '1D999' is unacceptable
!           (on a computer where this is so) until the entire number
!           has been decoded.

!     16    Certain highly unlikely combinations of mantissa &
!           exponent can cause arithmetic faults during the
!           decode, in some cases despite the fact that they
!           together could be construed as a valid number.

!     17    Decoding is left to right, one pass.

!     18    See also FLOTIN and INTIN

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: nstrt
DOUBLE PRECISION :: dreslt
INTEGER :: jflag

INTEGER :: nptr,msign,nexp,ndp,nvec,ndigit,isignx,j
DOUBLE PRECISION :: dmant,digit



!  Current character
nptr=nstrt

!  Set defaults: mantissa & sign, exponent & sign, decimal place count
dmant=0D0
msign=1
nexp=0
isignx=1
ndp=0

!  Look for sign
100  CONTINUE
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO ( 400,  100,  800,  500,  300,  200, 9110, 9100, 9110),nvec
!             0-9    SP   D/E    .     +     -     ,   ELSE   END

!  Negative
200  CONTINUE
msign=-1

!  Look for first leading decimal
300  CONTINUE
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO ( 400, 300,  800,  500, 9200, 9200, 9200, 9200, 9210),nvec
!             0-9   SP   D/E    .     +     -     ,   ELSE   END

!  Accept leading decimals
400  CONTINUE
dmant=dmant*1D1+digit
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO ( 400, 1310,  900,  600, 1300, 1300, 1300, 1300, 1310),nvec
!             0-9   SP    D/E    .     +     -     ,   ELSE   END

!  Look for decimal when none preceded the point
500  CONTINUE
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO ( 700, 500, 9200, 9200, 9200, 9200, 9200, 9200, 9210),nvec
!             0-9   SP   D/E    .     +     -     ,   ELSE   END

!  Look for trailing decimals
600  CONTINUE
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO ( 700, 1310,  900, 1300, 1300, 1300, 1300, 1300, 1310),nvec
!             0-9   SP    D/E    .     +     -     ,   ELSE   END

!  Accept trailing decimals
700  CONTINUE
ndp=ndp+1
dmant=dmant*1D1+digit
GO TO 600

!  Exponent symbol first in field: default mantissa to 1
800  CONTINUE
dmant=1D0

!  Look for sign of exponent
900  CONTINUE
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO (1200, 900, 9200, 9200, 1100, 1000, 9200, 9200, 9210),nvec
!             0-9   SP   D/E    .     +     -     ,   ELSE   END

!  Exponent negative
1000 CONTINUE
isignx=-1

!  Look for first digit of exponent
1100 CONTINUE
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO (1200, 1100, 9200, 9200, 9200, 9200, 9200, 9200, 9210),nvec
!             0-9   SP    D/E    .     +     -     ,   ELSE   END

!  Use exponent digit
1200 CONTINUE
nexp=nexp*10+ndigit
IF (nexp > 100) GO TO 9200

!  Look for subsequent digits of exponent
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO (1200, 1310, 1300, 1300, 1300, 1300, 1300, 1300, 1310),nvec
!             0-9   SP    D/E    .     +     -     ,   ELSE   END

!  Combine exponent and decimal place count
1300 CONTINUE
nptr=nptr-1
1310 CONTINUE
nexp=nexp*isignx-ndp

!  Skip if net exponent negative
IF (nexp < 0) GO TO 1500

!  Positive exponent: scale up
1400 CONTINUE
IF (nexp < 10) GO TO 1410
dmant=dmant*1D10
nexp=nexp-10
GO TO 1400
1410 CONTINUE
IF (nexp < 1) GO TO 1600
dmant=dmant*1D1
nexp=nexp-1
GO TO 1410

!  Negative exponent: scale down
1500 CONTINUE
IF (nexp > -10) GO TO 1510
dmant=dmant/1D10
nexp=nexp+10
GO TO 1500
1510 CONTINUE
IF (nexp > -1) GO TO 1600
dmant=dmant/1D1
nexp=nexp+1
GO TO 1510

!  Get result & status
1600 CONTINUE
j=0
IF (msign == 1) GO TO 1610
j=-1
dmant=-dmant
1610 CONTINUE
dreslt=dmant

!  Skip to end of field
1620 CONTINUE
CALL sla__idchf(string,nptr,nvec,ndigit,digit)
GO TO (1720, 1620, 1720, 1720, 1720, 1720, 9900, 1720, 9900),nvec
!             0-9   SP    D/E    .     +     -     ,   ELSE   END

1720 CONTINUE
nptr=nptr-1
GO TO 9900


!  Exits

!  Null field
9100 CONTINUE
nptr=nptr-1
9110 CONTINUE
j=1
GO TO 9900

!  Errors
9200 CONTINUE
nptr=nptr-1
9210 CONTINUE
j=2

!  Return
9900 CONTINUE
nstrt=nptr
jflag=j

END SUBROUTINE sla_dfltin
SUBROUTINE sla_dh2e (az, el, phi, ha, dec)
!+
!     - - - - -
!      D E 2 H
!     - - - - -

!  Horizon to equatorial coordinates:  Az,El to HA,Dec

!  (double precision)

!  Given:
!     AZ      d     azimuth
!     EL      d     elevation
!     PHI     d     observatory latitude

!  Returned:
!     HA      d     hour angle
!     DEC     d     declination

!  Notes:

!  1)  All the arguments are angles in radians.

!  2)  The sign convention for azimuth is north zero, east +pi/2.

!  3)  HA is returned in the range +/-pi.  Declination is returned
!      in the range +/-pi/2.

!  4)  The latitude is (in principle) geodetic.  In critical
!      applications, corrections for polar motion should be applied.

!  5)  In some applications it will be important to specify the
!      correct type of elevation in order to produce the required
!      type of HA,Dec.  In particular, it may be important to
!      distinguish between the elevation as affected by refraction,
!      which will yield the "observed" HA,Dec, and the elevation
!      in vacuo, which will yield the "topocentric" HA,Dec.  If the
!      effects of diurnal aberration can be neglected, the
!      topocentric HA,Dec may be used as an approximation to the
!      "apparent" HA,Dec.

!  6)  No range checking of arguments is done.

!  7)  In applications which involve many such calculations, rather
!      than calling the present routine it will be more efficient to
!      use inline code, having previously computed fixed terms such
!      as sine and cosine of latitude.

!  P.T.Wallace   Starlink   21 February 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: az,el,phi,ha,dec

DOUBLE PRECISION :: sa,ca,se,ce,sp,cp,x,y,z,r


!  Useful trig functions
sa=SIN(az)
ca=COS(az)
se=SIN(el)
ce=COS(el)
sp=SIN(phi)
cp=COS(phi)

!  HA,Dec as x,y,z
x=-ca*ce*sp+se*cp
y=-sa*ce
z=ca*ce*cp+se*sp

!  To HA,Dec
r=SQRT(x*x+y*y)
IF (r == 0D0) THEN
  ha=0D0
ELSE
  ha=ATAN2(y,x)
END IF
dec=ATAN2(z,r)

END SUBROUTINE sla_dh2e
SUBROUTINE sla_dimxv (dm, va, vb)
!+
!     - - - - - -
!      D I M X V
!     - - - - - -

!  Performs the 3-D backward unitary transformation:

!     vector VB = (inverse of matrix DM) * vector VA

!  (double precision)

!  (n.b.  the matrix must be unitary, as this routine assumes that
!   the inverse and transpose are identical)

!  Given:
!     DM       dp(3,3)    matrix
!     VA       dp(3)      vector

!  Returned:
!     VB       dp(3)      result vector

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dm(3,3),va(3),vb(3)

INTEGER :: i,j
DOUBLE PRECISION :: w,vw(3)



!  Inverse of matrix DM * vector VA -> vector VW
DO j=1,3
  w=0D0
  DO i=1,3
    w=w+dm(i,j)*va(i)
  END DO
  vw(j)=w
END DO

!  Vector VW -> vector VB
DO j=1,3
  vb(j)=vw(j)
END DO

END SUBROUTINE sla_dimxv
SUBROUTINE sla_djcal (ndp, djm, iymdf, j)
!+
!     - - - - - -
!      D J C A L
!     - - - - - -

!  Modified Julian Date to Gregorian Calendar, expressed
!  in a form convenient for formatting messages (namely
!  rounded to a specified precision, and with the fields
!  stored in a single array)

!  Given:
!     NDP      i      number of decimal places of days in fraction
!     DJM      d      modified Julian Date (JD-2400000.5)

!  Returned:
!     IYMDF    i(4)   year, month, day, fraction in Gregorian
!                     calendar
!     J        i      status:  nonzero = out of range

!  Any date after 4701BC March 1 is accepted.

!  NDP should be 4 or less if internal overflows are to be avoided
!  on machines which use 32-bit integers.

!  The algorithm is derived from that of Hatcher 1984
!  (QJRAS 25, 53-55).

!  P.T.Wallace   Starlink   6 October 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ndp
DOUBLE PRECISION :: djm
INTEGER :: iymdf(4),j

INTEGER :: nfd
DOUBLE PRECISION :: fd,df,f,d
INTEGER :: jd,n4,nd10



!  Validate
IF (djm <= -2395520D0.OR.djm >= 1D9) THEN
  j=-1
ELSE
  
!     Denominator of fraction
  nfd=10**MAX(ndp,0)
  fd=DBLE(nfd)
  
!     Round date and express in units of fraction
  df=nint(djm*fd)
  
!     Separate day and fraction
  f=MOD(df,fd)
  IF (f < 0D0) f=f+fd
  d=(df-f)/fd
  
!     Express day in Gregorian calendar
  jd=nint(d)+2400001
  
  n4=4*(jd+((2*((4*jd-17918)/146097)*3)/4+1)/2-37)
  nd10=10*(MOD(n4-237,1461)/4)+5
  
  iymdf(1)=n4/1461-4712
  iymdf(2)=MOD(nd10/306+2,12)+1
  iymdf(3)=MOD(nd10,306)/10+1
  iymdf(4)=nint(f)
  
  j=0
  
END IF

END SUBROUTINE sla_djcal
SUBROUTINE sla_djcl (djm, iy, im, id, fd, j)
!+
!     - - - - -
!      D J C L
!     - - - - -

!  Modified Julian Date to Gregorian year, month, day,
!  and fraction of a day.

!  Given:
!     DJM      dp     modified Julian Date (JD-2400000.5)

!  Returned:
!     IY       int    year
!     IM       int    month
!     ID       int    day
!     FD       dp     fraction of day
!     J        int    status:
!                       0 = OK
!                      -1 = unacceptable date (before 4701BC March 1)

!  The algorithm is derived from that of Hatcher 1984
!  (QJRAS 25, 53-55).

!  P.T.Wallace   Starlink   23 November 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: djm
INTEGER :: iy,im,id
DOUBLE PRECISION :: fd
INTEGER :: j

DOUBLE PRECISION :: f,d
INTEGER :: jd,n4,nd10



!  Check if date is acceptable
IF (djm <= -2395522D0.OR.djm >= 1D9) THEN
  j=-1
ELSE
  j=0
  
!     Separate day and fraction
  f=MOD(djm,1D0)
  IF (f < 0D0) f=f+1D0
  d=nint(djm-f)
  
!     Express day in Gregorian calendar
  jd=nint(d)+2400001
  
  n4=4*(jd+((6*((4*jd-17918)/146097))/4+1)/2-37)
  nd10=10*(MOD(n4-237,1461)/4)+5
  
  iy=n4/1461-4712
  im=MOD(nd10/306+2,12)+1
  id=MOD(nd10,306)/10+1
  fd=f
  
  j=0
  
END IF

END SUBROUTINE sla_djcl
SUBROUTINE sla_dm2av (rmat, axvec)
!+
!     - - - - - -
!      D M 2 A V
!     - - - - - -

!  From a rotation matrix, determine the corresponding axial vector.
!  (double precision)

!  A rotation matrix describes a rotation about some arbitrary axis.
!  The axis is called the Euler axis, and the angle through which the
!  reference frame rotates is called the Euler angle.  The axial
!  vector returned by this routine has the same direction as the
!  Euler axis, and its magnitude is the Euler angle in radians.  (The
!  magnitude and direction can be separated by means of the routine
!  sla_DVN.)

!  Given:
!    RMAT   d(3,3)   rotation matrix

!  Returned:
!    AXVEC  d(3)     axial vector (radians)

!  The reference frame rotates clockwise as seen looking along
!  the axial vector from the origin.

!  If RMAT is null, so is the result.

!  P.T.Wallace   Starlink   24 December 1992

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rmat(3,3),axvec(3)

DOUBLE PRECISION :: x,y,z,s2,c2,phi,f



x = rmat(2,3)-rmat(3,2)
y = rmat(3,1)-rmat(1,3)
z = rmat(1,2)-rmat(2,1)
s2 = SQRT(x*x+y*y+z*z)
IF (s2 /= 0D0) THEN
  c2 = (rmat(1,1)+rmat(2,2)+rmat(3,3)-1D0)
  phi = ATAN2(s2/2D0,c2/2D0)
  f = phi/s2
  axvec(1) = x*f
  axvec(2) = y*f
  axvec(3) = z*f
ELSE
  axvec(1) = 0D0
  axvec(2) = 0D0
  axvec(3) = 0D0
END IF

END SUBROUTINE sla_dm2av
SUBROUTINE sla_dmat (n, a, y, d, jf, iw)
!+
!     - - - - -
!      D M A T
!     - - - - -

!  Matrix inversion & solution of simultaneous equations
!  (double precision)

!  For the set of n simultaneous equations in n unknowns:
!     A.Y = X

!  where:
!     A is a non-singular N x N matrix
!     Y is the vector of N unknowns
!     X is the known vector

!  DMATRX computes:
!     the inverse of matrix A
!     the determinant of matrix A
!     the vector of N unknowns

!  Arguments:

!     symbol  type   dimension           before              after

!       N      i                    no. of unknowns       unchanged
!       A      d      (N,N)             matrix             inverse
!       Y      d       (N)              vector            solution
!       D      d                           -             determinant
!     * JF     i                           -           singularity flag
!       IW     i       (N)                 -              workspace

!  *  JF is the singularity flag.  If the matrix is non-singular,
!    JF=0 is returned.  If the matrix is singular, JF=-1 & D=0D0 are
!    returned.  In the latter case, the contents of array A on return
!    are undefined.

!  Algorithm:
!     Gaussian elimination with partial pivoting.

!  Speed:
!     Very fast.

!  Accuracy:
!     Fairly accurate - errors 1 to 4 times those of routines optimized
!     for accuracy.

!  P.T.Wallace   Starlink   7 February 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: n
DOUBLE PRECISION :: a(n,n),y(n),d
INTEGER :: jf
INTEGER :: iw(n)

DOUBLE PRECISION :: sfa
PARAMETER (sfa=1D-20)

INTEGER :: k,imx,i,j,np1mk,ki
DOUBLE PRECISION :: amx,t,akk,yk,aik


jf=0
d=1D0
DO k=1,n
  amx=DABS(a(k,k))
  imx=k
  IF (k /= n) THEN
    DO i=k+1,n
      t=DABS(a(i,k))
      IF (t > amx) THEN
        amx=t
        imx=i
      END IF
    END DO
  END IF
  IF (amx < sfa) THEN
    jf=-1
  ELSE
    IF (imx /= k) THEN
      DO j=1,n
        t=a(k,j)
        a(k,j)=a(imx,j)
        a(imx,j)=t
      END DO
      t=y(k)
      y(k)=y(imx)
      y(imx)=t
      d=-d
    END IF
    iw(k)=imx
    akk=a(k,k)
    d=d*akk
    IF (DABS(d) < sfa) THEN
      jf=-1
    ELSE
      akk=1D0/akk
      a(k,k)=akk
      DO j=1,n
        IF (j /= k) a(k,j)=a(k,j)*akk
      END DO
      yk=y(k)*akk
      y(k)=yk
      DO i=1,n
        aik=a(i,k)
        IF (i /= k) THEN
          DO j=1,n
            IF (j /= k) a(i,j)=a(i,j)-aik*a(k,j)
          END DO
          y(i)=y(i)-aik*yk
        END IF
      END DO
      DO i=1,n
        IF (i /= k) a(i,k)=-a(i,k)*akk
      END DO
    END IF
  END IF
END DO
IF (jf /= 0) THEN
  d=0D0
ELSE
  DO k=1,n
    np1mk=n+1-k
    ki=iw(np1mk)
    IF (np1mk /= ki) THEN
      DO i=1,n
        t=a(i,np1mk)
        a(i,np1mk)=a(i,ki)
        a(i,ki)=t
      END DO
    END IF
  END DO
END IF

END SUBROUTINE sla_dmat
SUBROUTINE sla_dmoon (date, pv)
!+
!     - - - - - -
!      D M O O N
!     - - - - - -

!  Approximate geocentric position and velocity of the Moon
!  (double precision)

!  Given:
!     DATE       D       TDB (loosely ET) as a Modified Julian Date
!                                                    (JD-2400000.5)

!  Returned:
!     PV         D(6)    Moon x,y,z,xdot,ydot,zdot, mean equator and
!                                         equinox of date (AU, AU/s)

!  Notes:

!  1  This routine is a full implementation of the algorithm
!     published by Meeus (see reference).

!  2  Meeus quotes accuracies of 10 arcsec in longitude, 3 arcsec in
!     latitude and 0.2 arcsec in HP (equivalent to about 20 km in
!     distance).  Comparison with JPL DE200 over the interval
!     1960-2025 gives RMS errors of 3.7 arcsec and 83 mas/hour in
!     longitude, 2.3 arcsec and 48 mas/hour in latitude, 11 km
!     and 81 mm/s in distance.  The maximum errors over the same
!     interval are 18 arcsec and 0.50 arcsec/hour in longitude,
!     11 arcsec and 0.24 arcsec/hour in latitude, 40 km and 0.29 m/s
!     in distance.

!  3  The original algorithm is expressed in terms of the obsolete
!     timescale Ephemeris Time.  Either TDB or TT can be used, but
!     not UT without incurring significant errors (30 arcsec at
!     the present time) due to the Moon's 0.5 arcsec/sec movement.

!  4  The algorithm is based on pre IAU 1976 standards.  However,
!     the result has been moved onto the new (FK5) equinox, an
!     adjustment which is in any case much smaller than the
!     intrinsic accuracy of the procedure.

!  5  Velocity is obtained by a complete analytical differentiation
!     of the Meeus model.

!  Reference:
!     Meeus, l'Astronomie, June 1984, p348.

!  P.T.Wallace   Starlink   22 January 1998

!  Copyright (C) 1998 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,pv(6)

!  Degrees, arcseconds and seconds of time to radians
DOUBLE PRECISION :: d2r,das2r,ds2r
PARAMETER (d2r=0.0174532925199432957692369D0,  &
das2r=4.848136811095359935899141D-6,ds2r=7.272205216643039903848712D-5)

!  Seconds per Julian century (86400*36525)
DOUBLE PRECISION :: cj
PARAMETER (cj=3155760000D0)

!  Julian epoch of B1950
DOUBLE PRECISION :: b1950
PARAMETER (b1950=1949.9997904423D0)

!  Earth equatorial radius in AU ( = 6378.137 / 149597870 )
DOUBLE PRECISION :: eradau
PARAMETER (eradau=4.2635212653763D-5)

DOUBLE PRECISION :: t,theta,sinom,cosom,domcom,wa,dwa,wb,dwb,wom,  &
dwom,sinwom,coswom,v,dv,coeff,emn,empn,dn,fn,en,  &
den,dtheta,ftheta,el,del,b,db,bf,dbf,p,dp,sp,r,  &
dr,x,y,z,xd,yd,zd,sel,cel,sb,cb,rcb,rbd,w,epj,eqcor,eps,sineps,coseps,es,ec
INTEGER :: n,i


!  Coefficients for fundamental arguments

!   at J1900:  T**0, T**1, T**2, T**3
!   at epoch:  T**0, T**1

!  Units are degrees for position and Julian centuries for time


!  Moon's mean longitude
DOUBLE PRECISION :: elp0,elp1,elp2,elp3,elp,delp
PARAMETER (elp0=270.434164D0,elp1=481267.8831D0,  &
elp2=-0.001133D0,elp3=0.0000019D0)

!  Sun's mean anomaly
DOUBLE PRECISION :: em0,em1,em2,em3,em,dem
PARAMETER (em0=358.475833D0,em1=35999.0498D0,  &
em2=-0.000150D0,em3=-0.0000033D0)

!  Moon's mean anomaly
DOUBLE PRECISION :: emp0,emp1,emp2,emp3,emp,demp
PARAMETER (emp0=296.104608D0,emp1=477198.8491D0,  &
emp2=0.009192D0,emp3=0.0000144D0)

!  Moon's mean elongation
DOUBLE PRECISION :: d0,d1,d2,d3,d,dd
PARAMETER (d0=350.737486D0,d1=445267.1142D0,  &
d2=-0.001436D0,d3=0.0000019D0)

!  Mean distance of the Moon from its ascending node
DOUBLE PRECISION :: f0,f1,f2,f3,f,df
PARAMETER (f0=11.250889D0,f1=483202.0251D0,  &
f2=-0.003211D0,f3=-0.0000003D0)

!  Longitude of the Moon's ascending node
DOUBLE PRECISION :: om0,om1,om2,om3,om,dom
PARAMETER (om0=259.183275D0,om1=-1934.1420D0,  &
om2=0.002078D0,om3=0.0000022D0)

!  Coefficients for (dimensionless) E factor
DOUBLE PRECISION :: e1,e2,e,de,esq,desq
PARAMETER (e1=-0.002495D0,e2=-0.00000752D0)

!  Coefficients for periodic variations etc
DOUBLE PRECISION :: pac,pa0,pa1
PARAMETER (pac=0.000233D0,pa0=51.2D0,pa1=20.2D0)
DOUBLE PRECISION :: pbc
PARAMETER (pbc=-0.001778D0)
DOUBLE PRECISION :: pcc
PARAMETER (pcc=0.000817D0)
DOUBLE PRECISION :: pdc
PARAMETER (pdc=0.002011D0)
DOUBLE PRECISION :: pec,pe0,pe1,pe2
PARAMETER (pec=0.003964D0,pe0=346.560D0,pe1=132.870D0,pe2=-0.0091731D0)
DOUBLE PRECISION :: pfc
PARAMETER (pfc=0.001964D0)
DOUBLE PRECISION :: pgc
PARAMETER (pgc=0.002541D0)
DOUBLE PRECISION :: phc
PARAMETER (phc=0.001964D0)
DOUBLE PRECISION :: pic
PARAMETER (pic=-0.024691D0)
DOUBLE PRECISION :: pjc,pj0,pj1
PARAMETER (pjc=-0.004328D0,pj0=275.05D0,pj1=-2.30D0)
DOUBLE PRECISION :: cw1
PARAMETER (cw1=0.0004664D0)
DOUBLE PRECISION :: cw2
PARAMETER (cw2=0.0000754D0)


!  Coefficients for Moon position

!   Tx(N)       = coefficient of L, B or P term (deg)
!   ITx(N,1-5)  = coefficients of M, M', D, F, E**n in argument

INTEGER :: nl,nb,np
PARAMETER (nl=50,nb=45,np=31)
DOUBLE PRECISION :: tl(nl),tb(nb),tp(np)
INTEGER :: itl(5,nl),itb(5,nb),itp(5,np)

!  Longitude
!                                         M   M'  D   F   n
DATA tl( 1)/            +6.288750D0                     /,  &
(itl(i, 1),i=1,5)/            +0, +1, +0, +0,  0   /
DATA tl( 2)/            +1.274018D0                     /,  &
(itl(i, 2),i=1,5)/            +0, -1, +2, +0,  0   /
DATA tl( 3)/            +0.658309D0                     /,  &
(itl(i, 3),i=1,5)/            +0, +0, +2, +0,  0   /
DATA tl( 4)/            +0.213616D0                     /,  &
(itl(i, 4),i=1,5)/            +0, +2, +0, +0,  0   /
DATA tl( 5)/            -0.185596D0                     /,  &
(itl(i, 5),i=1,5)/            +1, +0, +0, +0,  1   /
DATA tl( 6)/            -0.114336D0                     /,  &
(itl(i, 6),i=1,5)/            +0, +0, +0, +2,  0   /
DATA tl( 7)/            +0.058793D0                     /,  &
(itl(i, 7),i=1,5)/            +0, -2, +2, +0,  0   /
DATA tl( 8)/            +0.057212D0                     /,  &
(itl(i, 8),i=1,5)/            -1, -1, +2, +0,  1   /
DATA tl( 9)/            +0.053320D0                     /,  &
(itl(i, 9),i=1,5)/            +0, +1, +2, +0,  0   /
DATA tl(10)/            +0.045874D0                     /,  &
(itl(i,10),i=1,5)/            -1, +0, +2, +0,  1   /
DATA tl(11)/            +0.041024D0                     /,  &
(itl(i,11),i=1,5)/            -1, +1, +0, +0,  1   /
DATA tl(12)/            -0.034718D0                     /,  &
(itl(i,12),i=1,5)/            +0, +0, +1, +0,  0   /
DATA tl(13)/            -0.030465D0                     /,  &
(itl(i,13),i=1,5)/            +1, +1, +0, +0,  1   /
DATA tl(14)/            +0.015326D0                     /,  &
(itl(i,14),i=1,5)/            +0, +0, +2, -2,  0   /
DATA tl(15)/            -0.012528D0                     /,  &
(itl(i,15),i=1,5)/            +0, +1, +0, +2,  0   /
DATA tl(16)/            -0.010980D0                     /,  &
(itl(i,16),i=1,5)/            +0, -1, +0, +2,  0   /
DATA tl(17)/            +0.010674D0                     /,  &
(itl(i,17),i=1,5)/            +0, -1, +4, +0,  0   /
DATA tl(18)/            +0.010034D0                     /,  &
(itl(i,18),i=1,5)/            +0, +3, +0, +0,  0   /
DATA tl(19)/            +0.008548D0                     /,  &
(itl(i,19),i=1,5)/            +0, -2, +4, +0,  0   /
DATA tl(20)/            -0.007910D0                     /,  &
(itl(i,20),i=1,5)/            +1, -1, +2, +0,  1   /
DATA tl(21)/            -0.006783D0                     /,  &
(itl(i,21),i=1,5)/            +1, +0, +2, +0,  1   /
DATA tl(22)/            +0.005162D0                     /,  &
(itl(i,22),i=1,5)/            +0, +1, -1, +0,  0   /
DATA tl(23)/            +0.005000D0                     /,  &
(itl(i,23),i=1,5)/            +1, +0, +1, +0,  1   /
DATA tl(24)/            +0.004049D0                     /,  &
(itl(i,24),i=1,5)/            -1, +1, +2, +0,  1   /
DATA tl(25)/            +0.003996D0                     /,  &
(itl(i,25),i=1,5)/            +0, +2, +2, +0,  0   /
DATA tl(26)/            +0.003862D0                     /,  &
(itl(i,26),i=1,5)/            +0, +0, +4, +0,  0   /
DATA tl(27)/            +0.003665D0                     /,  &
(itl(i,27),i=1,5)/            +0, -3, +2, +0,  0   /
DATA tl(28)/            +0.002695D0                     /,  &
(itl(i,28),i=1,5)/            -1, +2, +0, +0,  1   /
DATA tl(29)/            +0.002602D0                     /,  &
(itl(i,29),i=1,5)/            +0, +1, -2, -2,  0   /
DATA tl(30)/            +0.002396D0                     /,  &
(itl(i,30),i=1,5)/            -1, -2, +2, +0,  1   /
DATA tl(31)/            -0.002349D0                     /,  &
(itl(i,31),i=1,5)/            +0, +1, +1, +0,  0   /
DATA tl(32)/            +0.002249D0                     /,  &
(itl(i,32),i=1,5)/            -2, +0, +2, +0,  2   /
DATA tl(33)/            -0.002125D0                     /,  &
(itl(i,33),i=1,5)/            +1, +2, +0, +0,  1   /
DATA tl(34)/            -0.002079D0                     /,  &
(itl(i,34),i=1,5)/            +2, +0, +0, +0,  2   /
DATA tl(35)/            +0.002059D0                     /,  &
(itl(i,35),i=1,5)/            -2, -1, +2, +0,  2   /
DATA tl(36)/            -0.001773D0                     /,  &
(itl(i,36),i=1,5)/            +0, +1, +2, -2,  0   /
DATA tl(37)/            -0.001595D0                     /,  &
(itl(i,37),i=1,5)/            +0, +0, +2, +2,  0   /
DATA tl(38)/            +0.001220D0                     /,  &
(itl(i,38),i=1,5)/            -1, -1, +4, +0,  1   /
DATA tl(39)/            -0.001110D0                     /,  &
(itl(i,39),i=1,5)/            +0, +2, +0, +2,  0   /
DATA tl(40)/            +0.000892D0                     /,  &
(itl(i,40),i=1,5)/            +0, +1, -3, +0,  0   /
DATA tl(41)/            -0.000811D0                     /,  &
(itl(i,41),i=1,5)/            +1, +1, +2, +0,  1   /
DATA tl(42)/            +0.000761D0                     /,  &
(itl(i,42),i=1,5)/            -1, -2, +4, +0,  1   /
DATA tl(43)/            +0.000717D0                     /,  &
(itl(i,43),i=1,5)/            -2, +1, +0, +0,  2   /
DATA tl(44)/            +0.000704D0                     /,  &
(itl(i,44),i=1,5)/            -2, +1, -2, +0,  2   /
DATA tl(45)/            +0.000693D0                     /,  &
(itl(i,45),i=1,5)/            +1, -2, +2, +0,  1   /
DATA tl(46)/            +0.000598D0                     /,  &
(itl(i,46),i=1,5)/            -1, +0, +2, -2,  1   /
DATA tl(47)/            +0.000550D0                     /,  &
(itl(i,47),i=1,5)/            +0, +1, +4, +0,  0   /
DATA tl(48)/            +0.000538D0                     /,  &
(itl(i,48),i=1,5)/            +0, +4, +0, +0,  0   /
DATA tl(49)/            +0.000521D0                     /,  &
(itl(i,49),i=1,5)/            -1, +0, +4, +0,  1   /
DATA tl(50)/            +0.000486D0                     /,  &
(itl(i,50),i=1,5)/            +0, +2, -1, +0,  0   /

!  Latitude
!                                         M   M'  D   F   n
DATA tb( 1)/            +5.128189D0                     /,  &
(itb(i, 1),i=1,5)/            +0, +0, +0, +1,  0   /
DATA tb( 2)/            +0.280606D0                     /,  &
(itb(i, 2),i=1,5)/            +0, +1, +0, +1,  0   /
DATA tb( 3)/            +0.277693D0                     /,  &
(itb(i, 3),i=1,5)/            +0, +1, +0, -1,  0   /
DATA tb( 4)/            +0.173238D0                     /,  &
(itb(i, 4),i=1,5)/            +0, +0, +2, -1,  0   /
DATA tb( 5)/            +0.055413D0                     /,  &
(itb(i, 5),i=1,5)/            +0, -1, +2, +1,  0   /
DATA tb( 6)/            +0.046272D0                     /,  &
(itb(i, 6),i=1,5)/            +0, -1, +2, -1,  0   /
DATA tb( 7)/            +0.032573D0                     /,  &
(itb(i, 7),i=1,5)/            +0, +0, +2, +1,  0   /
DATA tb( 8)/            +0.017198D0                     /,  &
(itb(i, 8),i=1,5)/            +0, +2, +0, +1,  0   /
DATA tb( 9)/            +0.009267D0                     /,  &
(itb(i, 9),i=1,5)/            +0, +1, +2, -1,  0   /
DATA tb(10)/            +0.008823D0                     /,  &
(itb(i,10),i=1,5)/            +0, +2, +0, -1,  0   /
DATA tb(11)/            +0.008247D0                     /,  &
(itb(i,11),i=1,5)/            -1, +0, +2, -1,  1   /
DATA tb(12)/            +0.004323D0                     /,  &
(itb(i,12),i=1,5)/            +0, -2, +2, -1,  0   /
DATA tb(13)/            +0.004200D0                     /,  &
(itb(i,13),i=1,5)/            +0, +1, +2, +1,  0   /
DATA tb(14)/            +0.003372D0                     /,  &
(itb(i,14),i=1,5)/            -1, +0, -2, +1,  1   /
DATA tb(15)/            +0.002472D0                     /,  &
(itb(i,15),i=1,5)/            -1, -1, +2, +1,  1   /
DATA tb(16)/            +0.002222D0                     /,  &
(itb(i,16),i=1,5)/            -1, +0, +2, +1,  1   /
DATA tb(17)/            +0.002072D0                     /,  &
(itb(i,17),i=1,5)/            -1, -1, +2, -1,  1   /
DATA tb(18)/            +0.001877D0                     /,  &
(itb(i,18),i=1,5)/            -1, +1, +0, +1,  1   /
DATA tb(19)/            +0.001828D0                     /,  &
(itb(i,19),i=1,5)/            +0, -1, +4, -1,  0   /
DATA tb(20)/            -0.001803D0                     /,  &
(itb(i,20),i=1,5)/            +1, +0, +0, +1,  1   /
DATA tb(21)/            -0.001750D0                     /,  &
(itb(i,21),i=1,5)/            +0, +0, +0, +3,  0   /
DATA tb(22)/            +0.001570D0                     /,  &
(itb(i,22),i=1,5)/            -1, +1, +0, -1,  1   /
DATA tb(23)/            -0.001487D0                     /,  &
(itb(i,23),i=1,5)/            +0, +0, +1, +1,  0   /
DATA tb(24)/            -0.001481D0                     /,  &
(itb(i,24),i=1,5)/            +1, +1, +0, +1,  1   /
DATA tb(25)/            +0.001417D0                     /,  &
(itb(i,25),i=1,5)/            -1, -1, +0, +1,  1   /
DATA tb(26)/            +0.001350D0                     /,  &
(itb(i,26),i=1,5)/            -1, +0, +0, +1,  1   /
DATA tb(27)/            +0.001330D0                     /,  &
(itb(i,27),i=1,5)/            +0, +0, -1, +1,  0   /
DATA tb(28)/            +0.001106D0                     /,  &
(itb(i,28),i=1,5)/            +0, +3, +0, +1,  0   /
DATA tb(29)/            +0.001020D0                     /,  &
(itb(i,29),i=1,5)/            +0, +0, +4, -1,  0   /
DATA tb(30)/            +0.000833D0                     /,  &
(itb(i,30),i=1,5)/            +0, -1, +4, +1,  0   /
DATA tb(31)/            +0.000781D0                     /,  &
(itb(i,31),i=1,5)/            +0, +1, +0, -3,  0   /
DATA tb(32)/            +0.000670D0                     /,  &
(itb(i,32),i=1,5)/            +0, -2, +4, +1,  0   /
DATA tb(33)/            +0.000606D0                     /,  &
(itb(i,33),i=1,5)/            +0, +0, +2, -3,  0   /
DATA tb(34)/            +0.000597D0                     /,  &
(itb(i,34),i=1,5)/            +0, +2, +2, -1,  0   /
DATA tb(35)/            +0.000492D0                     /,  &
(itb(i,35),i=1,5)/            -1, +1, +2, -1,  1   /
DATA tb(36)/            +0.000450D0                     /,  &
(itb(i,36),i=1,5)/            +0, +2, -2, -1,  0   /
DATA tb(37)/            +0.000439D0                     /,  &
(itb(i,37),i=1,5)/            +0, +3, +0, -1,  0   /
DATA tb(38)/            +0.000423D0                     /,  &
(itb(i,38),i=1,5)/            +0, +2, +2, +1,  0   /
DATA tb(39)/            +0.000422D0                     /,  &
(itb(i,39),i=1,5)/            +0, -3, +2, -1,  0   /
DATA tb(40)/            -0.000367D0                     /,  &
(itb(i,40),i=1,5)/            +1, -1, +2, +1,  1   /
DATA tb(41)/            -0.000353D0                     /,  &
(itb(i,41),i=1,5)/            +1, +0, +2, +1,  1   /
DATA tb(42)/            +0.000331D0                     /,  &
(itb(i,42),i=1,5)/            +0, +0, +4, +1,  0   /
DATA tb(43)/            +0.000317D0                     /,  &
(itb(i,43),i=1,5)/            -1, +1, +2, +1,  1   /
DATA tb(44)/            +0.000306D0                     /,  &
(itb(i,44),i=1,5)/            -2, +0, +2, -1,  2   /
DATA tb(45)/            -0.000283D0                     /,  &
(itb(i,45),i=1,5)/            +0, +1, +0, +3,  0   /

!  Parallax
!                                         M   M'  D   F   n
DATA tp( 1)/            +0.950724D0                     /,  &
(itp(i, 1),i=1,5)/            +0, +0, +0, +0,  0   /
DATA tp( 2)/            +0.051818D0                     /,  &
(itp(i, 2),i=1,5)/            +0, +1, +0, +0,  0   /
DATA tp( 3)/            +0.009531D0                     /,  &
(itp(i, 3),i=1,5)/            +0, -1, +2, +0,  0   /
DATA tp( 4)/            +0.007843D0                     /,  &
(itp(i, 4),i=1,5)/            +0, +0, +2, +0,  0   /
DATA tp( 5)/            +0.002824D0                     /,  &
(itp(i, 5),i=1,5)/            +0, +2, +0, +0,  0   /
DATA tp( 6)/            +0.000857D0                     /,  &
(itp(i, 6),i=1,5)/            +0, +1, +2, +0,  0   /
DATA tp( 7)/            +0.000533D0                     /,  &
(itp(i, 7),i=1,5)/            -1, +0, +2, +0,  1   /
DATA tp( 8)/            +0.000401D0                     /,  &
(itp(i, 8),i=1,5)/            -1, -1, +2, +0,  1   /
DATA tp( 9)/            +0.000320D0                     /,  &
(itp(i, 9),i=1,5)/            -1, +1, +0, +0,  1   /
DATA tp(10)/            -0.000271D0                     /,  &
(itp(i,10),i=1,5)/            +0, +0, +1, +0,  0   /
DATA tp(11)/            -0.000264D0                     /,  &
(itp(i,11),i=1,5)/            +1, +1, +0, +0,  1   /
DATA tp(12)/            -0.000198D0                     /,  &
(itp(i,12),i=1,5)/            +0, -1, +0, +2,  0   /
DATA tp(13)/            +0.000173D0                     /,  &
(itp(i,13),i=1,5)/            +0, +3, +0, +0,  0   /
DATA tp(14)/            +0.000167D0                     /,  &
(itp(i,14),i=1,5)/            +0, -1, +4, +0,  0   /
DATA tp(15)/            -0.000111D0                     /,  &
(itp(i,15),i=1,5)/            +1, +0, +0, +0,  1   /
DATA tp(16)/            +0.000103D0                     /,  &
(itp(i,16),i=1,5)/            +0, -2, +4, +0,  0   /
DATA tp(17)/            -0.000084D0                     /,  &
(itp(i,17),i=1,5)/            +0, +2, -2, +0,  0   /
DATA tp(18)/            -0.000083D0                     /,  &
(itp(i,18),i=1,5)/            +1, +0, +2, +0,  1   /
DATA tp(19)/            +0.000079D0                     /,  &
(itp(i,19),i=1,5)/            +0, +2, +2, +0,  0   /
DATA tp(20)/            +0.000072D0                     /,  &
(itp(i,20),i=1,5)/            +0, +0, +4, +0,  0   /
DATA tp(21)/            +0.000064D0                     /,  &
(itp(i,21),i=1,5)/            -1, +1, +2, +0,  1   /
DATA tp(22)/            -0.000063D0                     /,  &
(itp(i,22),i=1,5)/            +1, -1, +2, +0,  1   /
DATA tp(23)/            +0.000041D0                     /,  &
(itp(i,23),i=1,5)/            +1, +0, +1, +0,  1   /
DATA tp(24)/            +0.000035D0                     /,  &
(itp(i,24),i=1,5)/            -1, +2, +0, +0,  1   /
DATA tp(25)/            -0.000033D0                     /,  &
(itp(i,25),i=1,5)/            +0, +3, -2, +0,  0   /
DATA tp(26)/            -0.000030D0                     /,  &
(itp(i,26),i=1,5)/            +0, +1, +1, +0,  0   /
DATA tp(27)/            -0.000029D0                     /,  &
(itp(i,27),i=1,5)/            +0, +0, -2, +2,  0   /
DATA tp(28)/            -0.000029D0                     /,  &
(itp(i,28),i=1,5)/            +1, +2, +0, +0,  1   /
DATA tp(29)/            +0.000026D0                     /,  &
(itp(i,29),i=1,5)/            -2, +0, +2, +0,  2   /
DATA tp(30)/            -0.000023D0                     /,  &
(itp(i,30),i=1,5)/            +0, +1, -2, +2,  0   /
DATA tp(31)/            +0.000019D0                     /,  &
(itp(i,31),i=1,5)/            -1, -1, +4, +0,  1   /



!  Centuries since J1900
t=(date-15019.5D0)/36525D0


!  Fundamental arguments (radians) and derivatives (radians per
!  Julian century) for the current epoch


!  Moon's mean longitude
elp=d2r*MOD(elp0+(elp1+(elp2+elp3*t)*t)*t,360D0)
delp=d2r*(elp1+(2D0*elp2+3D0*elp3*t)*t)

!  Sun's mean anomaly
em=d2r*MOD(em0+(em1+(em2+em3*t)*t)*t,360D0)
dem=d2r*(em1+(2D0*em2+3D0*em3*t)*t)

!  Moon's mean anomaly
emp=d2r*MOD(emp0+(emp1+(emp2+emp3*t)*t)*t,360D0)
demp=d2r*(emp1+(2D0*emp2+3D0*emp3*t)*t)

!  Moon's mean elongation
d=d2r*MOD(d0+(d1+(d2+d3*t)*t)*t,360D0)
dd=d2r*(d1+(2D0*d2+3D0*d3*t)*t)

!  Mean distance of the Moon from its ascending node
f=d2r*MOD(f0+(f1+(f2+f3*t)*t)*t,360D0)
df=d2r*(f1+(2D0*f2+3D0*f3*t)*t)

!  Longitude of the Moon's ascending node
om=d2r*MOD(om0+(om1+(om2+om3*t)*t)*t,360D0)
dom=d2r*(om1+(2D0*om2+3D0*om3*t)*t)
sinom=SIN(om)
cosom=COS(om)
domcom=dom*cosom

!  Add the periodic variations
theta=d2r*(pa0+pa1*t)
wa=SIN(theta)
dwa=d2r*pa1*COS(theta)
theta=d2r*(pe0+(pe1+pe2*t)*t)
wb=pec*SIN(theta)
dwb=d2r*pec*(pe1+2D0*pe2*t)*COS(theta)
elp=elp+d2r*(pac*wa+wb+pfc*sinom)
delp=delp+d2r*(pac*dwa+dwb+pfc*domcom)
em=em+d2r*pbc*wa
dem=dem+d2r*pbc*dwa
emp=emp+d2r*(pcc*wa+wb+pgc*sinom)
demp=demp+d2r*(pcc*dwa+dwb+pgc*domcom)
d=d+d2r*(pdc*wa+wb+phc*sinom)
dd=dd+d2r*(pdc*dwa+dwb+phc*domcom)
wom=om+d2r*(pj0+pj1*t)
dwom=dom+d2r*pj1
sinwom=SIN(wom)
coswom=COS(wom)
f=f+d2r*(wb+pic*sinom+pjc*sinwom)
df=df+d2r*(dwb+pic*domcom+pjc*dwom*coswom)

!  E-factor, and square
e=1D0+(e1+e2*t)*t
de=e1+2D0*e2*t
esq=e*e
desq=2D0*e*de


!  Series expansions


!  Longitude
v=0D0
dv=0D0
DO n=nl,1,-1
  coeff=tl(n)
  emn=DBLE(itl(1,n))
  empn=DBLE(itl(2,n))
  dn=DBLE(itl(3,n))
  fn=DBLE(itl(4,n))
  i=itl(5,n)
  IF (i == 0) THEN
    en=1D0
    den=0D0
  ELSE IF (i == 1) THEN
    en=e
    den=de
  ELSE
    en=esq
    den=desq
  END IF
  theta=emn*em+empn*emp+dn*d+fn*f
  dtheta=emn*dem+empn*demp+dn*dd+fn*df
  ftheta=SIN(theta)
  v=v+coeff*ftheta*en
  dv=dv+coeff*(COS(theta)*dtheta*en+ftheta*den)
END DO
el=elp+d2r*v
del=(delp+d2r*dv)/cj

!  Latitude
v=0D0
dv=0D0
DO n=nb,1,-1
  coeff=tb(n)
  emn=DBLE(itb(1,n))
  empn=DBLE(itb(2,n))
  dn=DBLE(itb(3,n))
  fn=DBLE(itb(4,n))
  i=itb(5,n)
  IF (i == 0) THEN
    en=1D0
    den=0D0
  ELSE IF (i == 1) THEN
    en=e
    den=de
  ELSE
    en=esq
    den=desq
  END IF
  theta=emn*em+empn*emp+dn*d+fn*f
  dtheta=emn*dem+empn*demp+dn*dd+fn*df
  ftheta=SIN(theta)
  v=v+coeff*ftheta*en
  dv=dv+coeff*(COS(theta)*dtheta*en+ftheta*den)
END DO
bf=1D0-cw1*cosom-cw2*coswom
dbf=cw1*dom*sinom+cw2*dwom*sinwom
b=d2r*v*bf
db=d2r*(dv*bf+v*dbf)/cj

!  Parallax
v=0D0
dv=0D0
DO n=np,1,-1
  coeff=tp(n)
  emn=DBLE(itp(1,n))
  empn=DBLE(itp(2,n))
  dn=DBLE(itp(3,n))
  fn=DBLE(itp(4,n))
  i=itp(5,n)
  IF (i == 0) THEN
    en=1D0
    den=0D0
  ELSE IF (i == 1) THEN
    en=e
    den=de
  ELSE
    en=esq
    den=desq
  END IF
  theta=emn*em+empn*emp+dn*d+fn*f
  dtheta=emn*dem+empn*demp+dn*dd+fn*df
  ftheta=COS(theta)
  v=v+coeff*ftheta*en
  dv=dv+coeff*(-SIN(theta)*dtheta*en+ftheta*den)
END DO
p=d2r*v
dp=d2r*dv/cj


!  Transformation into final form


!  Parallax to distance (AU, AU/sec)
sp=SIN(p)
r=eradau/sp
dr=-r*dp*COS(p)/sp

!  Longitude, latitude to x,y,z (AU)
sel=SIN(el)
cel=COS(el)
sb=SIN(b)
cb=COS(b)
rcb=r*cb
rbd=r*db
w=rbd*sb-cb*dr
x=rcb*cel
y=rcb*sel
z=r*sb
xd=-y*del-w*cel
yd=x*del-w*sel
zd=rbd*cb+sb*dr

!  Julian centuries since J2000
t=(date-51544.5D0)/36525D0

!  Fricke equinox correction
epj=2000D0+t*100D0
eqcor=ds2r*(0.035D0+0.00085D0*(epj-b1950))

!  Mean obliquity (IAU 1976)
eps=das2r*(84381.448D0+(-46.8150D0+(-0.00059D0+0.001813D0*t)*t)*t)

!  To the equatorial system, mean of date, FK5 system
sineps=SIN(eps)
coseps=COS(eps)
es=eqcor*sineps
ec=eqcor*coseps
pv(1)=x-ec*y+es*z
pv(2)=eqcor*x+y*coseps-z*sineps
pv(3)=y*sineps+z*coseps
pv(4)=xd-ec*yd+es*zd
pv(5)=eqcor*xd+yd*coseps-zd*sineps
pv(6)=yd*sineps+zd*coseps

END SUBROUTINE sla_dmoon
SUBROUTINE sla_dmxm (a, b, c)
!+
!     - - - - -
!      D M X M
!     - - - - -

!  Product of two 3x3 matrices:

!      matrix C  =  matrix A  x  matrix B

!  (double precision)

!  Given:
!      A      dp(3,3)        matrix
!      B      dp(3,3)        matrix

!  Returned:
!      C      dp(3,3)        matrix result

!  To comply with the ANSI Fortran 77 standard, A, B and C must
!  be different arrays.  However, the routine is coded so as to
!  work properly on the VAX and many other systems even if this
!  rule is violated.

!  P.T.Wallace   Starlink   5 April 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: a(3,3),b(3,3),c(3,3)

INTEGER :: i,j,k
DOUBLE PRECISION :: w,wm(3,3)


!  Multiply into scratch matrix
DO i=1,3
  DO j=1,3
    w=0D0
    DO k=1,3
      w=w+a(i,k)*b(k,j)
    END DO
    wm(i,j)=w
  END DO
END DO

!  Return the result
DO j=1,3
  DO i=1,3
    c(i,j)=wm(i,j)
  END DO
END DO

END SUBROUTINE sla_dmxm
SUBROUTINE sla_dmxv (dm, va, vb)
!+
!     - - - - -
!      D M X V
!     - - - - -

!  Performs the 3-D forward unitary transformation:

!     vector VB = matrix DM * vector VA

!  (double precision)

!  Given:
!     DM       dp(3,3)    matrix
!     VA       dp(3)      vector

!  Returned:
!     VB       dp(3)      result vector

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dm(3,3),va(3),vb(3)

INTEGER :: i,j
DOUBLE PRECISION :: w,vw(3)


!  Matrix DM * vector VA -> vector VW
DO j=1,3
  w=0D0
  DO i=1,3
    w=w+dm(j,i)*va(i)
  END DO
  vw(j)=w
END DO

!  Vector VW -> vector VB
DO j=1,3
  vb(j)=vw(j)
END DO

END SUBROUTINE sla_dmxv
DOUBLE PRECISION FUNCTION sla_dpav ( v1, v2 )
!+
!     - - - - -
!      D P A V
!     - - - - -

!  Position angle of one celestial direction with respect to another.

!  (double precision)

!  Given:
!     V1    d(3)    direction cosines of one point
!     V2    d(3)    direction cosines of the other point

!  (The coordinate frames correspond to RA,Dec, Long,Lat etc.)

!  The result is the bearing (position angle), in radians, of point
!  V2 with respect to point V1.  It is in the range +/- pi.  The
!  sense is such that if V2 is a small distance east of V1, the
!  bearing is about +pi/2.  Zero is returned if the two points
!  are coincident.

!  V1 and V2 need not be unit vectors.

!  The routine sla_DBEAR performs an equivalent function except
!  that the points are specified in the form of spherical
!  coordinates.

!  Patrick Wallace   Starlink   13 July 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: v1(3),v2(3)

DOUBLE PRECISION :: x1,y1,z1,w,r,xu1,yu1,zu1,dx,dy,dz,sq,cq



!  Unit vector to point 1
x1=v1(1)
y1=v1(2)
z1=v1(3)
w=SQRT(x1*x1+y1*y1+z1*z1)
IF (w /= 0D0) THEN
  x1=x1/w
  y1=y1/w
  z1=z1/w
END IF

!  Unit vector "north" from point 1
r=SQRT(x1*x1+y1*y1)
IF (r == 0.0) r=1D-5
w=z1/r
xu1=-x1*w
yu1=-y1*w
zu1=r

!  Vector from point 1 to point 2
dx=v2(1)-x1
dy=v2(2)-y1
dz=v2(3)-z1

!  Position angle
sq=dx*yu1*z1+dy*zu1*x1+dz*xu1*y1-dz*yu1*x1-dy*xu1*z1-dx*zu1*y1
cq=dx*xu1+dy*yu1+dz*zu1
IF (sq == 0D0.AND.cq == 0D0) cq=1D0
sla_dpav=ATAN2(sq,cq)

END FUNCTION sla_dpav
SUBROUTINE sla_dr2af (ndp, angle, SIGN, idmsf)
!+
!     - - - - - -
!      D R 2 A F
!     - - - - - -

!  Convert an angle in radians to degrees, arcminutes, arcseconds
!  (double precision)

!  Given:
!     NDP       int      number of decimal places of arcseconds
!     ANGLE     dp       angle in radians

!  Returned:
!     SIGN      char     '+' or '-'
!     IDMSF     int(4)   degrees, arcminutes, arcseconds, fraction

!  Notes:

!     1)  NDP less than zero is interpreted as zero.

!     2)  The largest useful value for NDP is determined by the size
!         of ANGLE, the format of DOUBLE PRECISION floating-point
!         numbers on the target machine, and the risk of overflowing
!         IDMSF(4).  For example, on the VAX, for ANGLE up to 2pi, the
!         available floating-point precision corresponds roughly to
!         NDP=12.  However, the practical limit is NDP=9, set by the
!         capacity of the 32-bit integer IDMSF(4).

!     3)  The absolute value of ANGLE may exceed 2pi.  In cases where it
!         does not, it is up to the caller to test for and handle the
!         case where ANGLE is very nearly 2pi and rounds up to 360 deg,
!         by testing for IDMSF(1)=360 and setting IDMSF(1-4) to zero.

!  Called:
!     sla_DD2TF

!  P.T.Wallace   Starlink   11 September 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ndp
DOUBLE PRECISION :: angle
CHARACTER :: SIGN*(*)
INTEGER :: idmsf(4)

!  Hours to degrees * radians to turns
DOUBLE PRECISION :: f
PARAMETER (f=15D0/6.283185307179586476925287D0)



!  Scale then use days to h,m,s routine
CALL sla_dd2tf(ndp,angle*f,SIGN,idmsf)

END SUBROUTINE sla_dr2af
SUBROUTINE sla_dr2tf (ndp, angle, SIGN, ihmsf)
!+
!     - - - - - -
!      D R 2 T F
!     - - - - - -

!  Convert an angle in radians to hours, minutes, seconds
!  (double precision)

!  Given:
!     NDP       int      number of decimal places of seconds
!     ANGLE     dp       angle in radians

!  Returned:
!     SIGN      char     '+' or '-'
!     IHMSF     int(4)   hours, minutes, seconds, fraction

!  Notes:

!     1)  NDP less than zero is interpreted as zero.

!     2)  The largest useful value for NDP is determined by the size
!         of ANGLE, the format of DOUBLE PRECISION floating-point
!         numbers on the target machine, and the risk of overflowing
!         IHMSF(4).  For example, on the VAX, for ANGLE up to 2pi, the
!         available floating-point precision corresponds roughly to
!         NDP=12.  However, the practical limit is NDP=9, set by the
!         capacity of the 32-bit integer IHMSF(4).

!     3)  The absolute value of ANGLE may exceed 2pi.  In cases where it
!         does not, it is up to the caller to test for and handle the
!         case where ANGLE is very nearly 2pi and rounds up to 24 hours,
!         by testing for IHMSF(1)=24 and setting IHMSF(1-4) to zero.

!  Called:
!     sla_DD2TF

!  P.T.Wallace   Starlink   11 September 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ndp
DOUBLE PRECISION :: angle
CHARACTER :: SIGN*(*)
INTEGER :: ihmsf(4)

!  Turns to radians
DOUBLE PRECISION :: t2r
PARAMETER (t2r=6.283185307179586476925287D0)



!  Scale then use days to h,m,s routine
CALL sla_dd2tf(ndp,angle/t2r,SIGN,ihmsf)

END SUBROUTINE sla_dr2tf
DOUBLE PRECISION FUNCTION sla_drange (angle)
!+
!     - - - - - - -
!      D R A N G E
!     - - - - - - -

!  Normalize angle into range +/- pi  (double precision)

!  Given:
!     ANGLE     dp      the angle in radians

!  The result (double precision) is ANGLE expressed in the range +/- pi.

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: angle

DOUBLE PRECISION :: dpi,d2pi
PARAMETER (dpi=3.141592653589793238462643D0)
PARAMETER (d2pi=6.283185307179586476925287D0)


sla_drange=MOD(angle,d2pi)
IF (ABS(sla_drange) >= dpi) sla_drange=sla_drange-SIGN(d2pi,angle)

END FUNCTION sla_drange
DOUBLE PRECISION FUNCTION sla_dranrm (angle)
!+
!     - - - - - - -
!      D R A N R M
!     - - - - - - -

!  Normalize angle into range 0-2 pi  (double precision)

!  Given:
!     ANGLE     dp      the angle in radians

!  The result is ANGLE expressed in the range 0-2 pi (double
!  precision).

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: angle

DOUBLE PRECISION :: d2pi
PARAMETER (d2pi=6.283185307179586476925286766559D0)


sla_dranrm=MOD(angle,d2pi)
IF (sla_dranrm < 0D0) sla_dranrm=sla_dranrm+d2pi

END FUNCTION sla_dranrm
SUBROUTINE sla_ds2c6 (a, b, r, ad, bd, rd, v)
!+
!     - - - - - -
!      D S 2 C 6
!     - - - - - -

!  Conversion of position & velocity in spherical coordinates
!  to Cartesian coordinates

!  (double precision)

!  Given:
!     A     dp      longitude (radians)
!     B     dp      latitude (radians)
!     R     dp      radial coordinate
!     AD    dp      longitude derivative (radians per unit time)
!     BD    dp      latitude derivative (radians per unit time)
!     RD    dp      radial derivative

!  Returned:
!     V     dp(6)   Cartesian position & velocity vector

!  P.T.Wallace   Starlink   10 July 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: a,b,r,ad,bd,rd,v(6)

DOUBLE PRECISION :: sa,ca,sb,cb,rcb,x,y,rbd,w



!  Useful functions
sa=SIN(a)
ca=COS(a)
sb=SIN(b)
cb=COS(b)
rcb=r*cb
x=rcb*ca
y=rcb*sa
rbd=r*bd
w=rbd*sb-cb*rd

!  Position
v(1)=x
v(2)=y
v(3)=r*sb

!  Velocity
v(4)=-y*ad-w*ca
v(5)=x*ad-w*sa
v(6)=rbd*cb+sb*rd

END SUBROUTINE sla_ds2c6
SUBROUTINE sla_ds2tp (ra, dec, raz, decz, xi, eta, j)
!+
!     - - - - - -
!      D S 2 T P
!     - - - - - -

!  Projection of spherical coordinates onto tangent plane:
!  "gnomonic" projection - "standard coordinates" (double precision)

!  Given:
!     RA,DEC      dp   spherical coordinates of point to be projected
!     RAZ,DECZ    dp   spherical coordinates of tangent point

!  Returned:
!     XI,ETA      dp   rectangular coordinates on tangent plane
!     J           int  status:   0 = OK, star on tangent plane
!                                1 = error, star too far from axis
!                                2 = error, antistar on tangent plane
!                                3 = error, antistar too far from axis

!  P.T.Wallace   Starlink   18 July 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ra,dec,raz,decz,xi,eta
INTEGER :: j

DOUBLE PRECISION :: sdecz,sdec,cdecz,cdec,radif,sradif,cradif,denom

DOUBLE PRECISION :: tiny
PARAMETER (tiny=1D-6)


!  Trig functions
sdecz=SIN(decz)
sdec=SIN(dec)
cdecz=COS(decz)
cdec=COS(dec)
radif=ra-raz
sradif=SIN(radif)
cradif=COS(radif)

!  Reciprocal of star vector length to tangent plane
denom=sdec*sdecz+cdec*cdecz*cradif

!  Handle vectors too far from axis
IF (denom > tiny) THEN
  j=0
ELSE IF (denom >= 0D0) THEN
  j=1
  denom=tiny
ELSE IF (denom > -tiny) THEN
  j=2
  denom=-tiny
ELSE
  j=3
END IF

!  Compute tangent plane coordinates (even in dubious cases)
xi=cdec*sradif/denom
eta=(sdec*cdecz-cdec*sdecz*cradif)/denom

END SUBROUTINE sla_ds2tp
DOUBLE PRECISION FUNCTION sla_dsep (a1, b1, a2, b2)
!+
!     - - - - -
!      D S E P
!     - - - - -

!  Angle between two points on a sphere (double precision)

!  Given:
!     A1,B1    dp    spherical coordinates of one point
!     A2,B2    dp    spherical coordinates of the other point

!  (The spherical coordinates are RA,Dec, Long,Lat etc, in radians.)

!  The result is the angle, in radians, between the two points.  It
!  is always positive.

!  Called:  sla_DCS2C

!  P.T.Wallace   Starlink   April 1985

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: a1,b1,a2,b2

INTEGER :: i
DOUBLE PRECISION :: v1(3),v2(3),w



!  Convert coordinates from spherical to Cartesian
CALL sla_dcs2c(a1,b1,v1)
CALL sla_dcs2c(a2,b2,v2)

!  Modulus squared of half the difference vector
w=0D0
DO i=1,3
  w=w+(v1(i)-v2(i))**2
END DO
w=w/4D0

!  Angle between the vectors
sla_dsep=2D0*ATAN2(SQRT(w),SQRT(MAX(0D0,1D0-w)))

END FUNCTION sla_dsep
DOUBLE PRECISION FUNCTION sla_dt (epoch)
!+
!     - - -
!      D T
!     - - -

!  Estimate the offset between dynamical time and Universal Time
!  for a given historical epoch.

!  Given:
!     EPOCH       d        (Julian) epoch (e.g. 1850D0)

!  The result is a rough estimate of ET-UT (after 1984, TT-UT) at
!  the given epoch, in seconds.

!  Notes:

!  1  Depending on the epoch, one of three parabolic approximations
!     is used:

!      before 979    Stephenson & Morrison's 390 BC to AD 948 model
!      979 to 1708   Stephenson & Morrison's 948 to 1600 model
!      after 1708    McCarthy & Babcock's post-1650 model

!     The breakpoints are chosen to ensure continuity:  they occur
!     at places where the adjacent models give the same answer as
!     each other.

!  2  The accuracy is modest, with errors of up to 20 sec during
!     the interval since 1650, rising to perhaps 30 min by 1000 BC.
!     Comparatively accurate values from AD 1600 are tabulated in
!     the Astronomical Almanac (see section K8 of the 1995 AA).

!  3  The use of double-precision for both argument and result is
!     purely for compatibility with other SLALIB time routines.

!  4  The models used are based on a lunar tidal acceleration value
!     of -26.00 arcsec per century.

!  Reference:  Explanatory Supplement to the Astronomical Almanac,
!              ed P.K.Seidelmann, University Science Books (1992),
!              section 2.553, p83.  This contains references to
!              the Stephenson & Morrison and McCarthy & Babcock
!              papers.

!  P.T.Wallace   Starlink   1 March 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: epoch
DOUBLE PRECISION :: t,w,s


!  Centuries since 1800
t=(epoch-1800D0)/100D0

!  Select model
IF (epoch >= 1708.185161980887D0) THEN
  
!     Post-1708: use McCarthy & Babcock
  w=t-0.19D0
  s=5.156D0+13.3066D0*w*w
ELSE IF (epoch >= 979.0258204760233D0) THEN
  
!     979-1708: use Stephenson & Morrison's 948-1600 model
  s=25.5D0*t*t
ELSE
  
!     Pre-979: use Stephenson & Morrison's 390 BC to AD 948 model
  s=1360.0D0+(320D0+44.3D0*t)*t
END IF

!  Result
sla_dt=s

END FUNCTION sla_dt
SUBROUTINE sla_dtf2d (ihour, imin, sec, days, j)
!+
!     - - - - - -
!      D T F 2 D
!     - - - - - -

!  Convert hours, minutes, seconds to days (double precision)

!  Given:
!     IHOUR       int       hours
!     IMIN        int       minutes
!     SEC         dp        seconds

!  Returned:
!     DAYS        dp        interval in days
!     J           int       status:  0 = OK
!                                    1 = IHOUR outside range 0-23
!                                    2 = IMIN outside range 0-59
!                                    3 = SEC outside range 0-59.999...

!  Notes:

!     1)  The result is computed even if any of the range checks fail.

!     2)  The sign must be dealt with outside this routine.

!  P.T.Wallace   Starlink   July 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ihour,imin
DOUBLE PRECISION :: sec,days
INTEGER :: j

!  Seconds per day
DOUBLE PRECISION :: d2s
PARAMETER (d2s=86400D0)



!  Preset status
j=0

!  Validate sec, min, hour
IF (sec < 0D0.OR.sec >= 60D0) j=3
IF (imin < 0.OR.imin > 59) j=2
IF (ihour < 0.OR.ihour > 23) j=1

!  Compute interval
days=(60D0*(60D0*DBLE(ihour)+DBLE(imin))+sec)/d2s

END SUBROUTINE sla_dtf2d
SUBROUTINE sla_dtf2r (ihour, imin, sec, rad, j)
!+
!     - - - - - -
!      D T F 2 R
!     - - - - - -

!  Convert hours, minutes, seconds to radians (double precision)

!  Given:
!     IHOUR       int       hours
!     IMIN        int       minutes
!     SEC         dp        seconds

!  Returned:
!     RAD         dp        angle in radians
!     J           int       status:  0 = OK
!                                    1 = IHOUR outside range 0-23
!                                    2 = IMIN outside range 0-59
!                                    3 = SEC outside range 0-59.999...

!  Called:
!     sla_DTF2D

!  Notes:

!     1)  The result is computed even if any of the range checks fail.

!     2)  The sign must be dealt with outside this routine.

!  P.T.Wallace   Starlink   July 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: ihour,imin
DOUBLE PRECISION :: sec,rad
INTEGER :: j

DOUBLE PRECISION :: turns

!  Turns to radians
DOUBLE PRECISION :: t2r
PARAMETER (t2r=6.283185307179586476925287D0)



!  Convert to turns then radians
CALL sla_dtf2d(ihour,imin,sec,turns,j)
rad=t2r*turns

END SUBROUTINE sla_dtf2r
SUBROUTINE sla_dtp2s (xi, eta, raz, decz, ra, dec)
!+
!     - - - - - -
!      D T P 2 S
!     - - - - - -

!  Transform tangent plane coordinates into spherical
!  (double precision)

!  Given:
!     XI,ETA      dp   tangent plane rectangular coordinates
!     RAZ,DECZ    dp   spherical coordinates of tangent point

!  Returned:
!     RA,DEC      dp   spherical coordinates (0-2pi,+/-pi/2)

!  Called:        sla_DRANRM

!  P.T.Wallace   Starlink   24 July 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: xi,eta,raz,decz,ra,dec


DOUBLE PRECISION :: sdecz,cdecz,denom



sdecz=SIN(decz)
cdecz=COS(decz)

denom=cdecz-eta*sdecz

ra=sla_dranrm(ATAN2(xi,denom)+raz)
dec=ATAN2(sdecz+eta*cdecz,SQRT(xi*xi+denom*denom))

END SUBROUTINE sla_dtp2s
SUBROUTINE sla_dtp2v (xi, eta, v0, v)
!+
!     - - - - - -
!      D T P 2 V
!     - - - - - -

!  Given the tangent-plane coordinates of a star and the direction
!  cosines of the tangent point, determine the direction cosines
!  of the star.

!  (double precision)

!  Given:
!     XI,ETA    d       tangent plane coordinates of star
!     V0        d(3)    direction cosines of tangent point

!  Returned:
!     V         d(3)    direction cosines of star

!  Notes:

!  1  If vector V0 is not of unit length, the returned vector V will
!     be wrong.

!  2  If vector V0 points at a pole, the returned vector V will be
!     based on the arbitrary assumption that the RA of the tangent
!     point is zero.

!  3  This routine is the Cartesian equivalent of the routine sla_DTP2S.

!  P.T.Wallace   Starlink   11 February 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: xi,eta,v0(3),v(3)

DOUBLE PRECISION :: x,y,z,f,r


x=v0(1)
y=v0(2)
z=v0(3)
f=SQRT(1D0+xi*xi+eta*eta)
r=SQRT(x*x+y*y)
IF (r == 0D0) THEN
  r=1D-20
  x=r
END IF
v(1)=(x-(xi*y+eta*x*z)/r)/f
v(2)=(y+(xi*x-eta*y*z)/r)/f
v(3)=(z+eta*r)/f

END SUBROUTINE sla_dtp2v
SUBROUTINE sla_dtps2c (xi, eta, ra, dec, raz1, decz1,raz2, decz2, n)
!+
!     - - - - - - -
!      D T P S 2 C
!     - - - - - - -

!  From the tangent plane coordinates of a star of known RA,Dec,
!  determine the RA,Dec of the tangent point.

!  (double precision)

!  Given:
!     XI,ETA      d    tangent plane rectangular coordinates
!     RA,DEC      d    spherical coordinates

!  Returned:
!     RAZ1,DECZ1  d    spherical coordinates of tangent point, solution 1
!     RAZ2,DECZ2  d    spherical coordinates of tangent point, solution 2
!     N           i    number of solutions:
!                        0 = no solutions returned (note 2)
!                        1 = only the first solution is useful (note 3)
!                        2 = both solutions are useful (note 3)

!  Notes:

!  1  The RAZ1 and RAZ2 values are returned in the range 0-2pi.

!  2  Cases where there is no solution can only arise near the poles.
!     For example, it is clearly impossible for a star at the pole
!     itself to have a non-zero XI value, and hence it is
!     meaningless to ask where the tangent point would have to be
!     to bring about this combination of XI and DEC.

!  3  Also near the poles, cases can arise where there are two useful
!     solutions.  The argument N indicates whether the second of the
!     two solutions returned is useful.  N=1 indicates only one useful
!     solution, the usual case;  under these circumstances, the second
!     solution corresponds to the "over-the-pole" case, and this is
!     reflected in the values of RAZ2 and DECZ2 which are returned.

!  4  The DECZ1 and DECZ2 values are returned in the range +/-pi, but
!     in the usual, non-pole-crossing, case, the range is +/-pi/2.

!  5  This routine is the spherical equivalent of the routine sla_DTPV2C.

!  Called:  sla_DRANRM

!  P.T.Wallace   Starlink   5 June 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: xi,eta,ra,dec,raz1,decz1,raz2,decz2
INTEGER :: n

DOUBLE PRECISION :: x2,y2,sd,cd,sdf,r2,r,s,c



x2=xi*xi
y2=eta*eta
sd=SIN(dec)
cd=COS(dec)
sdf=sd*SQRT(1D0+x2+y2)
r2=cd*cd*(1D0+y2)-sd*sd*x2
IF (r2 >= 0D0) THEN
  r=SQRT(r2)
  s=sdf-eta*r
  c=sdf*eta+r
  IF (xi == 0D0.AND.r == 0D0) r=1D0
  raz1=sla_dranrm(ra-ATAN2(xi,r))
  decz1=ATAN2(s,c)
  r=-r
  s=sdf-eta*r
  c=sdf*eta+r
  raz2=sla_dranrm(ra-ATAN2(xi,r))
  decz2=ATAN2(s,c)
  IF (ABS(sdf) < 1D0) THEN
    n=1
  ELSE
    n=2
  END IF
ELSE
  n=0
END IF

END SUBROUTINE sla_dtps2c
SUBROUTINE sla_dtpv2c (xi, eta, v, v01, v02, n)
!+
!     - - - - - - -
!      D T P V 2 C
!     - - - - - - -

!  Given the tangent-plane coordinates of a star and its direction
!  cosines, determine the direction cosines of the tangent-point.

!  (double precision)

!  Given:
!     XI,ETA    d       tangent plane coordinates of star
!     V         d(3)    direction cosines of star

!  Returned:
!     V01       d(3)    direction cosines of tangent point, solution 1
!     V02       d(3)    direction cosines of tangent point, solution 2
!     N         i       number of solutions:
!                         0 = no solutions returned (note 2)
!                         1 = only the first solution is useful (note 3)
!                         2 = both solutions are useful (note 3)

!  Notes:

!  1  The vector V must be of unit length or the result will be wrong.

!  2  Cases where there is no solution can only arise near the poles.
!     For example, it is clearly impossible for a star at the pole
!     itself to have a non-zero XI value, and hence it is meaningless
!     to ask where the tangent point would have to be.

!  3  Also near the poles, cases can arise where there are two useful
!     solutions.  The argument N indicates whether the second of the
!     two solutions returned is useful.  N=1 indicates only one useful
!     solution, the usual case;  under these circumstances, the second
!     solution can be regarded as valid if the vector V02 is interpreted
!     as the "over-the-pole" case.

!  4  This routine is the Cartesian equivalent of the routine sla_DTPS2C.

!  P.T.Wallace   Starlink   5 June 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: xi,eta,v(3),v01(3),v02(3)
INTEGER :: n

DOUBLE PRECISION :: x,y,z,rxy2,xi2,eta2p1,sdf,r2,r,c


x=v(1)
y=v(2)
z=v(3)
rxy2=x*x+y*y
xi2=xi*xi
eta2p1=eta*eta+1D0
sdf=z*SQRT(xi2+eta2p1)
r2=rxy2*eta2p1-z*z*xi2
IF (r2 > 0D0) THEN
  r=SQRT(r2)
  c=(sdf*eta+r)/(eta2p1*SQRT(rxy2*(r2+xi2)))
  v01(1)=c*(x*r+y*xi)
  v01(2)=c*(y*r-x*xi)
  v01(3)=(sdf-eta*r)/eta2p1
  r=-r
  c=(sdf*eta+r)/(eta2p1*SQRT(rxy2*(r2+xi2)))
  v02(1)=c*(x*r+y*xi)
  v02(2)=c*(y*r-x*xi)
  v02(3)=(sdf-eta*r)/eta2p1
  IF (ABS(sdf) < 1D0) THEN
    n=1
  ELSE
    n=2
  END IF
ELSE
  n=0
END IF

END SUBROUTINE sla_dtpv2c
DOUBLE PRECISION FUNCTION sla_dtt (utc)
!+
!     - - - -
!      D T T
!     - - - -

!  Increment to be applied to Coordinated Universal Time UTC to give
!  Terrestrial Time TT (formerly Ephemeris Time ET)

!  (double precision)

!  Given:
!     UTC      d      UTC date as a modified JD (JD-2400000.5)

!  Result:  TT-UTC in seconds

!  Notes:

!  1  The UTC is specified to be a date rather than a time to indicate
!     that care needs to be taken not to specify an instant which lies
!     within a leap second.  Though in most cases UTC can include the
!     fractional part, correct behaviour on the day of a leap second
!     can only be guaranteed up to the end of the second 23:59:59.

!  2  Pre 1972 January 1 a fixed value of 10 + ET-TAI is returned.

!  3  See also the routine sla_DT, which roughly estimates ET-UT for
!     historical epochs.

!  Called:  sla_DAT

!  P.T.Wallace   Starlink   6 December 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: utc



sla_dtt=32.184D0+sla_dat(utc)

END FUNCTION sla_dtt
SUBROUTINE sla_dv2tp (v, v0, xi, eta, j)
!+
!     - - - - - -
!      D V 2 T P
!     - - - - - -

!  Given the direction cosines of a star and of the tangent point,
!  determine the star's tangent-plane coordinates.

!  (double precision)

!  Given:
!     V         d(3)    direction cosines of star
!     V0        d(3)    direction cosines of tangent point

!  Returned:
!     XI,ETA    d       tangent plane coordinates of star
!     J         i       status:   0 = OK
!                                 1 = error, star too far from axis
!                                 2 = error, antistar on tangent plane
!                                 3 = error, antistar too far from axis

!  Notes:

!  1  If vector V0 is not of unit length, or if vector V is of zero
!     length, the results will be wrong.

!  2  If V0 points at a pole, the returned XI,ETA will be based on the
!     arbitrary assumption that the RA of the tangent point is zero.

!  3  This routine is the Cartesian equivalent of the routine sla_DS2TP.

!  P.T.Wallace   Starlink   27 November 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: v(3),v0(3),xi,eta
INTEGER :: j

DOUBLE PRECISION :: x,y,z,x0,y0,z0,r2,r,w,d

DOUBLE PRECISION :: tiny
PARAMETER (tiny=1D-6)


x=v(1)
y=v(2)
z=v(3)
x0=v0(1)
y0=v0(2)
z0=v0(3)
r2=x0*x0+y0*y0
r=SQRT(r2)
IF (r == 0D0) THEN
  r=1D-20
  x0=r
END IF
w=x*x0+y*y0
d=w+z*z0
IF (d > tiny) THEN
  j=0
ELSE IF (d >= 0D0) THEN
  j=1
  d=tiny
ELSE IF (d > -tiny) THEN
  j=2
  d=-tiny
ELSE
  j=3
END IF
d=d*r
xi=(y*x0-x*y0)/d
eta=(z*r2-z0*w)/d

END SUBROUTINE sla_dv2tp
DOUBLE PRECISION FUNCTION sla_dvdv (va, vb)
!+
!     - - - - -
!      D V D V
!     - - - - -

!  Scalar product of two 3-vectors  (double precision)

!  Given:
!      VA      dp(3)     first vector
!      VB      dp(3)     second vector

!  The result is the scalar product VA.VB (double precision)

!  P.T.Wallace   Starlink   November 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: va(3),vb(3)


sla_dvdv=va(1)*vb(1)+va(2)*vb(2)+va(3)*vb(3)

END FUNCTION sla_dvdv
SUBROUTINE sla_dvn (v, uv, vm)
!+
!     - - - -
!      D V N
!     - - - -

!  Normalizes a 3-vector also giving the modulus (double precision)

!  Given:
!     V       dp(3)      vector

!  Returned:
!     UV      dp(3)      unit vector in direction of V
!     VM      dp         modulus of V

!  If the modulus of V is zero, UV is set to zero as well

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: v(3),uv(3),vm

INTEGER :: i
DOUBLE PRECISION :: w1,w2


!  Modulus
w1=0D0
DO i=1,3
  w2=v(i)
  w1=w1+w2*w2
END DO
w1=SQRT(w1)
vm=w1

!  Normalize the vector
IF (w1 <= 0D0) w1=1D0
DO i=1,3
  uv(i)=v(i)/w1
END DO

END SUBROUTINE sla_dvn
SUBROUTINE sla_dvxv (va, vb, vc)
!+
!     - - - - -
!      D V X V
!     - - - - -

!  Vector product of two 3-vectors  (double precision)

!  Given:
!      VA      dp(3)     first vector
!      VB      dp(3)     second vector

!  Returned:
!      VC      dp(3)     vector result

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: va(3),vb(3),vc(3)

DOUBLE PRECISION :: vw(3)
INTEGER :: i


!  Form the vector product VA cross VB
vw(1)=va(2)*vb(3)-va(3)*vb(2)
vw(2)=va(3)*vb(1)-va(1)*vb(3)
vw(3)=va(1)*vb(2)-va(2)*vb(1)

!  Return the result
DO i=1,3
  vc(i)=vw(i)
END DO

END SUBROUTINE sla_dvxv
SUBROUTINE sla_e2h (ha, dec, phi, az, el)
!+
!     - - - -
!      E 2 H
!     - - - -

!  Equatorial to horizon coordinates:  HA,Dec to Az,El

!  (single precision)

!  Given:
!     HA      r     hour angle
!     DEC     r     declination
!     PHI     r     observatory latitude

!  Returned:
!     AZ      r     azimuth
!     EL      r     elevation

!  Notes:

!  1)  All the arguments are angles in radians.

!  2)  Azimuth is returned in the range 0-2pi;  north is zero,
!      and east is +pi/2.  Elevation is returned in the range
!      +/-pi/2.

!  3)  The latitude must be geodetic.  In critical applications,
!      corrections for polar motion should be applied.

!  4)  In some applications it will be important to specify the
!      correct type of hour angle and declination in order to
!      produce the required type of azimuth and elevation.  In
!      particular, it may be important to distinguish between
!      elevation as affected by refraction, which would
!      require the "observed" HA,Dec, and the elevation
!      in vacuo, which would require the "topocentric" HA,Dec.
!      If the effects of diurnal aberration can be neglected, the
!      "apparent" HA,Dec may be used instead of the topocentric
!      HA,Dec.

!  5)  No range checking of arguments is carried out.

!  6)  In applications which involve many such calculations, rather
!      than calling the present routine it will be more efficient to
!      use inline code, having previously computed fixed terms such
!      as sine and cosine of latitude, and (for tracking a star)
!      sine and cosine of declination.

!  P.T.Wallace   Starlink   9 July 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: ha,dec,phi,az,el

REAL :: r2pi
PARAMETER (r2pi=6.283185307179586476925286766559)

REAL :: sh,ch,sd,cd,sp,cp,x,y,z,r,a


!  Useful trig functions
sh=SIN(ha)
ch=COS(ha)
sd=SIN(dec)
cd=COS(dec)
sp=SIN(phi)
cp=COS(phi)

!  Az,El as x,y,z
x=-ch*cd*sp+sd*cp
y=-sh*cd
z=ch*cd*cp+sd*sp

!  To spherical
r=SQRT(x*x+y*y)
IF (r == 0.0) THEN
  a=0.0
ELSE
  a=ATAN2(y,x)
END IF
IF (a < 0.0) a=a+r2pi
az=a
el=ATAN2(z,r)

END SUBROUTINE sla_e2h
SUBROUTINE sla_earth (iy, id, fd, pv)
!+
!     - - - - - -
!      E A R T H
!     - - - - - -

!  Approximate heliocentric position and velocity of the Earth

!  Given:
!     IY       I       year
!     ID       I       day in year (1 = Jan 1st)
!     FD       R       fraction of day

!  Returned:
!     PV       R(6)    Earth position & velocity vector

!  Notes:

!  1  The date and time is TDB (loosely ET) in a Julian calendar
!     which has been aligned to the ordinary Gregorian
!     calendar for the interval 1900 March 1 to 2100 February 28.
!     The year and day can be obtained by calling sla_CALYD or
!     sla_CLYD.

!  2  The Earth heliocentric 6-vector is mean equator and equinox
!     of date.  Position part, PV(1-3), is in AU;  velocity part,
!     PV(4-6), is in AU/sec.

!  3  Max/RMS errors 1950-2050:
!       13/5 E-5 AU = 19200/7600 km in position
!       47/26 E-10 AU/s = 0.0070/0.0039 km/s in speed

!  4  More precise results are obtainable with the routine sla_EVP.

!  P.T.Wallace   Starlink   23 November 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: iy,id
REAL :: fd,pv(6)

INTEGER :: iy4
REAL :: twopi,speed,remb,semb,yi,yf,t,elm,gamma,em,elt,eps0,  &
e,esq,v,r,elmm,coselt,sineps,coseps,w1,w2,selmm,celmm

PARAMETER (twopi=6.28318530718)

!  Mean orbital speed of Earth, AU/s
PARAMETER (speed=1.9913E-7)

!  Mean Earth:EMB distance and speed, AU and AU/s
PARAMETER (remb=3.12E-5,semb=8.31E-11)



!  Whole years & fraction of year, and years since J1900.0
yi=FLOAT(iy-1900)
iy4=MOD(MOD(iy,4)+4,4)
yf=(FLOAT(4*(id-1/(iy4+1))-iy4-2)+4.0*fd)/1461.0
t=yi+yf

!  Geometric mean longitude of Sun
!  (cf 4.881627938+6.283319509911*T MOD 2PI)
elm=MOD(4.881628+twopi*yf+0.00013420*t,twopi)

!  Mean longitude of perihelion
gamma=4.908230+3.0005E-4*t

!  Mean anomaly
em=elm-gamma

!  Mean obliquity
eps0=0.40931975-2.27E-6*t

!  Eccentricity
e=0.016751-4.2E-7*t
esq=e*e

!  True anomaly
v=em+2.0*e*SIN(em)+1.25*esq*SIN(2.0*em)

!  True ecliptic longitude
elt=v+gamma

!  True distance
r=(1.0-esq)/(1.0+e*COS(v))

!  Moon's mean longitude
elmm=MOD(4.72+83.9971*t,twopi)

!  Useful functions
coselt=COS(elt)
sineps=SIN(eps0)
coseps=COS(eps0)
w1=-r*SIN(elt)
w2=-speed*(coselt+e*COS(gamma))
selmm=SIN(elmm)
celmm=COS(elmm)

!  Earth position and velocity
pv(1)=-r*coselt-remb*celmm
pv(2)=(w1-remb*selmm)*coseps
pv(3)=w1*sineps
pv(4)=speed*(SIN(elt)+e*SIN(gamma))+semb*selmm
pv(5)=(w2-semb*celmm)*coseps
pv(6)=w2*sineps

END SUBROUTINE sla_earth
SUBROUTINE sla_ecleq (dl, db, date, dr, dd)
!+
!     - - - - - -
!      E C L E Q
!     - - - - - -

!  Transformation from ecliptic coordinates to
!  J2000.0 equatorial coordinates (double precision)

!  Given:
!     DL,DB       dp      ecliptic longitude and latitude
!                           (mean of date, IAU 1980 theory, radians)
!     DATE        dp      TDB (loosely ET) as Modified Julian Date
!                                              (JD-2400000.5)
!  Returned:
!     DR,DD       dp      J2000.0 mean RA,Dec (radians)

!  Called:
!     sla_DCS2C, sla_ECMAT, sla_DIMXV, sla_PREC, sla_EPJ, sla_DCC2S,
!     sla_DRANRM, sla_DRANGE

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dl,db,date,dr,dd

DOUBLE PRECISION :: rmat(3,3),v1(3),v2(3)



!  Spherical to Cartesian
CALL sla_dcs2c(dl,db,v1)

!  Ecliptic to equatorial
CALL sla_ecmat(date,rmat)
CALL sla_dimxv(rmat,v1,v2)

!  Mean of date to J2000
CALL sla_prec(2000D0,sla_epj(date),rmat)
CALL sla_dimxv(rmat,v2,v1)

!  Cartesian to spherical
CALL sla_dcc2s(v1,dr,dd)

!  Express in conventional ranges
dr=sla_dranrm(dr)
dd=sla_drange(dd)

END SUBROUTINE sla_ecleq
SUBROUTINE sla_ecmat (date, rmat)
!+
!     - - - - - -
!      E C M A T
!     - - - - - -

!  Form the equatorial to ecliptic rotation matrix - IAU 1980 theory
!  (double precision)

!  Given:
!     DATE     dp         TDB (loosely ET) as Modified Julian Date
!                                            (JD-2400000.5)
!  Returned:
!     RMAT     dp(3,3)    matrix

!  Reference:
!     Murray,C.A., Vectorial Astrometry, section 4.3.

!  Note:
!    The matrix is in the sense   V(ecl)  =  RMAT * V(equ);  the
!    equator, equinox and ecliptic are mean of date.

!  Called:  sla_DEULER

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,rmat(3,3)

!  Arc seconds to radians
DOUBLE PRECISION :: as2r
PARAMETER (as2r=0.484813681109535994D-5)

DOUBLE PRECISION :: t,eps0



!  Interval between basic epoch J2000.0 and current epoch (JC)
t = (date-51544.5D0)/36525D0

!  Mean obliquity
eps0 = as2r*(84381.448D0+(-46.8150D0+(-0.00059D0+0.001813D0*t)*t)*t)

!  Matrix
CALL sla_deuler('X',eps0,0D0,0D0,rmat)

END SUBROUTINE sla_ecmat
SUBROUTINE sla_ecor (rm, dm, iy, id, fd, rv, tl)
!+
!     - - - - -
!      E C O R
!     - - - - -

!  Component of Earth orbit velocity and heliocentric
!  light time in a given direction (single precision)

!  Given:
!     RM,DM    real    mean RA, Dec of date (radians)
!     IY       int     year
!     ID       int     day in year (1 = Jan 1st)
!     FD       real    fraction of day

!  Returned:
!     RV       real    component of Earth orbital velocity (km/sec)
!     TL       real    component of heliocentric light time (sec)

!  Notes:

!  1  The date and time is TDB (loosely ET) in a Julian calendar
!     which has been aligned to the ordinary Gregorian
!     calendar for the interval 1900 March 1 to 2100 February 28.
!     The year and day can be obtained by calling sla_CALYD or
!     sla_CLYD.

!  2  Sign convention:

!     The velocity component is +ve when the Earth is receding from
!     the given point on the sky.  The light time component is +ve
!     when the Earth lies between the Sun and the given point on
!     the sky.

!  3  Accuracy:

!     The velocity component is usually within 0.004 km/s of the
!     correct value and is never in error by more than 0.007 km/s.
!     The error in light time correction is about 0.03s at worst,
!     but is usually better than 0.01s. For applications requiring
!     higher accuracy, see the sla_EVP routine.

!  Called:  sla_EARTH, sla_CS2C, sla_VDV

!  P.T.Wallace   Starlink   24 November 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: rm,dm
INTEGER :: iy,id
REAL :: fd,rv,tl


REAL :: pv(6),v(3),aukm,ausec

!  AU to km and light sec (1985 Almanac)
PARAMETER (aukm=1.4959787066E8,ausec=499.0047837)



!  Sun:Earth position & velocity vector
CALL sla_earth(iy,id,fd,pv)

!  Star position vector
CALL sla_cs2c(rm,dm,v)

!  Velocity component
rv=-aukm*sla_vdv(pv(4),v)

!  Light time component
tl=ausec*sla_vdv(pv(1),v)

END SUBROUTINE sla_ecor
SUBROUTINE sla_eg50 (dr, dd, dl, db)
!+
!     - - - - -
!      E G 5 0
!     - - - - -

!  Transformation from B1950.0 'FK4' equatorial coordinates to
!  IAU 1958 galactic coordinates (double precision)

!  Given:
!     DR,DD       dp       B1950.0 'FK4' RA,Dec

!  Returned:
!     DL,DB       dp       galactic longitude and latitude L2,B2

!  (all arguments are radians)

!  Called:
!     sla_DCS2C, sla_DMXV, sla_DCC2S, sla_SUBET, sla_DRANRM, sla_DRANGE

!  Note:
!     The equatorial coordinates are B1950.0 'FK4'.  Use the
!     routine sla_EQGAL if conversion from J2000.0 coordinates
!     is required.

!  Reference:
!     Blaauw et al, Mon.Not.R.Astron.Soc.,121,123 (1960)

!  P.T.Wallace   Starlink   5 September 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dr,dd,dl,db

DOUBLE PRECISION :: v1(3),v2(3),r,d


!  L2,B2 system of galactic coordinates

!  P = 192.25       RA of galactic north pole (mean B1950.0)
!  Q =  62.6        inclination of galactic to mean B1950.0 equator
!  R =  33          longitude of ascending node

!  P,Q,R are degrees


!  Equatorial to galactic rotation matrix

!  The Euler angles are P, Q, 90-R, about the z then y then
!  z axes.

!         +CP.CQ.SR-SP.CR     +SP.CQ.SR+CP.CR     -SQ.SR

!         -CP.CQ.CR-SP.SR     -SP.CQ.CR+CP.SR     +SQ.CR

!         +CP.SQ              +SP.SQ              +CQ


DOUBLE PRECISION :: rmat(3,3)
DATA rmat(1,1),rmat(1,2),rmat(1,3),rmat(2,1),rmat(2,2),rmat(2,3),  &
rmat(3,1),rmat(3,2),rmat(3,3) /  &
-0.066988739415D0,-0.872755765852D0,-0.483538914632D0,  &
+0.492728466075D0,-0.450346958020D0,+0.744584633283D0,  &
-0.867600811151D0,-0.188374601723D0,+0.460199784784D0 /



!  Remove E-terms
CALL sla_subet(dr,dd,1950D0,r,d)

!  Spherical to Cartesian
CALL sla_dcs2c(r,d,v1)

!  Rotate to galactic
CALL sla_dmxv(rmat,v1,v2)

!  Cartesian to spherical
CALL sla_dcc2s(v2,dl,db)

!  Express angles in conventional ranges
dl=sla_dranrm(dl)
db=sla_drange(db)

END SUBROUTINE sla_eg50
DOUBLE PRECISION FUNCTION sla_epb (date)
!+
!     - - - -
!      E P B
!     - - - -

!  Conversion of Modified Julian Date to Besselian Epoch
!  (double precision)

!  Given:
!     DATE     dp       Modified Julian Date (JD - 2400000.5)

!  The result is the Besselian Epoch.

!  Reference:
!     Lieske,J.H., 1979. Astron.Astrophys.,73,282.

!  P.T.Wallace   Starlink   February 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date


sla_epb = 1900D0 + (date-15019.81352D0)/365.242198781D0

END FUNCTION sla_epb
DOUBLE PRECISION FUNCTION sla_epb2d (epb)
!+
!     - - - - - -
!      E P B 2 D
!     - - - - - -

!  Conversion of Besselian Epoch to Modified Julian Date
!  (double precision)

!  Given:
!     EPB      dp       Besselian Epoch

!  The result is the Modified Julian Date (JD - 2400000.5).

!  Reference:
!     Lieske,J.H., 1979. Astron.Astrophys.,73,282.

!  P.T.Wallace   Starlink   February 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: epb


sla_epb2d = 15019.81352D0 + (epb-1900D0)*365.242198781D0

END FUNCTION sla_epb2d
DOUBLE PRECISION FUNCTION sla_epco (k0, k, e)
!+
!     - - - - -
!      E P C O
!     - - - - -

!  Convert an epoch into the appropriate form - 'B' or 'J'

!  Given:
!     K0    char    form of result:  'B'=Besselian, 'J'=Julian
!     K     char    form of given epoch:  'B' or 'J'
!     E     dp      epoch

!  Called:  sla_EPB, sla_EPJ2D, sla_EPJ, sla_EPB2D

!  Notes:

!     1) The result is always either equal to or very close to
!        the given epoch E.  The routine is required only in
!        applications where punctilious treatment of heterogeneous
!        mixtures of star positions is necessary.

!     2) K0 and K are not validated.  They are interpreted as follows:

!        o  If K0 and K are the same the result is E.
!        o  If K0 is 'B' or 'b' and K isn't, the conversion is J to B.
!        o  In all other cases, the conversion is B to J.

!        Note that K0 and K won't match if their cases differ.

!  P.T.Wallace   Starlink   5 September 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) k0,k
DOUBLE PRECISION :: e


IF (k == k0) THEN
  sla_epco=e
ELSE IF (k0 == 'B'.OR.k0 == 'b') THEN
  sla_epco=sla_epb(sla_epj2d(e))
ELSE
  sla_epco=sla_epj(sla_epb2d(e))
END IF

END FUNCTION sla_epco
DOUBLE PRECISION FUNCTION sla_epj (date)
!+
!     - - - -
!      E P J
!     - - - -

!  Conversion of Modified Julian Date to Julian Epoch (double precision)

!  Given:
!     DATE     dp       Modified Julian Date (JD - 2400000.5)

!  The result is the Julian Epoch.

!  Reference:
!     Lieske,J.H., 1979. Astron.Astrophys.,73,282.

!  P.T.Wallace   Starlink   February 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date


sla_epj = 2000D0 + (date-51544.5D0)/365.25D0

END FUNCTION sla_epj
DOUBLE PRECISION FUNCTION sla_epj2d (epj)
!+
!     - - - - - -
!      E P J 2 D
!     - - - - - -

!  Conversion of Julian Epoch to Modified Julian Date (double precision)

!  Given:
!     EPJ      dp       Julian Epoch

!  The result is the Modified Julian Date (JD - 2400000.5).

!  Reference:
!     Lieske,J.H., 1979. Astron.Astrophys.,73,282.

!  P.T.Wallace   Starlink   February 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: epj


sla_epj2d = 51544.5D0 + (epj-2000D0)*365.25D0

END FUNCTION sla_epj2d
SUBROUTINE sla_eqecl (dr, dd, date, dl, db)
!+
!     - - - - - -
!      E Q E C L
!     - - - - - -

!  Transformation from J2000.0 equatorial coordinates to
!  ecliptic coordinates (double precision)

!  Given:
!     DR,DD       dp      J2000.0 mean RA,Dec (radians)
!     DATE        dp      TDB (loosely ET) as Modified Julian Date
!                                              (JD-2400000.5)
!  Returned:
!     DL,DB       dp      ecliptic longitude and latitude
!                         (mean of date, IAU 1980 theory, radians)

!  Called:
!     sla_DCS2C, sla_PREC, sla_EPJ, sla_DMXV, sla_ECMAT, sla_DCC2S,
!     sla_DRANRM, sla_DRANGE

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dr,dd,date,dl,db

DOUBLE PRECISION :: rmat(3,3),v1(3),v2(3)



!  Spherical to Cartesian
CALL sla_dcs2c(dr,dd,v1)

!  Mean J2000 to mean of date
CALL sla_prec(2000D0,sla_epj(date),rmat)
CALL sla_dmxv(rmat,v1,v2)

!  Equatorial to ecliptic
CALL sla_ecmat(date,rmat)
CALL sla_dmxv(rmat,v2,v1)

!  Cartesian to spherical
CALL sla_dcc2s(v1,dl,db)

!  Express in conventional ranges
dl=sla_dranrm(dl)
db=sla_drange(db)

END SUBROUTINE sla_eqecl
DOUBLE PRECISION FUNCTION sla_eqeqx (date)
!+
!     - - - - - -
!      E Q E Q X
!     - - - - - -

!  Equation of the equinoxes  (IAU 1994, double precision)

!  Given:
!     DATE    dp      TDB (loosely ET) as Modified Julian Date
!                                          (JD-2400000.5)

!  The result is the equation of the equinoxes (double precision)
!  in radians:

!     Greenwich apparent ST = GMST + sla_EQEQX

!  References:  IAU Resolution C7, Recommendation 3 (1994)
!               Capitaine, N. & Gontier, A.-M., Astron. Astrophys.,
!               275, 645-650 (1993)

!  Called:  sla_NUTC

!  Patrick Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date

!  Turns to arc seconds and arc seconds to radians
DOUBLE PRECISION :: t2as,as2r
PARAMETER (t2as=1296000D0,as2r=0.484813681109535994D-5)

DOUBLE PRECISION :: t,om,dpsi,deps,eps0



!  Interval between basic epoch J2000.0 and current epoch (JC)
t=(date-51544.5D0)/36525D0

!  Longitude of the mean ascending node of the lunar orbit on the
!   ecliptic, measured from the mean equinox of date
om=as2r*(450160.280D0+(-5D0*t2as-482890.539D0+(7.455D0+0.008D0*t)*t)*t)

!  Nutation
CALL sla_nutc(date,dpsi,deps,eps0)

!  Equation of the equinoxes
sla_eqeqx=dpsi*COS(eps0)+as2r*(0.00264D0*SIN(om)+0.000063D0*SIN(om+om))

END FUNCTION sla_eqeqx
SUBROUTINE sla_eqgal (dr, dd, dl, db)
!+
!     - - - - - -
!      E Q G A L
!     - - - - - -

!  Transformation from J2000.0 equatorial coordinates to
!  IAU 1958 galactic coordinates (double precision)

!  Given:
!     DR,DD       dp       J2000.0 RA,Dec

!  Returned:
!     DL,DB       dp       galactic longitude and latitude L2,B2

!  (all arguments are radians)

!  Called:
!     sla_DCS2C, sla_DMXV, sla_DCC2S, sla_DRANRM, sla_DRANGE

!  Note:
!     The equatorial coordinates are J2000.0.  Use the routine
!     sla_EG50 if conversion from B1950.0 'FK4' coordinates is
!     required.

!  Reference:
!     Blaauw et al, Mon.Not.R.Astron.Soc.,121,123 (1960)

!  P.T.Wallace   Starlink   November 1988

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dr,dd,dl,db

DOUBLE PRECISION :: v1(3),v2(3)


!  L2,B2 system of galactic coordinates

!  P = 192.25       RA of galactic north pole (mean B1950.0)
!  Q =  62.6        inclination of galactic to mean B1950.0 equator
!  R =  33          longitude of ascending node

!  P,Q,R are degrees

!  Equatorial to galactic rotation matrix (J2000.0), obtained by
!  applying the standard FK4 to FK5 transformation, for inertially
!  zero proper motion, to the columns of the B1950 equatorial to
!  galactic rotation matrix:

DOUBLE PRECISION :: rmat(3,3)
DATA rmat(1,1),rmat(1,2),rmat(1,3),rmat(2,1),rmat(2,2),rmat(2,3),  &
rmat(3,1),rmat(3,2),rmat(3,3)/  &
-0.054875539726D0,-0.873437108010D0,-0.483834985808D0,  &
+0.494109453312D0,-0.444829589425D0,+0.746982251810D0,  &
-0.867666135858D0,-0.198076386122D0,+0.455983795705D0/



!  Spherical to Cartesian
CALL sla_dcs2c(dr,dd,v1)

!  Equatorial to galactic
CALL sla_dmxv(rmat,v1,v2)

!  Cartesian to spherical
CALL sla_dcc2s(v2,dl,db)

!  Express in conventional ranges
dl=sla_dranrm(dl)
db=sla_drange(db)

END SUBROUTINE sla_eqgal
SUBROUTINE sla_etrms (ep, ev)
!+
!     - - - - - -
!      E T R M S
!     - - - - - -

!  Compute the E-terms (elliptic component of annual aberration)
!  vector (double precision)

!  Given:
!     EP      dp      Besselian epoch

!  Returned:
!     EV      dp(3)   E-terms as (dx,dy,dz)

!  Note the use of the J2000 aberration constant (20.49552 arcsec).
!  This is a reflection of the fact that the E-terms embodied in
!  existing star catalogues were computed from a variety of
!  aberration constants.  Rather than adopting one of the old
!  constants the latest value is used here.

!  References:
!     1  Smith, C.A. et al., 1989.  Astr.J. 97, 265.
!     2  Yallop, B.D. et al., 1989.  Astr.J. 97, 274.

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ep,ev(3)

!  Arcseconds to radians
DOUBLE PRECISION :: as2r
PARAMETER (as2r=0.484813681109535994D-5)

DOUBLE PRECISION :: t,e,e0,p,ek,cp



!  Julian centuries since B1950
t=(ep-1950D0)*1.00002135903D-2

!  Eccentricity
e=0.01673011D0-(0.00004193D0+0.000000126D0*t)*t

!  Mean obliquity
e0=(84404.836D0-(46.8495D0+(0.00319D0+0.00181D0*t)*t)*t)*as2r

!  Mean longitude of perihelion
p=(1015489.951D0+(6190.67D0+(1.65D0+0.012D0*t)*t)*t)*as2r

!  E-terms
ek=e*20.49552D0*as2r
cp=COS(p)
ev(1)= ek*SIN(p)
ev(2)=-ek*cp*COS(e0)
ev(3)=-ek*cp*SIN(e0)

END SUBROUTINE sla_etrms
SUBROUTINE sla_euler (order, phi, theta, psi, rmat)
!+
!     - - - - - -
!      E U L E R
!     - - - - - -

!  Form a rotation matrix from the Euler angles - three successive
!  rotations about specified Cartesian axes (single precision)

!  Given:
!    ORDER  c*(*)    specifies about which axes the rotations occur
!    PHI    r        1st rotation (radians)
!    THETA  r        2nd rotation (   "   )
!    PSI    r        3rd rotation (   "   )

!  Returned:
!    RMAT   r(3,3)   rotation matrix

!  A rotation is positive when the reference frame rotates
!  anticlockwise as seen looking towards the origin from the
!  positive region of the specified axis.

!  The characters of ORDER define which axes the three successive
!  rotations are about.  A typical value is 'ZXZ', indicating that
!  RMAT is to become the direction cosine matrix corresponding to
!  rotations of the reference frame through PHI radians about the
!  old Z-axis, followed by THETA radians about the resulting X-axis,
!  then PSI radians about the resulting Z-axis.

!  The axis names can be any of the following, in any order or
!  combination:  X, Y, Z, uppercase or lowercase, 1, 2, 3.  Normal
!  axis labelling/numbering conventions apply;  the xyz (=123)
!  triad is right-handed.  Thus, the 'ZXZ' example given above
!  could be written 'zxz' or '313' (or even 'ZxZ' or '3xZ').  ORDER
!  is terminated by length or by the first unrecognized character.

!  Fewer than three rotations are acceptable, in which case the later
!  angle arguments are ignored.  If all rotations are zero, the
!  identity matrix is produced.

!  Called:  sla_DEULER

!  P.T.Wallace   Starlink   23 May 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) order
REAL :: phi,theta,psi,rmat(3,3)

INTEGER :: j,i
DOUBLE PRECISION :: w(3,3)



!  Compute matrix in double precision
CALL sla_deuler(order,DBLE(phi),DBLE(theta),DBLE(psi),w)

!  Copy the result
DO j=1,3
  DO i=1,3
    rmat(i,j) = REAL(w(i,j))
  END DO
END DO

END SUBROUTINE sla_euler
SUBROUTINE sla_evp (date, deqx, dvb, dpb, dvh, dph)
!+
!     - - - -
!      E V P
!     - - - -

!  Barycentric and heliocentric velocity and position of the Earth

!  All arguments are double precision

!  Given:

!     DATE          TDB (loosely ET) as a Modified Julian Date
!                                         (JD-2400000.5)

!     DEQX          Julian Epoch (e.g. 2000.0D0) of mean equator and
!                   equinox of the vectors returned.  If DEQX .LE. 0D0,
!                   all vectors are referred to the mean equator and
!                   equinox (FK5) of epoch DATE.

!  Returned (all 3D Cartesian vectors):

!     DVB,DPB       barycentric velocity, position

!     DVH,DPH       heliocentric velocity, position

!  (Units are AU/s for velocity and AU for position)

!  Called:  sla_EPJ, sla_PREC

!  Accuracy:

!     The maximum deviations from the JPL DE96 ephemeris are as
!     follows:

!     barycentric velocity         0.42  m/s
!     barycentric position         6900  km

!     heliocentric velocity        0.42  m/s
!     heliocentric position        1600  km

!  This routine is adapted from the BARVEL and BARCOR
!  subroutines of P.Stumpff, which are described in
!  Astron. Astrophys. Suppl. Ser. 41, 1-8 (1980).  Most of the
!  changes are merely cosmetic and do not affect the results at
!  all.  However, some adjustments have been made so as to give
!  results that refer to the new (IAU 1976 'FK5') equinox
!  and precession, although the differences these changes make
!  relative to the results from Stumpff's original 'FK4' version
!  are smaller than the inherent accuracy of the algorithm.  One
!  minor shortcoming in the original routines that has NOT been
!  corrected is that better numerical accuracy could be achieved
!  if the various polynomial evaluations were nested.  Note also
!  that one of Stumpff's precession constants differs by 0.001 arcsec
!  from the value given in the Explanatory Supplement to the A.E.

!  P.T.Wallace   Starlink   21 November 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,deqx,dvb(3),dpb(3),dvh(3),dph(3)

INTEGER :: ideq,i,j,k

REAL :: cc2pi,ccsec3,ccsgd,cckm,ccmld,ccfdi,ccim,t,tsq,a,pertl,  &
pertld,pertr,pertrd,cosa,sina,esq,e,param,twoe,twog,g,  &
phi,f,sinf,cosf,phid,psid,pertp,pertpd,tl,sinlm,coslm,  &
sigma,b,plon,pomg,pecc,flatm,flat

DOUBLE PRECISION :: dc2pi,ds2r,dcsld,dc1mme,dt,dtsq,dlocal,dml,  &
deps,dparam,dpsi,d1pdro,drd,drld,dtl,dsinls,  &
dcosls,dxhd,dyhd,dzhd,dxbd,dybd,dzbd,dcosep,  &
dsinep,dyahd,dzahd,dyabd,dzabd,dr,dxh,dyh,dzh,dxb,dyb,dzb,dyah,dzah,dyab,  &
dzab,depj,deqcor,b1950

REAL :: sn(4),ccsel(3,17),ccamps(5,15),ccsec(3,4),ccampm(4,3),  &
ccpamv(4),ccpam(4),forbel(7),sorbel(17),sinlp(4),coslp(4)
EQUIVALENCE (sorbel(1),e),(forbel(1),g)

DOUBLE PRECISION :: dcfel(3,8),dceps(3),dcargs(2,15),dcargm(2,3),  &
dprema(3,3),w,vw(3)


PARAMETER (dc2pi=6.2831853071796D0,cc2pi=6.283185)
PARAMETER (ds2r=0.7272205216643D-4)
PARAMETER (b1950=1949.9997904423D0)


!   Constants DCFEL(I,K) of fast changing elements
!                     I=1                I=2              I=3
DATA dcfel/ 1.7400353D+00, 6.2833195099091D+02, 5.2796D-06,  &
6.2565836D+00, 6.2830194572674D+02,-2.6180D-06,  &
4.7199666D+00, 8.3997091449254D+03,-1.9780D-05,  &
1.9636505D-01, 8.4334662911720D+03,-5.6044D-05,  &
4.1547339D+00, 5.2993466764997D+01, 5.8845D-06,  &
4.6524223D+00, 2.1354275911213D+01, 5.6797D-06,  &
4.2620486D+00, 7.5025342197656D+00, 5.5317D-06,  &
1.4740694D+00, 3.8377331909193D+00, 5.6093D-06/


!   Constants DCEPS and CCSEL(I,K) of slowly changing elements
!                      I=1           I=2           I=3
DATA dceps/  4.093198D-01,-2.271110D-04,-2.860401D-08 /
DATA ccsel/  1.675104E-02,-4.179579E-05,-1.260516E-07,  &
2.220221E-01, 2.809917E-02, 1.852532E-05,  &
1.589963E+00, 3.418075E-02, 1.430200E-05,  &
2.994089E+00, 2.590824E-02, 4.155840E-06,  &
8.155457E-01, 2.486352E-02, 6.836840E-06,  &
1.735614E+00, 1.763719E-02, 6.370440E-06,  &
1.968564E+00, 1.524020E-02,-2.517152E-06,  &
1.282417E+00, 8.703393E-03, 2.289292E-05,  &
2.280820E+00, 1.918010E-02, 4.484520E-06,  &
4.833473E-02, 1.641773E-04,-4.654200E-07,  &
5.589232E-02,-3.455092E-04,-7.388560E-07,  &
4.634443E-02,-2.658234E-05, 7.757000E-08,  &
8.997041E-03, 6.329728E-06,-1.939256E-09,  &
2.284178E-02,-9.941590E-05, 6.787400E-08,  &
4.350267E-02,-6.839749E-05,-2.714956E-07,  &
1.348204E-02, 1.091504E-05, 6.903760E-07,  &
3.106570E-02,-1.665665E-04,-1.590188E-07/


!   Constants of the arguments of the short-period perturbations
!   by the planets:   DCARGS(I,K)
!                       I=1               I=2
DATA dcargs/ 5.0974222D+00,-7.8604195454652D+02,  &
3.9584962D+00,-5.7533848094674D+02,1.6338070D+00,-1.1506769618935D+03,  &
2.5487111D+00,-3.9302097727326D+02,4.9255514D+00,-5.8849265665348D+02,  &
1.3363463D+00,-5.5076098609303D+02,1.6072053D+00,-5.2237501616674D+02,  &
1.3629480D+00,-1.1790629318198D+03,5.5657014D+00,-1.0977134971135D+03,  &
5.0708205D+00,-1.5774000881978D+02,3.9318944D+00, 5.2963464780000D+01,  &
4.8989497D+00, 3.9809289073258D+01,1.3097446D+00, 7.7540959633708D+01,  &
3.5147141D+00, 7.9618578146517D+01,3.5413158D+00,-5.4868336758022D+02/


!   Amplitudes CCAMPS(N,K) of the short-period perturbations
!           N=1          N=2          N=3          N=4          N=5
DATA ccamps/  &
-2.279594E-5, 1.407414E-5, 8.273188E-6, 1.340565E-5,-2.490817E-7,  &
-3.494537E-5, 2.860401E-7, 1.289448E-7, 1.627237E-5,-1.823138E-7,  &
6.593466E-7, 1.322572E-5, 9.258695E-6,-4.674248E-7,-3.646275E-7,  &
1.140767E-5,-2.049792E-5,-4.747930E-6,-2.638763E-6,-1.245408E-7,  &
9.516893E-6,-2.748894E-6,-1.319381E-6,-4.549908E-6,-1.864821E-7,  &
7.310990E-6,-1.924710E-6,-8.772849E-7,-3.334143E-6,-1.745256E-7,  &
-2.603449E-6, 7.359472E-6, 3.168357E-6, 1.119056E-6,-1.655307E-7,  &
-3.228859E-6, 1.308997E-7, 1.013137E-7, 2.403899E-6,-3.736225E-7,  &
3.442177E-7, 2.671323E-6, 1.832858E-6,-2.394688E-7,-3.478444E-7,  &
8.702406E-6,-8.421214E-6,-1.372341E-6,-1.455234E-6,-4.998479E-8,  &
-1.488378E-6,-1.251789E-5, 5.226868E-7,-2.049301E-7, 0.0E0,  &
-8.043059E-6,-2.991300E-6, 1.473654E-7,-3.154542E-7, 0.0E0,  &
3.699128E-6,-3.316126E-6, 2.901257E-7, 3.407826E-7, 0.0E0,  &
2.550120E-6,-1.241123E-6, 9.901116E-8, 2.210482E-7, 0.0E0,  &
-6.351059E-7, 2.341650E-6, 1.061492E-6, 2.878231E-7, 0.0E0/


!   Constants of the secular perturbations in longitude
!   CCSEC3 and CCSEC(N,K)
!                      N=1           N=2           N=3
DATA ccsec3/-7.757020E-08/,ccsec/  1.289600E-06, 5.550147E-01, 2.076942E+00,  &
3.102810E-05, 4.035027E+00, 3.525565E-01,  &
9.124190E-06, 9.990265E-01, 2.622706E+00,  &
9.793240E-07, 5.508259E+00, 1.559103E+01/

!   Sidereal rate DCSLD in longitude, rate CCSGD in mean anomaly
DATA dcsld/1.990987D-07/,ccsgd/1.990969E-07/

!   Some constants used in the calculation of the lunar contribution
DATA cckm/3.122140E-05/,ccmld/2.661699E-06/,  &
ccfdi/2.399485E-07/


!   Constants DCARGM(I,K) of the arguments of the perturbations
!   of the motion of the Moon
!                       I=1               I=2
DATA dcargm/  5.1679830D+00, 8.3286911095275D+03,  &
5.4913150D+00,-7.2140632838100D+03,5.9598530D+00, 1.5542754389685D+04/


!   Amplitudes CCAMPM(N,K) of the perturbations of the Moon
!            N=1          N=2           N=3           N=4
DATA ccampm/1.097594E-01, 2.896773E-07, 5.450474E-02, 1.438491E-07,  &
-2.223581E-02, 5.083103E-08, 1.002548E-02,-2.291823E-08,  &
1.148966E-02, 5.658888E-08, 8.249439E-03, 4.063015E-08/


!   CCPAMV(K)=A*M*DL/DT (planets), DC1MME=1-MASS(Earth+Moon)
DATA ccpamv/8.326827E-11,1.843484E-11,1.988712E-12,1.881276E-12/
DATA dc1mme/0.99999696D0/

!   CCPAM(K)=A*M(planets), CCIM=INCLINATION(Moon)
DATA ccpam/4.960906E-3,2.727436E-3,8.392311E-4,1.556861E-3/
DATA ccim/8.978749E-2/





!   EXECUTION
!   ---------

!   Control parameter IDEQ, and time arguments
ideq = 0
IF (deqx > 0D0) ideq=1
dt = (date-15019.5D0)/36525D0
t = REAL(dt)
dtsq = dt*dt
tsq = REAL(dtsq)

!   Values of all elements for the instant DATE
DO k=1,8
  dlocal = MOD(dcfel(1,k)+dt*dcfel(2,k)+dtsq*dcfel(3,k), dc2pi)
  IF (k == 1) THEN
    dml = dlocal
  ELSE
    forbel(k-1) = REAL(dlocal)
  END IF
END DO
deps = MOD(dceps(1)+dt*dceps(2)+dtsq*dceps(3), dc2pi)
DO k=1,17
  sorbel(k) = MOD(ccsel(1,k)+t*ccsel(2,k)+tsq*ccsel(3,k),cc2pi)
END DO

!   Secular perturbations in longitude
DO k=1,4
  a = MOD(ccsec(2,k)+t*ccsec(3,k), cc2pi)
  sn(k) = SIN(a)
END DO

!   Periodic perturbations of the EMB (Earth-Moon barycentre)
pertl =  ccsec(1,1)          *sn(1) +ccsec(1,2)*sn(2)+  &
(ccsec(1,3)+t*ccsec3)*sn(3) +ccsec(1,4)*sn(4)
pertld = 0.0
pertr = 0.0
pertrd = 0.0
DO k=1,15
  a = SNGL(MOD(dcargs(1,k)+dt*dcargs(2,k), dc2pi))
  cosa = COS(a)
  sina = SIN(a)
  pertl = pertl + ccamps(1,k)*cosa+ccamps(2,k)*sina
  pertr = pertr + ccamps(3,k)*cosa+ccamps(4,k)*sina
  IF (k < 11) THEN
    pertld = pertld+(ccamps(2,k)*cosa-ccamps(1,k)*sina)*ccamps(5,k)
    pertrd = pertrd+(ccamps(4,k)*cosa-ccamps(3,k)*sina)*ccamps(5,k)
  END IF
END DO

!   Elliptic part of the motion of the EMB
esq = e*e
dparam = 1D0-DBLE(esq)
param = REAL(dparam)
twoe = e+e
twog = g+g
phi = twoe*((1.0-esq*0.125)*SIN(g)+e*0.625*SIN(twog)  &
+esq*0.5416667*SIN(g+twog) )
f = g+phi
sinf = SIN(f)
cosf = COS(f)
dpsi = dparam/(1D0+DBLE(e*cosf))
phid = twoe*ccsgd*((1.0+esq*1.5)*cosf+e*(1.25-sinf*sinf*0.5))
psid = ccsgd*e*sinf/SQRT(param)

!   Perturbed heliocentric motion of the EMB
d1pdro = 1D0+DBLE(pertr)
drd = d1pdro*(DBLE(psid)+dpsi*DBLE(pertrd))
drld = d1pdro*dpsi*(dcsld+DBLE(phid)+DBLE(pertld))
dtl = MOD(dml+DBLE(phi)+DBLE(pertl), dc2pi)
dsinls = SIN(dtl)
dcosls = COS(dtl)
dxhd = drd*dcosls-drld*dsinls
dyhd = drd*dsinls+drld*dcosls

!   Influence of eccentricity, evection and variation on the
!   geocentric motion of the Moon
pertl = 0.0
pertld = 0.0
pertp = 0.0
pertpd = 0.0
DO k=1,3
  a = SNGL(MOD(dcargm(1,k)+dt*dcargm(2,k), dc2pi))
  sina = SIN(a)
  cosa = COS(a)
  pertl = pertl +ccampm(1,k)*sina
  pertld = pertld+ccampm(2,k)*cosa
  pertp = pertp +ccampm(3,k)*cosa
  pertpd = pertpd-ccampm(4,k)*sina
END DO

!   Heliocentric motion of the Earth
tl = forbel(2)+pertl
sinlm = SIN(tl)
coslm = COS(tl)
sigma = cckm/(1.0+pertp)
a = sigma*(ccmld+pertld)
b = sigma*pertpd
dxhd = dxhd+DBLE(a*sinlm)+DBLE(b*coslm)
dyhd = dyhd-DBLE(a*coslm)+DBLE(b*sinlm)
dzhd =     -DBLE(sigma*ccfdi*COS(forbel(3)))

!   Barycentric motion of the Earth
dxbd = dxhd*dc1mme
dybd = dyhd*dc1mme
dzbd = dzhd*dc1mme
DO k=1,4
  plon = forbel(k+3)
  pomg = sorbel(k+1)
  pecc = sorbel(k+9)
  tl = MOD(plon+2.0*pecc*SIN(plon-pomg), cc2pi)
  sinlp(k) = SIN(tl)
  coslp(k) = COS(tl)
  dxbd = dxbd+DBLE(ccpamv(k)*(sinlp(k)+pecc*SIN(pomg)))
  dybd = dybd-DBLE(ccpamv(k)*(coslp(k)+pecc*COS(pomg)))
  dzbd = dzbd-DBLE(ccpamv(k)*sorbel(k+13)*COS(plon-sorbel(k+5)))
END DO

!   Transition to mean equator of date
dcosep = COS(deps)
dsinep = SIN(deps)
dyahd = dcosep*dyhd-dsinep*dzhd
dzahd = dsinep*dyhd+dcosep*dzhd
dyabd = dcosep*dybd-dsinep*dzbd
dzabd = dsinep*dybd+dcosep*dzbd

!   Heliocentric coordinates of the Earth
dr = dpsi*d1pdro
flatm = ccim*SIN(forbel(3))
a = sigma*COS(flatm)
dxh = dr*dcosls-DBLE(a*coslm)
dyh = dr*dsinls-DBLE(a*sinlm)
dzh =          -DBLE(sigma*SIN(flatm))

!   Barycentric coordinates of the Earth
dxb = dxh*dc1mme
dyb = dyh*dc1mme
dzb = dzh*dc1mme
DO k=1,4
  flat = sorbel(k+13)*SIN(forbel(k+3)-sorbel(k+5))
  a = ccpam(k)*(1.0-sorbel(k+9)*COS(forbel(k+3)-sorbel(k+1)))
  b = a*COS(flat)
  dxb = dxb-DBLE(b*coslp(k))
  dyb = dyb-DBLE(b*sinlp(k))
  dzb = dzb-DBLE(a*SIN(flat))
END DO

!   Transition to mean equator of date
dyah = dcosep*dyh-dsinep*dzh
dzah = dsinep*dyh+dcosep*dzh
dyab = dcosep*dyb-dsinep*dzb
dzab = dsinep*dyb+dcosep*dzb

!   Copy result components into vectors, correcting for FK4 equinox
depj=sla_epj(date)
deqcor = ds2r*(0.035D0+0.00085D0*(depj-b1950))
dvh(1) = dxhd-deqcor*dyahd
dvh(2) = dyahd+deqcor*dxhd
dvh(3) = dzahd
dvb(1) = dxbd-deqcor*dyabd
dvb(2) = dyabd+deqcor*dxbd
dvb(3) = dzabd
dph(1) = dxh-deqcor*dyah
dph(2) = dyah+deqcor*dxh
dph(3) = dzah
dpb(1) = dxb-deqcor*dyab
dpb(2) = dyab+deqcor*dxb
dpb(3) = dzab

!   Was precession to another equinox requested?
IF (ideq /= 0) THEN
  
!     Yes: compute precession matrix from MJD DATE to Julian epoch DEQX
  CALL sla_prec(depj,deqx,dprema)
  
!     Rotate DVH
  DO j=1,3
    w=0D0
    DO i=1,3
      w=w+dprema(j,i)*dvh(i)
    END DO
    vw(j)=w
  END DO
  DO j=1,3
    dvh(j)=vw(j)
  END DO
  
!     Rotate DVB
  DO j=1,3
    w=0D0
    DO i=1,3
      w=w+dprema(j,i)*dvb(i)
    END DO
    vw(j)=w
  END DO
  DO j=1,3
    dvb(j)=vw(j)
  END DO
  
!     Rotate DPH
  DO j=1,3
    w=0D0
    DO i=1,3
      w=w+dprema(j,i)*dph(i)
    END DO
    vw(j)=w
  END DO
  DO j=1,3
    dph(j)=vw(j)
  END DO
  
!     Rotate DPB
  DO j=1,3
    w=0D0
    DO i=1,3
      w=w+dprema(j,i)*dpb(i)
    END DO
    vw(j)=w
  END DO
  DO j=1,3
    dpb(j)=vw(j)
  END DO
END IF

END SUBROUTINE sla_evp
SUBROUTINE sla_fitxy (itype,np,xye,xym,coeffs,j)
!+
!     - - - - - -
!      F I T X Y
!     - - - - - -

!  Fit a linear model to relate two sets of [X,Y] coordinates.

!  Given:
!     ITYPE    i        type of model: 4 or 6 (note 1)
!     NP       i        number of samples (note 2)
!     XYE     d(2,np)   expected [X,Y] for each sample
!     XYM     d(2,np)   measured [X,Y] for each sample

!  Returned:
!     COEFFS  d(6)      coefficients of model (note 3)
!     J        i        status:  0 = OK
!                               -1 = illegal ITYPE
!                               -2 = insufficient data
!                               -3 = singular solution

!  Notes:

!  1)  ITYPE, which must be either 4 or 6, selects the type of model
!      fitted.  Both allowed ITYPE values produce a model COEFFS which
!      consists of six coefficients, namely the zero points and, for
!      each of XE and YE, the coefficient of XM and YM.  For ITYPE=6,
!      all six coefficients are independent, modelling squash and shear
!      as well as origin, scale, and orientation.  However, ITYPE=4
!      selects the "solid body rotation" option;  the model COEFFS
!      still consists of the same six coefficients, but now two of
!      them are used twice (appropriately signed).  Origin, scale
!      and orientation are still modelled, but not squash or shear -
!      the units of X and Y have to be the same.

!  2)  For NC=4, NP must be at least 2.  For NC=6, NP must be at
!      least 3.

!  3)  The model is returned in the array COEFFS.  Naming the
!      elements of COEFFS as follows:

!                  COEFFS(1) = A
!                  COEFFS(2) = B
!                  COEFFS(3) = C
!                  COEFFS(4) = D
!                  COEFFS(5) = E
!                  COEFFS(6) = F

!      the model is:

!            XE = A + B*XM + C*YM
!            YE = D + E*XM + F*YM

!      For the "solid body rotation" option (ITYPE=4), the
!      magnitudes of B and F, and of C and E, are equal.  The
!      signs of these coefficients depend on whether there is a
!      sign reversal between XE,YE and XM,YM;  fits are performed
!      with and without a sign reversal and the best one chosen.

!  4)  Error status values J=-1 and -2 leave COEFFS unchanged;
!      if J=-3 COEFFS may have been changed.

!  See also sla_PXY, sla_INVF, sla_XY2XY, sla_DCMPF

!  Called:  sla_DMAT, sla_DMXV

!  P.T.Wallace   Starlink   11 February 1991

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: itype,np
DOUBLE PRECISION :: xye(2,np),xym(2,np),coeffs(6)
INTEGER :: j

INTEGER :: i,jstat,iw(4),nsol
DOUBLE PRECISION :: p,sxe,sxexm,sxeym,sye,syeym,syexm,sxm,  &
sym,sxmxm,sxmym,symym,xe,ye,xm,ym,v(4),dm3(3,3),dm4(4,4),det,  &
sgn,sxxyy,sxyyx,sx2y2,a,b,c,d,sdr2,xr,yr,aold,bold,cold,dold,sold



!  Preset the status
j=0

!  Float the number of samples
p=DBLE(np)

!  Check ITYPE
IF (itype == 6) THEN
  
  
!     Six-coefficient linear model
!     ----------------------------
  
!     Check enough samples
  IF (np >= 3) THEN
    
!     Form summations
    sxe=0D0
    sxexm=0D0
    sxeym=0D0
    sye=0D0
    syeym=0D0
    syexm=0D0
    sxm=0D0
    sym=0D0
    sxmxm=0D0
    sxmym=0D0
    symym=0D0
    DO i=1,np
      xe=xye(1,i)
      ye=xye(2,i)
      xm=xym(1,i)
      ym=xym(2,i)
      sxe=sxe+xe
      sxexm=sxexm+xe*xm
      sxeym=sxeym+xe*ym
      sye=sye+ye
      syeym=syeym+ye*ym
      syexm=syexm+ye*xm
      sxm=sxm+xm
      sym=sym+ym
      sxmxm=sxmxm+xm*xm
      sxmym=sxmym+xm*ym
      symym=symym+ym*ym
    END DO
    
!        Solve for A,B,C in  XE = A + B*XM + C*YM
    v(1)=sxe
    v(2)=sxexm
    v(3)=sxeym
    dm3(1,1)=p
    dm3(1,2)=sxm
    dm3(1,3)=sym
    dm3(2,1)=sxm
    dm3(2,2)=sxmxm
    dm3(2,3)=sxmym
    dm3(3,1)=sym
    dm3(3,2)=sxmym
    dm3(3,3)=symym
    CALL sla_dmat(3,dm3,v,det,jstat,iw)
    IF (jstat == 0) THEN
      DO i=1,3
        coeffs(i)=v(i)
      END DO
      
!           Solve for D,E,F in  YE = D + E*XM + F*YM
      v(1)=sye
      v(2)=syexm
      v(3)=syeym
      CALL sla_dmxv(dm3,v,coeffs(4))
      
    ELSE
      
!           No 6-coefficient solution possible
      j=-3
      
    END IF
    
  ELSE
    
!        Insufficient data for 6-coefficient fit
    j=-2
    
  END IF
  
ELSE IF (itype == 4) THEN
  
  
!     Four-coefficient solid body rotation model
!     ------------------------------------------
  
!     Check enough samples
  IF (np >= 2) THEN
    
!        Try two solutions, first without then with flip in X
    DO nsol=1,2
      IF (nsol == 1) THEN
        sgn=1D0
      ELSE
        sgn=-1D0
      END IF
      
!           Form summations
      sxe=0D0
      sxxyy=0D0
      sxyyx=0D0
      sye=0D0
      sxm=0D0
      sym=0D0
      sx2y2=0D0
      DO i=1,np
        xe=xye(1,i)*sgn
        ye=xye(2,i)
        xm=xym(1,i)
        ym=xym(2,i)
        sxe=sxe+xe
        sxxyy=sxxyy+xe*xm+ye*ym
        sxyyx=sxyyx+xe*ym-ye*xm
        sye=sye+ye
        sxm=sxm+xm
        sym=sym+ym
        sx2y2=sx2y2+xm*xm+ym*ym
      END DO
      
      
!           Solve for A,B,C,D in:  +/- XE = A + B*XM - C*YM
!                                    + YE = D + C*XM + B*YM
      v(1)=sxe
      v(2)=sxxyy
      v(3)=sxyyx
      v(4)=sye
      dm4(1,1)=p
      dm4(1,2)=sxm
      dm4(1,3)=-sym
      dm4(1,4)=0D0
      dm4(2,1)=sxm
      dm4(2,2)=sx2y2
      dm4(2,3)=0D0
      dm4(2,4)=sym
      dm4(3,1)=sym
      dm4(3,2)=0D0
      dm4(3,3)=-sx2y2
      dm4(3,4)=-sxm
      dm4(4,1)=0D0
      dm4(4,2)=sym
      dm4(4,3)=sxm
      dm4(4,4)=p
      CALL sla_dmat(4,dm4,v,det,jstat,iw)
      IF (jstat == 0) THEN
        a=v(1)
        b=v(2)
        c=v(3)
        d=v(4)
        
!              Determine sum of radial errors squared
        sdr2=0D0
        DO i=1,np
          xm=xym(1,i)
          ym=xym(2,i)
          xr=a+b*xm-c*ym-xye(1,i)*sgn
          yr=d+c*xm+b*ym-xye(2,i)
          sdr2=sdr2+xr*xr+yr*yr
        END DO
        
      ELSE
        
!              Singular: set flag
        sdr2=-1D0
        
      END IF
      
!           If first pass and non-singular, save variables
      IF (nsol == 1.AND.jstat == 0) THEN
        aold=a
        bold=b
        cold=c
        dold=d
        sold=sdr2
      END IF
      
    END DO
    
!        Pick the best of the two solutions
    IF (sold >= 0D0.AND.(sold <= sdr2.OR.np == 2)) THEN
      coeffs(1)=aold
      coeffs(2)=bold
      coeffs(3)=-cold
      coeffs(4)=dold
      coeffs(5)=cold
      coeffs(6)=bold
    ELSE IF (jstat == 0) THEN
      coeffs(1)=-a
      coeffs(2)=-b
      coeffs(3)=c
      coeffs(4)=d
      coeffs(5)=c
      coeffs(6)=b
    ELSE
      
!           No 4-coefficient fit possible
      j=-3
    END IF
  ELSE
    
!        Insufficient data for 4-coefficient fit
    j=-2
  END IF
ELSE
  
!     Illegal ITYPE - not 4 or 6
  j=-1
END IF

END SUBROUTINE sla_fitxy
SUBROUTINE sla_fk425 (r1950,d1950,dr1950,dd1950,p1950,v1950,  &
r2000,d2000,dr2000,dd2000,p2000,v2000)
!+
!     - - - - - -
!      F K 4 2 5
!     - - - - - -

!  Convert B1950.0 FK4 star data to J2000.0 FK5 (double precision)

!  This routine converts stars from the old, Bessel-Newcomb, FK4
!  system to the new, IAU 1976, FK5, Fricke system.  The precepts
!  of Smith et al (Ref 1) are followed, using the implementation
!  by Yallop et al (Ref 2) of a matrix method due to Standish.
!  Kinoshita's development of Andoyer's post-Newcomb precession is
!  used.  The numerical constants from Seidelmann et al (Ref 3) are
!  used canonically.

!  Given:  (all B1950.0,FK4)
!     R1950,D1950     dp    B1950.0 RA,Dec (rad)
!     DR1950,DD1950   dp    B1950.0 proper motions (rad/trop.yr)
!     P1950           dp    parallax (arcsec)
!     V1950           dp    radial velocity (km/s, +ve = moving away)

!  Returned:  (all J2000.0,FK5)
!     R2000,D2000     dp    J2000.0 RA,Dec (rad)
!     DR2000,DD2000   dp    J2000.0 proper motions (rad/Jul.yr)
!     P2000           dp    parallax (arcsec)
!     V2000           dp    radial velocity (km/s, +ve = moving away)

!  Notes:

!  1)  The proper motions in RA are dRA/dt rather than
!      cos(Dec)*dRA/dt, and are per year rather than per century.

!  2)  Conversion from Besselian epoch 1950.0 to Julian epoch
!      2000.0 only is provided for.  Conversions involving other
!      epochs will require use of the appropriate precession,
!      proper motion, and E-terms routines before and/or
!      after FK425 is called.

!  3)  In the FK4 catalogue the proper motions of stars within
!      10 degrees of the poles do not embody the differential
!      E-term effect and should, strictly speaking, be handled
!      in a different manner from stars outside these regions.
!      However, given the general lack of homogeneity of the star
!      data available for routine astrometry, the difficulties of
!      handling positions that may have been determined from
!      astrometric fields spanning the polar and non-polar regions,
!      the likelihood that the differential E-terms effect was not
!      taken into account when allowing for proper motion in past
!      astrometry, and the undesirability of a discontinuity in
!      the algorithm, the decision has been made in this routine to
!      include the effect of differential E-terms on the proper
!      motions for all stars, whether polar or not.  At epoch 2000,
!      and measuring on the sky rather than in terms of dRA, the
!      errors resulting from this simplification are less than
!      1 milliarcsecond in position and 1 milliarcsecond per
!      century in proper motion.

!  References:

!     1  Smith, C.A. et al, 1989.  "The transformation of astrometric
!        catalog systems to the equinox J2000.0".  Astron.J. 97, 265.

!     2  Yallop, B.D. et al, 1989.  "Transformation of mean star places
!        from FK4 B1950.0 to FK5 J2000.0 using matrices in 6-space".
!        Astron.J. 97, 274.

!     3  Seidelmann, P.K. (ed), 1992.  "Explanatory Supplement to
!        the Astronomical Almanac", ISBN 0-935702-68-7.

!  P.T.Wallace   Starlink   19 December 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: r1950,d1950,dr1950,dd1950,p1950,v1950,  &
r2000,d2000,dr2000,dd2000,p2000,v2000


!  Miscellaneous
DOUBLE PRECISION :: r,d,ur,ud,px,rv,sr,cr,sd,cd,w,wd
DOUBLE PRECISION :: x,y,z,xd,yd,zd
DOUBLE PRECISION :: rxysq,rxyzsq,rxy,rxyz,spxy,spxyz
INTEGER :: i,j

!  Star position and velocity vectors
DOUBLE PRECISION :: r0(3),rd0(3)

!  Combined position and velocity vectors
DOUBLE PRECISION :: v1(6),v2(6)

!  2Pi
DOUBLE PRECISION :: d2pi
PARAMETER (d2pi=6.283185307179586476925287D0)

!  Radians per year to arcsec per century
DOUBLE PRECISION :: pmf
PARAMETER (pmf=100D0*60D0*60D0*360D0/d2pi)

!  Small number to avoid arithmetic problems
DOUBLE PRECISION :: tiny
PARAMETER (tiny=1D-30)



!  CANONICAL CONSTANTS  (see references)


!  Km per sec to AU per tropical century
!  = 86400 * 36524.2198782 / 149597870
DOUBLE PRECISION :: vf
PARAMETER (vf=21.095D0)

!  Constant vector and matrix (by columns)
DOUBLE PRECISION :: a(3),ad(3),em(6,6)
DATA a,ad/ -1.62557D-6,  -0.31919D-6, -0.13843D-6,  &
+1.245D-3,    -1.580D-3,   -0.659D-3/

DATA (em(i,1),i=1,6) / +0.9999256782D0,+0.0111820610D0,  &
+0.0048579479D0,-0.000551D0,  &
+0.238514D0,-0.435623D0 /

DATA (em(i,2),i=1,6) / -0.0111820611D0,+0.9999374784D0,  &
-0.0000271474D0,-0.238565D0,  &
-0.002667D0,+0.012254D0 /

DATA (em(i,3),i=1,6) / -0.0048579477D0,-0.0000271765D0,  &
+0.9999881997D0,+0.435739D0,  &
-0.008541D0,+0.002117D0 /

DATA (em(i,4),i=1,6) / +0.00000242395018D0,+0.00000002710663D0,  &
+0.00000001177656D0,+0.99994704D0,  &
+0.01118251D0,+0.00485767D0 /

DATA (em(i,5),i=1,6) / -0.00000002710663D0,+0.00000242397878D0,  &
-0.00000000006582D0,-0.01118251D0,  &
+0.99995883D0,-0.00002714D0 /

DATA (em(i,6),i=1,6) / -0.00000001177656D0,-0.00000000006587D0,  &
+0.00000242410173D0,-0.00485767D0,  &
-0.00002718D0,+1.00000956D0 /



!  Pick up B1950 data (units radians and arcsec/TC)
r=r1950
d=d1950
ur=dr1950*pmf
ud=dd1950*pmf
px=p1950
rv=v1950

!  Spherical to Cartesian
sr=SIN(r)
cr=COS(r)
sd=SIN(d)
cd=COS(d)

r0(1)=cr*cd
r0(2)=sr*cd
r0(3)=   sd

w=vf*rv*px

rd0(1)=-sr*cd*ur-cr*sd*ud+w*r0(1)
rd0(2)= cr*cd*ur-sr*sd*ud+w*r0(2)
rd0(3)=             cd*ud+w*r0(3)

!  Allow for e-terms and express as position+velocity 6-vector
w=r0(1)*a(1)+r0(2)*a(2)+r0(3)*a(3)
wd=r0(1)*ad(1)+r0(2)*ad(2)+r0(3)*ad(3)
DO i=1,3
  v1(i)=r0(i)-a(i)+w*r0(i)
  v1(i+3)=rd0(i)-ad(i)+wd*r0(i)
END DO

!  Convert position+velocity vector to Fricke system
DO i=1,6
  w=0D0
  DO j=1,6
    w=w+em(i,j)*v1(j)
  END DO
  v2(i)=w
END DO

!  Revert to spherical coordinates
x=v2(1)
y=v2(2)
z=v2(3)
xd=v2(4)
yd=v2(5)
zd=v2(6)

rxysq=x*x+y*y
rxyzsq=rxysq+z*z
rxy=SQRT(rxysq)
rxyz=SQRT(rxyzsq)

spxy=x*xd+y*yd
spxyz=spxy+z*zd

IF (x == 0D0.AND.y == 0D0) THEN
  r=0D0
ELSE
  r=ATAN2(y,x)
  IF (r < 0.0D0) r=r+d2pi
END IF
d=ATAN2(z,rxy)

IF (rxy > tiny) THEN
  ur=(x*yd-y*xd)/rxysq
  ud=(zd*rxysq-z*spxy)/(rxyzsq*rxy)
END IF

IF (px > tiny) THEN
  rv=spxyz/(px*rxyz*vf)
  px=px/rxyz
END IF

!  Return results
r2000=r
d2000=d
dr2000=ur/pmf
dd2000=ud/pmf
v2000=rv
p2000=px

END SUBROUTINE sla_fk425
SUBROUTINE sla_fk45z (r1950,d1950,bepoch,r2000,d2000)
!+
!     - - - - - -
!      F K 4 5 Z
!     - - - - - -

!  Convert B1950.0 FK4 star data to J2000.0 FK5 assuming zero
!  proper motion in an inertial frame (double precision)

!  This routine converts stars from the old, Bessel-Newcomb, FK4
!  system to the new, IAU 1976, FK5, Fricke system, in such a
!  way that the FK5 proper motion is zero.  Because such a star
!  has, in general, a non-zero proper motion in the FK4 system,
!  the routine requires the epoch at which the position in the
!  FK4 system was determined.

!  The method is from Appendix 2 of Ref 1, but using the constants
!  of Ref 4.

!  Given:
!     R1950,D1950     dp    B1950.0 FK4 RA,Dec at epoch (rad)
!     BEPOCH          dp    Besselian epoch (e.g. 1979.3D0)

!  Returned:
!     R2000,D2000     dp    J2000.0 FK5 RA,Dec (rad)

!  Notes:

!  1)  The epoch BEPOCH is strictly speaking Besselian, but
!      if a Julian epoch is supplied the result will be
!      affected only to a negligible extent.

!  2)  Conversion from Besselian epoch 1950.0 to Julian epoch
!      2000.0 only is provided for.  Conversions involving other
!      epochs will require use of the appropriate precession,
!      proper motion, and E-terms routines before and/or
!      after FK45Z is called.

!  3)  In the FK4 catalogue the proper motions of stars within
!      10 degrees of the poles do not embody the differential
!      E-term effect and should, strictly speaking, be handled
!      in a different manner from stars outside these regions.
!      However, given the general lack of homogeneity of the star
!      data available for routine astrometry, the difficulties of
!      handling positions that may have been determined from
!      astrometric fields spanning the polar and non-polar regions,
!      the likelihood that the differential E-terms effect was not
!      taken into account when allowing for proper motion in past
!      astrometry, and the undesirability of a discontinuity in
!      the algorithm, the decision has been made in this routine to
!      include the effect of differential E-terms on the proper
!      motions for all stars, whether polar or not.  At epoch 2000,
!      and measuring on the sky rather than in terms of dRA, the
!      errors resulting from this simplification are less than
!      1 milliarcsecond in position and 1 milliarcsecond per
!      century in proper motion.

!  References:

!     1  Aoki,S., et al, 1983.  Astron.Astrophys., 128, 263.

!     2  Smith, C.A. et al, 1989.  "The transformation of astrometric
!        catalog systems to the equinox J2000.0".  Astron.J. 97, 265.

!     3  Yallop, B.D. et al, 1989.  "Transformation of mean star places
!        from FK4 B1950.0 to FK5 J2000.0 using matrices in 6-space".
!        Astron.J. 97, 274.

!     4  Seidelmann, P.K. (ed), 1992.  "Explanatory Supplement to
!        the Astronomical Almanac", ISBN 0-935702-68-7.

!  Called:  sla_DCS2C, sla_EPJ, sla_EPB2D, sla_DCC2S, sla_DRANRM

!  P.T.Wallace   Starlink   5 November 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: r1950,d1950,bepoch,r2000,d2000

DOUBLE PRECISION :: d2pi
PARAMETER (d2pi=6.283185307179586476925287D0)

DOUBLE PRECISION :: w
INTEGER :: i,j

!  Position and position+velocity vectors
DOUBLE PRECISION :: r0(3),a1(3),v1(3),v2(6)

!  Radians per year to arcsec per century
DOUBLE PRECISION :: pmf
PARAMETER (pmf=100D0*60D0*60D0*360D0/d2pi)

!  Functions


!  CANONICAL CONSTANTS  (see references)


!  Vectors A and Adot, and matrix M (only half of which is needed here)
DOUBLE PRECISION :: a(3),ad(3),em(6,3)
DATA a,ad/ -1.62557D-6,  -0.31919D-6, -0.13843D-6,  &
+1.245D-3,    -1.580D-3,   -0.659D-3/

DATA (em(i,1),i=1,6) / +0.9999256782D0,+0.0111820610D0,  &
+0.0048579479D0,-0.000551D0,  &
+0.238514D0,-0.435623D0 /

DATA (em(i,2),i=1,6) / -0.0111820611D0,+0.9999374784D0,  &
-0.0000271474D0,-0.238565D0,  &
-0.002667D0,+0.012254D0 /

DATA (em(i,3),i=1,6) / -0.0048579477D0,-0.0000271765D0,  &
+0.9999881997D0,+0.435739D0,  &
-0.008541D0,+0.002117D0 /



!  Spherical to Cartesian
CALL sla_dcs2c(r1950,d1950,r0)

!  Adjust vector A to give zero proper motion in FK5
w=(bepoch-1950D0)/pmf
DO i=1,3
  a1(i)=a(i)+w*ad(i)
END DO

!  Remove e-terms
w=r0(1)*a1(1)+r0(2)*a1(2)+r0(3)*a1(3)
DO i=1,3
  v1(i)=r0(i)-a1(i)+w*r0(i)
END DO

!  Convert position vector to Fricke system
DO i=1,6
  w=0D0
  DO j=1,3
    w=w+em(i,j)*v1(j)
  END DO
  v2(i)=w
END DO

!  Allow for fictitious proper motion in FK4
w=(sla_epj(sla_epb2d(bepoch))-2000D0)/pmf
DO i=1,3
  v2(i)=v2(i)+w*v2(i+3)
END DO

!  Revert to spherical coordinates
CALL sla_dcc2s(v2,w,d2000)
r2000=sla_dranrm(w)

END SUBROUTINE sla_fk45z
SUBROUTINE sla_fk524 (r2000,d2000,dr2000,dd2000,p2000,v2000,  &
r1950,d1950,dr1950,dd1950,p1950,v1950)
!+
!     - - - - - -
!      F K 5 2 4
!     - - - - - -

!  Convert J2000.0 FK5 star data to B1950.0 FK4 (double precision)

!  This routine converts stars from the new, IAU 1976, FK5, Fricke
!  system, to the old, Bessel-Newcomb, FK4 system.  The precepts
!  of Smith et al (Ref 1) are followed, using the implementation
!  by Yallop et al (Ref 2) of a matrix method due to Standish.
!  Kinoshita's development of Andoyer's post-Newcomb precession is
!  used.  The numerical constants from Seidelmann et al (Ref 3) are
!  used canonically.

!  Given:  (all J2000.0,FK5)
!     R2000,D2000     dp    J2000.0 RA,Dec (rad)
!     DR2000,DD2000   dp    J2000.0 proper motions (rad/Jul.yr)
!     P2000           dp    parallax (arcsec)
!     V2000           dp    radial velocity (km/s, +ve = moving away)

!  Returned:  (all B1950.0,FK4)
!     R1950,D1950     dp    B1950.0 RA,Dec (rad)
!     DR1950,DD1950   dp    B1950.0 proper motions (rad/trop.yr)
!     P1950           dp    parallax (arcsec)
!     V1950           dp    radial velocity (km/s, +ve = moving away)

!  Notes:

!  1)  The proper motions in RA are dRA/dt rather than
!      cos(Dec)*dRA/dt, and are per year rather than per century.

!  2)  Note that conversion from Julian epoch 2000.0 to Besselian
!      epoch 1950.0 only is provided for.  Conversions involving
!      other epochs will require use of the appropriate precession,
!      proper motion, and E-terms routines before and/or after
!      FK524 is called.

!  3)  In the FK4 catalogue the proper motions of stars within
!      10 degrees of the poles do not embody the differential
!      E-term effect and should, strictly speaking, be handled
!      in a different manner from stars outside these regions.
!      However, given the general lack of homogeneity of the star
!      data available for routine astrometry, the difficulties of
!      handling positions that may have been determined from
!      astrometric fields spanning the polar and non-polar regions,
!      the likelihood that the differential E-terms effect was not
!      taken into account when allowing for proper motion in past
!      astrometry, and the undesirability of a discontinuity in
!      the algorithm, the decision has been made in this routine to
!      include the effect of differential E-terms on the proper
!      motions for all stars, whether polar or not.  At epoch 2000,
!      and measuring on the sky rather than in terms of dRA, the
!      errors resulting from this simplification are less than
!      1 milliarcsecond in position and 1 milliarcsecond per
!      century in proper motion.

!  References:

!     1  Smith, C.A. et al, 1989.  "The transformation of astrometric
!        catalog systems to the equinox J2000.0".  Astron.J. 97, 265.

!     2  Yallop, B.D. et al, 1989.  "Transformation of mean star places
!        from FK4 B1950.0 to FK5 J2000.0 using matrices in 6-space".
!        Astron.J. 97, 274.

!     3  Seidelmann, P.K. (ed), 1992.  "Explanatory Supplement to
!        the Astronomical Almanac", ISBN 0-935702-68-7.

!  P.T.Wallace   Starlink   19 December 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: r2000,d2000,dr2000,dd2000,p2000,v2000,  &
r1950,d1950,dr1950,dd1950,p1950,v1950


!  Miscellaneous
DOUBLE PRECISION :: r,d,ur,ud,px,rv
DOUBLE PRECISION :: sr,cr,sd,cd,x,y,z,w
DOUBLE PRECISION :: v1(6),v2(6)
DOUBLE PRECISION :: xd,yd,zd
DOUBLE PRECISION :: rxyz,wd,rxysq,rxy
INTEGER :: i,j

!  2Pi
DOUBLE PRECISION :: d2pi
PARAMETER (d2pi=6.283185307179586476925287D0)

!  Radians per year to arcsec per century
DOUBLE PRECISION :: pmf
PARAMETER (pmf=100D0*60D0*60D0*360D0/d2pi)

!  Small number to avoid arithmetic problems
DOUBLE PRECISION :: tiny
PARAMETER (tiny=1D-30)


!  CANONICAL CONSTANTS  (see references)


!  Km per sec to AU per tropical century
!  = 86400 * 36524.2198782 / 149597870
DOUBLE PRECISION :: vf
PARAMETER (vf=21.095D0)

!  Constant vector and matrix (by columns)
DOUBLE PRECISION :: a(6),emi(6,6)
DATA a/ -1.62557D-6,  -0.31919D-6, -0.13843D-6,  &
+1.245D-3,    -1.580D-3,   -0.659D-3/

DATA (emi(i,1),i=1,6) / +0.9999256795D0,-0.0111814828D0,  &
-0.0048590040D0,-0.000551D0,  &
-0.238560D0,+0.435730D0 /

DATA (emi(i,2),i=1,6) / +0.0111814828D0,+0.9999374849D0,  &
-0.0000271557D0,+0.238509D0,  &
-0.002667D0,-0.008541D0 /

DATA (emi(i,3),i=1,6) / +0.0048590039D0,-0.0000271771D0,  &
+0.9999881946D0,-0.435614D0,  &
+0.012254D0,+0.002117D0 /

DATA (emi(i,4),i=1,6) / -0.00000242389840D0,+0.00000002710544D0,  &
+0.00000001177742D0,+0.99990432D0,  &
-0.01118145D0,-0.00485852D0 /

DATA (emi(i,5),i=1,6) / -0.00000002710544D0,-0.00000242392702D0,  &
+0.00000000006585D0,+0.01118145D0,  &
+0.99991613D0,-0.00002716D0 /

DATA (emi(i,6),i=1,6) / -0.00000001177742D0,+0.00000000006585D0,  &
-0.00000242404995D0,+0.00485852D0,  &
-0.00002717D0,+0.99996684D0 /



!  Pick up J2000 data (units radians and arcsec/JC)
r=r2000
d=d2000
ur=dr2000*pmf
ud=dd2000*pmf
px=p2000
rv=v2000

!  Spherical to Cartesian
sr=SIN(r)
cr=COS(r)
sd=SIN(d)
cd=COS(d)

x=cr*cd
y=sr*cd
z=   sd

w=vf*rv*px

v1(1)=x
v1(2)=y
v1(3)=z

v1(4)=-ur*y-cr*sd*ud+w*x
v1(5)= ur*x-sr*sd*ud+w*y
v1(6)=         cd*ud+w*z

!  Convert position+velocity vector to BN system
DO i=1,6
  w=0D0
  DO j=1,6
    w=w+emi(i,j)*v1(j)
  END DO
  v2(i)=w
END DO

!  Position vector components and magnitude
x=v2(1)
y=v2(2)
z=v2(3)
rxyz=SQRT(x*x+y*y+z*z)

!  Apply E-terms to position
w=x*a(1)+y*a(2)+z*a(3)
x=x+a(1)*rxyz-w*x
y=y+a(2)*rxyz-w*y
z=z+a(3)*rxyz-w*z

!  Recompute magnitude
rxyz=SQRT(x*x+y*y+z*z)

!  Apply E-terms to both position and velocity
x=v2(1)
y=v2(2)
z=v2(3)
w=x*a(1)+y*a(2)+z*a(3)
wd=x*a(4)+y*a(5)+z*a(6)
x=x+a(1)*rxyz-w*x
y=y+a(2)*rxyz-w*y
z=z+a(3)*rxyz-w*z
xd=v2(4)+a(4)*rxyz-wd*x
yd=v2(5)+a(5)*rxyz-wd*y
zd=v2(6)+a(6)*rxyz-wd*z

!  Convert to spherical
rxysq=x*x+y*y
rxy=SQRT(rxysq)

IF (x == 0D0.AND.y == 0D0) THEN
  r=0D0
ELSE
  r=ATAN2(y,x)
  IF (r < 0.0D0) r=r+d2pi
END IF
d=ATAN2(z,rxy)

IF (rxy > tiny) THEN
  ur=(x*yd-y*xd)/rxysq
  ud=(zd*rxysq-z*(x*xd+y*yd))/((rxysq+z*z)*rxy)
END IF

!  Radial velocity and parallax
IF (px > tiny) THEN
  rv=(x*xd+y*yd+z*zd)/(px*vf*rxyz)
  px=px/rxyz
END IF

!  Return results
r1950=r
d1950=d
dr1950=ur/pmf
dd1950=ud/pmf
p1950=px
v1950=rv

END SUBROUTINE sla_fk524
SUBROUTINE sla_fk54z (r2000,d2000,bepoch,r1950,d1950,dr1950,dd1950)
!+
!     - - - - - -
!      F K 5 4 Z
!     - - - - - -

!  Convert a J2000.0 FK5 star position to B1950.0 FK4 assuming
!  zero proper motion and parallax (double precision)

!  This routine converts star positions from the new, IAU 1976,
!  FK5, Fricke system to the old, Bessel-Newcomb, FK4 system.

!  Given:
!     R2000,D2000     dp    J2000.0 FK5 RA,Dec (rad)
!     BEPOCH          dp    Besselian epoch (e.g. 1950D0)

!  Returned:
!     R1950,D1950     dp    B1950.0 FK4 RA,Dec (rad) at epoch BEPOCH
!     DR1950,DD1950   dp    B1950.0 FK4 proper motions (rad/trop.yr)

!  Notes:

!  1)  The proper motion in RA is dRA/dt rather than cos(Dec)*dRA/dt.

!  2)  Conversion from Julian epoch 2000.0 to Besselian epoch 1950.0
!      only is provided for.  Conversions involving other epochs will
!      require use of the appropriate precession routines before and
!      after this routine is called.

!  3)  Unlike in the sla_FK524 routine, the FK5 proper motions, the
!      parallax and the radial velocity are presumed zero.

!  4)  It is the intention that FK5 should be a close approximation
!      to an inertial frame, so that distant objects have zero proper
!      motion;  such objects have (in general) non-zero proper motion
!      in FK4, and this routine returns those fictitious proper
!      motions.

!  5)  The position returned by this routine is in the B1950
!      reference frame but at Besselian epoch BEPOCH.  For
!      comparison with catalogues the BEPOCH argument will
!      frequently be 1950D0.

!  Called:  sla_FK524, sla_PM

!  P.T.Wallace   Starlink   10 April 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: r2000,d2000,bepoch,r1950,d1950,dr1950,dd1950

DOUBLE PRECISION :: r,d,px,rv



!  FK5 equinox J2000 (any epoch) to FK4 equinox B1950 epoch B1950
CALL sla_fk524(r2000,d2000,0D0,0D0,0D0,0D0,r,d,dr1950,dd1950,px,rv)

!  Fictitious proper motion to epoch BEPOCH
CALL sla_pm(r,d,dr1950,dd1950,0D0,0D0,1950D0,bepoch,r1950,d1950)

END SUBROUTINE sla_fk54z
SUBROUTINE sla_flotin (string, nstrt, reslt, jflag)
!+
!     - - - - - - -
!      F L O T I N
!     - - - - - - -

!  Convert free-format input into single precision floating point

!  Given:
!     STRING     c     string containing number to be decoded
!     NSTRT      i     pointer to where decoding is to start
!     RESLT      r     current value of result

!  Returned:
!     NSTRT      i      advanced to next number
!     RESLT      r      result
!     JFLAG      i      status: -1 = -OK, 0 = +OK, 1 = null, 2 = error

!  Called:  sla_DFLTIN

!  Notes:

!     1     The reason FLOTIN has separate OK status values for +
!           and - is to enable minus zero to be detected.   This is
!           of crucial importance when decoding mixed-radix numbers.
!           For example, an angle expressed as deg, arcmin, arcsec
!           may have a leading minus sign but a zero degrees field.

!     2     A TAB is interpreted as a space, and lowercase characters
!           are interpreted as uppercase.

!     3     The basic format is the sequence of fields #^.^@#^, where
!           # is a sign character + or -, ^ means a string of decimal
!           digits, and @, which indicates an exponent, means D or E.
!           Various combinations of these fields can be omitted, and
!           embedded blanks are permissible in certain places.

!     4     Spaces:

!             .  Leading spaces are ignored.

!             .  Embedded spaces are allowed only after +, -, D or E,
!                and after the decomal point if the first sequence of
!                digits is absent.

!             .  Trailing spaces are ignored;  the first signifies
!                end of decoding and subsequent ones are skipped.

!     5     Delimiters:

!             .  Any character other than +,-,0-9,.,D,E or space may be
!                used to signal the end of the number and terminate
!                decoding.

!             .  Comma is recognized by FLOTIN as a special case;  it
!                is skipped, leaving the pointer on the next character.
!                See 13, below.

!     6     Both signs are optional.  The default is +.

!     7     The mantissa ^.^ defaults to 1.

!     8     The exponent @#^ defaults to E0.

!     9     The strings of decimal digits may be of any length.

!     10    The decimal point is optional for whole numbers.

!     11    A "null result" occurs when the string of characters being
!           decoded does not begin with +,-,0-9,.,D or E, or consists
!           entirely of spaces.  When this condition is detected, JFLAG
!           is set to 1 and RESLT is left untouched.

!     12    NSTRT = 1 for the first character in the string.

!     13    On return from FLOTIN, NSTRT is set ready for the next
!           decode - following trailing blanks and any comma.  If a
!           delimiter other than comma is being used, NSTRT must be
!           incremented before the next call to FLOTIN, otherwise
!           all subsequent calls will return a null result.

!     14    Errors (JFLAG=2) occur when:

!             .  a +, -, D or E is left unsatisfied;  or

!             .  the decimal point is present without at least
!                one decimal digit before or after it;  or

!             .  an exponent more than 100 has been presented.

!     15    When an error has been detected, NSTRT is left
!           pointing to the character following the last
!           one used before the error came to light.  This
!           may be after the point at which a more sophisticated
!           program could have detected the error.  For example,
!           FLOTIN does not detect that '1E999' is unacceptable
!           (on a computer where this is so) until the entire number
!           has been decoded.

!     16    Certain highly unlikely combinations of mantissa &
!           exponent can cause arithmetic faults during the
!           decode, in some cases despite the fact that they
!           together could be construed as a valid number.

!     17    Decoding is left to right, one pass.

!     18    See also DFLTIN and INTIN

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: nstrt
REAL :: reslt
INTEGER :: jflag

DOUBLE PRECISION :: dreslt


!  Call the double precision version
CALL sla_dfltin(string,nstrt,dreslt,jflag)
IF (jflag <= 0) reslt=REAL(dreslt)

END SUBROUTINE sla_flotin
SUBROUTINE sla_galeq (dl, db, dr, dd)
!+
!     - - - - - -
!      G A L E Q
!     - - - - - -

!  Transformation from IAU 1958 galactic coordinates to
!  J2000.0 equatorial coordinates (double precision)

!  Given:
!     DL,DB       dp       galactic longitude and latitude L2,B2

!  Returned:
!     DR,DD       dp       J2000.0 RA,Dec

!  (all arguments are radians)

!  Called:
!     sla_DCS2C, sla_DIMXV, sla_DCC2S, sla_DRANRM, sla_DRANGE

!  Note:
!     The equatorial coordinates are J2000.0.  Use the routine
!     sla_GE50 if conversion to B1950.0 'FK4' coordinates is
!     required.

!  Reference:
!     Blaauw et al, Mon.Not.R.Astron.Soc.,121,123 (1960)

!  P.T.Wallace   Starlink   November 1988

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dl,db,dr,dd

DOUBLE PRECISION :: v1(3),v2(3)


!  L2,B2 system of galactic coordinates

!  P = 192.25       RA of galactic north pole (mean B1950.0)
!  Q =  62.6        inclination of galactic to mean B1950.0 equator
!  R =  33          longitude of ascending node

!  P,Q,R are degrees

!  Equatorial to galactic rotation matrix (J2000.0), obtained by
!  applying the standard FK4 to FK5 transformation, for inertially
!  zero proper motion, to the columns of the B1950 equatorial to
!  galactic rotation matrix:

DOUBLE PRECISION :: rmat(3,3)
DATA rmat(1,1),rmat(1,2),rmat(1,3),rmat(2,1),rmat(2,2),rmat(2,3),  &
rmat(3,1),rmat(3,2),rmat(3,3)/  &
-0.054875539726D0,-0.873437108010D0,-0.483834985808D0,  &
+0.494109453312D0,-0.444829589425D0,+0.746982251810D0,  &
-0.867666135858D0,-0.198076386122D0,+0.455983795705D0/



!  Spherical to Cartesian
CALL sla_dcs2c(dl,db,v1)

!  Galactic to equatorial
CALL sla_dimxv(rmat,v1,v2)

!  Cartesian to spherical
CALL sla_dcc2s(v2,dr,dd)

!  Express in conventional ranges
dr=sla_dranrm(dr)
dd=sla_drange(dd)

END SUBROUTINE sla_galeq
SUBROUTINE sla_galsup (dl, db, dsl, dsb)
!+
!     - - - - - - -
!      G A L S U P
!     - - - - - - -

!  Transformation from IAU 1958 galactic coordinates to
!  de Vaucouleurs supergalactic coordinates (double precision)

!  Given:
!     DL,DB       dp       galactic longitude and latitude L2,B2

!  Returned:
!     DSL,DSB     dp       supergalactic longitude and latitude

!  (all arguments are radians)

!  Called:
!     sla_DCS2C, sla_DIMXV, sla_DCC2S, sla_DRANRM, sla_DRANGE

!  References:

!     de Vaucouleurs, de Vaucouleurs, & Corwin, Second Reference
!     Catalogue of Bright Galaxies, U. Texas, page 8.

!     Systems & Applied Sciences Corp., Documentation for the
!     machine-readable version of the above catalogue,
!     Contract NAS 5-26490.

!    (These two references give different values for the galactic
!     longitude of the supergalactic origin.  Both are wrong;  the
!     correct value is L2=137.37.)

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dl,db,dsl,dsb

DOUBLE PRECISION :: v1(3),v2(3)


!  System of supergalactic coordinates:

!    SGL   SGB        L2     B2      (deg)
!     -    +90      47.37  +6.32
!     0     0         -      0

!  Galactic to supergalactic rotation matrix:

DOUBLE PRECISION :: rmat(3,3)
DATA rmat(1,1),rmat(1,2),rmat(1,3),rmat(2,1),rmat(2,2),rmat(2,3),  &
rmat(3,1),rmat(3,2),rmat(3,3)/  &
-0.735742574804D0,+0.677261296414D0,+0.000000000000D0,  &
-0.074553778365D0,-0.080991471307D0,+0.993922590400D0,  &
+0.673145302109D0,+0.731271165817D0,+0.110081262225D0/



!  Spherical to Cartesian
CALL sla_dcs2c(dl,db,v1)

!  Galactic to supergalactic
CALL sla_dmxv(rmat,v1,v2)

!  Cartesian to spherical
CALL sla_dcc2s(v2,dsl,dsb)

!  Express in conventional ranges
dsl=sla_dranrm(dsl)
dsb=sla_drange(dsb)

END SUBROUTINE sla_galsup
SUBROUTINE sla_ge50 (dl, db, dr, dd)
!+
!     - - - - -
!      G E 5 0
!     - - - - -

!  Transformation from IAU 1958 galactic coordinates to
!  B1950.0 'FK4' equatorial coordinates (double precision)

!  Given:
!     DL,DB       dp       galactic longitude and latitude L2,B2

!  Returned:
!     DR,DD       dp       B1950.0 'FK4' RA,Dec

!  (all arguments are radians)

!  Called:
!     sla_DCS2C, sla_DIMXV, sla_DCC2S, sla_ADDET, sla_DRANRM, sla_DRANGE

!  Note:
!     The equatorial coordinates are B1950.0 'FK4'.  Use the
!     routine sla_GALEQ if conversion to J2000.0 coordinates
!     is required.

!  Reference:
!     Blaauw et al, Mon.Not.R.Astron.Soc.,121,123 (1960)

!  P.T.Wallace   Starlink   5 September 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dl,db,dr,dd

DOUBLE PRECISION :: v1(3),v2(3),r,d,re,de


!  L2,B2 system of galactic coordinates

!  P = 192.25       RA of galactic north pole (mean B1950.0)
!  Q =  62.6        inclination of galactic to mean B1950.0 equator
!  R =  33          longitude of ascending node

!  P,Q,R are degrees


!  Equatorial to galactic rotation matrix

!  The Euler angles are P, Q, 90-R, about the z then y then
!  z axes.

!         +CP.CQ.SR-SP.CR     +SP.CQ.SR+CP.CR     -SQ.SR

!         -CP.CQ.CR-SP.SR     -SP.CQ.CR+CP.SR     +SQ.CR

!         +CP.SQ              +SP.SQ              +CQ


DOUBLE PRECISION :: rmat(3,3)
DATA rmat(1,1),rmat(1,2),rmat(1,3),rmat(2,1),rmat(2,2),rmat(2,3),  &
rmat(3,1),rmat(3,2),rmat(3,3) /  &
-0.066988739415D0,-0.872755765852D0,-0.483538914632D0,  &
+0.492728466075D0,-0.450346958020D0,+0.744584633283D0,  &
-0.867600811151D0,-0.188374601723D0,+0.460199784784D0 /



!  Spherical to Cartesian
CALL sla_dcs2c(dl,db,v1)

!  Rotate to mean B1950.0
CALL sla_dimxv(rmat,v1,v2)

!  Cartesian to spherical
CALL sla_dcc2s(v2,r,d)

!  Introduce E-terms
CALL sla_addet(r,d,1950D0,re,de)

!  Express in conventional ranges
dr=sla_dranrm(re)
dd=sla_drange(de)

END SUBROUTINE sla_ge50
SUBROUTINE sla_geoc (p, h, r, z)
!+
!     - - - - -
!      G E O C
!     - - - - -

!  Convert geodetic position to geocentric (double precision)

!  Given:
!     P     dp     latitude (geodetic, radians)
!     H     dp     height above reference spheroid (geodetic, metres)

!  Returned:
!     R     dp     distance from Earth axis (AU)
!     Z     dp     distance from plane of Earth equator (AU)

!  Notes:
!     1)  Geocentric latitude can be obtained by evaluating ATAN2(Z,R).
!     2)  IAU 1976 constants are used.

!  Reference:
!     Green,R.M., Spherical Astronomy, CUP 1985, p98.

!  P.T.Wallace   Starlink   4th October 1989

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: p,h,r,z

!  Earth equatorial radius (metres)
DOUBLE PRECISION :: a0
PARAMETER (a0=6378140D0)

!  Reference spheroid flattening factor and useful function
DOUBLE PRECISION :: f,b
PARAMETER (f=1D0/298.257D0,b=(1D0-f)**2)

!  Astronomical unit in metres
DOUBLE PRECISION :: au
PARAMETER (au=1.49597870D11)

DOUBLE PRECISION :: sp,cp,c,s



!  Geodetic to geocentric conversion
sp=SIN(p)
cp=COS(p)
c=1D0/SQRT(cp*cp+b*sp*sp)
s=b*c
r=(a0*c+h)*cp/au
z=(a0*s+h)*sp/au

END SUBROUTINE sla_geoc
DOUBLE PRECISION FUNCTION sla_gmst (ut1)
!+
!     - - - - -
!      G M S T
!     - - - - -

!  Conversion from universal time to sidereal time (double precision)

!  Given:
!    UT1    dp     universal time (strictly UT1) expressed as
!                  modified Julian Date (JD-2400000.5)

!  The result is the Greenwich mean sidereal time (double
!  precision, radians).

!  The IAU 1982 expression (see page S15 of 1984 Astronomical
!  Almanac) is used, but rearranged to reduce rounding errors.
!  This expression is always described as giving the GMST at
!  0 hours UT.  In fact, it gives the difference between the
!  GMST and the UT, which happens to equal the GMST (modulo
!  24 hours) at 0 hours UT each day.  In this routine, the
!  entire UT is used directly as the argument for the
!  standard formula, and the fractional part of the UT is
!  added separately;  note that the factor 1.0027379... does
!  not appear.

!  See also the routine sla_GMSTA, which delivers better numerical
!  precision by accepting the UT date and time as separate arguments.

!  Called:  sla_DRANRM

!  P.T.Wallace   Starlink   14 September 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ut1

DOUBLE PRECISION :: d2pi,s2r
PARAMETER (d2pi=6.283185307179586476925286766559D0,  &
s2r=7.272205216643039903848711535369D-5)

DOUBLE PRECISION :: tu



!  Julian centuries from fundamental epoch J2000 to this UT
tu=(ut1-51544.5D0)/36525D0

!  GMST at this UT
sla_gmst=sla_dranrm(MOD(ut1,1D0)*d2pi+(24110.54841D0+  &
(8640184.812866D0+(0.093104D0-6.2D-6*tu)*tu)*tu)*s2r)

END FUNCTION sla_gmst
DOUBLE PRECISION FUNCTION sla_gmsta (date, ut)
!+
!     - - - - - -
!      G M S T A
!     - - - - - -

!  Conversion from universal time to sidereal time with
!  rounding errors minimized.

!  double precision

!  Given:
!    DATE    d      UT1 date (MJD: integer part of JD-2400000.5))
!    UT      d      UT1 time (fraction of a day)

!  The result is the Greenwich mean sidereal time (double
!  precision, radians).

!  The algorithm is derived from the IAU 1982 expression (see page
!  S15 of the 1984 Astronomical Almanac).

!  There is no restriction on how the UT is apportioned between the
!  DATE and UT arguments.  Either of the two arguments could, for
!  example, be zero and the entire date+time supplied in the other.
!  However, the routine is designed to deliver maximum accuracy when
!  the DATE argument is a whole number and the UT lies in the range
!  0 to 1.

!  See also the routine sla_GMST, which accepts the UT as a single
!  argument.  Compared with sla_GMST, the extra numerical precision
!  delivered by the present routine is unlikely to be important in
!  an absolute sense, but may be useful when critically comparing
!  algorithms and in applications where two sidereal times close
!  together are differenced.

!  Called:  sla_DRANRM

!  P.T.Wallace   Starlink   14 September 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,ut

DOUBLE PRECISION :: d2pi,s2r
PARAMETER (d2pi=6.283185307179586476925286766559D0,  &
s2r=7.272205216643039903848711535369D-5)

DOUBLE PRECISION :: f,d,t



!  Fractional part of date (if any)
f=DMOD(date,1D0)

!  Days from fundamental epoch J2000 to 0h UT on this date
d=date-51544.5D0

!  Julian centuries from fundamental epoch J2000 to this UT
t=(d+ut)/36525D0

!  GMST at this UT
sla_gmsta=sla_dranrm(s2r*(24110.54841D0+86636.555367909D0*ut  &
+236.555367909D0*d+86400D0*f  &
+(0.093104D0-6.2D-6*t)*t*t))

END FUNCTION sla_gmsta
SUBROUTINE sla_h2e (az, el, phi, ha, dec)
!+
!     - - - - -
!      D E 2 H
!     - - - - -

!  Horizon to equatorial coordinates:  Az,El to HA,Dec

!  (single precision)

!  Given:
!     AZ      r     azimuth
!     EL      r     elevation
!     PHI     r     observatory latitude

!  Returned:
!     HA      r     hour angle
!     DEC     r     declination

!  Notes:

!  1)  All the arguments are angles in radians.

!  2)  The sign convention for azimuth is north zero, east +pi/2.

!  3)  HA is returned in the range +/-pi.  Declination is returned
!      in the range +/-pi/2.

!  4)  The latitude is (in principle) geodetic.  In critical
!      applications, corrections for polar motion should be applied.

!  5)  In some applications it will be important to specify the
!      correct type of elevation in order to produce the required
!      type of HA,Dec.  In particular, it may be important to
!      distinguish between the elevation as affected by refraction,
!      which will yield the "observed" HA,Dec, and the elevation
!      in vacuo, which will yield the "topocentric" HA,Dec.  If the
!      effects of diurnal aberration can be neglected, the
!      topocentric HA,Dec may be used as an approximation to the
!      "apparent" HA,Dec.

!  6)  No range checking of arguments is done.

!  7)  In applications which involve many such calculations, rather
!      than calling the present routine it will be more efficient to
!      use inline code, having previously computed fixed terms such
!      as sine and cosine of latitude.

!  P.T.Wallace   Starlink   21 February 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: az,el,phi,ha,dec

DOUBLE PRECISION :: sa,ca,se,ce,sp,cp,x,y,z,r


!  Useful trig functions
sa=SIN(az)
ca=COS(az)
se=SIN(el)
ce=COS(el)
sp=SIN(phi)
cp=COS(phi)

!  HA,Dec as x,y,z
x=-ca*ce*sp+se*cp
y=-sa*ce
z=ca*ce*cp+se*sp

!  To HA,Dec
r=SQRT(x*x+y*y)
IF (r == 0.0) THEN
  ha=0.0
ELSE
  ha=ATAN2(y,x)
END IF
dec=ATAN2(z,r)

END SUBROUTINE sla_h2e
SUBROUTINE sla__idchf (string, nptr, nvec, ndigit, digit)
!+
!     - - - - - -
!      I D C H F
!     - - - - - -

!  Internal routine used by DFLTIN

!  Identify next character in string

!  Given:
!     STRING      char        string
!     NPTR        int         pointer to character to be identified

!  Returned:
!     NPTR        int         incremented unless end of field
!     NVEC        int         vector for identified character
!     NDIGIT      int         0-9 if character was a numeral
!     DIGIT       double      equivalent of NDIGIT

!     NVEC takes the following values:

!      1     0-9
!      2     space or TAB   !!! n.b. ASCII TAB assumed !!!
!      3     D,d,E or e
!      4     .
!      5     +
!      6     -
!      7     ,
!      8     else
!      9     outside field

!  If the character is not 0-9, NDIGIT and DIGIT are either not
!  altered or are set to arbitrary values.

!  P.T.Wallace   Starlink   22 December 1992

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: nptr,nvec,ndigit
DOUBLE PRECISION :: digit

CHARACTER :: k
INTEGER :: nchar

!  Character/vector tables
INTEGER :: ncrec
PARAMETER (ncrec=19)
CHARACTER :: kctab(ncrec)
INTEGER :: kvtab(ncrec)
DATA kctab/'0','1','2','3','4','5','6','7','8','9',  &
' ','D','d','E','e','.','+','-',','/
DATA kvtab/10*1,2,4*3,4,5,6,7/


!  Handle pointer outside field
IF (nptr < 1.OR.nptr > LEN(string)) THEN
  nvec=9
ELSE
  
!     Not end of field: identify the character
  k=string(nptr:nptr)
  DO nchar=1,ncrec
    IF (k == kctab(nchar)) THEN
      
!           Recognized
      nvec=kvtab(nchar)
      ndigit=nchar-1
      digit=DBLE(ndigit)
      GO TO 2300
    END IF
  END DO
  
!     Not recognized: check for TAB    !!! n.b. ASCII assumed !!!
  IF (k == CHAR(9)) THEN
    
!        TAB: treat as space
    nvec=2
  ELSE
    
!        Unrecognized
    nvec=8
  END IF
  
!     Increment pointer
  2300    CONTINUE
  nptr=nptr+1
END IF

END SUBROUTINE sla__idchf
SUBROUTINE sla__idchi (string, nptr, nvec, digit)
!+
!     - - - - - -
!      I D C H I
!     - - - - - -

!  Internal routine used by INTIN

!  Identify next character in string

!  Given:
!     STRING      char        string
!     NPTR        int         pointer to character to be identified

!  Returned:
!     NPTR        int         incremented unless end of field
!     NVEC        int         vector for identified character
!     DIGIT       double      double precision digit if 0-9

!     NVEC takes the following values:

!      1     0-9
!      2     space or TAB   !!! n.b. ASCII TAB assumed !!!
!      3     +
!      4     -
!      5     ,
!      6     else
!      7     outside string

!  If the character is not 0-9, DIGIT is either unaltered or
!  is set to an arbitrary value.

!  P.T.Wallace   Starlink   22 December 1992

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: nptr,nvec
DOUBLE PRECISION :: digit

CHARACTER :: k
INTEGER :: nchar

!  Character/vector tables
INTEGER :: ncrec
PARAMETER (ncrec=14)
CHARACTER :: kctab(ncrec)
INTEGER :: kvtab(ncrec)
DATA kctab/'0','1','2','3','4','5','6','7','8','9',' ', '+','-',','/
DATA kvtab/10*1,2,3,4,5/



!  Handle pointer outside field
IF (nptr < 1.OR.nptr > LEN(string)) THEN
  nvec=7
ELSE
  
!     Not end of field: identify character
  k=string(nptr:nptr)
  DO nchar=1,ncrec
    IF (k == kctab(nchar)) THEN
      
!           Recognized
      nvec=kvtab(nchar)
      digit=DBLE(nchar-1)
      GO TO 2300
    END IF
  END DO
  
!     Not recognized: check for TAB   !!! n.b. ASCII assumed !!!
  IF (k == CHAR(9)) THEN
    
!        TAB: treat as space
    nvec=2
  ELSE
    
!        Unrecognized
    nvec=6
  END IF
  
!     Increment pointer
  2300    CONTINUE
  nptr=nptr+1
END IF

END SUBROUTINE sla__idchi
SUBROUTINE sla_imxv (rm, va, vb)
!+
!     - - - - -
!      I M X V
!     - - - - -

!  Performs the 3-D backward unitary transformation:

!     vector VB = (inverse of matrix RM) * vector VA

!  (single precision)

!  (n.b.  the matrix must be unitary, as this routine assumes that
!   the inverse and transpose are identical)

!  Given:
!     RM       real(3,3)    matrix
!     VA       real(3)      vector

!  Returned:
!     VB       real(3)      result vector

!  P.T.Wallace   Starlink   November 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: rm(3,3),va(3),vb(3)

INTEGER :: i,j
REAL :: w,vw(3)



!  Inverse of matrix RM * vector VA -> vector VW
DO j=1,3
  w=0.0
  DO i=1,3
    w=w+rm(i,j)*va(i)
  END DO
  vw(j)=w
END DO

!  Vector VW -> vector VB
DO j=1,3
  vb(j)=vw(j)
END DO

END SUBROUTINE sla_imxv
SUBROUTINE sla_intin (string, nstrt, ireslt, jflag)
!+
!     - - - - - -
!      I N T I N
!     - - - - - -

!  Convert free-format input into an integer

!  Given:
!     STRING     c     string containing number to be decoded
!     NSTRT      i     pointer to where decoding is to start
!     IRESLT     i     current value of result

!  Returned:
!     NSTRT      i      advanced to next number
!     IRESLT     i      result
!     JFLAG      i      status: -1 = -OK, 0 = +OK, 1 = null, 2 = error

!  Called:  sla__IDCHI

!  Notes:

!     1     The reason INTIN has separate OK status values for +
!           and - is to enable minus zero to be detected.   This is
!           of crucial importance when decoding mixed-radix numbers.
!           For example, an angle expressed as deg, arcmin, arcsec
!           may have a leading minus sign but a zero degrees field.

!     2     A TAB is interpreted as a space.

!     3     The basic format is the sequence of fields #^, where
!           # is a sign character + or -, and ^ means a string of
!           decimal digits.

!     4     Spaces:

!             .  Leading spaces are ignored.

!             .  Spaces between the sign and the number are allowed.

!             .  Trailing spaces are ignored;  the first signifies
!                end of decoding and subsequent ones are skipped.

!     5     Delimiters:

!             .  Any character other than +,-,0-9 or space may be
!                used to signal the end of the number and terminate
!                decoding.

!             .  Comma is recognized by INTIN as a special case;  it
!                is skipped, leaving the pointer on the next character.
!                See 9, below.

!     6     The sign is optional.  The default is +.

!     7     A "null result" occurs when the string of characters being
!           decoded does not begin with +,- or 0-9, or consists
!           entirely of spaces.  When this condition is detected, JFLAG
!           is set to 1 and IRESLT is left untouched.

!     8     NSTRT = 1 for the first character in the string.

!     9     On return from INTIN, NSTRT is set ready for the next
!           decode - following trailing blanks and any comma.  If a
!           delimiter other than comma is being used, NSTRT must be
!           incremented before the next call to INTIN, otherwise
!           all subsequent calls will return a null result.

!     10    Errors (JFLAG=2) occur when:

!             .  there is a + or - but no number;  or

!             .  the number is greater than BIG (defined below).

!     11    When an error has been detected, NSTRT is left
!           pointing to the character following the last
!           one used before the error came to light.

!     12    See also FLOTIN and DFLTIN.

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) string
INTEGER :: nstrt,ireslt,jflag

!  Maximum allowed value
DOUBLE PRECISION :: big
PARAMETER (big=2147483647D0)

INTEGER :: jptr,msign,nvec,j
DOUBLE PRECISION :: dres,digit



!  Current character
jptr=nstrt

!  Set defaults
dres=0D0
msign=1

!  Look for sign
100  CONTINUE
CALL sla__idchi(string,jptr,nvec,digit)
GO TO ( 400, 100,  300,  200, 9110, 9100, 9110),nvec
!             0-9   SP     +     -     ,   ELSE   END

!  Negative
200  CONTINUE
msign=-1

!  Look for first decimal
300  CONTINUE
CALL sla__idchi(string,jptr,nvec,digit)
GO TO ( 400, 300, 9200, 9200, 9200, 9200, 9210),nvec
!             0-9   SP     +     -     ,   ELSE   END

!  Accept decimals
400  CONTINUE
dres=dres*1D1+digit

!  Test for overflow
IF (dres > big) GO TO 9200

!  Look for subsequent decimals
CALL sla__idchi(string,jptr,nvec,digit)
GO TO ( 400, 1610, 1600, 1600, 1600, 1600, 1610),nvec
!             0-9   SP     +     -     ,   ELSE   END

!  Get result & status
1600 CONTINUE
jptr=jptr-1
1610 CONTINUE
j=0
IF (msign == 1) GO TO 1620
j=-1
dres=-dres
1620 CONTINUE
ireslt=INT(DNINT(dres))

!  Skip to end of field
1630 CONTINUE
CALL sla__idchi(string,jptr,nvec,digit)
GO TO (1720, 1630, 1720, 1720, 9900, 1720, 9900),nvec
!             0-9   SP     +     -     ,   ELSE   END

1720 CONTINUE
jptr=jptr-1
GO TO 9900

!  Exits

!  Null field
9100 CONTINUE
jptr=jptr-1
9110 CONTINUE
j=1
GO TO 9900

!  Errors
9200 CONTINUE
jptr=jptr-1
9210 CONTINUE
j=2

!  Return
9900 CONTINUE
nstrt=jptr
jflag=j

END SUBROUTINE sla_intin
SUBROUTINE sla_invf (fwds,bkwds,j)
!+
!     - - - - -
!      I N V F
!     - - - - -

!  Invert a linear model of the type produced by the
!  sla_FITXY routine.

!  Given:
!     FWDS    d(6)      model coefficients

!  Returned:
!     BKWDS   d(6)      inverse model
!     J        i        status:  0 = OK, -1 = no inverse

!  The models relate two sets of [X,Y] coordinates as follows.
!  Naming the elements of FWDS:

!     FWDS(1) = A
!     FWDS(2) = B
!     FWDS(3) = C
!     FWDS(4) = D
!     FWDS(5) = E
!     FWDS(6) = F

!  where two sets of coordinates [X1,Y1] and [X2,Y1] are related
!  thus:

!     X2 = A + B*X1 + C*Y1
!     Y2 = D + E*X1 + F*Y1

!  the present routine generates a new set of coefficients:

!     BKWDS(1) = P
!     BKWDS(2) = Q
!     BKWDS(3) = R
!     BKWDS(4) = S
!     BKWDS(5) = T
!     BKWDS(6) = U

!  such that:

!     X1 = P + Q*X2 + R*Y2
!     Y1 = S + T*X2 + U*Y2

!  Two successive calls to sla_INVF will thus deliver a set
!  of coefficients equal to the starting values.

!  To comply with the ANSI Fortran standard, FWDS and BKWDS must
!  not be the same array, even though the routine is coded to
!  work on the VAX and most other computers even if this rule
!  is violated.

!  See also sla_FITXY, sla_PXY, sla_XY2XY, sla_DCMPF

!  P.T.Wallace   Starlink   11 April 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: fwds(6),bkwds(6)
INTEGER :: j

DOUBLE PRECISION :: a,b,c,d,e,f,det



a=fwds(1)
b=fwds(2)
c=fwds(3)
d=fwds(4)
e=fwds(5)
f=fwds(6)
det=b*f-c*e
IF (det /= 0D0) THEN
  bkwds(1)=(c*d-a*f)/det
  bkwds(2)=f/det
  bkwds(3)=-c/det
  bkwds(4)=(a*e-b*d)/det
  bkwds(5)=-e/det
  bkwds(6)=b/det
  j=0
ELSE
  j=-1
END IF

END SUBROUTINE sla_invf
SUBROUTINE sla_kbj (jb, e, k, j)
!+
!     - - - -
!      K B J
!     - - - -

!  Select epoch prefix 'B' or 'J'

!  Given:
!     JB     int     sla_DBJIN prefix status:  0=none, 1='B', 2='J'
!     E      dp      epoch - Besselian or Julian

!  Returned:
!     K      char    'B' or 'J'
!     J      int     status:  0=OK

!  If JB=0, B is assumed for E < 1984D0, otherwise J.

!  P.T.Wallace   Starlink   31 July 1989

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: jb
DOUBLE PRECISION :: e
CHARACTER :: k*(*)
INTEGER :: j

!  Preset status
j=0

!  If prefix given expressly, use it
IF (jb == 1) THEN
  k='B'
ELSE IF (jb == 2) THEN
  k='J'
  
!  If no prefix, examine the epoch
ELSE IF (jb == 0) THEN
  
!     If epoch is pre-1984.0, assume Besselian;  otherwise Julian
  IF (e < 1984D0) THEN
    k='B'
  ELSE
    k='J'
  END IF
  
!  If illegal prefix, return error status
ELSE
  k=' '
  j=1
END IF

END SUBROUTINE sla_kbj
SUBROUTINE sla_m2av (rmat, axvec)
!+
!     - - - - -
!      M 2 A V
!     - - - - -

!  From a rotation matrix, determine the corresponding axial vector
!  (single precision)

!  A rotation matrix describes a rotation about some arbitrary axis.
!  The axis is called the Euler axis, and the angle through which the
!  reference frame rotates is called the Euler angle.  The axial
!  vector returned by this routine has the same direction as the
!  Euler axis, and its magnitude is the Euler angle in radians.  (The
!  magnitude and direction can be separated by means of the routine
!  sla_VN.)

!  Given:
!    RMAT   r(3,3)   rotation matrix

!  Returned:
!    AXVEC  r(3)     axial vector (radians)

!  The reference frame rotates clockwise as seen looking along
!  the axial vector from the origin.

!  If RMAT is null, so is the result.

!  P.T.Wallace   Starlink   11 April 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: rmat(3,3),axvec(3)

REAL :: x,y,z,s2,c2,phi,f



x = rmat(2,3)-rmat(3,2)
y = rmat(3,1)-rmat(1,3)
z = rmat(1,2)-rmat(2,1)
s2 = SQRT(x*x+y*y+z*z)
IF (s2 /= 0.0) THEN
  c2 = (rmat(1,1)+rmat(2,2)+rmat(3,3)-1.0)
  phi = ATAN2(s2/2.0,c2/2.0)
  f = phi/s2
  axvec(1) = x*f
  axvec(2) = y*f
  axvec(3) = z*f
ELSE
  axvec(1) = 0.0
  axvec(2) = 0.0
  axvec(3) = 0.0
END IF

END SUBROUTINE sla_m2av
SUBROUTINE sla_map (rm, dm, pr, pd, px, rv, EQ, date, ra, da)
!+
!     - - - -
!      M A P
!     - - - -

!  Transform star RA,Dec from mean place to geocentric apparent

!  The reference frames and timescales used are post IAU 1976.

!  References:
!     1984 Astronomical Almanac, pp B39-B41.
!     (also Lederle & Schwan, Astron. Astrophys. 134,
!      1-6, 1984)

!  Given:
!     RM,DM    dp     mean RA,Dec (rad)
!     PR,PD    dp     proper motions:  RA,Dec changes per Julian year
!     PX       dp     parallax (arcsec)
!     RV       dp     radial velocity (km/sec, +ve if receding)
!     EQ       dp     epoch and equinox of star data (Julian)
!     DATE     dp     TDB for apparent place (JD-2400000.5)

!  Returned:
!     RA,DA    dp     apparent RA,Dec (rad)

!  Called:
!     sla_MAPPA       star-independent parameters
!     sla_MAPQK       quick mean to apparent

!  Notes:

!  1)  EQ is the Julian epoch specifying both the reference
!      frame and the epoch of the position - usually 2000.
!      For positions where the epoch and equinox are
!      different, use the routine sla_PM to apply proper
!      motion corrections before using this routine.

!  2)  The distinction between the required TDB and TT is
!      always negligible.  Moreover, for all but the most
!      critical applications UTC is adequate.

!  3)  The proper motions in RA are dRA/dt rather than
!      cos(Dec)*dRA/dt.

!  4)  This routine may be wasteful for some applications
!      because it recomputes the Earth position/velocity and
!      the precession/nutation matrix each time, and because
!      it allows for parallax and proper motion.  Where
!      multiple transformations are to be carried out for one
!      epoch, a faster method is to call the sla_MAPPA routine
!      once and then either the sla_MAPQK routine (which includes
!      parallax and proper motion) or sla_MAPQKZ (which assumes
!      zero parallax and proper motion).

!  P.T.Wallace   Starlink   19 January 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rm,dm,pr,pd,px,rv,EQ,date,ra,da

DOUBLE PRECISION :: amprms(21)



!  Star-independent parameters
CALL sla_mappa(EQ,date,amprms)

!  Mean to apparent
CALL sla_mapqk(rm,dm,pr,pd,px,rv,amprms,ra,da)

END SUBROUTINE sla_map
SUBROUTINE sla_mappa (EQ, date, amprms)
!+
!     - - - - - -
!      M A P P A
!     - - - - - -

!  Compute star-independent parameters in preparation for
!  conversions between mean place and geocentric apparent place.

!  The parameters produced by this routine are required in the
!  parallax, light deflection, aberration, and precession/nutation
!  parts of the mean/apparent transformations.

!  The reference frames and timescales used are post IAU 1976.

!  Given:
!     EQ       d      epoch of mean equinox to be used (Julian)
!     DATE     d      TDB (JD-2400000.5)

!  Returned:
!     AMPRMS   d(21)  star-independent mean-to-apparent parameters:

!       (1)      time interval for proper motion (Julian years)
!       (2-4)    barycentric position of the Earth (AU)
!       (5-7)    heliocentric direction of the Earth (unit vector)
!       (8)      (grav rad Sun)*2/(Sun-Earth distance)
!       (9-11)   ABV: barycentric Earth velocity in units of c
!       (12)     sqrt(1-v**2) where v=modulus(ABV)
!       (13-21)  precession/nutation (3,3) matrix

!  References:
!     1984 Astronomical Almanac, pp B39-B41.
!     (also Lederle & Schwan, Astron. Astrophys. 134,
!      1-6, 1984)

!  Notes:

!  1)  For DATE, the distinction between the required TDB and TT
!      is always negligible.  Moreover, for all but the most
!      critical applications UTC is adequate.

!  2)  The accuracy of the routines using the parameters AMPRMS is
!      limited by the routine sla_EVP, used here to compute the
!      Earth position and velocity by the methods of Stumpff.
!      The maximum error in the resulting aberration corrections is
!      about 0.3 milliarcsecond.

!  3)  The vectors AMPRMS(2-4) and AMPRMS(5-7) are referred to
!      the mean equinox and equator of epoch EQ.

!  4)  The parameters AMPRMS produced by this routine are used by
!      sla_MAPQK and sla_MAPQKZ.

!  Called:
!     sla_EPJ         MDJ to Julian epoch
!     sla_EVP         earth position & velocity
!     sla_DVN         normalize vector
!     sla_PRENUT      precession/nutation matrix

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: EQ,date,amprms(21)

!  Light time for 1 AU (sec)
DOUBLE PRECISION :: cr
PARAMETER (cr=499.004782D0)

!  Gravitational radius of the sun x 2 (2*mu/c**2, AU)
DOUBLE PRECISION :: gr2
PARAMETER (gr2=2D0*9.87063D-9)

INTEGER :: i

DOUBLE PRECISION :: ebd(3),ehd(3),eh(3),e,vn(3),vm




!  Time interval for proper motion correction
amprms(1) = sla_epj(date)-EQ

!  Get Earth barycentric and heliocentric position and velocity
CALL sla_evp(date,EQ,ebd,amprms(2),ehd,eh)

!  Heliocentric direction of earth (normalized) and modulus
CALL sla_dvn(eh,amprms(5),e)

!  Light deflection parameter
amprms(8) = gr2/e

!  Aberration parameters
DO i=1,3
  amprms(i+8) = ebd(i)*cr
END DO
CALL sla_dvn(amprms(9),vn,vm)
amprms(12) = SQRT(1D0-vm*vm)

!  Precession/nutation matrix
CALL sla_prenut(EQ,date,amprms(13))

END SUBROUTINE sla_mappa
SUBROUTINE sla_mapqk (rm, dm, pr, pd, px, rv, amprms, ra, da)
!+
!     - - - - - -
!      M A P Q K
!     - - - - - -

!  Quick mean to apparent place:  transform a star RA,Dec from
!  mean place to geocentric apparent place, given the
!  star-independent parameters.

!  Use of this routine is appropriate when efficiency is important
!  and where many star positions, all referred to the same equator
!  and equinox, are to be transformed for one epoch.  The
!  star-independent parameters can be obtained by calling the
!  sla_MAPPA routine.

!  If the parallax and proper motions are zero the sla_MAPQKZ
!  routine can be used instead.

!  The reference frames and timescales used are post IAU 1976.

!  Given:
!     RM,DM    d      mean RA,Dec (rad)
!     PR,PD    d      proper motions:  RA,Dec changes per Julian year
!     PX       d      parallax (arcsec)
!     RV       d      radial velocity (km/sec, +ve if receding)

!     AMPRMS   d(21)  star-independent mean-to-apparent parameters:

!       (1)      time interval for proper motion (Julian years)
!       (2-4)    barycentric position of the Earth (AU)
!       (5-7)    heliocentric direction of the Earth (unit vector)
!       (8)      (grav rad Sun)*2/(Sun-Earth distance)
!       (9-11)   barycentric Earth velocity in units of c
!       (12)     sqrt(1-v**2) where v=modulus(ABV)
!       (13-21)  precession/nutation (3,3) matrix

!  Returned:
!     RA,DA    d      apparent RA,Dec (rad)

!  References:
!     1984 Astronomical Almanac, pp B39-B41.
!     (also Lederle & Schwan, Astron. Astrophys. 134,
!      1-6, 1984)

!  Notes:

!  1)  The vectors AMPRMS(2-4) and AMPRMS(5-7) are referred to
!      the mean equinox and equator of epoch EQ.

!  2)  Strictly speaking, the routine is not valid for solar-system
!      sources, though the error will usually be extremely small.
!      However, to prevent gross errors in the case where the
!      position of the Sun is specified, the gravitational
!      deflection term is restrained within about 920 arcsec of the
!      centre of the Sun's disc.  The term has a maximum value of
!      about 1.85 arcsec at this radius, and decreases to zero as
!      the centre of the disc is approached.

!  Called:
!     sla_DCS2C       spherical to Cartesian
!     sla_DVDV        dot product
!     sla_DMXV        matrix x vector
!     sla_DCC2S       Cartesian to spherical
!     sla_DRANRM      normalize angle 0-2Pi

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rm,dm,pr,pd,px,rv,amprms(21),ra,da

!  Arc seconds to radians
DOUBLE PRECISION :: as2r
PARAMETER (as2r=0.484813681109535994D-5)

!  Km/s to AU/year
DOUBLE PRECISION :: vf
PARAMETER (vf=0.21094502D0)

INTEGER :: i

DOUBLE PRECISION :: pmt,gr2e,ab1,eb(3),ehn(3),abv(3),  &
q(3),pxr,w,em(3),p(3),pn(3),pde,pdep1,p1(3),p1dv,p1dvp1,p2(3),p3(3)




!  Unpack scalar and vector parameters
pmt = amprms(1)
gr2e = amprms(8)
ab1 = amprms(12)
DO i=1,3
  eb(i) = amprms(i+1)
  ehn(i) = amprms(i+4)
  abv(i) = amprms(i+8)
END DO

!  Spherical to x,y,z
CALL sla_dcs2c(rm,dm,q)

!  Space motion (radians per year)
pxr = px*as2r
w = vf*rv*pxr
em(1) = -pr*q(2)-pd*COS(rm)*SIN(dm)+w*q(1)
em(2) =  pr*q(1)-pd*SIN(rm)*SIN(dm)+w*q(2)
em(3) =          pd*COS(dm)        +w*q(3)

!  Geocentric direction of star (normalized)
DO i=1,3
  p(i) = q(i)+pmt*em(i)-pxr*eb(i)
END DO
CALL sla_dvn(p,pn,w)

!  Light deflection (restrained within the Sun's disc)
pde = sla_dvdv(pn,ehn)
pdep1 = pde+1D0
w = gr2e/MAX(pdep1,1D-5)
DO i=1,3
  p1(i) = pn(i)+w*(ehn(i)-pde*pn(i))
END DO

!  Aberration
p1dv = sla_dvdv(p1,abv)
p1dvp1 = p1dv+1D0
w = 1D0+p1dv/(ab1+1D0)
DO i=1,3
  p2(i) = (ab1*p1(i)+w*abv(i))/p1dvp1
END DO

!  Precession and nutation
CALL sla_dmxv(amprms(13),p2,p3)

!  Geocentric apparent RA,Dec
CALL sla_dcc2s(p3,ra,da)
ra = sla_dranrm(ra)

END SUBROUTINE sla_mapqk
SUBROUTINE sla_mapqkz (rm, dm, amprms, ra, da)
!+
!     - - - - - - -
!      M A P Q K Z
!     - - - - - - -

!  Quick mean to apparent place:  transform a star RA,Dec from
!  mean place to geocentric apparent place, given the
!  star-independent parameters, and assuming zero parallax
!  and proper motion.

!  Use of this routine is appropriate when efficiency is important
!  and where many star positions, all with parallax and proper
!  motion either zero or already allowed for, and all referred to
!  the same equator and equinox, are to be transformed for one
!  epoch.  The star-independent parameters can be obtained by
!  calling the sla_MAPPA routine.

!  The corresponding routine for the case of non-zero parallax
!  and proper motion is sla_MAPQK.

!  The reference frames and timescales used are post IAU 1976.

!  Given:
!     RM,DM    d      mean RA,Dec (rad)
!     AMPRMS   d(21)  star-independent mean-to-apparent parameters:

!       (1-4)    not used
!       (5-7)    heliocentric direction of the Earth (unit vector)
!       (8)      (grav rad Sun)*2/(Sun-Earth distance)
!       (9-11)   ABV: barycentric Earth velocity in units of c
!       (12)     sqrt(1-v**2) where v=modulus(ABV)
!       (13-21)  precession/nutation (3,3) matrix

!  Returned:
!     RA,DA    d      apparent RA,Dec (rad)

!  References:
!     1984 Astronomical Almanac, pp B39-B41.
!     (also Lederle & Schwan, Astron. Astrophys. 134,
!      1-6, 1984)

!  Notes:

!  1)  The vectors AMPRMS(2-4) and AMPRMS(5-7) are referred to the
!      mean equinox and equator of epoch EQ.

!  2)  Strictly speaking, the routine is not valid for solar-system
!      sources, though the error will usually be extremely small.
!      However, to prevent gross errors in the case where the
!      position of the Sun is specified, the gravitational
!      deflection term is restrained within about 920 arcsec of the
!      centre of the Sun's disc.  The term has a maximum value of
!      about 1.85 arcsec at this radius, and decreases to zero as
!      the centre of the disc is approached.

!  Called:
!     sla_DCS2C       spherical to Cartesian
!     sla_DVDV        dot product
!     sla_DMXV        matrix x vector
!     sla_DCC2S       Cartesian to spherical
!     sla_DRANRM      normalize angle 0-2Pi

!  P.T.Wallace   Starlink   28 April 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rm,dm,amprms(21),ra,da

INTEGER :: i

DOUBLE PRECISION :: gr2e,ab1,ehn(3),abv(3),p(3),pde,pdep1,w,p1(3),p1dv,  &
p1dvp1,p2(3),p3(3)





!  Unpack scalar and vector parameters
gr2e = amprms(8)
ab1 = amprms(12)
DO i=1,3
  ehn(i) = amprms(i+4)
  abv(i) = amprms(i+8)
END DO

!  Spherical to x,y,z
CALL sla_dcs2c(rm,dm,p)

!  Light deflection
pde = sla_dvdv(p,ehn)
pdep1 = pde+1D0
w = gr2e/MAX(pdep1,1D-5)
DO i=1,3
  p1(i) = p(i)+w*(ehn(i)-pde*p(i))
END DO

!  Aberration
p1dv = sla_dvdv(p1,abv)
p1dvp1 = p1dv+1D0
w = 1D0+p1dv/(ab1+1D0)
DO i=1,3
  p2(i) = (ab1*p1(i)+w*abv(i))/p1dvp1
END DO

!  Precession and nutation
CALL sla_dmxv(amprms(13),p2,p3)

!  Geocentric apparent RA,Dec
CALL sla_dcc2s(p3,ra,da)
ra = sla_dranrm(ra)

END SUBROUTINE sla_mapqkz
SUBROUTINE sla_moon (iy, id, fd, pv)
!+
!     - - - - -
!      M O O N
!     - - - - -

!  Approximate geocentric position and velocity of the Moon
!  (single precision).

!  Given:
!     IY       i       year
!     ID       i       day in year (1 = Jan 1st)
!     FD       r       fraction of day

!  Returned:
!     PV       r(6)    Moon position & velocity vector

!  Notes:

!  1  The date and time is TDB (loosely ET) in a Julian calendar
!     which has been aligned to the ordinary Gregorian
!     calendar for the interval 1900 March 1 to 2100 February 28.
!     The year and day can be obtained by calling sla_CALYD or
!     sla_CLYD.

!  2  The Moon 6-vector is Moon centre relative to Earth centre,
!     mean equator and equinox of date.  Position part, PV(1-3),
!     is in AU;  velocity part, PV(4-6), is in AU/sec.

!  3  The position is accurate to better than 0.5 arcminute
!     in direction and 1000 km in distance.  The velocity
!     is accurate to better than 0.5"/hour in direction and
!     4 m/s in distance.  (RMS figures with respect to JPL DE200
!     for the interval 1960-2025 are 14 arcsec and 0.2 arcsec/hour in
!     longitude, 9 arcsec and 0.2 arcsec/hour in latitude, 350 km and
!     2 m/s in distance.)  Note that the distance accuracy is
!     comparatively poor because this routine is principally intended
!     for computing topocentric direction.

!  4  This routine is only a partial implementation of the original
!     Meeus algorithm (reference below), which offers 4 times the
!     accuracy in direction and 30 times the accuracy in distance
!     when fully implemented (as it is in sla_DMOON).

!  Reference:
!     Meeus, l'Astronomie, June 1984, p348.

!  Called:  sla_CS2C6

!  P.T.Wallace   Starlink   8 December 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: iy,id
REAL :: fd,pv(6)

INTEGER :: itp(4,4),itl(4,39),itb(4,29),i,iy4,n
REAL :: d2r,ratcon,eradau
REAL :: elp0,elp1,elp1i,elp1f
REAL :: em0,em1,em1f
REAL :: emp0,emp1,emp1i,emp1f
REAL :: d0,d1,d1i,d1f
REAL :: f0,f1,f1i,f1f
REAL :: tl(39)
REAL :: tb(29)
REAL :: tp(4)
REAL :: yi,yf,t,elp,em,emp,d,f,el,eld,coeff,cem,cemp
REAL :: cd,cf,theta,thetad,b,bd,p,pd,sp,r,rd
REAL :: v(6),eps,sineps,coseps

!  Degrees to radians
PARAMETER (d2r=1.745329252E-2)

!  Rate conversion factor:  D2R**2/(86400*365.25)
PARAMETER (ratcon=9.652743551E-12)

!  Earth radius in AU:  6378.137/149597870
PARAMETER (eradau=4.2635212653763E-5)


!  Coefficients for fundamental arguments

!  Fixed term (deg), term in T (deg & whole revs + fraction per year)

!  Moon's mean longitude
DATA elp0,elp1,elp1i,elp1f /270.434164, 4812.678831, 4680., 132.678831 /

!  Sun's mean anomaly
DATA em0,em1,em1f /358.475833,  359.990498,        359.990498 /

!  Moon's mean anomaly
DATA emp0,emp1,emp1i,emp1f /296.104608, 4771.988491, 4680.,  91.988491 /

!  Moon's mean elongation
DATA d0,d1,d1i,d1f /350.737486,  4452.671142, 4320., 132.671142 /

!  Mean distance of the Moon from its ascending node
DATA f0,f1,f1i,f1f /11.250889, 4832.020251, 4680., 152.020251 /


!  Coefficients for Moon position

!   T(N)       = coefficient of term (deg)
!   IT(N,1-4)  = coefficients of M, M', D, F in argument

!  Longitude
!                                         M   M'  D   F
DATA tl( 1)/            +6.288750                     /,  &
(itl(i, 1),i=1,4)/             0, +1,  0,  0     /
DATA tl( 2)/            +1.274018                     /,  &
(itl(i, 2),i=1,4)/             0, -1, +2,  0     /
DATA tl( 3)/            +0.658309                     /,  &
(itl(i, 3),i=1,4)/             0,  0, +2,  0     /
DATA tl( 4)/            +0.213616                     /,  &
(itl(i, 4),i=1,4)/             0, +2,  0,  0     /
DATA tl( 5)/            -0.185596                     /,  &
(itl(i, 5),i=1,4)/            +1,  0,  0,  0     /
DATA tl( 6)/            -0.114336                     /,  &
(itl(i, 6),i=1,4)/             0,  0,  0, +2     /
DATA tl( 7)/            +0.058793                     /,  &
(itl(i, 7),i=1,4)/             0, -2, +2,  0     /
DATA tl( 8)/            +0.057212                     /,  &
(itl(i, 8),i=1,4)/            -1, -1, +2,  0     /
DATA tl( 9)/            +0.053320                     /,  &
(itl(i, 9),i=1,4)/             0, +1, +2,  0     /
DATA tl(10)/            +0.045874                     /,  &
(itl(i,10),i=1,4)/            -1,  0, +2,  0     /
DATA tl(11)/            +0.041024                     /,  &
(itl(i,11),i=1,4)/            -1, +1,  0,  0     /
DATA tl(12)/            -0.034718                     /,  &
(itl(i,12),i=1,4)/             0,  0, +1,  0     /
DATA tl(13)/            -0.030465                     /,  &
(itl(i,13),i=1,4)/            +1, +1,  0,  0     /
DATA tl(14)/            +0.015326                     /,  &
(itl(i,14),i=1,4)/             0,  0, +2, -2     /
DATA tl(15)/            -0.012528                     /,  &
(itl(i,15),i=1,4)/             0, +1,  0, +2     /
DATA tl(16)/            -0.010980                     /,  &
(itl(i,16),i=1,4)/             0, -1,  0, +2     /
DATA tl(17)/            +0.010674                     /,  &
(itl(i,17),i=1,4)/             0, -1, +4,  0     /
DATA tl(18)/            +0.010034                     /,  &
(itl(i,18),i=1,4)/             0, +3,  0,  0     /
DATA tl(19)/            +0.008548                     /,  &
(itl(i,19),i=1,4)/             0, -2, +4,  0     /
DATA tl(20)/            -0.007910                     /,  &
(itl(i,20),i=1,4)/            +1, -1, +2,  0     /
DATA tl(21)/            -0.006783                     /,  &
(itl(i,21),i=1,4)/            +1,  0, +2,  0     /
DATA tl(22)/            +0.005162                     /,  &
(itl(i,22),i=1,4)/             0, +1, -1,  0     /
DATA tl(23)/            +0.005000                     /,  &
(itl(i,23),i=1,4)/            +1,  0, +1,  0     /
DATA tl(24)/            +0.004049                     /,  &
(itl(i,24),i=1,4)/            -1, +1, +2,  0     /
DATA tl(25)/            +0.003996                     /,  &
(itl(i,25),i=1,4)/             0, +2, +2,  0     /
DATA tl(26)/            +0.003862                     /,  &
(itl(i,26),i=1,4)/             0,  0, +4,  0     /
DATA tl(27)/            +0.003665                     /,  &
(itl(i,27),i=1,4)/             0, -3, +2,  0     /
DATA tl(28)/            +0.002695                     /,  &
(itl(i,28),i=1,4)/            -1, +2,  0,  0     /
DATA tl(29)/            +0.002602                     /,  &
(itl(i,29),i=1,4)/             0, +1, -2, -2     /
DATA tl(30)/            +0.002396                     /,  &
(itl(i,30),i=1,4)/            -1, -2, +2,  0     /
DATA tl(31)/            -0.002349                     /,  &
(itl(i,31),i=1,4)/             0, +1, +1,  0     /
DATA tl(32)/            +0.002249                     /,  &
(itl(i,32),i=1,4)/            -2,  0, +2,  0     /
DATA tl(33)/            -0.002125                     /,  &
(itl(i,33),i=1,4)/            +1, +2,  0,  0     /
DATA tl(34)/            -0.002079                     /,  &
(itl(i,34),i=1,4)/            +2,  0,  0,  0     /
DATA tl(35)/            +0.002059                     /,  &
(itl(i,35),i=1,4)/            -2, -1, +2,  0     /
DATA tl(36)/            -0.001773                     /,  &
(itl(i,36),i=1,4)/             0, +1, +2, -2     /
DATA tl(37)/            -0.001595                     /,  &
(itl(i,37),i=1,4)/             0,  0, +2, +2     /
DATA tl(38)/            +0.001220                     /,  &
(itl(i,38),i=1,4)/            -1, -1, +4,  0     /
DATA tl(39)/            -0.001110                     /,  &
(itl(i,39),i=1,4)/             0, +2,  0, +2     /

!  Latitude
!                                         M   M'  D   F
DATA tb( 1)/            +5.128189                     /,  &
(itb(i, 1),i=1,4)/             0,  0,  0, +1     /
DATA tb( 2)/            +0.280606                     /,  &
(itb(i, 2),i=1,4)/             0, +1,  0, +1     /
DATA tb( 3)/            +0.277693                     /,  &
(itb(i, 3),i=1,4)/             0, +1,  0, -1     /
DATA tb( 4)/            +0.173238                     /,  &
(itb(i, 4),i=1,4)/             0,  0, +2, -1     /
DATA tb( 5)/            +0.055413                     /,  &
(itb(i, 5),i=1,4)/             0, -1, +2, +1     /
DATA tb( 6)/            +0.046272                     /,  &
(itb(i, 6),i=1,4)/             0, -1, +2, -1     /
DATA tb( 7)/            +0.032573                     /,  &
(itb(i, 7),i=1,4)/             0,  0, +2, +1     /
DATA tb( 8)/            +0.017198                     /,  &
(itb(i, 8),i=1,4)/             0, +2,  0, +1     /
DATA tb( 9)/            +0.009267                     /,  &
(itb(i, 9),i=1,4)/             0, +1, +2, -1     /
DATA tb(10)/            +0.008823                     /,  &
(itb(i,10),i=1,4)/             0, +2,  0, -1     /
DATA tb(11)/            +0.008247                     /,  &
(itb(i,11),i=1,4)/            -1,  0, +2, -1     /
DATA tb(12)/            +0.004323                     /,  &
(itb(i,12),i=1,4)/             0, -2, +2, -1     /
DATA tb(13)/            +0.004200                     /,  &
(itb(i,13),i=1,4)/             0, +1, +2, +1     /
DATA tb(14)/            +0.003372                     /,  &
(itb(i,14),i=1,4)/            -1,  0, -2, +1     /
DATA tb(15)/            +0.002472                     /,  &
(itb(i,15),i=1,4)/            -1, -1, +2, +1     /
DATA tb(16)/            +0.002222                     /,  &
(itb(i,16),i=1,4)/            -1,  0, +2, +1     /
DATA tb(17)/            +0.002072                     /,  &
(itb(i,17),i=1,4)/            -1, -1, +2, -1     /
DATA tb(18)/            +0.001877                     /,  &
(itb(i,18),i=1,4)/            -1, +1,  0, +1     /
DATA tb(19)/            +0.001828                     /,  &
(itb(i,19),i=1,4)/             0, -1, +4, -1     /
DATA tb(20)/            -0.001803                     /,  &
(itb(i,20),i=1,4)/            +1,  0,  0, +1     /
DATA tb(21)/            -0.001750                     /,  &
(itb(i,21),i=1,4)/             0,  0,  0, +3     /
DATA tb(22)/            +0.001570                     /,  &
(itb(i,22),i=1,4)/            -1, +1,  0, -1     /
DATA tb(23)/            -0.001487                     /,  &
(itb(i,23),i=1,4)/             0,  0, +1, +1     /
DATA tb(24)/            -0.001481                     /,  &
(itb(i,24),i=1,4)/            +1, +1,  0, +1     /
DATA tb(25)/            +0.001417                     /,  &
(itb(i,25),i=1,4)/            -1, -1,  0, +1     /
DATA tb(26)/            +0.001350                     /,  &
(itb(i,26),i=1,4)/            -1,  0,  0, +1     /
DATA tb(27)/            +0.001330                     /,  &
(itb(i,27),i=1,4)/             0,  0, -1, +1     /
DATA tb(28)/            +0.001106                     /,  &
(itb(i,28),i=1,4)/             0, +3,  0, +1     /
DATA tb(29)/            +0.001020                     /,  &
(itb(i,29),i=1,4)/             0,  0, +4, -1     /

!  Parallax
!                                         M   M'  D   F
DATA tp( 1)/            +0.051818                     /,  &
(itp(i, 1),i=1,4)/             0, +1,  0,  0     /
DATA tp( 2)/            +0.009531                     /,  &
(itp(i, 2),i=1,4)/             0, -1, +2,  0     /
DATA tp( 3)/            +0.007843                     /,  &
(itp(i, 3),i=1,4)/             0,  0, +2,  0     /
DATA tp( 4)/            +0.002824                     /,  &
(itp(i, 4),i=1,4)/             0, +2,  0,  0     /



!  Whole years & fraction of year, and years since J1900.0
yi=FLOAT(iy-1900)
iy4=MOD(MOD(iy,4)+4,4)
yf=(FLOAT(4*(id-1/(iy4+1))-iy4-2)+4.0*fd)/1461.0
t=yi+yf

!  Moon's mean longitude
elp=d2r*MOD(elp0+elp1i*yf+elp1f*t,360.0)

!  Sun's mean anomaly
em=d2r*MOD(em0+em1f*t,360.0)

!  Moon's mean anomaly
emp=d2r*MOD(emp0+emp1i*yf+emp1f*t,360.0)

!  Moon's mean elongation
d=d2r*MOD(d0+d1i*yf+d1f*t,360.0)

!  Mean distance of the moon from its ascending node
f=d2r*MOD(f0+f1i*yf+f1f*t,360.0)

!  Longitude
el=0.0
eld=0.0
DO n=39,1,-1
  coeff=tl(n)
  cem=FLOAT(itl(1,n))
  cemp=FLOAT(itl(2,n))
  cd=FLOAT(itl(3,n))
  cf=FLOAT(itl(4,n))
  theta=cem*em+cemp*emp+cd*d+cf*f
  thetad=cem*em1+cemp*emp1+cd*d1+cf*f1
  el=el+coeff*SIN(theta)
  eld=eld+coeff*COS(theta)*thetad
END DO
el=el*d2r+elp
eld=ratcon*(eld+elp1/d2r)

!  Latitude
b=0.0
bd=0.0
DO n=29,1,-1
  coeff=tb(n)
  cem=FLOAT(itb(1,n))
  cemp=FLOAT(itb(2,n))
  cd=FLOAT(itb(3,n))
  cf=FLOAT(itb(4,n))
  theta=cem*em+cemp*emp+cd*d+cf*f
  thetad=cem*em1+cemp*emp1+cd*d1+cf*f1
  b=b+coeff*SIN(theta)
  bd=bd+coeff*COS(theta)*thetad
END DO
b=b*d2r
bd=ratcon*bd

!  Parallax
p=0.0
pd=0.0
DO n=4,1,-1
  coeff=tp(n)
  cem=FLOAT(itp(1,n))
  cemp=FLOAT(itp(2,n))
  cd=FLOAT(itp(3,n))
  cf=FLOAT(itp(4,n))
  theta=cem*em+cemp*emp+cd*d+cf*f
  thetad=cem*em1+cemp*emp1+cd*d1+cf*f1
  p=p+coeff*COS(theta)
  pd=pd-coeff*SIN(theta)*thetad
END DO
p=(p+0.950724)*d2r
pd=ratcon*pd

!  Transform parallax to distance (AU, AU/sec)
sp=SIN(p)
r=eradau/sp
rd=-r*pd/sp

!  Longitude, latitude to x,y,z (AU)
CALL sla_cs2c6(el,b,r,eld,bd,rd,v)

!  Mean obliquity
eps=d2r*(23.45229-0.00013*t)
sineps=SIN(eps)
coseps=COS(eps)

!  Rotate Moon position and velocity into equatorial system
pv(1)=v(1)
pv(2)=v(2)*coseps-v(3)*sineps
pv(3)=v(2)*sineps+v(3)*coseps
pv(4)=v(4)
pv(5)=v(5)*coseps-v(6)*sineps
pv(6)=v(5)*sineps+v(6)*coseps

END SUBROUTINE sla_moon
SUBROUTINE sla_mxm (a, b, c)
!+
!     - - - -
!      M X M
!     - - - -

!  Product of two 3x3 matrices:
!      matrix C  =  matrix A  x  matrix B

!  (single precision)

!  Given:
!      A      real(3,3)        matrix
!      B      real(3,3)        matrix

!  Returned:
!      C      real(3,3)        matrix result

!  To comply with the ANSI Fortran 77 standard, A, B and C must
!  be different arrays.  However, the routine is coded so as to
!  work properly on the VAX and many other systems even if this
!  rule is violated.

!  P.T.Wallace   Starlink   5 April 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: a(3,3),b(3,3),c(3,3)

INTEGER :: i,j,k
REAL :: w,wm(3,3)


!  Multiply into scratch matrix
DO i=1,3
  DO j=1,3
    w=0.0
    DO k=1,3
      w=w+a(i,k)*b(k,j)
    END DO
    wm(i,j)=w
  END DO
END DO

!  Return the result
DO j=1,3
  DO i=1,3
    c(i,j)=wm(i,j)
  END DO
END DO

END SUBROUTINE sla_mxm
SUBROUTINE sla_mxv (rm, va, vb)
!+
!     - - - -
!      M X V
!     - - - -

!  Performs the 3-D forward unitary transformation:

!     vector VB = matrix RM * vector VA

!  (single precision)

!  Given:
!     RM       real(3,3)    matrix
!     VA       real(3)      vector

!  Returned:
!     VB       real(3)      result vector

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: rm(3,3),va(3),vb(3)

INTEGER :: i,j
REAL :: w,vw(3)


!  Matrix RM * vector VA -> vector VW
DO j=1,3
  w=0.0
  DO i=1,3
    w=w+rm(j,i)*va(i)
  END DO
  vw(j)=w
END DO

!  Vector VW -> vector VB
DO j=1,3
  vb(j)=vw(j)
END DO

END SUBROUTINE sla_mxv
SUBROUTINE sla_nut (date, rmatn)
!+
!     - - - -
!      N U T
!     - - - -

!  Form the matrix of nutation for a given date - IAU 1980 theory
!  (double precision)

!  References:
!     Final report of the IAU Working Group on Nutation,
!      chairman P.K.Seidelmann, 1980.
!     Kaplan,G.H., 1981, USNO circular no. 163, pA3-6.

!  Given:
!     DATE   dp         TDB (loosely ET) as Modified Julian Date
!                                           (=JD-2400000.5)
!  Returned:
!     RMATN  dp(3,3)    nutation matrix

!  The matrix is in the sense   V(true)  =  RMATN * V(mean)

!  Called:   sla_NUTC, sla_DEULER

!  P.T.Wallace   Starlink   1 January 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,rmatn(3,3)

DOUBLE PRECISION :: dpsi,deps,eps0



!  Nutation components and mean obliquity
CALL sla_nutc(date,dpsi,deps,eps0)

!  Rotation matrix
CALL sla_deuler('XZX',eps0,-dpsi,-(eps0+deps),rmatn)

END SUBROUTINE sla_nut
SUBROUTINE sla_nutc (date, dpsi, deps, eps0)
!+
!     - - - - -
!      N U T C
!     - - - - -

!  Nutation:  longitude & obliquity components and mean
!  obliquity - IAU 1980 theory (double precision)

!  Given:

!     DATE        dp    TDB (loosely ET) as Modified Julian Date
!                                            (JD-2400000.5)
!  Returned:

!     DPSI,DEPS   dp    nutation in longitude,obliquity
!     EPS0        dp    mean obliquity

!  References:
!     Final report of the IAU Working Group on Nutation,
!      chairman P.K.Seidelmann, 1980.
!     Kaplan,G.H., 1981, USNO circular no. 163, pA3-6.

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,dpsi,deps,eps0

DOUBLE PRECISION :: t2as,as2r,u2r
DOUBLE PRECISION :: t,el,el2,el3
DOUBLE PRECISION :: elp,elp2
DOUBLE PRECISION :: f,f2,f4
DOUBLE PRECISION :: d,d2,d4
DOUBLE PRECISION :: om,om2
DOUBLE PRECISION :: dp,de
DOUBLE PRECISION :: a


!  Turns to arc seconds
PARAMETER (t2as=1296000D0)
!  Arc seconds to radians
PARAMETER (as2r=0.484813681109535994D-5)
!  Units of 0.0001 arcsec to radians
PARAMETER (u2r=as2r/1D4)




!  Interval between basic epoch J2000.0 and current epoch (JC)
t=(date-51544.5D0)/36525D0


!  FUNDAMENTAL ARGUMENTS in the FK5 reference system


!  Mean longitude of the Moon minus mean longitude of the Moon's perigee
el=as2r*(485866.733D0+(1325D0*t2as+715922.633D0+(31.310D0+0.064D0*t)*t)*t)

!  Mean longitude of the Sun minus mean longitude of the Sun's perigee
elp=as2r*(1287099.804D0+(99D0*t2as+1292581.224D0+(-0.577D0-0.012D0*t)*t)*t)

!  Mean longitude of the Moon minus mean longitude of the Moon's node
f=as2r*(335778.877D0+(1342D0*t2as+295263.137D0+(-13.257D0+0.011D0*t)*t)*t)

!  Mean elongation of the Moon from the Sun
d=as2r*(1072261.307D0+(1236D0*t2as+1105601.328D0+(-6.891D0+0.019D0*t)*t)*t)

!  Longitude of the mean ascending node of the lunar orbit on the
!   ecliptic, measured from the mean equinox of date
om=as2r*(450160.280D0+(-5D0*t2as-482890.539D0+(7.455D0+0.008D0*t)*t)*t)

!  Multiples of arguments
el2=el+el
el3=el2+el
elp2=elp+elp
f2=f+f
f4=f2+f2
d2=d+d
d4=d2+d2
om2=om+om



!  SERIES FOR THE NUTATION

dp=0D0
de=0D0

!  106
dp=dp+SIN(elp+d)
!  105
dp=dp-SIN(f2+d4+om2)
!  104
dp=dp+SIN(el2+d2)
!  103
dp=dp-SIN(el-f2+d2)
!  102
dp=dp-SIN(el+elp-d2+om)
!  101
dp=dp-SIN(-elp+f2+om)
!  100
dp=dp-SIN(el-f2-d2)
!  99
dp=dp-SIN(elp+d2)
!  98
dp=dp-SIN(f2-d+om2)
!  97
dp=dp-SIN(-f2+om)
!  96
dp=dp+SIN(-el-elp+d2+om)
!  95
dp=dp+SIN(elp+f2+om)
!  94
dp=dp-SIN(el+f2-d2)
!  93
dp=dp+SIN(el3+f2-d2+om2)
!  92
dp=dp+SIN(f4-d2+om2)
!  91
dp=dp-SIN(el+d2+om)
!  90
dp=dp-SIN(el2+f2+d2+om2)
!  89
a=el2+f2-d2+om
dp=dp+SIN(a)
de=de-COS(a)
!  88
dp=dp+SIN(el-elp-d2)
!  87
dp=dp+SIN(-el+f4+om2)
!  86
a=-el2+f2+d4+om2
dp=dp-SIN(a)
de=de+COS(a)
!  85
a=el+f2+d2+om
dp=dp-SIN(a)
de=de+COS(a)
!  84
a=el+elp+f2-d2+om2
dp=dp+SIN(a)
de=de-COS(a)
!  83
dp=dp-SIN(el2-d4)
!  82
a=-el+f2+d4+om2
dp=dp-2D0*SIN(a)
de=de+COS(a)
!  81
a=-el2+f2+d2+om2
dp=dp+SIN(a)
de=de-COS(a)
!  80
dp=dp-SIN(el-d4)
!  79
a=-el+om2
dp=dp+SIN(a)
de=de-COS(a)
!  78
a=f2+d+om2
dp=dp+2D0*SIN(a)
de=de-COS(a)
!  77
dp=dp+2D0*SIN(el3)
!  76
a=el+om2
dp=dp-2D0*SIN(a)
de=de+COS(a)
!  75
a=el2+om
dp=dp+2D0*SIN(a)
de=de-COS(a)
!  74
a=-el+f2-d2+om
dp=dp-2D0*SIN(a)
de=de+COS(a)
!  73
a=el+elp+f2+om2
dp=dp+2D0*SIN(a)
de=de-COS(a)
!  72
a=-elp+f2+d2+om2
dp=dp-3D0*SIN(a)
de=de+COS(a)
!  71
a=el3+f2+om2
dp=dp-3D0*SIN(a)
de=de+COS(a)
!  70
a=-el2+om
dp=dp-2D0*SIN(a)
de=de+COS(a)
!  69
a=-el-elp+f2+d2+om2
dp=dp-3D0*SIN(a)
de=de+COS(a)
!  68
a=el-elp+f2+om2
dp=dp-3D0*SIN(a)
de=de+COS(a)
!  67
dp=dp+3D0*SIN(el+f2)
!  66
dp=dp-3D0*SIN(el+elp)
!  65
dp=dp-4D0*SIN(d)
!  64
dp=dp+4D0*SIN(el-f2)
!  63
dp=dp-4D0*SIN(elp-d2)
!  62
a=el2+f2+om
dp=dp-5D0*SIN(a)
de=de+3D0*COS(a)
!  61
dp=dp+5D0*SIN(el-elp)
!  60
a=-d2+om
dp=dp-5D0*SIN(a)
de=de+3D0*COS(a)
!  59
a=el+f2-d2+om
dp=dp+6D0*SIN(a)
de=de-3D0*COS(a)
!  58
a=f2+d2+om
dp=dp-7D0*SIN(a)
de=de+3D0*COS(a)
!  57
a=d2+om
dp=dp-6D0*SIN(a)
de=de+3D0*COS(a)
!  56
a=el2+f2-d2+om2
dp=dp+6D0*SIN(a)
de=de-3D0*COS(a)
!  55
dp=dp+6D0*SIN(el+d2)
!  54
a=el+f2+d2+om2
dp=dp-8D0*SIN(a)
de=de+3D0*COS(a)
!  53
a=-elp+f2+om2
dp=dp-7D0*SIN(a)
de=de+3D0*COS(a)
!  52
a=elp+f2+om2
dp=dp+7D0*SIN(a)
de=de-3D0*COS(a)
!  51
dp=dp-7D0*SIN(el+elp-d2)
!  50
a=-el+f2+d2+om
dp=dp-10D0*SIN(a)
de=de+5D0*COS(a)
!  49
a=el-d2+om
dp=dp-13D0*SIN(a)
de=de+7D0*COS(a)
!  48
a=-el+d2+om
dp=dp+16D0*SIN(a)
de=de-8D0*COS(a)
!  47
a=-el+f2+om
dp=dp+21D0*SIN(a)
de=de-10D0*COS(a)
!  46
dp=dp+26D0*SIN(f2)
de=de-COS(f2)
!  45
a=el2+f2+om2
dp=dp-31D0*SIN(a)
de=de+13D0*COS(a)
!  44
a=el+f2-d2+om2
dp=dp+29D0*SIN(a)
de=de-12D0*COS(a)
!  43
dp=dp+29D0*SIN(el2)
de=de-COS(el2)
!  42
a=f2+d2+om2
dp=dp-38D0*SIN(a)
de=de+16D0*COS(a)
!  41
a=el+f2+om
dp=dp-51D0*SIN(a)
de=de+27D0*COS(a)
!  40
a=-el+f2+d2+om2
dp=dp-59D0*SIN(a)
de=de+26D0*COS(a)
!  39
a=-el+om
dp=dp+(-58D0-0.1D0*t)*SIN(a)
de=de+32D0*COS(a)
!  38
a=el+om
dp=dp+(63D0+0.1D0*t)*SIN(a)
de=de-33D0*COS(a)
!  37
dp=dp+63D0*SIN(d2)
de=de-2D0*COS(d2)
!  36
a=-el+f2+om2
dp=dp+123D0*SIN(a)
de=de-53D0*COS(a)
!  35
a=el-d2
dp=dp-158D0*SIN(a)
de=de-COS(a)
!  34
a=el+f2+om2
dp=dp-301D0*SIN(a)
de=de+(129D0-0.1D0*t)*COS(a)
!  33
a=f2+om
dp=dp+(-386D0-0.4D0*t)*SIN(a)
de=de+200D0*COS(a)
!  32
dp=dp+(712D0+0.1D0*t)*SIN(el)
de=de-7D0*COS(el)
!  31
a=f2+om2
dp=dp+(-2274D0-0.2D0*t)*SIN(a)
de=de+(977D0-0.5D0*t)*COS(a)
!  30
dp=dp-SIN(elp+f2-d2)
!  29
dp=dp+SIN(-el+d+om)
!  28
dp=dp+SIN(elp+om2)
!  27
dp=dp-SIN(elp-f2+d2)
!  26
dp=dp+SIN(-f2+d2+om)
!  25
dp=dp+SIN(el2+elp-d2)
!  24
dp=dp-4D0*SIN(el-d)
!  23
a=elp+f2-d2+om
dp=dp+4D0*SIN(a)
de=de-2D0*COS(a)
!  22
a=el2-d2+om
dp=dp+4D0*SIN(a)
de=de-2D0*COS(a)
!  21
a=-elp+f2-d2+om
dp=dp-5D0*SIN(a)
de=de+3D0*COS(a)
!  20
a=-el2+d2+om
dp=dp-6D0*SIN(a)
de=de+3D0*COS(a)
!  19
a=-elp+om
dp=dp-12D0*SIN(a)
de=de+6D0*COS(a)
!  18
a=elp2+f2-d2+om2
dp=dp+(-16D0+0.1D0*t)*SIN(a)
de=de+7D0*COS(a)
!  17
a=elp+om
dp=dp-15D0*SIN(a)
de=de+9D0*COS(a)
!  16
dp=dp+(17D0-0.1D0*t)*SIN(elp2)
!  15
dp=dp-22D0*SIN(f2-d2)
!  14
a=el2-d2
dp=dp+48D0*SIN(a)
de=de+COS(a)
!  13
a=f2-d2+om
dp=dp+(129D0+0.1D0*t)*SIN(a)
de=de-70D0*COS(a)
!  12
a=-elp+f2-d2+om2
dp=dp+(217D0-0.5D0*t)*SIN(a)
de=de+(-95D0+0.3D0*t)*COS(a)
!  11
a=elp+f2-d2+om2
dp=dp+(-517D0+1.2D0*t)*SIN(a)
de=de+(224D0-0.6D0*t)*COS(a)
!  10
dp=dp+(1426D0-3.4D0*t)*SIN(elp)
de=de+(54D0-0.1D0*t)*COS(elp)
!  9
a=f2-d2+om2
dp=dp+(-13187D0-1.6D0*t)*SIN(a)
de=de+(5736D0-3.1D0*t)*COS(a)
!  8
dp=dp+SIN(el2-f2+om)
!  7
a=-elp2+f2-d2+om
dp=dp-2D0*SIN(a)
de=de+1D0*COS(a)
!  6
dp=dp-3D0*SIN(el-elp-d)
!  5
a=-el2+f2+om2
dp=dp-3D0*SIN(a)
de=de+1D0*COS(a)
!  4
dp=dp+11D0*SIN(el2-f2)
!  3
a=-el2+f2+om
dp=dp+46D0*SIN(a)
de=de-24D0*COS(a)
!  2
dp=dp+(2062D0+0.2D0*t)*SIN(om2)
de=de+(-895D0+0.5D0*t)*COS(om2)
!  1
dp=dp+(-171996D0-174.2D0*t)*SIN(om)
de=de+(92025D0+8.9D0*t)*COS(om)

!  Convert results to radians
dpsi=dp*u2r
deps=de*u2r

!  Mean obliquity
eps0=as2r*(84381.448D0+(-46.8150D0+  &
(-0.00059D0+0.001813D0*t)*t)*t)

END SUBROUTINE sla_nutc
SUBROUTINE sla_oap (TYPE, ob1, ob2, date, dut, elongm, phim,  &
hm, xp, yp, tdk, pmb, rh, wl, tlr,rap, dap)
!+
!     - - - -
!      O A P
!     - - - -

!  Observed to apparent place

!  Given:
!     TYPE   c*(*)  type of coordinates - 'R', 'H' or 'A' (see below)
!     OB1    d      observed Az, HA or RA (radians; Az is N=0,E=90)
!     OB2    d      observed ZD or Dec (radians)
!     DATE   d      UTC date/time (modified Julian Date, JD-2400000.5)
!     DUT    d      delta UT:  UT1-UTC (UTC seconds)
!     ELONGM d      mean longitude of the observer (radians, east +ve)
!     PHIM   d      mean geodetic latitude of the observer (radians)
!     HM     d      observer's height above sea level (metres)
!     XP     d      polar motion x-coordinate (radians)
!     YP     d      polar motion y-coordinate (radians)
!     TDK    d      local ambient temperature (DegK; std=273.155D0)
!     PMB    d      local atmospheric pressure (mB; std=1013.25D0)
!     RH     d      local relative humidity (in the range 0D0-1D0)
!     WL     d      effective wavelength (micron, e.g. 0.55D0)
!     TLR    d      tropospheric lapse rate (DegK/metre, e.g. 0.0065D0)

!  Returned:
!     RAP    d      geocentric apparent right ascension
!     DAP    d      geocentric apparent declination

!  Notes:

!  1)  Only the first character of the TYPE argument is significant.
!      'R' or 'r' indicates that OBS1 and OBS2 are the observed Right
!      Ascension and Declination;  'H' or 'h' indicates that they are
!      Hour Angle (West +ve) and Declination;  anything else ('A' or
!      'a' is recommended) indicates that OBS1 and OBS2 are Azimuth
!      (North zero, East is 90 deg) and zenith distance.  (Zenith
!      distance is used rather than elevation in order to reflect the
!      fact that no allowance is made for depression of the horizon.)

!  2)  The accuracy of the result is limited by the corrections for
!      refraction.  Providing the meteorological parameters are
!      known accurately and there are no gross local effects, the
!      predicted apparent RA,Dec should be within about 0.1 arcsec
!      for a zenith distance of less than 70 degrees.  Even at a
!      topocentric zenith distance of 90 degrees, the accuracy in
!      elevation should be better than 1 arcmin;  useful results
!      are available for a further 3 degrees, beyond which the
!      sla_REFRO routine returns a fixed value of the refraction.
!      The complementary routines sla_AOP (or sla_AOPQK) and sla_OAP
!      (or sla_OAPQK) are self-consistent to better than 1 micro-
!      arcsecond all over the celestial sphere.

!  3)  It is advisable to take great care with units, as even
!      unlikely values of the input parameters are accepted and
!      processed in accordance with the models used.

!  4)  "Observed" Az,El means the position that would be seen by a
!      perfect theodolite located at the observer.  This is
!      related to the observed HA,Dec via the standard rotation, using
!      the geodetic latitude (corrected for polar motion), while the
!      observed HA and RA are related simply through the local
!      apparent ST.  "Observed" RA,Dec or HA,Dec thus means the
!      position that would be seen by a perfect equatorial located
!      at the observer and with its polar axis aligned to the
!      Earth's axis of rotation (n.b. not to the refracted pole).
!      By removing from the observed place the effects of
!      atmospheric refraction and diurnal aberration, the
!      geocentric apparent RA,Dec is obtained.

!  5)  Frequently, mean rather than apparent RA,Dec will be required,
!      in which case further transformations will be necessary.  The
!      sla_AMP etc routines will convert the apparent RA,Dec produced
!      by the present routine into an "FK5" (J2000) mean place, by
!      allowing for the Sun's gravitational lens effect, annual
!      aberration, nutation and precession.  Should "FK4" (1950)
!      coordinates be needed, the routines sla_FK524 etc will also
!      need to be applied.

!  6)  To convert to apparent RA,Dec the coordinates read from a
!      real telescope, corrections would have to be applied for
!      encoder zero points, gear and encoder errors, tube flexure,
!      the position of the rotator axis and the pointing axis
!      relative to it, non-perpendicularity between the mounting
!      axes, and finally for the tilt of the azimuth or polar axis
!      of the mounting (with appropriate corrections for mount
!      flexures).  Some telescopes would, of course, exhibit other
!      properties which would need to be accounted for at the
!      appropriate point in the sequence.

!  7)  The star-independent apparent-to-observed-place parameters
!      in AOPRMS may be computed by means of the sla_AOPPA routine.
!      If nothing has changed significantly except the time, the
!      sla_AOPPAT routine may be used to perform the requisite
!      partial recomputation of AOPRMS.

!  8)  The DATE argument is UTC expressed as an MJD.  This is,
!      strictly speaking, wrong, because of leap seconds.  However,
!      as long as the delta UT and the UTC are consistent there
!      are no difficulties, except during a leap second.  In this
!      case, the start of the 61st second of the final minute should
!      begin a new MJD day and the old pre-leap delta UT should
!      continue to be used.  As the 61st second completes, the MJD
!      should revert to the start of the day as, simultaneously,
!      the delta UTC changes by one second to its post-leap new value.

!  9)  The delta UT (UT1-UTC) is tabulated in IERS circulars and
!      elsewhere.  It increases by exactly one second at the end of
!      each UTC leap second, introduced in order to keep delta UT
!      within +/- 0.9 seconds.

!  10) IMPORTANT -- TAKE CARE WITH THE LONGITUDE SIGN CONVENTION.
!      The longitude required by the present routine is east-positive,
!      in accordance with geographical convention (and right-handed).
!      In particular, note that the longitudes returned by the
!      sla_OBS routine are west-positive, following astronomical
!      usage, and must be reversed in sign before use in the present
!      routine.

!  11) The polar coordinates XP,YP can be obtained from IERS
!      circulars and equivalent publications.  The maximum amplitude
!      is about 0.3 arcseconds.  If XP,YP values are unavailable,
!      use XP=YP=0D0.  See page B60 of the 1988 Astronomical Almanac
!      for a definition of the two angles.

!  12) The height above sea level of the observing station, HM,
!      can be obtained from the Astronomical Almanac (Section J
!      in the 1988 edition), or via the routine sla_OBS.  If P,
!      the pressure in millibars, is available, an adequate
!      estimate of HM can be obtained from the expression

!            HM ~ -8149.9415D0*LOG(P/1013.25D0).

!      (See Astrophysical Quantities, C.W.Allen, 3rd edition,
!      section 52.)  Similarly, if the pressure P is not known,
!      it can be estimated from the height of the observing
!      station, HM as follows:

!            P ~ 1013.25D0*EXP(-HM/8149.9415D0).

!      Note, however, that the refraction is proportional to the
!      pressure and that an accurate P value is important for
!      precise work.

!  13) The azimuths etc used by the present routine are with respect
!      to the celestial pole.  Corrections from the terrestrial pole
!      can be computed using sla_POLMO.

!  Called:  sla_AOPPA, sla_OAPQK

!  P.T.Wallace   Starlink   22 February 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) TYPE
DOUBLE PRECISION :: ob1,ob2,date,dut,elongm,phim,hm,  &
xp,yp,tdk,pmb,rh,wl,tlr,rap,dap

DOUBLE PRECISION :: aoprms(14)


CALL sla_aoppa(date,dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr,aoprms)
CALL sla_oapqk(TYPE,ob1,ob2,aoprms,rap,dap)

END SUBROUTINE sla_oap
SUBROUTINE sla_oapqk (TYPE, ob1, ob2, aoprms, rap, dap)
!+
!     - - - - - -
!      O A P Q K
!     - - - - - -

!  Quick observed to apparent place

!  Given:
!     TYPE   c*(*)  type of coordinates - 'R', 'H' or 'A' (see below)
!     OB1    d      observed Az, HA or RA (radians; Az is N=0,E=90)
!     OB2    d      observed ZD or Dec (radians)
!     AOPRMS d(14)  star-independent apparent-to-observed parameters:

!       (1)      geodetic latitude (radians)
!       (2,3)    sine and cosine of geodetic latitude
!       (4)      magnitude of diurnal aberration vector
!       (5)      height (HM)
!       (6)      ambient temperature (T)
!       (7)      pressure (P)
!       (8)      relative humidity (RH)
!       (9)      wavelength (WL)
!       (10)     lapse rate (TLR)
!       (11,12)  refraction constants A and B (radians)
!       (13)     longitude + eqn of equinoxes + sidereal DUT (radians)
!       (14)     local apparent sidereal time (radians)

!  Returned:
!     RAP    d      geocentric apparent right ascension
!     DAP    d      geocentric apparent declination

!  Notes:

!  1)  Only the first character of the TYPE argument is significant.
!      'R' or 'r' indicates that OBS1 and OBS2 are the observed Right
!      Ascension and Declination;  'H' or 'h' indicates that they are
!      Hour Angle (West +ve) and Declination;  anything else ('A' or
!      'a' is recommended) indicates that OBS1 and OBS2 are Azimuth
!      (North zero, East is 90 deg) and zenith distance.  (Zenith
!      distance is used rather than elevation in order to reflect the
!      fact that no allowance is made for depression of the horizon.)

!  2)  The accuracy of the result is limited by the corrections for
!      refraction.  Providing the meteorological parameters are
!      known accurately and there are no gross local effects, the
!      predicted apparent RA,Dec should be within about 0.1 arcsec
!      for a zenith distance of less than 70 degrees.  Even at a
!      topocentric zenith distance of 90 degrees, the accuracy in
!      elevation should be better than 1 arcmin;  useful results
!      are available for a further 3 degrees, beyond which the
!      sla_REFRO routine returns a fixed value of the refraction.
!      The complementary routines sla_AOP (or sla_AOPQK) and sla_OAP
!      (or sla_OAPQK) are self-consistent to better than 1 micro-
!      arcsecond all over the celestial sphere.

!  3)  It is advisable to take great care with units, as even
!      unlikely values of the input parameters are accepted and
!      processed in accordance with the models used.

!  5)  "Observed" Az,El means the position that would be seen by a
!      perfect theodolite located at the observer.  This is
!      related to the observed HA,Dec via the standard rotation, using
!      the geodetic latitude (corrected for polar motion), while the
!      observed HA and RA are related simply through the local
!      apparent ST.  "Observed" RA,Dec or HA,Dec thus means the
!      position that would be seen by a perfect equatorial located
!      at the observer and with its polar axis aligned to the
!      Earth's axis of rotation (n.b. not to the refracted pole).
!      By removing from the observed place the effects of
!      atmospheric refraction and diurnal aberration, the
!      geocentric apparent RA,Dec is obtained.

!  5)  Frequently, mean rather than apparent RA,Dec will be required,
!      in which case further transformations will be necessary.  The
!      sla_AMP etc routines will convert the apparent RA,Dec produced
!      by the present routine into an "FK5" (J2000) mean place, by
!      allowing for the Sun's gravitational lens effect, annual
!      aberration, nutation and precession.  Should "FK4" (1950)
!      coordinates be needed, the routines sla_FK524 etc will also
!      need to be applied.

!  6)  To convert to apparent RA,Dec the coordinates read from a
!      real telescope, corrections would have to be applied for
!      encoder zero points, gear and encoder errors, tube flexure,
!      the position of the rotator axis and the pointing axis
!      relative to it, non-perpendicularity between the mounting
!      axes, and finally for the tilt of the azimuth or polar axis
!      of the mounting (with appropriate corrections for mount
!      flexures).  Some telescopes would, of course, exhibit other
!      properties which would need to be accounted for at the
!      appropriate point in the sequence.

!  7)  The star-independent apparent-to-observed-place parameters
!      in AOPRMS may be computed by means of the sla_AOPPA routine.
!      If nothing has changed significantly except the time, the
!      sla_AOPPAT routine may be used to perform the requisite
!      partial recomputation of AOPRMS.

!  8) The azimuths etc used by the present routine are with respect
!     to the celestial pole.  Corrections from the terrestrial pole
!     can be computed using sla_POLMO.

!  Called:  sla_DCS2C, sla_DCC2S, sla_REFRO, sla_DRANRM

!  P.T.Wallace   Starlink   23 June 1997

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER*(*) TYPE
DOUBLE PRECISION :: ob1,ob2,aoprms(14),rap,dap

!  Breakpoint for fast/slow refraction algorithm:
!  ZD greater than arctan(4), (see sla_REFCO routine)
!  or vector Z less than cosine(arctan(Z)) = 1/sqrt(17)
DOUBLE PRECISION :: zbreak
PARAMETER (zbreak=0.242535625D0)

CHARACTER :: c
DOUBLE PRECISION :: c1,c2,sphi,cphi,st,ce,xaeo,yaeo,zaeo,v(3),  &
xmhdo,ymhdo,zmhdo,az,sz,zdo,tz,dref,zdt,  &
xaet,yaet,zaet,xmhda,ymhda,zmhda,diurab,f,hma




!  Coordinate type
c = TYPE(1:1)

!  Coordinates
c1 = ob1
c2 = ob2

!  Sin, cos of latitude
sphi = aoprms(2)
cphi = aoprms(3)

!  Local apparent sidereal time
st = aoprms(14)

!  Standardise coordinate type
IF (c == 'R'.OR.c == 'r') THEN
  c = 'R'
ELSE IF (c == 'H'.OR.c == 'h') THEN
  c = 'H'
ELSE
  c = 'A'
END IF

!  If Az,ZD convert to Cartesian (S=0,E=90)
IF (c == 'A') THEN
  ce = SIN(c2)
  xaeo = -COS(c1)*ce
  yaeo = SIN(c1)*ce
  zaeo = COS(c2)
ELSE
  
!     If RA,Dec convert to HA,Dec
  IF (c == 'R') THEN
    c1 = st-c1
  END IF
  
!     To Cartesian -HA,Dec
  CALL sla_dcs2c(-c1,c2,v)
  xmhdo = v(1)
  ymhdo = v(2)
  zmhdo = v(3)
  
!     To Cartesian Az,El (S=0,E=90)
  xaeo = sphi*xmhdo-cphi*zmhdo
  yaeo = ymhdo
  zaeo = cphi*xmhdo+sphi*zmhdo
END IF

!  Azimuth (S=0,E=90)
IF (xaeo /= 0D0.OR.yaeo /= 0D0) THEN
  az = ATAN2(yaeo,xaeo)
ELSE
  az = 0D0
END IF

!  Sine of observed ZD, and observed ZD
sz = SQRT(xaeo*xaeo+yaeo*yaeo)
zdo = ATAN2(sz,zaeo)


!  Refraction
!  ----------

!  Large zenith distance?
IF (zaeo >= zbreak) THEN
  
!     Fast algorithm using two constant model
  tz = sz/zaeo
  dref = aoprms(11)*tz+aoprms(12)*tz*tz*tz
  
ELSE
  
!     Rigorous algorithm for large ZD
  CALL sla_refro(zdo,aoprms(5),aoprms(6),aoprms(7),aoprms(8),  &
  aoprms(9),aoprms(1),aoprms(10),1D-8,dref)
END IF

zdt = zdo+dref

!  To Cartesian Az,ZD
ce = SIN(zdt)
xaet = COS(az)*ce
yaet = SIN(az)*ce
zaet = COS(zdt)

!  Cartesian Az,ZD to Cartesian -HA,Dec
xmhda = sphi*xaet+cphi*zaet
ymhda = yaet
zmhda = -cphi*xaet+sphi*zaet

!  Diurnal aberration
diurab = -aoprms(4)
f = (1D0-diurab*ymhda)
v(1) = f*xmhda
v(2) = f*(ymhda+diurab)
v(3) = f*zmhda

!  To spherical -HA,Dec
CALL sla_dcc2s(v,hma,dap)

!  Right Ascension
rap = sla_dranrm(st+hma)

END SUBROUTINE sla_oapqk
SUBROUTINE sla_obs (n, c, NAME, w, p, h)
!+
!     - - - -
!      O B S
!     - - - -

!  Parameters of selected groundbased observing stations

!  Given:
!     N       int     number specifying observing station

!  Either given or returned
!     C       c*(*)   identifier specifying observing station

!  Returned:
!     NAME    c*(*)   name of specified observing station
!     W       dp      longitude (radians, West +ve)
!     P       dp      geodetic latitude (radians, North +ve)
!     H       dp      height above sea level (metres)

!  Notes:

!     Station identifiers C may be up to 10 characters long,
!     and station names NAME may be up to 40 characters long.

!     C and N are alternative ways of specifying the observing
!     station.  The C option, which is the most generally useful,
!     may be selected by specifying an N value of zero or less.
!     If N is 1 or more, the parameters of the Nth station
!     in the currently supported list are interrogated, and
!     the station identifier C is returned as well as NAME, W,
!     P and H.

!     If the station parameters are not available, either because
!     the station identifier C is not recognized, or because an
!     N value greater than the number of stations supported is
!     given, a name of '?' is returned and C, W, P and H are left
!     in their current states.

!     Programs can obtain a list of all currently supported
!     stations by calling the routine repeatedly, with N=1,2,3...
!     When NAME='?' is seen, the list of stations has been
!     exhausted.

!     Station numbers, identifiers, names and other details are
!     subject to change and should not be hardwired into
!     application programs.

!     All station identifiers C are uppercase only;  lowercase
!     characters must be converted to uppercase by the calling
!     program.  The station names returned may contain both upper-
!     and lowercase.  All characters up to the first space are
!     checked;  thus an abbreviated ID will return the parameters
!     for the first station in the list which matches the
!     abbreviation supplied, and no station in the list will ever
!     contain embedded spaces.  C must not have leading spaces.

!     IMPORTANT -- BEWARE OF THE LONGITUDE SIGN CONVENTION.  The
!     longitude returned by sla_OBS is west-positive in accordance
!     with astronomical usage.  However, this sign convention is
!     left-handed and is the opposite of the one used by geographers;
!     elsewhere in SLALIB the preferable east-positive convention is
!     used.  In particular, note that for use in sla_AOP, sla_AOPPA
!     and sla_OAP the sign of the longitude must be reversed.

!     Users are urged to inform the author of any improvements
!     they would like to see made.  For example:

!         typographical corrections
!         more accurate parameters
!         better station identifiers or names
!         additional stations

!  P.T.Wallace   Starlink   2 September 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: n
CHARACTER :: c*(*),NAME*(*)
DOUBLE PRECISION :: w,p,h

INTEGER :: nmax,m,ns,i
CHARACTER*10 cc

DOUBLE PRECISION :: as2r,west,north,east,south
INTEGER :: id,iam
REAL :: as
PARAMETER (as2r=0.484813681109535994D-5)

!  Table of station identifiers
PARAMETER (nmax=69)
CHARACTER*10 ctab(nmax)
DATA ctab  (1) /'AAT       '/
DATA ctab  (2) /'LPO4.2    '/
DATA ctab  (3) /'LPO2.5    '/
DATA ctab  (4) /'LPO1      '/
DATA ctab  (5) /'LICK120   '/
DATA ctab  (6) /'MMT       '/
DATA ctab  (7) /'DAO72     '/
DATA ctab  (8) /'DUPONT    '/
DATA ctab  (9) /'MTHOP1.5  '/
DATA ctab (10) /'STROMLO74 '/
DATA ctab (11) /'ANU2.3    '/
DATA ctab (12) /'GBVA140   '/
DATA ctab (13) /'TOLOLO4M  '/
DATA ctab (14) /'TOLOLO1.5M'/
DATA ctab (15) /'TIDBINBLA '/
DATA ctab (16) /'BLOEMF    '/
DATA ctab (17) /'BOSQALEGRE'/
DATA ctab (18) /'FLAGSTF61 '/
DATA ctab (19) /'LOWELL72  '/
DATA ctab (20) /'HARVARD   '/
DATA ctab (21) /'OKAYAMA   '/
DATA ctab (22) /'KPNO158   '/
DATA ctab (23) /'KPNO90    '/
DATA ctab (24) /'KPNO84    '/
DATA ctab (25) /'KPNO36FT  '/
DATA ctab (26) /'KOTTAMIA  '/
DATA ctab (27) /'ESO3.6    '/
DATA ctab (28) /'MAUNAK88  '/
DATA ctab (29) /'UKIRT     '/
DATA ctab (30) /'QUEBEC1.6 '/
DATA ctab (31) /'MTEKAR    '/
DATA ctab (32) /'MTLEMMON60'/
DATA ctab (33) /'MCDONLD2.7'/
DATA ctab (34) /'MCDONLD2.1'/
DATA ctab (35) /'PALOMAR200'/
DATA ctab (36) /'PALOMAR60 '/
DATA ctab (37) /'DUNLAP74  '/
DATA ctab (38) /'HPROV1.93 '/
DATA ctab (39) /'HPROV1.52 '/
DATA ctab (40) /'SANPM83   '/
DATA ctab (41) /'SAAO74    '/
DATA ctab (42) /'TAUTNBG   '/
DATA ctab (43) /'CATALINA61'/
DATA ctab (44) /'STEWARD90 '/
DATA ctab (45) /'USSR6     '/
DATA ctab (46) /'ARECIBO   '/
DATA ctab (47) /'CAMB5KM   '/
DATA ctab (48) /'CAMB1MILE '/
DATA ctab (49) /'EFFELSBERG'/
DATA ctab (50) /'GBVA300   '/
DATA ctab (51) /'JODRELL1  '/
DATA ctab (52) /'PARKES    '/
DATA ctab (53) /'VLA       '/
DATA ctab (54) /'SUGARGROVE'/
DATA ctab (55) /'USSR600   '/
DATA ctab (56) /'NOBEYAMA  '/
DATA ctab (57) /'JCMT      '/
DATA ctab (58) /'ESONTT    '/
DATA ctab (59) /'ST.ANDREWS'/
DATA ctab (60) /'APO3.5    '/
DATA ctab (61) /'KECK1     '/
DATA ctab (62) /'TAUTSCHM  '/
DATA ctab (63) /'PALOMAR48 '/
DATA ctab (64) /'UKST      '/
DATA ctab (65) /'KISO      '/
DATA ctab (66) /'ESOSCHM   '/
DATA ctab (67) /'ATCA      '/
DATA ctab (68) /'MOPRA     '/
DATA ctab (69) /'SUBARU    '/

!  Degrees, arcminutes, arcseconds to radians
west(id,iam,as)=as2r*(DBLE(60*(60*id+iam))+DBLE(as))
north(id,iam,as)=west(id,iam,as)
east(id,iam,as)=-west(id,iam,as)
south(id,iam,as)=-west(id,iam,as)




!  Station specified by number or identifier?
IF (n > 0) THEN
  
!     Station specified by number
  m=n
  IF (m <= nmax) c=ctab(m)
  
ELSE
  
!     Station specified by identifier:  determine corresponding number
  cc=c
  DO ns=1,nmax
    DO i=1,10
      IF (cc(i:i) == ' ') GO TO 5
      IF (cc(i:i) /= ctab(ns)(i:i)) GO TO 1
    END DO
    GO TO 5
    1          CONTINUE
  END DO
  ns=nmax+1
  5       CONTINUE
  IF (c(1:1) /= ' ') THEN
    m=ns
  ELSE
    m=nmax+1
  END IF
  
END IF


!  Return parameters of Mth station
!  --------------------------------

GO TO (10,20,30,40,50,60,70,80,90,100,  &
110,120,130,140,150,160,170,180,190,200,  &
210,220,230,240,250,260,270,280,290,300,  &
310,320,330,340,350,360,370,380,390,400,  &
410,420,430,440,450,460,470,480,490,500,  &
510,520,530,540,550,560,570,580,590,600,610,620,630,640,650,660,670) m
GO TO 9000

!  AAT (Observer's Guide)                                            AAT
10   CONTINUE
NAME='Anglo-Australian 3.9m Telescope'
w=east(149,03,57.91)
p=south(31,16,37.34)
h=1164D0
GO TO 9999

!  WHT (Gemini, April 1987)                                       LPO4.2
20   CONTINUE
NAME='William Herschel 4.2m Telescope'
w=west(17,52,53.9)
p=north(28,45,38.1)
h=2332D0
GO TO 9999

!  INT (Gemini, April 1987)                                       LPO2.5
30   CONTINUE
NAME='Isaac Newton 2.5m Telescope'
w=west(17,52,39.5)
p=north(28,45,43.2)
h=2336D0
GO TO 9999

!  JKT (Gemini, April 1987)                                         LPO1
40   CONTINUE
NAME='Jacobus Kapteyn 1m Telescope'
w=west(17,52,41.2)
p=north(28,45,39.9)
h=2364D0
GO TO 9999

!  Lick 120" (1984 Almanac)                                      LICK120
50   CONTINUE
NAME='Lick 120 inch'
w=west(121,38,09.9)
p=north(37,20,35.2)
h=1290D0
GO TO 9999

!  MMT (1984 Almanac)                                                MMT
60   CONTINUE
NAME='MMT, Mt Hopkins'
w=west(110,53,04.3)
p=north(31,41,19.6)
h=2608D0
GO TO 9999

!  Victoria B.C. 1.85m (1984 Almanac)                              DAO72
70   CONTINUE
NAME='DAO Victoria BC 1.85 metre'
w=west(123,25,01.18)
p=north(48,31,11.9)
h=238D0
GO TO 9999

!  Las Campanas (1983 Almanac)                                    DUPONT
80   CONTINUE
NAME='Du Pont 2.5m Telescope, Las Campanas'
w=west(70,42,9.)
p=south(29,00,11.)
h=2280D0
GO TO 9999

!  Mt Hopkins 1.5m (1983 Almanac)                               MTHOP1.5
90   CONTINUE
NAME='Mt Hopkins 1.5 metre'
w=west(110,52,39.00)
p=north(31,40,51.4)
h=2344D0
GO TO 9999

!  Mt Stromlo 74" (1983 Almanac)                               STROMLO74
100  CONTINUE
NAME='Mount Stromlo 74 inch'
w=east(149,00,27.59)
p=south(35,19,14.3)
h=767D0
GO TO 9999

!  ANU 2.3m, SSO (Gary Hovey)                                     ANU2.3
110  CONTINUE
NAME='Siding Spring 2.3 metre'
w=east(149,03,40.3)
p=south(31,16,24.1)
h=1149D0
GO TO 9999

!  Greenbank 140' (1983 Almanac)                                 GBVA140
120  CONTINUE
NAME='Greenbank 140 foot'
w=west(79,50,09.61)
p=north(38,26,15.4)
h=881D0
GO TO 9999

!  Cerro Tololo 4m (1982 Almanac)                               TOLOLO4M
130  CONTINUE
NAME='Cerro Tololo 4 metre'
w=west(70,48,53.6)
p=south(30,09,57.8)
h=2235D0
GO TO 9999

!  Cerro Tololo 1.5m (1982 Almanac)                           TOLOLO1.5M
140  CONTINUE
NAME='Cerro Tololo 1.5 metre'
w=west(70,48,54.5)
p=south(30,09,56.3)
h=2225D0
GO TO 9999

!  Tidbinbilla 64m (1982 Almanac)                              TIDBINBLA
150  CONTINUE
NAME='Tidbinbilla 64 metre'
w=east(148,58,48.20)
p=south(35,24,14.3)
h=670D0
GO TO 9999

!  Bloemfontein 1.52m (1981 Almanac)                              BLOEMF
160  CONTINUE
NAME='Bloemfontein 1.52 metre'
w=east(26,24,18.)
p=south(29,02,18.)
h=1387D0
GO TO 9999

!  Bosque Alegre 1.54m (1981 Almanac)                         BOSQALEGRE
170  CONTINUE
NAME='Bosque Alegre 1.54 metre'
w=west(64,32,48.0)
p=south(31,35,53.)
h=1250D0
GO TO 9999

!  USNO 61" astrographic reflector, Flagstaff (1981 Almanac)   FLAGSTF61
180  CONTINUE
NAME='USNO 61 inch astrograph, Flagstaff'
w=west(111,44,23.6)
p=north(35,11,02.5)
h=2316D0
GO TO 9999

!  Lowell 72" (1981 Almanac)                                    LOWELL72
190  CONTINUE
NAME='Perkins 72 inch, Lowell'
w=west(111,32,09.3)
p=north(35,05,48.6)
h=2198D0
GO TO 9999

!  Harvard 1.55m (1981 Almanac)                                  HARVARD
200  CONTINUE
NAME='Harvard College Observatory 1.55m'
w=west(71,33,29.32)
p=north(42,30,19.0)
h=185D0
GO TO 9999

!  Okayama 1.88m (1981 Almanac)                                  OKAYAMA
210  CONTINUE
NAME='Okayama 1.88 metre'
w=east(133,35,47.29)
p=north(34,34,26.1)
h=372D0
GO TO 9999

!  Kitt Peak Mayall 4m (1981 Almanac)                            KPNO158
220  CONTINUE
NAME='Kitt Peak 158 inch'
w=west(111,35,57.61)
p=north(31,57,50.3)
h=2120D0
GO TO 9999

!  Kitt Peak 90 inch (1981 Almanac)                               KPNO90
230  CONTINUE
NAME='Kitt Peak 90 inch'
w=west(111,35,58.24)
p=north(31,57,46.9)
h=2071D0
GO TO 9999

!  Kitt Peak 84 inch (1981 Almanac)                               KPNO84
240  CONTINUE
NAME='Kitt Peak 84 inch'
w=west(111,35,51.56)
p=north(31,57,29.2)
h=2096D0
GO TO 9999

!  Kitt Peak 36 foot (1981 Almanac)                             KPNO36FT
250  CONTINUE
NAME='Kitt Peak 36 foot'
w=west(111,36,51.12)
p=north(31,57,12.1)
h=1939D0
GO TO 9999

!  Kottamia 74" (1981 Almanac)                                  KOTTAMIA
260  CONTINUE
NAME='Kottamia 74 inch'
w=east(31,49,30.)
p=north(29,55,54.)
h=476D0
GO TO 9999

!  La Silla 3.6m (1981 Almanac)                                   ESO3.6
270  CONTINUE
NAME='ESO 3.6 metre'
w=west(70,43,36.)
p=south(29,15,36.)
h=2428D0
GO TO 9999

!  Mauna Kea 88 inch (1981 Almanac)                             MAUNAK88
280  CONTINUE
NAME='Mauna Kea 88 inch'
w=west(155,28,20.)
p=north(19,49,34.)
h=4215D0
GO TO 9999

!  UKIRT (1981 Almanac)                                            UKIRT
290  CONTINUE
NAME='UK Infra Red Telescope'
w=west(155,28,18.)
p=north(19,49,35.)
h=4200D0
GO TO 9999

!  Quebec 1.6m (1981 Almanac)                                  QUEBEC1.6
300  CONTINUE
NAME='Quebec 1.6 metre'
w=west(71,09,09.7)
p=north(45,27,20.6)
h=1114D0
GO TO 9999

!  Mt Ekar 1.82m (1981 Almanac)                                   MTEKAR
310  CONTINUE
NAME='Mt Ekar 1.82 metre'
w=east(11,34,15.)
p=north(45,50,48.)
h=1365D0
GO TO 9999

!  Mt Lemmon 60" (1981 Almanac)                               MTLEMMON60
320  CONTINUE
NAME='Mt Lemmon 60 inch'
w=west(110,42,16.9)
p=north(32,26,33.9)
h=2790D0
GO TO 9999

!  Mt Locke 2.7m (1981 Almanac)                               MCDONLD2.7
330  CONTINUE
NAME='McDonald 2.7 metre'
w=west(104,01,17.60)
p=north(30,40,17.7)
h=2075D0
GO TO 9999

!  Mt Locke 2.1m (1981 Almanac)                               MCDONLD2.1
340  CONTINUE
NAME='McDonald 2.1 metre'
w=west(104,01,20.10)
p=north(30,40,17.7)
h=2075D0
GO TO 9999

!  Palomar 200" (1981 Almanac)                                PALOMAR200
350  CONTINUE
NAME='Palomar 200 inch'
w=west(116,51,50.)
p=north(33,21,22.)
h=1706D0
GO TO 9999

!  Palomar 60" (1981 Almanac)                                  PALOMAR60
360  CONTINUE
NAME='Palomar 60 inch'
w=west(116,51,31.)
p=north(33,20,56.)
h=1706D0
GO TO 9999

!  David Dunlap 74" (1981 Almanac)                              DUNLAP74
370  CONTINUE
NAME='David Dunlap 74 inch'
w=west(79,25,20.)
p=north(43,51,46.)
h=244D0
GO TO 9999

!  Haute Provence 1.93m (1981 Almanac)                         HPROV1.93
380  CONTINUE
NAME='Haute Provence 1.93 metre'
w=east(5,42,46.75)
p=north(43,55,53.3)
h=665D0
GO TO 9999

!  Haute Provence 1.52m (1981 Almanac)                         HPROV1.52
390  CONTINUE
NAME='Haute Provence 1.52 metre'
w=east(5,42,43.82)
p=north(43,56,00.2)
h=667D0
GO TO 9999

!  San Pedro Martir 83" (1981 Almanac)                           SANPM83
400  CONTINUE
NAME='San Pedro Martir 83 inch'
w=west(115,27,47.)
p=north(31,02,38.)
h=2830D0
GO TO 9999

!  Sutherland 74" (1981 Almanac)                                  SAAO74
410  CONTINUE
NAME='Sutherland 74 inch'
w=east(20,48,44.3)
p=south(32,22,43.4)
h=1771D0
GO TO 9999

!  Tautenburg 2m (1981 Almanac)                                  TAUTNBG
420  CONTINUE
NAME='Tautenburg 2 metre'
w=east(11,42,45.)
p=north(50,58,51.)
h=331D0
GO TO 9999

!  Catalina 61" (1981 Almanac)                                CATALINA61
430  CONTINUE
NAME='Catalina 61 inch'
w=west(110,43,55.1)
p=north(32,25,00.7)
h=2510D0
GO TO 9999

!  Steward 90" (1981 Almanac)                                  STEWARD90
440  CONTINUE
NAME='Steward 90 inch'
w=west(111,35,58.24)
p=north(31,57,46.9)
h=2071D0
GO TO 9999

!  Russian 6m (1981 Almanac)                                       USSR6
450  CONTINUE
NAME='USSR 6 metre'
w=east(41,26,30.0)
p=north(43,39,12.)
h=2100D0
GO TO 9999

!  Arecibo 1000' (1981 Almanac)                                  ARECIBO
460  CONTINUE
NAME='Arecibo 1000 foot'
w=west(66,45,11.1)
p=north(18,20,36.6)
h=496D0
GO TO 9999

!  Cambridge 5km (1981 Almanac)                                  CAMB5KM
470  CONTINUE
NAME='Cambridge 5km'
w=east(0,02,37.23)
p=north(52,10,12.2)
h=17D0
GO TO 9999

!  Cambridge 1 mile (1981 Almanac)                             CAMB1MILE
480  CONTINUE
NAME='Cambridge 1 mile'
w=east(0,02,21.64)
p=north(52,09,47.3)
h=17D0
GO TO 9999

!  Bonn 100m (1981 Almanac)                                   EFFELSBERG
490  CONTINUE
NAME='Effelsberg 100 metre'
w=east(6,53,01.5)
p=north(50,31,28.6)
h=366D0
GO TO 9999

!  Greenbank 300' (1981 Almanac)                        GBVA300 (R.I.P.)
500  CONTINUE
NAME='Greenbank 300 foot'
w=west(79,50,56.36)
p=north(38,25,46.3)
h=894D0
GO TO 9999

!  Jodrell Bank Mk 1 (1981 Almanac)                             JODRELL1
510  CONTINUE
NAME='Jodrell Bank 250 foot'
w=west(2,18,25.)
p=north(53,14,10.5)
h=78D0
GO TO 9999

!  Australia Telescope Parkes Observatory                         PARKES
!  (Peter te Lintel Hekkert, private comm)
520  CONTINUE
NAME='Parkes 64 metre'
w=east(148,15,44.3591)
p=south(32,59,59.8657)
h=391.79D0
GO TO 9999

!  VLA (1981 Almanac)                                                VLA
530  CONTINUE
NAME='Very Large Array'
w=west(107,37,03.82)
p=north(34,04,43.5)
h=2124D0
GO TO 9999

!  Sugar Grove 150' (1981 Almanac)                            SUGARGROVE
540  CONTINUE
NAME='Sugar Grove 150 foot'
w=west(79,16,23.)
p=north(38,31,14.)
h=705D0
GO TO 9999

!  Russian 600' (1981 Almanac)                                   USSR600
550  CONTINUE
NAME='USSR 600 foot'
w=east(41,35,25.5)
p=north(43,49,32.)
h=973D0
GO TO 9999

!  Nobeyama 45 metre mm dish (based on 1981 Almanac entry)      NOBEYAMA
560  CONTINUE
NAME='Nobeyama 45 metre'
w=east(138,29,12.)
p=north(35,56,19.)
h=1350D0
GO TO 9999

!  James Clerk Maxwell 15 metre mm telescope, Mauna Kea (I.Coulson) JCMT
570  CONTINUE
NAME='JCMT 15 metre'
w=west(155,28,47.)
p=north(19,49,33.)
h=4111D0
GO TO 9999

!  ESO 3.5 metre NTT, La Silla (K.Wirenstrand)                    ESONTT
580  CONTINUE
NAME='ESO 3.5 metre NTT'
w=west(70,43,07.)
p=south(29,15,30.)
h=2377D0
GO TO 9999

!  St Andrews University Observatory (1982 Almanac)           ST.ANDREWS
590  CONTINUE
NAME='St Andrews'
w=west(2,48,52.5)
p=north(56,20,12.)
h=30D0
GO TO 9999

!  Apache Point 3.5 metre (R.Owen)                                APO3.5
600  CONTINUE
NAME='Apache Point 3.5m'
w=west(105,49,11.56)
p=north(32,46,48.96)
h=2809D0
GO TO 9999

!  W.M.Keck Observatory, Telescope 1 (site survey)                 KECK1
610  CONTINUE
NAME='Keck 10m Telescope #1'
w=west(155,28,39.0)
p=north(19,49,44.4)
h=4160D0
GO TO 9999

!  Tautenberg Schmidt (1983 Almanac)                            TAUTSCHM
620  CONTINUE
NAME='Tautenberg 1.34 metre Schmidt'
w=east(11,42,45.0)
p=north(50,58,51.0)
h=331D0
GO TO 9999

!  Palomar Schmidt (1981 Almanac)                              PALOMAR48
630  CONTINUE
NAME='Palomar 48-inch Schmidt'
w=west(116,51,32.0)
p=north(33,21,26.0)
h=1706D0
GO TO 9999

!  UK Schmidt, Siding Spring (1983 Almanac)                         UKST
640  CONTINUE
NAME='UK 1.2 metre Schmidt, Siding Spring'
w=east(149,04,12.8)
p=south(31,16,27.8)
h=1145D0
GO TO 9999

!  Kiso Schmidt, Japan (1981 Almanac)                               KISO
650  CONTINUE
NAME='Kiso 1.05 metre Schmidt, Japan'
w=east(137,37,42.2)
p=north(35,47,38.7)
h=1130D0
GO TO 9999

!  ESO Schmidt, La Silla (1981 Almanac)                          ESOSCHM
660  CONTINUE
NAME='ESO 1 metre Schmidt, La Silla'
w=west(70,43,46.5)
p=south(29,15,25.8)
h=2347D0
GO TO 9999

!  Australia Telescope Compact Array                                ATCA
!  (WGS84 coordinates of Station 35, Mark Calabretta, private comm)
670  CONTINUE
NAME='Australia Telescope Compact Array'
w=east(149,33,00.500)
p=south(30,18,46.385)
h=236.9D0
GO TO 9999

!  Australia Telescope Mopra Observatory                           MOPRA
!  (Peter te Lintel Hekkert, private comm)
680  CONTINUE
NAME='ATNF Mopra Observatory'
w=east(149,05,58.732)
p=south(31,16,04.451)
h=850D0
GO TO 9999

!  Subaru telescope, Mauna Kea                                     SUBARU
!  (poster display, 1997 IAU General Assembly)
690  CONTINUE
NAME='Subaru 8 metre'
w=west(155,28,50.0)
p=north(19,49,43.0)
h=4139D0
GO TO 9999

!  Unrecognized station
9000 CONTINUE
NAME='?'

!  Exit
9999 CONTINUE

END SUBROUTINE sla_obs
DOUBLE PRECISION FUNCTION sla_pa (ha, dec, phi)
!+
!     - - -
!      P A
!     - - -

!  HA, Dec to Parallactic Angle (double precision)

!  Given:
!     HA     d     hour angle in radians (geocentric apparent)
!     DEC    d     declination in radians (geocentric apparent)
!     PHI    d     observatory latitude in radians (geodetic)

!  The result is in the range -pi to +pi

!  Notes:

!  1)  The parallactic angle at a point in the sky is the position
!      angle of the vertical, i.e. the angle between the direction to
!      the pole and to the zenith.  In precise applications care must
!      be taken only to use geocentric apparent HA,Dec and to consider
!      separately the effects of atmospheric refraction and telescope
!      mount errors.

!  2)  At the pole a zero result is returned.

!  P.T.Wallace   Starlink   16 August 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ha,dec,phi

DOUBLE PRECISION :: cp,sqsz,cqsz



cp=COS(phi)
sqsz=cp*SIN(ha)
cqsz=SIN(phi)*COS(dec)-cp*SIN(dec)*COS(ha)
IF (sqsz == 0D0.AND.cqsz == 0D0) cqsz=1D0
sla_pa=ATAN2(sqsz,cqsz)

END FUNCTION sla_pa
REAL FUNCTION sla_pav ( v1, v2 )
!+
!     - - - -
!      P A V
!     - - - -

!  Position angle of one celestial direction with respect to another.

!  (single precision)

!  Given:
!     V1    r(3)    direction cosines of one point
!     V2    r(3)    direction cosines of the other point

!  (The coordinate frames correspond to RA,Dec, Long,Lat etc.)

!  The result is the bearing (position angle), in radians, of point
!  V2 with respect to point V1.  It is in the range +/- pi.  The
!  sense is such that if V2 is a small distance east of V1, the
!  bearing is about +pi/2.  Zero is returned if the two points
!  are coincident.

!  V1 and V2 do not have to be unit vectors.

!  The routine sla_BEAR performs an equivalent function except
!  that the points are specified in the form of spherical
!  coordinates.

!  Called:  sla_DPAV

!  Patrick Wallace   Starlink   23 May 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: v1(3),v2(3)

INTEGER :: i
DOUBLE PRECISION :: d1(3),d2(3)



!  Call the double precision version
DO i=1,3
  d1(i)=v1(i)
  d2(i)=v2(i)
END DO
sla_pav=sla_dpav(d1,d2)

END FUNCTION sla_pav
SUBROUTINE sla_pcd (disco,x,y)
!+
!     - - - -
!      P C D
!     - - - -

!  Apply pincushion/barrel distortion to a tangent-plane [x,y].

!  Given:
!     DISCO    d      pincushion/barrel distortion coefficient
!     X,Y      d      tangent-plane coordinates

!  Returned:
!     X,Y      d      distorted coordinates

!  Notes:

!  1)  The distortion is of the form RP = R*(1 + C*R**2), where R is
!      the radial distance from the tangent point, C is the DISCO
!      argument, and RP is the radial distance in the presence of
!      the distortion.

!  2)  For pincushion distortion, C is +ve;  for barrel distortion,
!      C is -ve.

!  3)  For X,Y in units of one projection radius (in the case of
!      a photographic plate, the focal length), the following
!      DISCO values apply:

!          Geometry          DISCO

!          astrograph         0.0
!          Schmidt           -0.3333
!          AAT PF doublet  +147.069
!          AAT PF triplet  +178.585
!          AAT f/8          +21.20
!          JKT f/8          +13.32

!  4)  There is a companion routine, sla_UNPCD, which performs
!      an approximately inverse operation.

!  P.T.Wallace   Starlink   31 December 1992

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: disco,x,y

DOUBLE PRECISION :: f



f=1D0+disco*(x*x+y*y)
x=x*f
y=y*f

END SUBROUTINE sla_pcd
SUBROUTINE sla_pda2h (p, d, a, h1, j1, h2, j2)
!+
!     - - - - - -
!      P D A 2 H
!     - - - - - -

!  Hour Angle corresponding to a given azimuth

!  (double precision)

!  Given:
!     P       d        latitude
!     D       d        declination
!     A       d        azimuth

!  Returned:
!     H1      d        hour angle:  first solution if any
!     J1      i        flag: 0 = solution 1 is valid
!     H2      d        hour angle:  second solution if any
!     J2      i        flag: 0 = solution 2 is valid

!  Called:  sla_DRANGE

!  P.T.Wallace   Starlink   6 October 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: p,d,a,h1
INTEGER :: j1
DOUBLE PRECISION :: h2
INTEGER :: j2

DOUBLE PRECISION :: dpi
PARAMETER (dpi=3.141592653589793238462643D0)
DOUBLE PRECISION :: d90
PARAMETER (d90=dpi/2D0)
DOUBLE PRECISION :: tiny
PARAMETER (tiny=1D-12)
DOUBLE PRECISION :: pn,an,dn,sa,ca,sasp,qt,qb,hpt,t


!  Preset status flags to OK
j1=0
j2=0

!  Adjust latitude, azimuth, declination to avoid critical values
pn=sla_drange(p)
IF (ABS(ABS(pn)-d90) < tiny) THEN
  pn=pn-SIGN(tiny,pn)
ELSE IF (ABS(pn) < tiny) THEN
  pn=tiny
END IF
an=sla_drange(a)
IF (ABS(ABS(an)-dpi) < tiny) THEN
  an=an-SIGN(tiny,an)
ELSE IF (ABS(an) < tiny) THEN
  an=tiny
END IF
dn=sla_drange(d)
IF (ABS(ABS(dn)-ABS(p)) < tiny) THEN
  dn=dn-SIGN(tiny,dn)
ELSE IF (ABS(ABS(dn)-d90) < tiny) THEN
  dn=dn-SIGN(tiny,dn)
ELSE IF (ABS(dn) < tiny) THEN
  dn=tiny
END IF

!  Useful functions
sa=SIN(an)
ca=COS(an)
sasp=sa*SIN(pn)

!  Quotient giving sin(h+t)
qt=SIN(dn)*sa*COS(pn)
qb=COS(dn)*SQRT(ca*ca+sasp*sasp)

!  Any solutions?
IF (ABS(qt) <= qb) THEN
  
!     Yes: find h+t and t
  hpt=ASIN(qt/qb)
  t=ATAN2(sasp,-ca)
  
!     The two solutions
  h1=sla_drange(hpt-t)
  h2=sla_drange(-hpt-(t+dpi))
  
!     Reject unless h and A different signs
  IF (h1*an > 0D0) j1=-1
  IF (h2*an > 0D0) j2=-1
ELSE
  j1=-1
  j2=-1
END IF

END SUBROUTINE sla_pda2h
SUBROUTINE sla_pdq2h (p, d, q, h1, j1, h2, j2)
!+
!     - - - - - -
!      P D Q 2 H
!     - - - - - -

!  Hour Angle corresponding to a given parallactic angle

!  (double precision)

!  Given:
!     P       d        latitude
!     D       d        declination
!     Q       d        parallactic angle

!  Returned:
!     H1      d        hour angle:  first solution if any
!     J1      i        flag: 0 = solution 1 is valid
!     H2      d        hour angle:  second solution if any
!     J2      i        flag: 0 = solution 2 is valid

!  Called:  sla_DRANGE

!  P.T.Wallace   Starlink   6 October 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: p,d,q,h1
INTEGER :: j1
DOUBLE PRECISION :: h2
INTEGER :: j2

DOUBLE PRECISION :: dpi
PARAMETER (dpi=3.141592653589793238462643D0)
DOUBLE PRECISION :: d90
PARAMETER (d90=dpi/2D0)
DOUBLE PRECISION :: tiny
PARAMETER (tiny=1D-12)
DOUBLE PRECISION :: pn,qn,dn,sq,cq,sqsd,qt,qb,hpt,t


!  Preset status flags to OK
j1=0
j2=0

!  Adjust latitude, declination, parallactic angle to avoid critical values
pn=sla_drange(p)
IF (ABS(ABS(pn)-d90) < tiny) THEN
  pn=pn-SIGN(tiny,pn)
ELSE IF (ABS(pn) < tiny) THEN
  pn=tiny
END IF
qn=sla_drange(q)
IF (ABS(ABS(qn)-dpi) < tiny) THEN
  qn=qn-SIGN(tiny,qn)
ELSE IF (ABS(qn) < tiny) THEN
  qn=tiny
END IF
dn=sla_drange(d)
IF (ABS(ABS(d)-ABS(p)) < tiny) THEN
  dn=dn-SIGN(tiny,dn)
ELSE IF (ABS(ABS(d)-d90) < tiny) THEN
  dn=dn-SIGN(tiny,dn)
END IF

!  Useful functions
sq=SIN(qn)
cq=COS(qn)
sqsd=sq*SIN(dn)

!  Quotient giving sin(h+t)
qt=SIN(pn)*sq*COS(dn)
qb=COS(pn)*SQRT(cq*cq+sqsd*sqsd)

!  Any solutions?
IF (ABS(qt) <= qb) THEN
  
!     Yes: find h+t and t
  hpt=ASIN(qt/qb)
  t=ATAN2(sqsd,cq)
  
!     The two solutions
  h1=sla_drange(hpt-t)
  h2=sla_drange(-hpt-(t+dpi))
  
!     Reject if h and Q different signs
  IF (h1*qn < 0D0) j1=-1
  IF (h2*qn < 0D0) j2=-1
ELSE
  j1=-1
  j2=-1
END IF

END SUBROUTINE sla_pdq2h
SUBROUTINE sla_planel (date, jform, epoch, orbinc, anode, perih,  &
aorq, e, aorl, dm, pv, jstat)
!+
!     - - - - - - -
!      P L A N E L
!     - - - - - - -

!  Heliocentric position and velocity of a planet, asteroid
!  or comet, starting from orbital elements.

!  Given:
!     DATE      d      date, Modified Julian Date (JD - 2400000.5)
!     JFORM     i      choice of element set (1-3, see Note 3, below)
!     EPOCH     d      epoch of elements (TT MJD)
!     ORBINC    d      inclination (radians)
!     ANODE     d      longitude of the ascending node (radians)
!     PERIH     d      longitude or argument of perihelion (radians)
!     AORQ      d      mean distance or perihelion distance (AU)
!     E         d      eccentricity
!     AORL      d      mean anomaly or longitude (radians, JFORM=1,2 only)
!     DM        d      daily motion (radians, JFORM=1 only)

!  Returned:
!     PV        d(6)   heliocentric x,y,z,xdot,ydot,zdot of date,
!                                       J2000 equatorial triad (AU,AU/s)
!     JSTAT     i      status:  0 = OK
!                              -1 = illegal JFORM
!                              -2 = illegal E
!                              -3 = illegal AORQ
!                              -4 = illegal DM
!                              -5 = failed to converge

!  Notes

!  1  DATE is the instant for which the prediction is required.  It is
!     in the TT timescale (formerly Ephemeris Time, ET) and is a
!     Modified Julian Date (JD-2400000.5).

!  2  The elements are with respect to the J2000 ecliptic and
!     equinox.

!  3  Three different element-format options are available:

!     Option JFORM=1, suitable for the major planets:

!     EPOCH  = epoch of elements (TT MJD)
!     ORBINC = inclination i (radians)
!     ANODE  = longitude of the ascending node, big omega (radians)
!     PERIH  = longitude of perihelion, curly pi (radians)
!     AORQ   = mean distance, a (AU)
!     E      = eccentricity, e
!     AORL   = mean longitude L (radians)
!     DM     = daily motion (radians)

!     Option JFORM=2, suitable for minor planets:

!     EPOCH  = epoch of elements (TT MJD)
!     ORBINC = inclination i (radians)
!     ANODE  = longitude of the ascending node, big omega (radians)
!     PERIH  = argument of perihelion, little omega (radians)
!     AORQ   = mean distance, a (AU)
!     E      = eccentricity, e
!     AORL   = mean anomaly M (radians)

!     Option JFORM=3, suitable for comets:

!     EPOCH  = epoch of perihelion (TT MJD)
!     ORBINC = inclination i (radians)
!     ANODE  = longitude of the ascending node, big omega (radians)
!     PERIH  = argument of perihelion, little omega (radians)
!     AORQ   = perihelion distance, q (AU)
!     E      = eccentricity, e

!  4  Unused elements (DM for JFORM=2, AORL and DM for JFORM=3) are
!     not accessed.

!  5  The reference frame for the result is equatorial and is with
!     respect to the mean equinox and ecliptic of epoch J2000.

!  6  The algorithm is adapted from the EPHSLA program of
!     D.H.P.Jones (private communication, 1996).  The method is
!     based on Stumpff's Universal Variables;  see Everhart and
!     Pitkin (1983, Am.J.Phys.51,712).

!  P.T.Wallace   Starlink   26 June 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date
INTEGER :: jform
DOUBLE PRECISION :: epoch,orbinc,anode,perih,aorq,e,aorl,dm,pv(6)
INTEGER :: jstat

!  Gaussian gravitational constant (exact)
DOUBLE PRECISION :: gcon
PARAMETER (gcon=0.01720209895D0)

!  Canonical days to seconds
DOUBLE PRECISION :: cd2s
PARAMETER (cd2s=gcon/86400D0)

!  Sin and cos of J2000 mean obliquity (IAU 1976)
DOUBLE PRECISION :: se,ce
PARAMETER (se=0.3977771559319137D0,ce=0.9174820620691818D0)

!  Test value for solution and maximum number of iterations
DOUBLE PRECISION :: test
INTEGER :: nitmax
PARAMETER (test=1D-11,nitmax=20)

INTEGER :: nit,n

DOUBLE PRECISION :: am,w,pht,argph,q,tf,alpha,aba,sqa,  &
dt,fc,fp,psi,psj,beta,bs0,bs1,bs2,bs3,ff,fdot,phs,sw,cw,si,ci,so,co,  &
x,y,z,xdot,ydot,zdot,vf



!  Validate arguments.
IF (jform < 1.OR.jform > 3) THEN
  jstat = -1
  GO TO 999
END IF
IF (e < 0D0.OR.(e >= 1D0.AND.jform /= 3)) THEN
  jstat = -2
  GO TO 999
END IF
IF (aorq <= 0D0) THEN
  jstat = -3
  GO TO 999
END IF
IF (jform == 1.AND.dm <= 0D0) THEN
  jstat = -4
  GO TO 999
END IF


!  Transform elements into standard form:

!  AM    = mean anomaly (M)
!  PHT   = epoch of perihelion passage
!  ARGPH = argument of perihelion (little omega)
!  Q     = perihelion distance (q)

!  Also computed is TF, the ratio of the daily motion to
!  the "theoretical" value.

IF (jform == 1) THEN
  am = aorl-perih
  w = SQRT(aorq*aorq*aorq)/gcon
  pht = epoch-am*w
  argph = perih-anode
  q = aorq*(1D0-e)
  tf = dm*w
ELSE IF (jform == 2) THEN
  am = aorl
  pht = epoch-am*SQRT(aorq*aorq*aorq)/gcon
  argph = perih
  q = aorq*(1D0-e)
  tf = 1D0
ELSE IF (jform == 3) THEN
  am = aorl
  pht = epoch
  argph = perih
  q = aorq
  tf = 1D0
END IF

!  The universal variable alpha.  This is proportional to the total
!  energy of the orbit:  -ve for an ellipse, zero for a parabola,
!  +ve for a hyperbola.

alpha = (e-1D0)/q
aba = ABS(alpha)
sqa = SQRT(aba)

!  Time from perihelion to date (in Canonical Days:  a canonical day
!  is 58.1324409... days, defined as 1/GCON).  The portion of the
!  time from the epoch to date is adjusted to take account of any
!  difference between the supplied and theoretical daily motion.

dt = ((date-epoch)*tf+epoch-pht)*gcon

!  First Approximation to the Universal Eccentric Anomaly, PSI,
!  based on the circle (FC) and parabola (FP) values.

fc = dt/q
w = (3D0*dt+SQRT(9D0*dt*dt+8D0*q*q*q))**(1D0/3D0)
fp = w-2D0*q/w
psi = (1D0-e)*fc+e*fp

!  Successive Approximations to Universal Eccentric Anomaly.

nit=1
w=1D0
DO WHILE (ABS(w) >= test)
  
!     Compute the Universal Variables BS0, BS1, BS2, BS3.
  
!     Form half angles until BETA below maximum (0.7).
  n = 0
  psj = psi
  beta = alpha*psi*psi
  DO WHILE (ABS(beta) > 0.7D0)
    n = n+1
    beta = beta/4D0
    psj = psj/2D0
  END DO
  
!     Calculate Universal Variables by nested series.
  bs3 = psj*psj*psj*((((((beta/210D0+1D0)*beta/156D0+1D0)  &
  *beta/110D0+1D0)*beta/72D0+1D0)  &
  *beta/42D0+1D0)*beta/20D0+1D0)/6D0
  bs2 = psj*psj*((((((beta/182D0+1D0)*beta/132D0+1D0)  &
  *beta/90D0+1D0)*beta/56D0+1D0)  &
  *beta/30D0+1D0)*beta/12D0+1D0)/2D0
  bs1 = psj+alpha*bs3
  bs0 = 1D0+alpha*bs2
  
!     Double angles until N vanishes.
  DO WHILE (n > 0)
    bs3 = 2D0*(bs0*bs3+psj*bs2)
    bs2 = 2D0*bs1*bs1
    bs1 = 2D0*bs0*bs1
    bs0 = 2D0*bs0*bs0-1D0
    n = n-1
    psj = 2D0*psj
  END DO
  
!     Improve the approximation.
  ff = q*bs1+bs3-dt
  fdot = q*bs0+bs2
  w = ff/fdot
  psi = psi-w
  IF (nit < nitmax) THEN
    nit = nit+1
  ELSE
    jstat = -5
    GO TO 999
  END IF
END DO

!  Speed at perihelion.

phs = SQRT(alpha+2D0/q)

!  In a Cartesian coordinate system which has the x-axis pointing
!  to perihelion and the z-axis normal to the orbit (such that the
!  object orbits counter-clockwise as seen from +ve z), the
!  perihelion position and velocity vectors are:

!    position   [Q,0,0]
!    velocity   [0,PHS,0]

!  Using the Universal Variables we project these vectors to the
!  given date:

x = q-bs2
y = phs*q*bs1
xdot = -bs1/fdot
ydot = phs*(1D0-bs2/fdot)

!  To express the results in J2000 equatorial coordinates we make a
!  series of four rotations of the Cartesian axes:

!            axis     Euler angle

!     1      z        argument of perihelion (little omega)
!     2      x        inclination (i)
!     3      z        longitude of the ascending node (big omega)
!     4      x        J2000 obliquity (epsilon)

!  In each case the rotation is clockwise as seen from the +ve end
!  of the axis concerned.

!  Functions of the Euler angles.
sw = SIN(argph)
cw = COS(argph)
si = SIN(orbinc)
ci = COS(orbinc)
so = SIN(anode)
co = COS(anode)

!  Position.
w = x*cw-y*sw
y = x*sw+y*cw
x = w
z = y*si
y = y*ci
w = x*co-y*so
y = x*so+y*co
pv(1) = w
pv(2) = y*ce-z*se
pv(3) = y*se+z*ce

!  Velocity (scaled to AU/s and adjusted for daily motion).
w = xdot*cw-ydot*sw
ydot = xdot*sw+ydot*cw
xdot = w
zdot = ydot*si
ydot = ydot*ci
w = xdot*co-ydot*so
ydot = xdot*so+ydot*co
vf = tf*cd2s
pv(4) = vf*w
pv(5) = vf*(ydot*ce-zdot*se)
pv(6) = vf*(ydot*se+zdot*ce)

!  Wrap up.
jstat = 0
999  CONTINUE

END SUBROUTINE sla_planel
SUBROUTINE sla_planet (date, np, pv, jstat)
!+
!     - - - - - - -
!      P L A N E T
!     - - - - - - -

!  Approximate heliocentric position and velocity of a specified
!  major planet.

!  Given:
!     DATE      d      Modified Julian Date (JD - 2400000.5)
!     NP        i      planet (1=Mercury, 2=Venus, 3=EMB ... 9=Pluto)

!  Returned:
!     PV        d(6)   heliocentric x,y,z,xdot,ydot,zdot, J2000
!                                           equatorial triad (AU,AU/s)
!     JSTAT     i      status: +1 = warning: date out of range
!                               0 = OK
!                              -1 = illegal NP (outside 1-9)
!                              -2 = solution didn't converge

!  Called:  sla_PLANEL

!  Notes

!  1  The epoch, DATE, is in the TDB timescale and is a Modified
!     Julian Date (JD-2400000.5).

!  2  The reference frame is equatorial and is with respect to the
!     mean equinox and ecliptic of epoch J2000.

!  3  If an NP value outside the range 1-9 is supplied, an error
!     status (JSTAT = -1) is returned and the PV vector set to zeroes.

!  4  The algorithm for obtaining the mean elements of the planets
!     from Mercury to Neptune is due to J.L. Simon, P. Bretagnon,
!     J. Chapront, M. Chapront-Touze, G. Francou and J. Laskar
!     (Bureau des Longitudes, Paris).  The (completely different)
!     algorithm for calculating the ecliptic coordinates of Pluto
!     is by Meeus.

!  5  Comparisons of the present routine with the JPL DE200 ephemeris
!     give the following RMS errors over the interval 1960-2025:

!                      position (km)     speed (metre/sec)

!        Mercury            334               0.437
!        Venus             1060               0.855
!        EMB               2010               0.815
!        Mars              7690               1.98
!        Jupiter          71700               7.70
!        Saturn          199000              19.4
!        Uranus          564000              16.4
!        Neptune         158000              14.4
!        Pluto            36400               0.137

!     From comparisons with DE102, Simon et al quote the following
!     longitude accuracies over the interval 1800-2200:

!        Mercury                 4"
!        Venus                   5"
!        EMB                     6"
!        Mars                   17"
!        Jupiter                71"
!        Saturn                 81"
!        Uranus                 86"
!        Neptune                11"

!     In the case of Pluto, Meeus quotes an accuracy of 0.6 arcsec
!     in longitude and 0.2 arcsec in latitude for the period
!     1885-2099.

!     For all except Pluto, over the period 1000-3000 the accuracy
!     is better than 1.5 times that over 1800-2200.  Outside the
!     period 1000-3000 the accuracy declines.  For Pluto the
!     accuracy declines rapidly outside the period 1885-2099.
!     Outside these ranges (1885-2099 for Pluto, 1000-3000 for
!     the rest) a "date out of range" warning status (JSTAT=+1)
!     is returned.

!  6  The algorithms for (i) Mercury through Neptune and (ii) Pluto
!     are completely independent.  In the Mercury through Neptune
!     case, the present SLALIB implementation differs from the
!     original Simon et al Fortran code in the following respects.

!     *  The date is supplied as a Modified Julian Date rather
!        than a Julian Date (MJD = JD - 2400000.5).

!     *  The result is returned only in equatorial Cartesian form;
!        the ecliptic longitude, latitude and radius vector are not
!        returned.

!     *  The velocity is in AU per second, not AU per day.

!     *  Different error/warning status values are used.

!     *  Kepler's equation is not solved inline.

!     *  Polynomials in T are nested to minimize rounding errors.

!     *  Explicit double-precision constants are used to avoid
!        mixed-mode expressions.

!     *  There are other, cosmetic, changes to comply with
!        Starlink/SLALIB style guidelines.

!     None of the above changes affects the result significantly.

!  7  For NP=3 the result is for the Earth-Moon Barycentre.  To
!     obtain the heliocentric position and velocity of the Earth,
!     either use the SLALIB routine sla_EVP or call sla_DMOON and
!     subtract 0.012150581 times the geocentric Moon vector from
!     the EMB vector produced by the present routine.  (The Moon
!     vector should be precessed to J2000 first, but this can
!     be omitted for modern epochs without introducing significant
!     inaccuracy.)

!  References:  Simon et al., Astron. Astrophys. 282, 663 (1994).
!               Meeus, Astronomical Algorithms, Willmann-Bell (1991).

!  P.T.Wallace   Starlink   27 May 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date
INTEGER :: np
DOUBLE PRECISION :: pv(6)
INTEGER :: jstat

!  2Pi, deg to radians, arcsec to radians
DOUBLE PRECISION :: d2pi,d2r,as2r
PARAMETER (d2pi=6.283185307179586476925286766559D0,  &
d2r=0.017453292519943295769236907684886D0,  &
as2r=4.848136811095359935899141023579D-6)

!  Gaussian gravitational constant (exact)
DOUBLE PRECISION :: gcon
PARAMETER (gcon=0.01720209895D0)

!  Seconds per Julian century
DOUBLE PRECISION :: spc
PARAMETER (spc=36525D0*86400D0)

!  Sin and cos of J2000 mean obliquity (IAU 1976)
DOUBLE PRECISION :: se,ce
PARAMETER (se=0.3977771559319137D0,ce=0.9174820620691818D0)

INTEGER :: i,j,ijsp(3,43)
DOUBLE PRECISION :: amas(8),a(3,8),dlm(3,8),e(3,8),  &
pi(3,8),dinc(3,8),omega(3,8),dkp(9,8),ca(9,8),sa(9,8),  &
dkq(10,8),clo(10,8),slo(10,8),t,da,de,dpe,di,DO,dmu,arga,argl,dm,  &
ab(2,3,43),dj0,ds0,dp0,dl0,dld0,db0,dr0,  &
dj,ds,dp,djd,dsd,dpd,wlbr(3),wlbrd(3),wj,ws,wp,al,ald,sal,cal,  &
ac,bc,dl,dld,db,dbd,dr,drd,sl,cl,sb,cb,slcb,clcb,x,y,z,xd,yd,zd

!  -----------------------
!  Mercury through Neptune
!  -----------------------

!  Planetary inverse masses
DATA amas / 6023600D0,408523.5D0,328900.5D0,3098710D0,  &
1047.355D0,3498.5D0,22869D0,19314D0 /


!  Tables giving the mean Keplerian elements, limited to T**2 terms:

!         A       semi-major axis (AU)
!         DLM     mean longitude (degree and arcsecond)
!         E       eccentricity
!         PI      longitude of the perihelion (degree and arcsecond)
!         DINC    inclination (degree and arcsecond)
!         OMEGA   longitude of the ascending node (degree and arcsecond)

DATA a /0.3870983098D0,             0D0,      0D0,  &
0.7233298200D0,             0D0,      0D0,  &
1.0000010178D0,             0D0,      0D0,  &
1.5236793419D0,           3D-10,      0D0,  &
5.2026032092D0,       19132D-10,  -39D-10,  &
9.5549091915D0, -0.0000213896D0,  444D-10,  &
19.2184460618D0,       -3716D-10,  979D-10,  &
30.1103868694D0,      -16635D-10,  686D-10 /

DATA dlm /252.25090552D0, 5381016286.88982D0,  -1.92789D0,  &
181.97980085D0, 2106641364.33548D0,   0.59381D0,  &
100.46645683D0, 1295977422.83429D0,  -2.04411D0,  &
355.43299958D0,  689050774.93988D0,   0.94264D0,  &
34.35151874D0,  109256603.77991D0, -30.60378D0,  &
50.07744430D0,   43996098.55732D0,  75.61614D0,  &
314.05500511D0,   15424811.93933D0,  -1.75083D0,  &
304.34866548D0,    7865503.20744D0,   0.21103D0/

DATA e /0.2056317526D0,  0.0002040653D0,      -28349D-10,  &
0.0067719164D0, -0.0004776521D0,       98127D-10,  &
0.0167086342D0, -0.0004203654D0, -0.0000126734D0,  &
0.0934006477D0,  0.0009048438D0,      -80641D-10,  &
0.0484979255D0,  0.0016322542D0, -0.0000471366D0,  &
0.0555481426D0, -0.0034664062D0, -0.0000643639D0,  &
0.0463812221D0, -0.0002729293D0,  0.0000078913D0,  &
0.0094557470D0,  0.0000603263D0,            0D0 /

DATA pi /77.45611904D0,  5719.11590D0,   -4.83016D0,  &
131.56370300D0,   175.48640D0, -498.48184D0,  &
102.93734808D0, 11612.35290D0,   53.27577D0,  &
336.06023395D0, 15980.45908D0,  -62.32800D0,  &
14.33120687D0,  7758.75163D0,  259.95938D0,  &
93.05723748D0, 20395.49439D0,  190.25952D0,  &
173.00529106D0,  3215.56238D0,  -34.09288D0,  &
48.12027554D0,  1050.71912D0,   27.39717D0 /

DATA dinc /7.00498625D0, -214.25629D0,   0.28977D0,  &
3.39466189D0,  -30.84437D0, -11.67836D0,0D0,  469.97289D0,  -3.35053D0,  &
1.84972648D0, -293.31722D0,  -8.11830D0,  &
1.30326698D0,  -71.55890D0,  11.95297D0,  &
2.48887878D0,   91.85195D0, -17.66225D0,  &
0.77319689D0,  -60.72723D0,   1.25759D0,  &
1.76995259D0,    8.12333D0,   0.08135D0 /

DATA omega /48.33089304D0,  -4515.21727D0,  -31.79892D0,  &
76.67992019D0, -10008.48154D0,  -51.32614D0,  &
174.87317577D0,  -8679.27034D0,   15.34191D0,  &
49.55809321D0, -10620.90088D0, -230.57416D0,  &
100.46440702D0,   6362.03561D0,  326.52178D0,  &
113.66550252D0,  -9240.19942D0,  -66.23743D0,  &
74.00595701D0,   2669.15033D0,  145.93964D0,  &
131.78405702D0,   -221.94322D0,   -0.78728D0 /

!  Tables for trigonometric terms to be added to the mean elements
!  of the semi-major axes.

DATA dkp /69613, 75645, 88306, 59899, 15746, 71087, 142173,  3086,    0,  &
21863, 32794, 26934, 10931, 26250, 43725,  53867, 28939,    0,  &
16002, 21863, 32004, 10931, 14529, 16368,  15318, 32794,    0,  &
6345,   7818, 15636,  7077,  8184, 14163,   1107,  4872,    0,  &
1760,   1454,  1167,   880,   287,  2640,     19,  2047, 1454,  &
574,      0,   880,   287,    19,  1760,   1167,   306,  574,  &
204,      0,   177,  1265,     4,   385,    200,   208,  204,  &
0,    102,   106,     4,    98,  1367,    487,   204,    0 /

DATA ca /4,    -13,    11,    -9,    -9,    -3,    -1,     4,    0,  &
-156,     59,   -42,     6,    19,   -20,   -10,   -12,    0,  &
64,   -152,    62,    -8,    32,   -41,    19,   -11,    0,  &
124,    621,  -145,   208,    54,   -57,    30,    15,    0,  &
-23437,  -2634,  6601,  6259, -1507, -1821,  2620, -2115,-1489,  &
62911,-119919, 79336, 17814,-24241, 12068,  8306, -4893, 8902,  &
389061,-262125,-44088,  8387,-22976, -2093,  -615, -9720, 6633,  &
-412235,-157046,-31430, 37817, -9740,   -13, -7449,  9644,    0 /

DATA sa /-29,    -1,     9,     6,    -6,     5,     4,     0,    0,  &
-48,  -125,   -26,   -37,    18,   -13,   -20,    -2,    0,  &
-150,   -46,    68,    54,    14,    24,   -28,    22,    0,  &
-621,   532,  -694,   -20,   192,   -94,    71,   -73,    0,  &
-14614,-19828, -5869,  1881, -4372, -2255,   782,   930,  913,  &
139737,     0, 24667, 51123, -5102,  7429, -4095, -1976,-9566,  &
-138081,     0, 37205,-49039,-41901,-33872,-27037,-12474,18797,  &
0, 28492,133236, 69654, 52322,-49577,-26430, -3593,    0 /

!  Tables giving the trigonometric terms to be added to the mean
!  elements of the mean longitudes.

DATA dkq /3086, 15746, 69613, 59899, 75645, 88306, 12661, 2658,  0,   0,  &
21863, 32794, 10931,    73,  4387, 26934,  1473, 2157,  0,   0,  &
10, 16002, 21863, 10931,  1473, 32004,  4387,   73,  0,   0,  &
10,  6345,  7818,  1107, 15636,  7077,  8184,  532, 10,   0,  &
19,  1760,  1454,   287,  1167,   880,   574, 2640, 19,1454,  &
19,   574,   287,   306,  1760,    12,    31,   38, 19, 574,  &
4,   204,   177,     8,    31,   200,  1265,  102,  4, 204,  &
4,   102,   106,     8,    98,  1367,   487,  204,  4, 102 /

DATA clo /21,   -95, -157,   41,   -5,   42,   23,   30,     0,    0,  &
-160,  -313, -235,   60,  -74,  -76,  -27,   34,     0,    0,  &
-325,  -322,  -79,  232,  -52,   97,   55,  -41,     0,    0,  &
2268,  -979,  802,  602, -668,  -33,  345,  201,   -55,    0,  &
7610, -4997,-7689,-5841,-2617, 1115, -748, -607,  6074,  354,  &
-18549, 30125,20012, -730,  824,   23, 1289, -352,-14767,-2062,  &
-135245,-14594, 4197,-4030,-5630,-2898, 2540, -306,  2939, 1986,  &
89948,  2103, 8963, 2695, 3682, 1648,  866, -154, -1963, -283 /

DATA slo /-342,   136,  -23,   62,   66,  -52,  -33,   17,     0,    0,  &
524,  -149,  -35,  117,  151,  122,  -71,  -62,     0,    0,  &
-105,  -137,  258,   35, -116,  -88, -112,  -80,     0,    0,  &
854,  -205, -936, -240,  140, -341,  -97, -232,   536,    0,  &
-56980,  8016, 1012, 1448,-3024,-3710,  318,  503,  3767,  577,  &
138606,-13478,-4964, 1441,-1319,-1482,  427, 1236, -9167,-1918,  &
71234,-41116, 5334,-4935,-1848,   66,  434,-1748,  3780, -701,  &
-47645, 11647, 2166, 3194,  679,    0, -244, -419, -2531,   48 /

!  -----
!  Pluto
!  -----


!  Coefficients for fundamental arguments:  mean longitudes
!  (degrees) and mean rate of change of longitude (degrees per
!  Julian century) for Jupiter, Saturn and Pluto

DATA dj0, djd / 34.35D0, 3034.9057D0 /
DATA ds0, dsd / 50.08D0, 1222.1138D0 /
DATA dp0, dpd / 238.96D0, 144.9600D0 /

!  Coefficients for latitude, longitude, radius vector
DATA dl0,dld0 / 238.956785D0, 144.96D0 /
DATA db0 / -3.908202D0 /
DATA dr0 / 40.7247248D0 /


!  Coefficients for periodic terms (Meeus's Table 36.A)

!  The coefficients for term n in the series are:

!    IJSP(1,n)     J
!    IJSP(2,n)     S
!    IJSP(3,n)     P
!    AB(1,1,n)     longitude sine (degrees)
!    AB(2,1,n)     longitude cosine (degrees)
!    AB(1,2,n)     latitude sine (degrees)
!    AB(2,2,n)     latitude cosine (degrees)
!    AB(1,3,n)     radius vector sine (AU)
!    AB(2,3,n)     radius vector cosine (AU)

DATA (ijsp(i, 1),i=1,3),((ab(j,i, 1),j=1,2),i=1,3) /0,  0,  1,  &
-19798886D-6,  19848454D-6,-5453098D-6, -14974876D-6,  &
66867334D-7,  68955876D-7 /
DATA (ijsp(i, 2),i=1,3),((ab(j,i, 2),j=1,2),i=1,3) /0,  0,  2,  &
897499D-6,  -4955707D-6,3527363D-6,   1672673D-6,  &
-11826086D-7,   -333765D-7 /
DATA (ijsp(i, 3),i=1,3),((ab(j,i, 3),j=1,2),i=1,3) /0,  0,  3,  &
610820D-6,   1210521D-6,-1050939D-6,    327763D-6,  &
1593657D-7,  -1439953D-7 /
DATA (ijsp(i, 4),i=1,3),((ab(j,i, 4),j=1,2),i=1,3) /0,  0,  4,  &
-341639D-6,   -189719D-6,178691D-6,   -291925D-6,  &
-18948D-7,    482443D-7 /
DATA (ijsp(i, 5),i=1,3),((ab(j,i, 5),j=1,2),i=1,3) /0,  0,  5,  &
129027D-6,    -34863D-6,18763D-6,    100448D-6,  &
-66634D-7,    -85576D-7 /
DATA (ijsp(i, 6),i=1,3),((ab(j,i, 6),j=1,2),i=1,3) /0,  0,  6,  &
-38215D-6,     31061D-6,-30594D-6,    -25838D-6,  &
30841D-7,     -5765D-7 /
DATA (ijsp(i, 7),i=1,3),((ab(j,i, 7),j=1,2),i=1,3) /0,  1, -1,  &
20349D-6,     -9886D-6,4965D-6,     11263D-6,  &
-6140D-7,     22254D-7 /
DATA (ijsp(i, 8),i=1,3),((ab(j,i, 8),j=1,2),i=1,3) /0,  1,  0,  &
-4045D-6,     -4904D-6,310D-6,      -132D-6,  &
4434D-7,      4443D-7 /
DATA (ijsp(i, 9),i=1,3),((ab(j,i, 9),j=1,2),i=1,3) /0,  1,  1,  &
-5885D-6,     -3238D-6,2036D-6,      -947D-6,  &
-1518D-7,       641D-7 /
DATA (ijsp(i,10),i=1,3),((ab(j,i,10),j=1,2),i=1,3) /0,  1,  2,  &
-3812D-6,      3011D-6,-2D-6,      -674D-6,  &
-5D-7,       792D-7 /
DATA (ijsp(i,11),i=1,3),((ab(j,i,11),j=1,2),i=1,3) /0,  1,  3,  &
-601D-6,      3468D-6,-329D-6,      -563D-6,  &
518D-7,       518D-7 /
DATA (ijsp(i,12),i=1,3),((ab(j,i,12),j=1,2),i=1,3) /0,  2, -2,  &
1237D-6,       463D-6,-64D-6,        39D-6,  &
-13D-7,      -221D-7 /
DATA (ijsp(i,13),i=1,3),((ab(j,i,13),j=1,2),i=1,3) /0,  2, -1,  &
1086D-6,      -911D-6,-94D-6,       210D-6,  &
837D-7,      -494D-7 /
DATA (ijsp(i,14),i=1,3),((ab(j,i,14),j=1,2),i=1,3) /0,  2,  0,  &
595D-6,     -1229D-6,-8D-6,      -160D-6,  &
-281D-7,       616D-7 /
DATA (ijsp(i,15),i=1,3),((ab(j,i,15),j=1,2),i=1,3) /1, -1,  0,  &
2484D-6,      -485D-6,-177D-6,       259D-6,  &
260D-7,      -395D-7 /
DATA (ijsp(i,16),i=1,3),((ab(j,i,16),j=1,2),i=1,3) /1, -1,  1,  &
839D-6,     -1414D-6,17D-6,       234D-6,  &
-191D-7,      -396D-7 /
DATA (ijsp(i,17),i=1,3),((ab(j,i,17),j=1,2),i=1,3) /1,  0, -3,  &
-964D-6,      1059D-6,582D-6,      -285D-6,  &
-3218D-7,       370D-7 /
DATA (ijsp(i,18),i=1,3),((ab(j,i,18),j=1,2),i=1,3) /1,  0, -2,  &
-2303D-6,     -1038D-6,-298D-6,       692D-6,  &
8019D-7,     -7869D-7 /
DATA (ijsp(i,19),i=1,3),((ab(j,i,19),j=1,2),i=1,3) /1,  0, -1,  &
7049D-6,       747D-6,157D-6,       201D-6,  &
105D-7,     45637D-7 /
DATA (ijsp(i,20),i=1,3),((ab(j,i,20),j=1,2),i=1,3) /1,  0,  0,  &
1179D-6,      -358D-6,304D-6,       825D-6,  &
8623D-7,      8444D-7 /
DATA (ijsp(i,21),i=1,3),((ab(j,i,21),j=1,2),i=1,3) /1,  0,  1,  &
393D-6,       -63D-6,-124D-6,       -29D-6,  &
-896D-7,      -801D-7 /
DATA (ijsp(i,22),i=1,3),((ab(j,i,22),j=1,2),i=1,3) /1,  0,  2,  &
111D-6,      -268D-6,15D-6,         8D-6,  &
208D-7,      -122D-7 /
DATA (ijsp(i,23),i=1,3),((ab(j,i,23),j=1,2),i=1,3) /1,  0,  3,  &
-52D-6,      -154D-6,7D-6,        15D-6,  &
-133D-7,        65D-7 /
DATA (ijsp(i,24),i=1,3),((ab(j,i,24),j=1,2),i=1,3) /1,  0,  4,  &
-78D-6,       -30D-6,2D-6,         2D-6,  &
-16D-7,         1D-7 /
DATA (ijsp(i,25),i=1,3),((ab(j,i,25),j=1,2),i=1,3) /1,  1, -3,  &
-34D-6,       -26D-6,4D-6,         2D-6,  &
-22D-7,         7D-7 /
DATA (ijsp(i,26),i=1,3),((ab(j,i,26),j=1,2),i=1,3) /1,  1, -2,  &
-43D-6,         1D-6,3D-6,         0D-6,  &
-8D-7,        16D-7 /
DATA (ijsp(i,27),i=1,3),((ab(j,i,27),j=1,2),i=1,3) /1,  1, -1,  &
-15D-6,        21D-6,1D-6,        -1D-6,  &
2D-7,         9D-7 /
DATA (ijsp(i,28),i=1,3),((ab(j,i,28),j=1,2),i=1,3) /1,  1,  0,  &
-1D-6,        15D-6,0D-6,        -2D-6,  &
12D-7,         5D-7 /
DATA (ijsp(i,29),i=1,3),((ab(j,i,29),j=1,2),i=1,3) /1,  1,  1,  &
4D-6,         7D-6,1D-6,         0D-6,  &
1D-7,        -3D-7 /
DATA (ijsp(i,30),i=1,3),((ab(j,i,30),j=1,2),i=1,3) /1,  1,  3,  &
1D-6,         5D-6,1D-6,        -1D-6,  &
1D-7,         0D-7 /
DATA (ijsp(i,31),i=1,3),((ab(j,i,31),j=1,2),i=1,3) /2,  0, -6,  &
8D-6,         3D-6,-2D-6,        -3D-6,  &
9D-7,         5D-7 /
DATA (ijsp(i,32),i=1,3),((ab(j,i,32),j=1,2),i=1,3) /2,  0, -5,  &
-3D-6,         6D-6,1D-6,         2D-6,  &
2D-7,        -1D-7 /
DATA (ijsp(i,33),i=1,3),((ab(j,i,33),j=1,2),i=1,3) /2,  0, -4,  &
6D-6,       -13D-6,-8D-6,         2D-6,  &
14D-7,        10D-7 /
DATA (ijsp(i,34),i=1,3),((ab(j,i,34),j=1,2),i=1,3) /2,  0, -3,  &
10D-6,        22D-6,10D-6,        -7D-6,  &
-65D-7,        12D-7 /
DATA (ijsp(i,35),i=1,3),((ab(j,i,35),j=1,2),i=1,3) /2,  0, -2,  &
-57D-6,       -32D-6,0D-6,        21D-6,  &
126D-7,      -233D-7 /
DATA (ijsp(i,36),i=1,3),((ab(j,i,36),j=1,2),i=1,3) /2,  0, -1,  &
157D-6,       -46D-6,8D-6,         5D-6,  &
270D-7,      1068D-7 /
DATA (ijsp(i,37),i=1,3),((ab(j,i,37),j=1,2),i=1,3) /2,  0,  0,  &
12D-6,       -18D-6,13D-6,        16D-6,  &
254D-7,       155D-7 /
DATA (ijsp(i,38),i=1,3),((ab(j,i,38),j=1,2),i=1,3) /2,  0,  1,  &
-4D-6,         8D-6,-2D-6,        -3D-6,  &
-26D-7,        -2D-7 /
DATA (ijsp(i,39),i=1,3),((ab(j,i,39),j=1,2),i=1,3) /2,  0,  2,  &
-5D-6,         0D-6,0D-6,         0D-6,  &
7D-7,         0D-7 /
DATA (ijsp(i,40),i=1,3),((ab(j,i,40),j=1,2),i=1,3) /2,  0,  3,  &
3D-6,         4D-6,0D-6,         1D-6,  &
-11D-7,         4D-7 /
DATA (ijsp(i,41),i=1,3),((ab(j,i,41),j=1,2),i=1,3) /3,  0, -2,  &
-1D-6,        -1D-6,0D-6,         1D-6,  &
4D-7,       -14D-7 /
DATA (ijsp(i,42),i=1,3),((ab(j,i,42),j=1,2),i=1,3) /3,  0, -1,  &
6D-6,        -3D-6,0D-6,         0D-6,  &
18D-7,        35D-7 /
DATA (ijsp(i,43),i=1,3),((ab(j,i,43),j=1,2),i=1,3) /3,  0,  0,  &
-1D-6,        -2D-6,0D-6,         1D-6,  &
13D-7,         3D-7 /


!  Validate the planet number.
IF (np < 1.OR.np > 9) THEN
  jstat=-1
  DO i=1,6
    pv(i)=0D0
  END DO
ELSE
  
!     Separate algorithms for Pluto and the rest.
  IF (np /= 9) THEN
    
!        -----------------------
!        Mercury through Neptune
!        -----------------------
    
!        Time: Julian millennia since J2000.
    t=(date-51544.5D0)/365250D0
    
!        OK status unless remote epoch.
    IF (ABS(t) <= 1D0) THEN
      jstat=0
    ELSE
      jstat=1
    END IF
    
!        Compute the mean elements.
    da=a(1,np)+(a(2,np)+a(3,np)*t)*t
    dl=(3600D0*dlm(1,np)+(dlm(2,np)+dlm(3,np)*t)*t)*as2r
    de=e(1,np)+(e(2,np)+e(3,np)*t)*t
    dpe=MOD((3600D0*pi(1,np)+(pi(2,np)+pi(3,np)*t)*t)*as2r,d2pi)
    di=(3600D0*dinc(1,np)+(dinc(2,np)+dinc(3,np)*t)*t)*as2r
    DO=MOD((3600D0*omega(1,np)+(omega(2,np)+omega(3,np)*t)*t)*as2r,d2pi)
    
!        Apply the trigonometric terms.
    dmu=0.35953620D0*t
    DO j=1,8
      arga=dkp(j,np)*dmu
      argl=dkq(j,np)*dmu
      da=da+(ca(j,np)*COS(arga)+sa(j,np)*SIN(arga))*1D-7
      dl=dl+(clo(j,np)*COS(argl)+slo(j,np)*SIN(argl))*1D-7
    END DO
    arga=dkp(9,np)*dmu
    da=da+t*(ca(9,np)*COS(arga)+sa(9,np)*SIN(arga))*1D-7
    DO j=9,10
      argl=dkq(j,np)*dmu
      dl=dl+t*(clo(j,np)*COS(argl)+slo(j,np)*SIN(argl))*1D-7
    END DO
    dl=MOD(dl,d2pi)
    
!        Daily motion.
    dm=gcon*SQRT((1D0+1D0/amas(np))/(da*da*da))
    
!        Make the prediction.
    CALL sla_planel(date,1,date,di,DO,dpe,da,de,dl,dm,pv,j)
    IF (j < 0) jstat=-2
    
  ELSE
    
!        -----
!        Pluto
!        -----
    
!        Time: Julian centuries since J2000.
    t=(date-51544.5D0)/36525D0
    
!        OK status unless remote epoch.
    IF (t >= -1.15D0.AND.t <= 1D0) THEN
      jstat=0
    ELSE
      jstat=1
    END IF
    
!        Fundamental arguments (radians).
    dj=(dj0+djd*t)*d2r
    ds=(ds0+dsd*t)*d2r
    dp=(dp0+dpd*t)*d2r
    
!        Initialize coefficients and derivatives.
    DO i=1,3
      wlbr(i)=0D0
      wlbrd(i)=0D0
    END DO
    
!        Term by term through Meeus Table 36.A.
    DO j=1,43
      
!           Argument and derivative (radians, radians per century).
      wj=DBLE(ijsp(1,j))
      ws=DBLE(ijsp(2,j))
      wp=DBLE(ijsp(3,j))
      al=wj*dj+ws*ds+wp*dp
      ald=(wj*djd+ws*dsd+wp*dpd)*d2r
      
!           Functions of argument.
      sal=SIN(al)
      cal=COS(al)
      
!           Periodic terms in longitude, latitude, radius vector.
      DO i=1,3
        
!              A and B coefficients (deg, AU).
        ac=ab(1,i,j)
        bc=ab(2,i,j)
        
!              Periodic terms (deg, AU, deg/Jc, AU/Jc).
        wlbr(i)=wlbr(i)+ac*sal+bc*cal
        wlbrd(i)=wlbrd(i)+(ac*cal-bc*sal)*ald
      END DO
    END DO
    
!        Heliocentric longitude and derivative (radians, radians/sec).
    dl=(dl0+dld0*t+wlbr(1))*d2r
    dld=(dld0+wlbrd(1))*d2r/spc
    
!        Heliocentric latitude and derivative (radians, radians/sec).
    db=(db0+wlbr(2))*d2r
    dbd=wlbrd(2)*d2r/spc
    
!        Heliocentric radius vector and derivative (AU, AU/sec).
    dr=dr0+wlbr(3)
    drd=wlbrd(3)/spc
    
!        Functions of latitude, longitude, radius vector.
    sl=SIN(dl)
    cl=COS(dl)
    sb=SIN(db)
    cb=COS(db)
    slcb=sl*cb
    clcb=cl*cb
    
!        Heliocentric vector and derivative, J2000 ecliptic and equinox.
    x=dr*clcb
    y=dr*slcb
    z=dr*sb
    xd=drd*clcb-dr*(cl*sb*dbd+slcb*dld)
    yd=drd*slcb+dr*(-sl*sb*dbd+clcb*dld)
    zd=drd*sb+dr*cb*dbd
    
!        Transform to J2000 equator and equinox.
    pv(1)=x
    pv(2)=y*ce-z*se
    pv(3)=y*se+z*ce
    pv(4)=xd
    pv(5)=yd*ce-zd*se
    pv(6)=yd*se+zd*ce
  END IF
END IF

END SUBROUTINE sla_planet
SUBROUTINE sla_plante (date, elong, phi, jform, epoch,  &
orbinc, anode, perih, aorq, e,aorl, dm, ra, dec, r, jstat)
!+
!     - - - - - - -
!      P L A N T E
!     - - - - - - -

!  Topocentric apparent RA,Dec of a Solar-System object whose
!  heliocentric orbital elements are known.

!  Given:
!     DATE      d     MJD of observation (JD - 2400000.5)
!     ELONG     d     observer's east longitude (radians)
!     PHI       d     observer's geodetic latitude (radians)
!     JFORM     i     choice of element set (1-3, see Note 4, below)
!     EPOCH     d     epoch of elements (TT MJD)
!     ORBINC    d     inclination (radians)
!     ANODE     d     longitude of the ascending node (radians)
!     PERIH     d     longitude or argument of perihelion (radians)
!     AORQ      d     mean distance or perihelion distance (AU)
!     E         d     eccentricity
!     AORL      d     mean anomaly or longitude (radians, JFORM=1,2 only)
!     DM        d     daily motion (radians, JFORM=1 only )

!  Returned:
!     RA,DEC    d     RA, Dec (topocentric apparent, radians)
!     R         d     distance from observer (AU)
!     JSTAT     i     status:  0 = OK
!                             -1 = illegal JFORM
!                             -2 = illegal E
!                             -3 = illegal AORQ
!                             -4 = illegal DM
!                             -5 = failed to converge

!  Notes:

!  1  DATE is the instant for which the prediction is required.  It is
!     in the TT timescale (formerly Ephemeris Time, ET) and is a
!     Modified Julian Date (JD-2400000.5).

!  2  The longitude and latitude allow correction for geocentric
!     parallax.  This is usually a small effect, but can become
!     important for Earth-crossing asteroids.  Geocentric positions
!     can be generated by appropriate use of routines sla_EVP and
!     sla_PLANEL.

!  3  The elements are with respect to the J2000 ecliptic and equinox.

!  4  Three different element-format options are available:

!     Option JFORM=1, suitable for the major planets:

!     EPOCH  = epoch of elements (TT MJD)
!     ORBINC = inclination i (radians)
!     ANODE  = longitude of the ascending node, big omega (radians)
!     PERIH  = longitude of perihelion, curly pi (radians)
!     AORQ   = mean distance, a (AU)
!     E      = eccentricity, e
!     AORL   = mean longitude L (radians)
!     DM     = daily motion (radians)

!     Option JFORM=2, suitable for minor planets:

!     EPOCH  = epoch of elements (TT MJD)
!     ORBINC = inclination i (radians)
!     ANODE  = longitude of the ascending node, big omega (radians)
!     PERIH  = argument of perihelion, little omega (radians)
!     AORQ   = mean distance, a (AU)
!     E      = eccentricity, e
!     AORL   = mean anomaly M (radians)

!     Option JFORM=3, suitable for comets:

!     EPOCH  = epoch of perihelion (TT MJD)
!     ORBINC = inclination i (radians)
!     ANODE  = longitude of the ascending node, big omega (radians)
!     PERIH  = argument of perihelion, little omega (radians)
!     AORQ   = perihelion distance, q (AU)
!     E      = eccentricity, e

!  5  Unused elements (DM for JFORM=2, AORL and DM for JFORM=3) are
!     not accessed.

!  Called: sla_GMST, sla_DT, sla_EPJ, sla_PVOBS, sla_PRENUT,
!          sla_PLANEL, sla_DMXV, sla_DCC2S, sla_DRANRM

!  P.T.Wallace   Starlink   25 June 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date,elong,phi
INTEGER :: jform
DOUBLE PRECISION :: epoch,orbinc,anode,perih,aorq,e,aorl,dm,ra,dec,r
INTEGER :: jstat

!  Light time for unit distance (sec)
DOUBLE PRECISION :: tau
PARAMETER (tau=499.004782D0)

INTEGER :: i
DOUBLE PRECISION :: dvb(3),dpb(3),vsg(6),vsp(6),v(6),rmat(3,3),  &
vgp(6),stl,vgo(6),dx,dy,dz,d,tl



!  Sun to geocentre (J2000).
CALL sla_evp(date,2000D0,dvb,dpb,vsg(4),vsg)

!  Sun to planet (J2000).
CALL sla_planel(date,jform,epoch,orbinc,anode,perih,aorq,e,aorl,dm,vsp,jstat)

!  Geocentre to planet (J2000).
DO i=1,6
  v(i)=vsp(i)-vsg(i)
END DO

!  Precession and nutation to date.
CALL sla_prenut(2000D0,date,rmat)
CALL sla_dmxv(rmat,v,vgp)
CALL sla_dmxv(rmat,v(4),vgp(4))

!  Geocentre to observer (date).
stl=sla_gmst(date-sla_dt(sla_epj(date))/86400D0)+elong
CALL sla_pvobs(phi,0D0,stl,vgo)

!  Observer to planet (date).
DO i=1,6
  v(i)=vgp(i)-vgo(i)
END DO

!  Geometric distance (AU).
dx=v(1)
dy=v(2)
dz=v(3)
d=SQRT(dx*dx+dy*dy+dz*dz)

!  Light time (sec).
tl=tau*d

!  Correct position for planetary aberration
DO i=1,3
  v(i)=v(i)-tl*v(i+3)
END DO

!  To RA,Dec.
CALL sla_dcc2s(v,ra,dec)
ra=sla_dranrm(ra)
r=d

END SUBROUTINE sla_plante
SUBROUTINE sla_pm (r0, d0, pr, pd, px, rv, ep0, ep1, r1, d1)
!+
!     - - -
!      P M
!     - - -

!  Apply corrections for proper motion to a star RA,Dec
!  (double precision)

!  References:
!     1984 Astronomical Almanac, pp B39-B41.
!     (also Lederle & Schwan, Astron. Astrophys. 134,
!      1-6, 1984)

!  Given:
!     R0,D0    dp     RA,Dec at epoch EP0 (rad)
!     PR,PD    dp     proper motions:  RA,Dec changes per year of epoch
!     PX       dp     parallax (arcsec)
!     RV       dp     radial velocity (km/sec, +ve if receding)
!     EP0      dp     start epoch in years (e.g. Julian epoch)
!     EP1      dp     end epoch in years (same system as EP0)

!  Returned:
!     R1,D1    dp     RA,Dec at epoch EP1 (rad)

!  Called:
!     sla_DCS2C       spherical to Cartesian
!     sla_DCC2S       Cartesian to spherical
!     sla_DRANRM      normalize angle 0-2Pi

!  Note:
!     The proper motions in RA are dRA/dt rather than
!     cos(Dec)*dRA/dt, and are in the same coordinate
!     system as R0,D0.

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: r0,d0,pr,pd,px,rv,ep0,ep1,r1,d1

!  Km/s to AU/year multiplied by arc seconds to radians
DOUBLE PRECISION :: vfr
PARAMETER (vfr=0.21094502D0*0.484813681109535994D-5)

INTEGER :: i
DOUBLE PRECISION :: w,em(3),t,p(3)



!  Spherical to Cartesian
CALL sla_dcs2c(r0,d0,p)

!  Space motion (radians per year)
w=vfr*rv*px
em(1)=-pr*p(2)-pd*COS(r0)*SIN(d0)+w*p(1)
em(2)= pr*p(1)-pd*SIN(r0)*SIN(d0)+w*p(2)
em(3)=         pd*COS(d0)        +w*p(3)

!  Apply the motion
t=ep1-ep0
DO i=1,3
  p(i)=p(i)+t*em(i)
END DO

!  Cartesian to spherical
CALL sla_dcc2s(p,r1,d1)
r1=sla_dranrm(r1)

END SUBROUTINE sla_pm
SUBROUTINE sla_polmo ( elongm, phim, xp, yp, elong, phi, daz )
!+
!     - - - - - -
!      P O L M O
!     - - - - - -

!  Polar motion:  correct site longitude and latitude for polar
!  motion and calculate azimuth difference between celestial and
!  terrestrial poles.

!  Given:
!     ELONGM   d      mean longitude of the observer (radians, east +ve)
!     PHIM     d      mean geodetic latitude of the observer (radians)
!     XP       d      polar motion x-coordinate (radians)
!     YP       d      polar motion y-coordinate (radians)

!  Returned:
!     ELONG    d      true longitude of the observer (radians, east +ve)
!     PHI      d      true geodetic latitude of the observer (radians)
!     DAZ      d      azimuth correction (terrestrial-celestial, radians)

!  Notes:

!   1)  "Mean" longitude and latitude are the (fixed) values for the
!       site's location with respect to the IERS terrestrial reference
!       frame;  the latitude is geodetic.  TAKE CARE WITH THE LONGITUDE
!       SIGN CONVENTION.  The longitudes used by the present routine
!       are east-positive, in accordance with geographical convention
!       (and right-handed).  In particular, note that the longitudes
!       returned by the sla_OBS routine are west-positive, following
!       astronomical usage, and must be reversed in sign before use in
!       the present routine.

!   2)  XP and YP are the (changing) coordinates of the Celestial
!       Ephemeris Pole with respect to the IERS Reference Pole.
!       XP is positive along the meridian at longitude 0 degrees,
!       and YP is positive along the meridian at longitude
!       270 degrees (i.e. 90 degrees west).  Values for XP,YP can
!       be obtained from IERS circulars and equivalent publications;
!       the maximum amplitude observed so far is about 0.3 arcseconds.

!   3)  "True" longitude and latitude are the (moving) values for
!       the site's location with respect to the celestial ephemeris
!       pole and the meridian which corresponds to the Greenwich
!       apparent sidereal time.  The true longitude and latitude
!       link the terrestrial coordinates with the standard celestial
!       models (for precession, nutation, sidereal time etc).

!   4)  The azimuths produced by sla_AOP and sla_AOPQK are with
!       respect to due north as defined by the Celestial Ephemeris
!       Pole, and can therefore be called "celestial azimuths".
!       However, a telescope fixed to the Earth measures azimuth
!       essentially with respect to due north as defined by the
!       IERS Reference Pole, and can therefore be called "terrestrial
!       azimuth".  Uncorrected, this would manifest itself as a
!       changing "azimuth zero-point error".  The value DAZ is the
!       correction to be added to a celestial azimuth to produce
!       a terrestrial azimuth.

!   5)  The present routine is rigorous.  For most practical
!       purposes, the following simplified formulae provide an
!       adequate approximation:

!       ELONG = ELONGM+XP*COS(ELONGM)-YP*SIN(ELONGM)
!       PHI   = PHIM+(XP*SIN(ELONGM)+YP*COS(ELONGM))*TAN(PHIM)
!       DAZ   = -SQRT(XP*XP+YP*YP)*COS(ELONGM-ATAN2(XP,YP))/COS(PHIM)

!       An alternative formulation for DAZ is:

!       X = COS(ELONGM)*COS(PHIM)
!       Y = SIN(ELONGM)*COS(PHIM)
!       DAZ = ATAN2(-X*YP-Y*XP,X*X+Y*Y)

!   Reference:  Seidelmann, P.K. (ed), 1992.  "Explanatory Supplement
!               to the Astronomical Almanac", ISBN 0-935702-68-7,
!               sections 3.27, 4.25, 4.52.

!  P.T.Wallace   Starlink   22 February 1996

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: elongm,phim,xp,yp,elong,phi,daz

DOUBLE PRECISION :: sel,cel,sph,cph,xm,ym,zm,xnm,ynm,znm,  &
sxp,cxp,syp,cyp,zw,xt,yt,zt,xnt,ynt



!  Site mean longitude and mean geodetic latitude as a Cartesian vector
sel=SIN(elongm)
cel=COS(elongm)
sph=SIN(phim)
cph=COS(phim)

xm=cel*cph
ym=sel*cph
zm=sph

!  Rotate site vector by polar motion, Y-component then X-component
sxp=SIN(xp)
cxp=COS(xp)
syp=SIN(yp)
cyp=COS(yp)

zw=(-ym*syp+zm*cyp)

xt=xm*cxp-zw*sxp
yt=ym*cyp+zm*syp
zt=xm*sxp+zw*cxp

!  Rotate also the geocentric direction of the terrestrial pole (0,0,1)
xnm=-sxp*cyp
ynm=syp
znm=cxp*cyp

cph=SQRT(xt*xt+yt*yt)
IF (cph == 0D0) xt=1D0
sel=yt/cph
cel=xt/cph

!  Return true longitude and true geodetic latitude of site
elong=ATAN2(yt,xt)
phi=ATAN2(zt,cph)

!  Return current azimuth of terrestrial pole seen from site position
xnt=(xnm*cel+ynm*sel)*zt-znm*cph
ynt=-xnm*sel+ynm*cel
daz=ATAN2(-ynt,-xnt)

END SUBROUTINE sla_polmo
SUBROUTINE sla_prebn (bep0, bep1, rmatp)
!+
!     - - - - - -
!      P R E B N
!     - - - - - -

!  Generate the matrix of precession between two epochs,
!  using the old, pre-IAU1976, Bessel-Newcomb model, using
!  Kinoshita's formulation (double precision)

!  Given:
!     BEP0    dp         beginning Besselian epoch
!     BEP1    dp         ending Besselian epoch

!  Returned:
!     RMATP  dp(3,3)    precession matrix

!  The matrix is in the sense   V(BEP1)  =  RMATP * V(BEP0)

!  Reference:
!     Kinoshita, H. (1975) 'Formulas for precession', SAO Special
!     Report No. 364, Smithsonian Institution Astrophysical
!     Observatory, Cambridge, Massachusetts.

!  Called:  sla_DEULER

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: bep0,bep1,rmatp(3,3)

!  Arc seconds to radians
DOUBLE PRECISION :: as2r
PARAMETER (as2r=0.484813681109535994D-5)

DOUBLE PRECISION :: bigt,t,tas2r,w,zeta,z,theta



!  Interval between basic epoch B1850.0 and beginning epoch in TC
bigt = (bep0-1850D0)/100D0

!  Interval over which precession required, in tropical centuries
t = (bep1-bep0)/100D0

!  Euler angles
tas2r = t*as2r
w = 2303.5548D0+(1.39720D0+0.000059D0*bigt)*bigt

zeta = (w+(0.30242D0-0.000269D0*bigt+0.017996D0*t)*t)*tas2r
z = (w+(1.09478D0+0.000387D0*bigt+0.018324D0*t)*t)*tas2r
theta = (2005.1125D0+(-0.85294D0-0.000365D0*bigt)*bigt+  &
(-0.42647D0-0.000365D0*bigt-0.041802D0*t)*t)*tas2r

!  Rotation matrix
CALL sla_deuler('ZYZ',-zeta,theta,-z,rmatp)

END SUBROUTINE sla_prebn
SUBROUTINE sla_prec (ep0, ep1, rmatp)
!+
!     - - - - -
!      P R E C
!     - - - - -

!  Form the matrix of precession between two epochs (IAU 1976, FK5)
!  (double precision)

!  Given:
!     EP0    dp         beginning epoch
!     EP1    dp         ending epoch

!  Returned:
!     RMATP  dp(3,3)    precession matrix

!  Notes:

!     1)  The epochs are TDB (loosely ET) Julian epochs.

!     2)  The matrix is in the sense   V(EP1)  =  RMATP * V(EP0)

!     3)  Though the matrix method itself is rigorous, the precession
!         angles are expressed through canonical polynomials which are
!         valid only for a limited time span.  There are also known
!         errors in the IAU precession rate.  The absolute accuracy
!         of the present formulation is better than 0.1 arcsec from
!         1960AD to 2040AD, better than 1 arcsec from 1640AD to 2360AD,
!         and remains below 3 arcsec for the whole of the period
!         500BC to 3000AD.  The errors exceed 10 arcsec outside the
!         range 1200BC to 3900AD, exceed 100 arcsec outside 4200BC to
!         5600AD and exceed 1000 arcsec outside 6800BC to 8200AD.
!         The SLALIB routine sla_PRECL implements a more elaborate
!         model which is suitable for problems spanning several
!         thousand years.

!  References:
!     Lieske,J.H., 1979. Astron.Astrophys.,73,282.
!      equations (6) & (7), p283.
!     Kaplan,G.H., 1981. USNO circular no. 163, pA2.

!  Called:  sla_DEULER

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ep0,ep1,rmatp(3,3)

!  Arc seconds to radians
DOUBLE PRECISION :: as2r
PARAMETER (as2r=0.484813681109535994D-5)

DOUBLE PRECISION :: t0,t,tas2r,w,zeta,z,theta



!  Interval between basic epoch J2000.0 and beginning epoch (JC)
t0 = (ep0-2000D0)/100D0

!  Interval over which precession required (JC)
t = (ep1-ep0)/100D0

!  Euler angles
tas2r = t*as2r
w = 2306.2181D0+(1.39656D0-0.000139D0*t0)*t0

zeta = (w+((0.30188D0-0.000344D0*t0)+0.017998D0*t)*t)*tas2r
z = (w+((1.09468D0+0.000066D0*t0)+0.018203D0*t)*t)*tas2r
theta = ((2004.3109D0+(-0.85330D0-0.000217D0*t0)*t0)  &
+((-0.42665D0-0.000217D0*t0)-0.041833D0*t)*t)*tas2r

!  Rotation matrix
CALL sla_deuler('ZYZ',-zeta,theta,-z,rmatp)

END SUBROUTINE sla_prec
SUBROUTINE sla_preces (system, ep0, ep1, ra, dc)
!+
!     - - - - - - -
!      P R E C E S
!     - - - - - - -

!  Precession - either FK4 (Bessel-Newcomb, pre IAU 1976) or
!  FK5 (Fricke, post IAU 1976) as required.

!  Given:
!     SYSTEM     char   precession to be applied: 'FK4' or 'FK5'
!     EP0,EP1    dp     starting and ending epoch
!     RA,DC      dp     RA,Dec, mean equator & equinox of epoch EP0

!  Returned:
!     RA,DC      dp     RA,Dec, mean equator & equinox of epoch EP1

!  Called:    sla_DRANRM, sla_PREBN, sla_PREC, sla_DCS2C,
!             sla_DMXV, sla_DCC2S

!  Notes:

!     1)  Lowercase characters in SYSTEM are acceptable.

!     2)  The epochs are Besselian if SYSTEM='FK4' and Julian if 'FK5'.
!         For example, to precess coordinates in the old system from
!         equinox 1900.0 to 1950.0 the call would be:
!             CALL sla_PRECES ('FK4', 1900D0, 1950D0, RA, DC)

!     3)  This routine will NOT correctly convert between the old and
!         the new systems - for example conversion from B1950 to J2000.
!         For these purposes see sla_FK425, sla_FK524, sla_FK45Z and
!         sla_FK54Z.

!     4)  If an invalid SYSTEM is supplied, values of -99D0,-99D0 will
!         be returned for both RA and DC.

!  P.T.Wallace   Starlink   20 April 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

CHARACTER :: system*(*)
DOUBLE PRECISION :: ep0,ep1,ra,dc

DOUBLE PRECISION :: pm(3,3),v1(3),v2(3)
CHARACTER :: sysuc*3





!  Convert to uppercase and validate SYSTEM
sysuc=system
IF (sysuc(1:1) == 'f') sysuc(1:1)='F'
IF (sysuc(2:2) == 'k') sysuc(2:2)='K'
IF (sysuc /= 'FK4'.AND.sysuc /= 'FK5') THEN
  ra=-99D0
  dc=-99D0
ELSE
  
!     Generate appropriate precession matrix
  IF (sysuc == 'FK4') THEN
    CALL sla_prebn(ep0,ep1,pm)
  ELSE
    CALL sla_prec(ep0,ep1,pm)
  END IF
  
!     Convert RA,Dec to x,y,z
  CALL sla_dcs2c(ra,dc,v1)
  
!     Precess
  CALL sla_dmxv(pm,v1,v2)
  
!     Back to RA,Dec
  CALL sla_dcc2s(v2,ra,dc)
  ra=sla_dranrm(ra)
  
END IF

END SUBROUTINE sla_preces
SUBROUTINE sla_precl (ep0, ep1, rmatp)
!+
!     - - - - - -
!      P R E C L
!     - - - - - -

!  Form the matrix of precession between two epochs, using the
!  model of Simon et al (1994), which is suitable for long
!  periods of time.

!  (double precision)

!  Given:
!     EP0    dp         beginning epoch
!     EP1    dp         ending epoch

!  Returned:
!     RMATP  dp(3,3)    precession matrix

!  Notes:

!     1)  The epochs are TDB Julian epochs.

!     2)  The matrix is in the sense   V(EP1)  =  RMATP * V(EP0)

!     3)  The absolute accuracy of the model is limited by the
!         uncertainty in the general precession, about 0.3 arcsec per
!         1000 years.  The remainder of the formulation provides a
!         precision of 1 mas over the interval from 1000AD to 3000AD,
!         0.1 arcsec from 1000BC to 5000AD and 1 arcsec from
!         4000BC to 8000AD.

!  Reference:
!     Simon, J.L. et al., 1994. Astron.Astrophys., 282, 663-683.

!  Called:  sla_DEULER

!  P.T.Wallace   Starlink   23 August 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ep0,ep1,rmatp(3,3)

!  Arc seconds to radians
DOUBLE PRECISION :: as2r
PARAMETER (as2r=0.484813681109535994D-5)

DOUBLE PRECISION :: t0,t,tas2r,w,zeta,z,theta



!  Interval between basic epoch J2000.0 and beginning epoch (1000JY)
t0 = (ep0-2000D0)/1000D0

!  Interval over which precession required (1000JY)
t = (ep1-ep0)/1000D0

!  Euler angles
tas2r = t*as2r
w =      23060.9097D0+(139.7459D0+  &
(-0.0038D0+(-0.5918D0+  &
(-0.0037D0+0.0007D0*t0)*t0)*t0)*t0)*t0

zeta =   (w+(30.2226D0+(-0.2523D0+  &
(-0.3840D0+(-0.0014D0+  &
0.0007D0*t0)*t0)*t0)*t0+(18.0183D0+  &
(-0.1326D0+(0.0006D0+  &
0.0005D0*t0)*t0)*t0+(-0.0583D0+  &
(-0.0001D0+0.0007D0*t0)*t0+  &
(-0.0285D0+(-0.0002D0)*t)*t)*t)*t)*t)*tas2r

z =     (w+(109.5270D0+(0.2446D0+  &
(-1.3913D0+(-0.0134D0+  &
0.0026D0*t0)*t0)*t0)*t0+(18.2667D0+  &
(-1.1400D0+(-0.0173D0+  &
0.0044D0*t0)*t0)*t0+(-0.2821D0+  &
(-0.0093D0+0.0032D0*t0)*t0+  &
(-0.0301D0+0.0006D0*t0  &
-0.0001D0*t)*t)*t)*t)*t)*tas2r

theta =  (20042.0207D0+(-85.3131D0+  &
(-0.2111D0+(0.3642D0+  &
(0.0008D0+(-0.0005D0)*t0)*t0)*t0)*t0)*t0+  &
(-42.6566D0+(-0.2111D0+  &
(0.5463D0+(0.0017D0+  &
(-0.0012D0)*t0)*t0)*t0)*t0+(-41.8238D0+  &
(0.0359D0+(0.0027D0+  &
(-0.0001D0)*t0)*t0)*t0+(-0.0731D0+  &
(0.0019D0+0.0009D0*t0)*t0+  &
(-0.0127D0+0.0011D0*t0+0.0004D0*t)*t)*t)*t)*t)*tas2r

!  Rotation matrix
CALL sla_deuler('ZYZ',-zeta,theta,-z,rmatp)

END SUBROUTINE sla_precl
SUBROUTINE sla_prenut (epoch, date, rmatpn)
!+
!     - - - - - - -
!      P R E N U T
!     - - - - - - -

!  Form the matrix of precession and nutation (IAU1976/FK5)
!  (double precision)

!  Given:
!     EPOCH   dp         Julian Epoch for mean coordinates
!     DATE    dp         Modified Julian Date (JD-2400000.5)
!                        for true coordinates

!  Returned:
!     RMATPN  dp(3,3)    combined precession/nutation matrix

!  Called:  sla_PREC, sla_EPJ, sla_NUT, sla_DMXM

!  Notes:

!  1)  The epoch and date are TDB (loosely ET).

!  2)  The matrix is in the sense   V(true)  =  RMATPN * V(mean)

!  P.T.Wallace   Starlink   April 1987

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: epoch,date,rmatpn(3,3)

DOUBLE PRECISION :: rmatp(3,3),rmatn(3,3)



!  Precession
CALL sla_prec(epoch,sla_epj(date),rmatp)

!  Nutation
CALL sla_nut(date,rmatn)

!  Combine the matrices:  PN = N x P
CALL sla_dmxm(rmatn,rmatp,rmatpn)

END SUBROUTINE sla_prenut
SUBROUTINE sla_pvobs (p, h, stl, pv)
!+
!     - - - - - -
!      P V O B S
!     - - - - - -

!  Position and velocity of an observing station (double precision)

!  Given:
!     P     dp     latitude (geodetic, radians)
!     H     dp     height above reference spheroid (geodetic, metres)
!     STL   dp     local apparent sidereal time (radians)

!  Returned:
!     PV    dp(6)  position/velocity 6-vector (AU, AU/s, true equator
!                                              and equinox of date)

!  Called:  sla_GEOC

!  IAU 1976 constants are used.

!  P.T.Wallace   Starlink   14 November 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: p,h,stl,pv(6)

DOUBLE PRECISION :: r,z,s,c,v

!  Mean sidereal rate (at J2000) in radians per (UT1) second
DOUBLE PRECISION :: sr
PARAMETER (sr=7.292115855306589D-5)



!  Geodetic to geocentric conversion
CALL sla_geoc(p,h,r,z)

!  Functions of ST
s=SIN(stl)
c=COS(stl)

!  Speed
v=sr*r

!  Position
pv(1)=r*c
pv(2)=r*s
pv(3)=z

!  Velocity
pv(4)=-v*s
pv(5)=v*c
pv(6)=0D0

END SUBROUTINE sla_pvobs
SUBROUTINE sla_pxy (np,xye,xym,coeffs,xyp,xrms,yrms,rrms)
!+
!     - - - -
!      P X Y
!     - - - -

!  Given arrays of "expected" and "measured" [X,Y] coordinates, and a
!  linear model relating them (as produced by sla_FITXY), compute
!  the array of "predicted" coordinates and the RMS residuals.

!  Given:
!     NP       i        number of samples
!     XYE     d(2,np)   expected [X,Y] for each sample
!     XYM     d(2,np)   measured [X,Y] for each sample
!     COEFFS  d(6)      coefficients of model (see below)

!  Returned:
!     XYP     d(2,np)   predicted [X,Y] for each sample
!     XRMS     d        RMS in X
!     YRMS     d        RMS in Y
!     RRMS     d        total RMS (vector sum of XRMS and YRMS)

!  The model is supplied in the array COEFFS.  Naming the
!  elements of COEFF as follows:

!     COEFFS(1) = A
!     COEFFS(2) = B
!     COEFFS(3) = C
!     COEFFS(4) = D
!     COEFFS(5) = E
!     COEFFS(6) = F

!  the model is applied thus:

!     XP = A + B*XM + C*YM
!     YP = D + E*XM + F*YM

!  The residuals are (XP-XE) and (YP-YE).

!  If NP is less than or equal to zero, no coordinates are
!  transformed, and the RMS residuals are all zero.

!  See also sla_FITXY, sla_INVF, sla_XY2XY, sla_DCMPF

!  Called:  sla_XY2XY

!  P.T.Wallace   Starlink   22 May 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: np
DOUBLE PRECISION :: xye(2,np),xym(2,np),coeffs(6),xyp(2,np),xrms,yrms,rrms

INTEGER :: i
DOUBLE PRECISION :: sdx2,sdy2,xp,yp,dx,dy,dx2,dy2,p



!  Initialize summations
sdx2=0D0
sdy2=0D0

!  Loop by sample
DO i=1,np
  
!     Transform "measured" [X,Y] to "predicted" [X,Y]
  CALL sla_xy2xy(xym(1,i),xym(2,i),coeffs,xp,yp)
  xyp(1,i)=xp
  xyp(2,i)=yp
  
!     Compute residuals in X and Y, and update summations
  dx=xye(1,i)-xp
  dy=xye(2,i)-yp
  dx2=dx*dx
  dy2=dy*dy
  sdx2=sdx2+dx2
  sdy2=sdy2+dy2
  
!     Next sample
END DO

!  Compute RMS values
p=MAX(1D0,DBLE(np))
xrms=SQRT(sdx2/p)
yrms=SQRT(sdy2/p)
rrms=SQRT(xrms*xrms+yrms*yrms)

END SUBROUTINE sla_pxy
REAL FUNCTION sla_range (angle)
!+
!     - - - - - -
!      R A N G E
!     - - - - - -

!  Normalize angle into range +/- pi  (single precision)

!  Given:
!     ANGLE     dp      the angle in radians

!  The result is ANGLE expressed in the +/- pi (single
!  precision).

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: angle

REAL :: api,a2pi
PARAMETER (api=3.141592653589793238462643)
PARAMETER (a2pi=6.283185307179586476925287)


sla_range=MOD(angle,a2pi)
IF (ABS(sla_range) >= api) sla_range=sla_range-SIGN(a2pi,angle)

END FUNCTION sla_range
REAL FUNCTION sla_ranorm (angle)
!+
!     - - - - - - -
!      R A N O R M
!     - - - - - - -

!  Normalize angle into range 0-2 pi  (single precision)

!  Given:
!     ANGLE     dp      the angle in radians

!  The result is ANGLE expressed in the range 0-2 pi (single
!  precision).

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: angle

REAL :: a2pi
PARAMETER (a2pi=6.283185307179586476925287)


sla_ranorm=MOD(angle,a2pi)
IF (sla_ranorm < 0.0) sla_ranorm=sla_ranorm+a2pi

END FUNCTION sla_ranorm
!!$DOUBLE PRECISION FUNCTION sla_rcc (tdb, ut1, wl, u, v)
!!$!+
!!$!     - - - -
!!$!      R C C
!!$!     - - - -
!!$
!!$!  Relativistic clock correction:  the difference between proper time at
!!$!  a point on the surface of the Earth and coordinate time in the Solar
!!$!  System barycentric space-time frame of reference.
!!$
!!$!  The proper time is Terrestrial Time TT;  the coordinate
!!$!  time is an implementation of the Barycentric Dynamical Time TDB.
!!$
!!$!  Given:
!!$!    TDB   dp   coordinate time (MJD: JD-2400000.5)
!!$!    UT1   dp   universal time (fraction of one day)
!!$!    WL    dp   clock longitude (radians west)
!!$!    U     dp   clock distance from Earth spin axis (km)
!!$!    V     dp   clock distance north of Earth equatorial plane (km)
!!$
!!$!  Returned:
!!$!    The clock correction, TDB-TT, in seconds.  TDB may be considered
!!$!    to be the coordinate time in the Solar System barycentre frame of
!!$!    reference, and TT is the proper time given by clocks at mean sea
!!$!    level on the Earth.
!!$
!!$!    The result has a main (annual) sinusoidal term of amplitude
!!$!    approximately 0.00166 seconds, plus planetary terms up to about
!!$!    20 microseconds, and lunar and diurnal terms up to 2 microseconds.
!!$!    The variation arises from the transverse Doppler effect and the
!!$!    gravitational red-shift as the observer varies in speed and moves
!!$!    through different gravitational potentials.
!!$
!!$!  The argument TDB is, strictly, the barycentric coordinate time;
!!$!  however, the terrestrial proper time (TT) can in practice be used.
!!$
!!$!  The geocentric model is that of Fairhead & Bretagnon (1990), in its
!!$!  full form.  It was supplied by Fairhead (private communication) as a
!!$!  FORTRAN subroutine.  The original Fairhead routine used explicit
!!$!  formulae, in such large numbers that problems were experienced with
!!$!  certain compilers (Microsoft Fortran on PC aborted with stack
!!$!  overflow, Convex compiled successfully but extremely slowly).  The
!!$!  present implementation is a complete recoding, with the original
!!$!  Fairhead coefficients held in a table.  To optimise arithmetic
!!$!  precision, the terms are accumulated in reverse order, smallest
!!$!  first.  A number of other coding changes were made, in order to match
!!$!  the calling sequence of previous versions of the present routine, and
!!$!  to comply with Starlink programming standards.  Under VAX/VMS, the
!!$!  numerical results compared with those from the Fairhead form are
!!$!  essentially unaffected by the changes, the differences being at the
!!$!  10^-20 sec level.
!!$
!!$!  The topocentric part of the model is from Moyer (1981) and
!!$!  Murray (1983).
!!$
!!$!  During the interval 1950-2050, the absolute accuracy is better
!!$!  than +/- 3 nanoseconds relative to direct numerical integrations
!!$!  using the JPL DE200/LE200 solar system ephemeris.
!!$
!!$!  The IAU definition of TDB is that it must differ from TT only by
!!$!  periodic terms.  Though practical, this is an imprecise definition
!!$!  which ignores the existence of very long-period and secular effects
!!$!  in the dynamics of the solar system.  As a consequence, different
!!$!  implementations of TDB will, in general, differ in zero-point and
!!$!  will drift linearly relative to one other.
!!$
!!$!  References:
!!$!    Bretagnon P, 1982 Astron. Astrophys., 114, 278-288.
!!$!    Fairhead L & Bretagnon P, 1990, Astron. Astrophys., 229, 240-247.
!!$!    Meeus J, 1984, l'Astronomie, 348-354.
!!$!    Moyer T D, 1981, Cel. Mech., 23, 33.
!!$!    Murray C A, 1983, Vectorial Astrometry, Adam Hilger.
!!$
!!$!  P.T.Wallace   Starlink   10 November 1995
!!$
!!$!  Copyright (C) 1995 Rutherford Appleton Laboratory
!!$!-
!!$
!!$IMPLICIT NONE
!!$
!!$DOUBLE PRECISION :: tdb,ut1,wl,u,v
!!$
!!$DOUBLE PRECISION :: d2pi
!!$PARAMETER (d2pi=6.283185307179586476925287D0)
!!$
!!$DOUBLE PRECISION :: t,tsol,rle,rge,rtj,rts,rdmoon,wt,w0,w1,w2,w3,w4,wf,wj
!!$
!!$! -----------------------------------------------------------------------
!!$
!!$!  Fairhead and Bretagnon canonical coefficients
!!$
!!$!  787 sets of three coefficients.
!!$
!!$!  Each set is amplitude (microseconds)
!!$!              frequency (radians per Julian millenium since J2000),
!!$!              phase (radians).
!!$
!!$!  Sets   1-474 are the T**0 terms,
!!$!   "   475-679  "   "  T**1   "
!!$!   "   680-764  "   "  T**2   "
!!$!   "   765-784  "   "  T**3   "
!!$!   "   785-787  "   "  T**4   "  .
!!$
!!$DOUBLE PRECISION :: fairhd(3,787)
!!$INTEGER :: i,j
!!$DATA ((fairhd(i,j),i=1,3),j=  1, 10) /  &
!!$1656.674564D-6,    6283.075849991D0, 6.240054195D0,  &
!!$22.417471D-6,    5753.384884897D0, 4.296977442D0,  &
!!$13.839792D-6,   12566.151699983D0, 6.196904410D0,  &
!!$4.770086D-6,     529.690965095D0, 0.444401603D0,  &
!!$4.676740D-6,    6069.776754553D0, 4.021195093D0,  &
!!$2.256707D-6,     213.299095438D0, 5.543113262D0,  &
!!$1.694205D-6,      -3.523118349D0, 5.025132748D0,  &
!!$1.554905D-6,   77713.771467920D0, 5.198467090D0,  &
!!$1.276839D-6,    7860.419392439D0, 5.988822341D0,  &
!!$1.193379D-6,    5223.693919802D0, 3.649823730D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 11, 20) /  &
!!$1.115322D-6,    3930.209696220D0, 1.422745069D0,  &
!!$0.794185D-6,   11506.769769794D0, 2.322313077D0,  &
!!$0.447061D-6,      26.298319800D0, 3.615796498D0,  &
!!$0.435206D-6,    -398.149003408D0, 4.349338347D0,  &
!!$0.600309D-6,    1577.343542448D0, 2.678271909D0,  &
!!$0.496817D-6,    6208.294251424D0, 5.696701824D0,  &
!!$0.486306D-6,    5884.926846583D0, 0.520007179D0,  &
!!$0.432392D-6,      74.781598567D0, 2.435898309D0,  &
!!$0.468597D-6,    6244.942814354D0, 5.866398759D0,  &
!!$0.375510D-6,    5507.553238667D0, 4.103476804D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 21, 30) /  &
!!$0.243085D-6,    -775.522611324D0, 3.651837925D0,  &
!!$0.173435D-6,   18849.227549974D0, 6.153743485D0,  &
!!$0.230685D-6,    5856.477659115D0, 4.773852582D0,  &
!!$0.203747D-6,   12036.460734888D0, 4.333987818D0,  &
!!$0.143935D-6,    -796.298006816D0, 5.957517795D0,  &
!!$0.159080D-6,   10977.078804699D0, 1.890075226D0,  &
!!$0.119979D-6,      38.133035638D0, 4.551585768D0,  &
!!$0.118971D-6,    5486.777843175D0, 1.914547226D0,  &
!!$0.116120D-6,    1059.381930189D0, 0.873504123D0,  &
!!$0.137927D-6,   11790.629088659D0, 1.135934669D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 31, 40) /  &
!!$0.098358D-6,    2544.314419883D0, 0.092793886D0,  &
!!$0.101868D-6,   -5573.142801634D0, 5.984503847D0,  &
!!$0.080164D-6,     206.185548437D0, 2.095377709D0,  &
!!$0.079645D-6,    4694.002954708D0, 2.949233637D0,  &
!!$0.062617D-6,      20.775395492D0, 2.654394814D0,  &
!!$0.075019D-6,    2942.463423292D0, 4.980931759D0,  &
!!$0.064397D-6,    5746.271337896D0, 1.280308748D0,  &
!!$0.063814D-6,    5760.498431898D0, 4.167901731D0,  &
!!$0.048042D-6,    2146.165416475D0, 1.495846011D0,  &
!!$0.048373D-6,     155.420399434D0, 2.251573730D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 41, 50) /  &
!!$0.058844D-6,     426.598190876D0, 4.839650148D0,  &
!!$0.046551D-6,      -0.980321068D0, 0.921573539D0,  &
!!$0.054139D-6,   17260.154654690D0, 3.411091093D0,  &
!!$0.042411D-6,    6275.962302991D0, 2.869567043D0,  &
!!$0.040184D-6,      -7.113547001D0, 3.565975565D0,  &
!!$0.036564D-6,    5088.628839767D0, 3.324679049D0,  &
!!$0.040759D-6,   12352.852604545D0, 3.981496998D0,  &
!!$0.036507D-6,     801.820931124D0, 6.248866009D0,  &
!!$0.036955D-6,    3154.687084896D0, 5.071801441D0,  &
!!$0.042732D-6,     632.783739313D0, 5.720622217D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 51, 60) /  &
!!$0.042560D-6,  161000.685737473D0, 1.270837679D0,  &
!!$0.040480D-6,   15720.838784878D0, 2.546610123D0,  &
!!$0.028244D-6,   -6286.598968340D0, 5.069663519D0,  &
!!$0.033477D-6,    6062.663207553D0, 4.144987272D0,  &
!!$0.034867D-6,     522.577418094D0, 5.210064075D0,  &
!!$0.032438D-6,    6076.890301554D0, 0.749317412D0,  &
!!$0.030215D-6,    7084.896781115D0, 3.389610345D0,  &
!!$0.029247D-6,  -71430.695617928D0, 4.183178762D0,  &
!!$0.033529D-6,    9437.762934887D0, 2.404714239D0,  &
!!$0.032423D-6,    8827.390269875D0, 5.541473556D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 61, 70) /  &
!!$0.027567D-6,    6279.552731642D0, 5.040846034D0,  &
!!$0.029862D-6,   12139.553509107D0, 1.770181024D0,  &
!!$0.022509D-6,   10447.387839604D0, 1.460726241D0,  &
!!$0.020937D-6,    8429.241266467D0, 0.652303414D0,  &
!!$0.020322D-6,     419.484643875D0, 3.735430632D0,  &
!!$0.024816D-6,   -1194.447010225D0, 1.087136918D0,  &
!!$0.025196D-6,    1748.016413067D0, 2.901883301D0,  &
!!$0.021691D-6,   14143.495242431D0, 5.952658009D0,  &
!!$0.017673D-6,    6812.766815086D0, 3.186129845D0,  &
!!$0.022567D-6,    6133.512652857D0, 3.307984806D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 71, 80) /  &
!!$0.016155D-6,   10213.285546211D0, 1.331103168D0,  &
!!$0.014751D-6,    1349.867409659D0, 4.308933301D0,  &
!!$0.015949D-6,    -220.412642439D0, 4.005298270D0,  &
!!$0.015974D-6,   -2352.866153772D0, 6.145309371D0,  &
!!$0.014223D-6,   17789.845619785D0, 2.104551349D0,  &
!!$0.017806D-6,      73.297125859D0, 3.475975097D0,  &
!!$0.013671D-6,    -536.804512095D0, 5.971672571D0,  &
!!$0.011942D-6,    8031.092263058D0, 2.053414715D0,  &
!!$0.014318D-6,   16730.463689596D0, 3.016058075D0,  &
!!$0.012462D-6,     103.092774219D0, 1.737438797D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 81, 90) /  &
!!$0.010962D-6,       3.590428652D0, 2.196567739D0,  &
!!$0.015078D-6,   19651.048481098D0, 3.969480770D0,  &
!!$0.010396D-6,     951.718406251D0, 5.717799605D0,  &
!!$0.011707D-6,   -4705.732307544D0, 2.654125618D0,  &
!!$0.010453D-6,    5863.591206116D0, 1.913704550D0,  &
!!$0.012420D-6,    4690.479836359D0, 4.734090399D0,  &
!!$0.011847D-6,    5643.178563677D0, 5.489005403D0,  &
!!$0.008610D-6,    3340.612426700D0, 3.661698944D0,  &
!!$0.011622D-6,    5120.601145584D0, 4.863931876D0,  &
!!$0.010825D-6,     553.569402842D0, 0.842715011D0 /
!!$DATA ((fairhd(i,j),i=1,3),j= 91,100) /  &
!!$0.008666D-6,    -135.065080035D0, 3.293406547D0,  &
!!$0.009963D-6,     149.563197135D0, 4.870690598D0,  &
!!$0.009858D-6,    6309.374169791D0, 1.061816410D0,  &
!!$0.007959D-6,     316.391869657D0, 2.465042647D0,  &
!!$0.010099D-6,     283.859318865D0, 1.942176992D0,  &
!!$0.007147D-6,    -242.728603974D0, 3.661486981D0,  &
!!$0.007505D-6,    5230.807466803D0, 4.920937029D0,  &
!!$0.008323D-6,   11769.853693166D0, 1.229392026D0,  &
!!$0.007490D-6,   -6256.777530192D0, 3.658444681D0,  &
!!$0.009370D-6,  149854.400134205D0, 0.673880395D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=101,110) /  &
!!$0.007117D-6,      38.027672636D0, 5.294249518D0,  &
!!$0.007857D-6,   12168.002696575D0, 0.525733528D0,  &
!!$0.007019D-6,    6206.809778716D0, 0.837688810D0,  &
!!$0.006056D-6,     955.599741609D0, 4.194535082D0,  &
!!$0.008107D-6,   13367.972631107D0, 3.793235253D0,  &
!!$0.006731D-6,    5650.292110678D0, 5.639906583D0,  &
!!$0.007332D-6,      36.648562930D0, 0.114858677D0,  &
!!$0.006366D-6,    4164.311989613D0, 2.262081818D0,  &
!!$0.006858D-6,    5216.580372801D0, 0.642063318D0,  &
!!$0.006919D-6,    6681.224853400D0, 6.018501522D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=111,120) /  &
!!$0.006826D-6,    7632.943259650D0, 3.458654112D0,  &
!!$0.005308D-6,   -1592.596013633D0, 2.500382359D0,  &
!!$0.005096D-6,   11371.704689758D0, 2.547107806D0,  &
!!$0.004841D-6,    5333.900241022D0, 0.437078094D0,  &
!!$0.005582D-6,    5966.683980335D0, 2.246174308D0,  &
!!$0.006304D-6,   11926.254413669D0, 2.512929171D0,  &
!!$0.006603D-6,   23581.258177318D0, 5.393136889D0,  &
!!$0.005123D-6,      -1.484472708D0, 2.999641028D0,  &
!!$0.004648D-6,    1589.072895284D0, 1.275847090D0,  &
!!$0.005119D-6,    6438.496249426D0, 1.486539246D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=121,130) /  &
!!$0.004521D-6,    4292.330832950D0, 6.140635794D0,  &
!!$0.005680D-6,   23013.539539587D0, 4.557814849D0,  &
!!$0.005488D-6,      -3.455808046D0, 0.090675389D0,  &
!!$0.004193D-6,    7234.794256242D0, 4.869091389D0,  &
!!$0.003742D-6,    7238.675591600D0, 4.691976180D0,  &
!!$0.004148D-6,    -110.206321219D0, 3.016173439D0,  &
!!$0.004553D-6,   11499.656222793D0, 5.554998314D0,  &
!!$0.004892D-6,    5436.993015240D0, 1.475415597D0,  &
!!$0.004044D-6,    4732.030627343D0, 1.398784824D0,  &
!!$0.004164D-6,   12491.370101415D0, 5.650931916D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=131,140) /  &
!!$0.004349D-6,   11513.883316794D0, 2.181745369D0,  &
!!$0.003919D-6,   12528.018664345D0, 5.823319737D0,  &
!!$0.003129D-6,    6836.645252834D0, 0.003844094D0,  &
!!$0.004080D-6,   -7058.598461315D0, 3.690360123D0,  &
!!$0.003270D-6,      76.266071276D0, 1.517189902D0,  &
!!$0.002954D-6,    6283.143160294D0, 4.447203799D0,  &
!!$0.002872D-6,      28.449187468D0, 1.158692983D0,  &
!!$0.002881D-6,     735.876513532D0, 0.349250250D0,  &
!!$0.003279D-6,    5849.364112115D0, 4.893384368D0,  &
!!$0.003625D-6,    6209.778724132D0, 1.473760578D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=141,150) /  &
!!$0.003074D-6,     949.175608970D0, 5.185878737D0,  &
!!$0.002775D-6,    9917.696874510D0, 1.030026325D0,  &
!!$0.002646D-6,   10973.555686350D0, 3.918259169D0,  &
!!$0.002575D-6,   25132.303399966D0, 6.109659023D0,  &
!!$0.003500D-6,     263.083923373D0, 1.892100742D0,  &
!!$0.002740D-6,   18319.536584880D0, 4.320519510D0,  &
!!$0.002464D-6,     202.253395174D0, 4.698203059D0,  &
!!$0.002409D-6,       2.542797281D0, 5.325009315D0,  &
!!$0.003354D-6,  -90955.551694697D0, 1.942656623D0,  &
!!$0.002296D-6,    6496.374945429D0, 5.061810696D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=151,160) /  &
!!$0.003002D-6,    6172.869528772D0, 2.797822767D0,  &
!!$0.003202D-6,   27511.467873537D0, 0.531673101D0,  &
!!$0.002954D-6,   -6283.008539689D0, 4.533471191D0,  &
!!$0.002353D-6,     639.897286314D0, 3.734548088D0,  &
!!$0.002401D-6,   16200.772724501D0, 2.605547070D0,  &
!!$0.003053D-6,  233141.314403759D0, 3.029030662D0,  &
!!$0.003024D-6,   83286.914269554D0, 2.355556099D0,  &
!!$0.002863D-6,   17298.182327326D0, 5.240963796D0,  &
!!$0.002103D-6,   -7079.373856808D0, 5.756641637D0,  &
!!$0.002303D-6,   83996.847317911D0, 2.013686814D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=161,170) /  &
!!$0.002303D-6,   18073.704938650D0, 1.089100410D0,  &
!!$0.002381D-6,      63.735898303D0, 0.759188178D0,  &
!!$0.002493D-6,    6386.168624210D0, 0.645026535D0,  &
!!$0.002366D-6,       3.932153263D0, 6.215885448D0,  &
!!$0.002169D-6,   11015.106477335D0, 4.845297676D0,  &
!!$0.002397D-6,    6243.458341645D0, 3.809290043D0,  &
!!$0.002183D-6,    1162.474704408D0, 6.179611691D0,  &
!!$0.002353D-6,    6246.427287062D0, 4.781719760D0,  &
!!$0.002199D-6,    -245.831646229D0, 5.956152284D0,  &
!!$0.001729D-6,    3894.181829542D0, 1.264976635D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=171,180) /  &
!!$0.001896D-6,   -3128.388765096D0, 4.914231596D0,  &
!!$0.002085D-6,      35.164090221D0, 1.405158503D0,  &
!!$0.002024D-6,   14712.317116458D0, 2.752035928D0,  &
!!$0.001737D-6,    6290.189396992D0, 5.280820144D0,  &
!!$0.002229D-6,     491.557929457D0, 1.571007057D0,  &
!!$0.001602D-6,   14314.168113050D0, 4.203664806D0,  &
!!$0.002186D-6,     454.909366527D0, 1.402101526D0,  &
!!$0.001897D-6,   22483.848574493D0, 4.167932508D0,  &
!!$0.001825D-6,   -3738.761430108D0, 0.545828785D0,  &
!!$0.001894D-6,    1052.268383188D0, 5.817167450D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=181,190) /  &
!!$0.001421D-6,      20.355319399D0, 2.419886601D0,  &
!!$0.001408D-6,   10984.192351700D0, 2.732084787D0,  &
!!$0.001847D-6,   10873.986030480D0, 2.903477885D0,  &
!!$0.001391D-6,   -8635.942003763D0, 0.593891500D0,  &
!!$0.001388D-6,      -7.046236698D0, 1.166145902D0,  &
!!$0.001810D-6,  -88860.057071188D0, 0.487355242D0,  &
!!$0.001288D-6,   -1990.745017041D0, 3.913022880D0,  &
!!$0.001297D-6,   23543.230504682D0, 3.063805171D0,  &
!!$0.001335D-6,    -266.607041722D0, 3.995764039D0,  &
!!$0.001376D-6,   10969.965257698D0, 5.152914309D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=191,200) /  &
!!$0.001745D-6,  244287.600007027D0, 3.626395673D0,  &
!!$0.001649D-6,   31441.677569757D0, 1.952049260D0,  &
!!$0.001416D-6,    9225.539273283D0, 4.996408389D0,  &
!!$0.001238D-6,    4804.209275927D0, 5.503379738D0,  &
!!$0.001472D-6,    4590.910180489D0, 4.164913291D0,  &
!!$0.001169D-6,    6040.347246017D0, 5.841719038D0,  &
!!$0.001039D-6,    5540.085789459D0, 2.769753519D0,  &
!!$0.001004D-6,    -170.672870619D0, 0.755008103D0,  &
!!$0.001284D-6,   10575.406682942D0, 5.306538209D0,  &
!!$0.001278D-6,      71.812653151D0, 4.713486491D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=201,210) /  &
!!$0.001321D-6,   18209.330263660D0, 2.624866359D0,  &
!!$0.001297D-6,   21228.392023546D0, 0.382603541D0,  &
!!$0.000954D-6,    6282.095528923D0, 0.882213514D0,  &
!!$0.001145D-6,    6058.731054289D0, 1.169483931D0,  &
!!$0.000979D-6,    5547.199336460D0, 5.448375984D0,  &
!!$0.000987D-6,   -6262.300454499D0, 2.656486959D0,  &
!!$0.001070D-6, -154717.609887482D0, 1.827624012D0,  &
!!$0.000991D-6,    4701.116501708D0, 4.387001801D0,  &
!!$0.001155D-6,     -14.227094002D0, 3.042700750D0,  &
!!$0.001176D-6,     277.034993741D0, 3.335519004D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=211,220) /  &
!!$0.000890D-6,   13916.019109642D0, 5.601498297D0,  &
!!$0.000884D-6,   -1551.045222648D0, 1.088831705D0,  &
!!$0.000876D-6,    5017.508371365D0, 3.969902609D0,  &
!!$0.000806D-6,   15110.466119866D0, 5.142876744D0,  &
!!$0.000773D-6,   -4136.910433516D0, 0.022067765D0,  &
!!$0.001077D-6,     175.166059800D0, 1.844913056D0,  &
!!$0.000954D-6,   -6284.056171060D0, 0.968480906D0,  &
!!$0.000737D-6,    5326.786694021D0, 4.923831588D0,  &
!!$0.000845D-6,    -433.711737877D0, 4.749245231D0,  &
!!$0.000819D-6,    8662.240323563D0, 5.991247817D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=221,230) /  &
!!$0.000852D-6,     199.072001436D0, 2.189604979D0,  &
!!$0.000723D-6,   17256.631536341D0, 6.068719637D0,  &
!!$0.000940D-6,    6037.244203762D0, 6.197428148D0,  &
!!$0.000885D-6,   11712.955318231D0, 3.280414875D0,  &
!!$0.000706D-6,   12559.038152982D0, 2.824848947D0,  &
!!$0.000732D-6,    2379.164473572D0, 2.501813417D0,  &
!!$0.000764D-6,   -6127.655450557D0, 2.236346329D0,  &
!!$0.000908D-6,     131.541961686D0, 2.521257490D0,  &
!!$0.000907D-6,   35371.887265976D0, 3.370195967D0,  &
!!$0.000673D-6,    1066.495477190D0, 3.876512374D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=231,240) /  &
!!$0.000814D-6,   17654.780539750D0, 4.627122566D0,  &
!!$0.000630D-6,      36.027866677D0, 0.156368499D0,  &
!!$0.000798D-6,     515.463871093D0, 5.151962502D0,  &
!!$0.000798D-6,     148.078724426D0, 5.909225055D0,  &
!!$0.000806D-6,     309.278322656D0, 6.054064447D0,  &
!!$0.000607D-6,     -39.617508346D0, 2.839021623D0,  &
!!$0.000601D-6,     412.371096874D0, 3.984225404D0,  &
!!$0.000646D-6,   11403.676995575D0, 3.852959484D0,  &
!!$0.000704D-6,   13521.751441591D0, 2.300991267D0,  &
!!$0.000603D-6,  -65147.619767937D0, 4.140083146D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=241,250) /  &
!!$0.000609D-6,   10177.257679534D0, 0.437122327D0,  &
!!$0.000631D-6,    5767.611978898D0, 4.026532329D0,  &
!!$0.000576D-6,   11087.285125918D0, 4.760293101D0,  &
!!$0.000674D-6,   14945.316173554D0, 6.270510511D0,  &
!!$0.000726D-6,    5429.879468239D0, 6.039606892D0,  &
!!$0.000710D-6,   28766.924424484D0, 5.672617711D0,  &
!!$0.000647D-6,   11856.218651625D0, 3.397132627D0,  &
!!$0.000678D-6,   -5481.254918868D0, 6.249666675D0,  &
!!$0.000618D-6,   22003.914634870D0, 2.466427018D0,  &
!!$0.000738D-6,    6134.997125565D0, 2.242668890D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=251,260) /  &
!!$0.000660D-6,     625.670192312D0, 5.864091907D0,  &
!!$0.000694D-6,    3496.032826134D0, 2.668309141D0,  &
!!$0.000531D-6,    6489.261398429D0, 1.681888780D0,  &
!!$0.000611D-6, -143571.324284214D0, 2.424978312D0,  &
!!$0.000575D-6,   12043.574281889D0, 4.216492400D0,  &
!!$0.000553D-6,   12416.588502848D0, 4.772158039D0,  &
!!$0.000689D-6,    4686.889407707D0, 6.224271088D0,  &
!!$0.000495D-6,    7342.457780181D0, 3.817285811D0,  &
!!$0.000567D-6,    3634.621024518D0, 1.649264690D0,  &
!!$0.000515D-6,   18635.928454536D0, 3.945345892D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=261,270) /  &
!!$0.000486D-6,    -323.505416657D0, 4.061673868D0,  &
!!$0.000662D-6,   25158.601719765D0, 1.794058369D0,  &
!!$0.000509D-6,     846.082834751D0, 3.053874588D0,  &
!!$0.000472D-6,  -12569.674818332D0, 5.112133338D0,  &
!!$0.000461D-6,    6179.983075773D0, 0.513669325D0,  &
!!$0.000641D-6,   83467.156352816D0, 3.210727723D0,  &
!!$0.000520D-6,   10344.295065386D0, 2.445597761D0,  &
!!$0.000493D-6,   18422.629359098D0, 1.676939306D0,  &
!!$0.000478D-6,    1265.567478626D0, 5.487314569D0,  &
!!$0.000472D-6,     -18.159247265D0, 1.999707589D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=271,280) /  &
!!$0.000559D-6,   11190.377900137D0, 5.783236356D0,  &
!!$0.000494D-6,    9623.688276691D0, 3.022645053D0,  &
!!$0.000463D-6,    5739.157790895D0, 1.411223013D0,  &
!!$0.000432D-6,   16858.482532933D0, 1.179256434D0,  &
!!$0.000574D-6,   72140.628666286D0, 1.758191830D0,  &
!!$0.000484D-6,   17267.268201691D0, 3.290589143D0,  &
!!$0.000550D-6,    4907.302050146D0, 0.864024298D0,  &
!!$0.000399D-6,      14.977853527D0, 2.094441910D0,  &
!!$0.000491D-6,     224.344795702D0, 0.878372791D0,  &
!!$0.000432D-6,   20426.571092422D0, 6.003829241D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=281,290) /  &
!!$0.000481D-6,    5749.452731634D0, 4.309591964D0,  &
!!$0.000480D-6,    5757.317038160D0, 1.142348571D0,  &
!!$0.000485D-6,    6702.560493867D0, 0.210580917D0,  &
!!$0.000426D-6,    6055.549660552D0, 4.274476529D0,  &
!!$0.000480D-6,    5959.570433334D0, 5.031351030D0,  &
!!$0.000466D-6,   12562.628581634D0, 4.959581597D0,  &
!!$0.000520D-6,   39302.096962196D0, 4.788002889D0,  &
!!$0.000458D-6,   12132.439962106D0, 1.880103788D0,  &
!!$0.000470D-6,   12029.347187887D0, 1.405611197D0,  &
!!$0.000416D-6,   -7477.522860216D0, 1.082356330D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=291,300) /  &
!!$0.000449D-6,   11609.862544012D0, 4.179989585D0,  &
!!$0.000465D-6,   17253.041107690D0, 0.353496295D0,  &
!!$0.000362D-6,   -4535.059436924D0, 1.583849576D0,  &
!!$0.000383D-6,   21954.157609398D0, 3.747376371D0,  &
!!$0.000389D-6,      17.252277143D0, 1.395753179D0,  &
!!$0.000331D-6,   18052.929543158D0, 0.566790582D0,  &
!!$0.000430D-6,   13517.870106233D0, 0.685827538D0,  &
!!$0.000368D-6,   -5756.908003246D0, 0.731374317D0,  &
!!$0.000330D-6,   10557.594160824D0, 3.710043680D0,  &
!!$0.000332D-6,   20199.094959633D0, 1.652901407D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=301,310) /  &
!!$0.000384D-6,   11933.367960670D0, 5.827781531D0,  &
!!$0.000387D-6,   10454.501386605D0, 2.541182564D0,  &
!!$0.000325D-6,   15671.081759407D0, 2.178850542D0,  &
!!$0.000318D-6,     138.517496871D0, 2.253253037D0,  &
!!$0.000305D-6,    9388.005909415D0, 0.578340206D0,  &
!!$0.000352D-6,    5749.861766548D0, 3.000297967D0,  &
!!$0.000311D-6,    6915.859589305D0, 1.693574249D0,  &
!!$0.000297D-6,   24072.921469776D0, 1.997249392D0,  &
!!$0.000363D-6,    -640.877607382D0, 5.071820966D0,  &
!!$0.000323D-6,   12592.450019783D0, 1.072262823D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=311,320) /  &
!!$0.000341D-6,   12146.667056108D0, 4.700657997D0,  &
!!$0.000290D-6,    9779.108676125D0, 1.812320441D0,  &
!!$0.000342D-6,    6132.028180148D0, 4.322238614D0,  &
!!$0.000329D-6,    6268.848755990D0, 3.033827743D0,  &
!!$0.000374D-6,   17996.031168222D0, 3.388716544D0,  &
!!$0.000285D-6,    -533.214083444D0, 4.687313233D0,  &
!!$0.000338D-6,    6065.844601290D0, 0.877776108D0,  &
!!$0.000276D-6,      24.298513841D0, 0.770299429D0,  &
!!$0.000336D-6,   -2388.894020449D0, 5.353796034D0,  &
!!$0.000290D-6,    3097.883822726D0, 4.075291557D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=321,330) /  &
!!$0.000318D-6,     709.933048357D0, 5.941207518D0,  &
!!$0.000271D-6,   13095.842665077D0, 3.208912203D0,  &
!!$0.000331D-6,    6073.708907816D0, 4.007881169D0,  &
!!$0.000292D-6,     742.990060533D0, 2.714333592D0,  &
!!$0.000362D-6,   29088.811415985D0, 3.215977013D0,  &
!!$0.000280D-6,   12359.966151546D0, 0.710872502D0,  &
!!$0.000267D-6,   10440.274292604D0, 4.730108488D0,  &
!!$0.000262D-6,     838.969287750D0, 1.327720272D0,  &
!!$0.000250D-6,   16496.361396202D0, 0.898769761D0,  &
!!$0.000325D-6,   20597.243963041D0, 0.180044365D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=331,340) /  &
!!$0.000268D-6,    6148.010769956D0, 5.152666276D0,  &
!!$0.000284D-6,    5636.065016677D0, 5.655385808D0,  &
!!$0.000301D-6,    6080.822454817D0, 2.135396205D0,  &
!!$0.000294D-6,    -377.373607916D0, 3.708784168D0,  &
!!$0.000236D-6,    2118.763860378D0, 1.733578756D0,  &
!!$0.000234D-6,    5867.523359379D0, 5.575209112D0,  &
!!$0.000268D-6, -226858.238553767D0, 0.069432392D0,  &
!!$0.000265D-6,  167283.761587465D0, 4.369302826D0,  &
!!$0.000280D-6,   28237.233459389D0, 5.304829118D0,  &
!!$0.000292D-6,   12345.739057544D0, 4.096094132D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=341,350) /  &
!!$0.000223D-6,   19800.945956225D0, 3.069327406D0,  &
!!$0.000301D-6,   43232.306658416D0, 6.205311188D0,  &
!!$0.000264D-6,   18875.525869774D0, 1.417263408D0,  &
!!$0.000304D-6,   -1823.175188677D0, 3.409035232D0,  &
!!$0.000301D-6,     109.945688789D0, 0.510922054D0,  &
!!$0.000260D-6,     813.550283960D0, 2.389438934D0,  &
!!$0.000299D-6,  316428.228673312D0, 5.384595078D0,  &
!!$0.000211D-6,    5756.566278634D0, 3.789392838D0,  &
!!$0.000209D-6,    5750.203491159D0, 1.661943545D0,  &
!!$0.000240D-6,   12489.885628707D0, 5.684549045D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=351,360) /  &
!!$0.000216D-6,    6303.851245484D0, 3.862942261D0,  &
!!$0.000203D-6,    1581.959348283D0, 5.549853589D0,  &
!!$0.000200D-6,    5642.198242609D0, 1.016115785D0,  &
!!$0.000197D-6,     -70.849445304D0, 4.690702525D0,  &
!!$0.000227D-6,    6287.008003254D0, 2.911891613D0,  &
!!$0.000197D-6,     533.623118358D0, 1.048982898D0,  &
!!$0.000205D-6,   -6279.485421340D0, 1.829362730D0,  &
!!$0.000209D-6,  -10988.808157535D0, 2.636140084D0,  &
!!$0.000208D-6,    -227.526189440D0, 4.127883842D0,  &
!!$0.000191D-6,     415.552490612D0, 4.401165650D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=361,370) /  &
!!$0.000190D-6,   29296.615389579D0, 4.175658539D0,  &
!!$0.000264D-6,   66567.485864652D0, 4.601102551D0,  &
!!$0.000256D-6,   -3646.350377354D0, 0.506364778D0,  &
!!$0.000188D-6,   13119.721102825D0, 2.032195842D0,  &
!!$0.000185D-6,    -209.366942175D0, 4.694756586D0,  &
!!$0.000198D-6,   25934.124331089D0, 3.832703118D0,  &
!!$0.000195D-6,    4061.219215394D0, 3.308463427D0,  &
!!$0.000234D-6,    5113.487598583D0, 1.716090661D0,  &
!!$0.000188D-6,    1478.866574064D0, 5.686865780D0,  &
!!$0.000222D-6,   11823.161639450D0, 1.942386641D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=371,380) /  &
!!$0.000181D-6,   10770.893256262D0, 1.999482059D0,  &
!!$0.000171D-6,    6546.159773364D0, 1.182807992D0,  &
!!$0.000206D-6,      70.328180442D0, 5.934076062D0,  &
!!$0.000169D-6,   20995.392966449D0, 2.169080622D0,  &
!!$0.000191D-6,   10660.686935042D0, 5.405515999D0,  &
!!$0.000228D-6,   33019.021112205D0, 4.656985514D0,  &
!!$0.000184D-6,   -4933.208440333D0, 3.327476868D0,  &
!!$0.000220D-6,    -135.625325010D0, 1.765430262D0,  &
!!$0.000166D-6,   23141.558382925D0, 3.454132746D0,  &
!!$0.000191D-6,    6144.558353121D0, 5.020393445D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=381,390) /  &
!!$0.000180D-6,    6084.003848555D0, 0.602182191D0,  &
!!$0.000163D-6,   17782.732072784D0, 4.960593133D0,  &
!!$0.000225D-6,   16460.333529525D0, 2.596451817D0,  &
!!$0.000222D-6,    5905.702242076D0, 3.731990323D0,  &
!!$0.000204D-6,     227.476132789D0, 5.636192701D0,  &
!!$0.000159D-6,   16737.577236597D0, 3.600691544D0,  &
!!$0.000200D-6,    6805.653268085D0, 0.868220961D0,  &
!!$0.000187D-6,   11919.140866668D0, 2.629456641D0,  &
!!$0.000161D-6,     127.471796607D0, 2.862574720D0,  &
!!$0.000205D-6,    6286.666278643D0, 1.742882331D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=391,400) /  &
!!$0.000189D-6,     153.778810485D0, 4.812372643D0,  &
!!$0.000168D-6,   16723.350142595D0, 0.027860588D0,  &
!!$0.000149D-6,   11720.068865232D0, 0.659721876D0,  &
!!$0.000189D-6,    5237.921013804D0, 5.245313000D0,  &
!!$0.000143D-6,    6709.674040867D0, 4.317625647D0,  &
!!$0.000146D-6,    4487.817406270D0, 4.815297007D0,  &
!!$0.000144D-6,    -664.756045130D0, 5.381366880D0,  &
!!$0.000175D-6,    5127.714692584D0, 4.728443327D0,  &
!!$0.000162D-6,    6254.626662524D0, 1.435132069D0,  &
!!$0.000187D-6,   47162.516354635D0, 1.354371923D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=401,410) /  &
!!$0.000146D-6,   11080.171578918D0, 3.369695406D0,  &
!!$0.000180D-6,    -348.924420448D0, 2.490902145D0,  &
!!$0.000148D-6,     151.047669843D0, 3.799109588D0,  &
!!$0.000157D-6,    6197.248551160D0, 1.284375887D0,  &
!!$0.000167D-6,     146.594251718D0, 0.759969109D0,  &
!!$0.000133D-6,   -5331.357443741D0, 5.409701889D0,  &
!!$0.000154D-6,      95.979227218D0, 3.366890614D0,  &
!!$0.000148D-6,   -6418.140930027D0, 3.384104996D0,  &
!!$0.000128D-6,   -6525.804453965D0, 3.803419985D0,  &
!!$0.000130D-6,   11293.470674356D0, 0.939039445D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=411,420) /  &
!!$0.000152D-6,   -5729.506447149D0, 0.734117523D0,  &
!!$0.000138D-6,     210.117701700D0, 2.564216078D0,  &
!!$0.000123D-6,    6066.595360816D0, 4.517099537D0,  &
!!$0.000140D-6,   18451.078546566D0, 0.642049130D0,  &
!!$0.000126D-6,   11300.584221356D0, 3.485280663D0,  &
!!$0.000119D-6,   10027.903195729D0, 3.217431161D0,  &
!!$0.000151D-6,    4274.518310832D0, 4.404359108D0,  &
!!$0.000117D-6,    6072.958148291D0, 0.366324650D0,  &
!!$0.000165D-6,   -7668.637425143D0, 4.298212528D0,  &
!!$0.000117D-6,   -6245.048177356D0, 5.379518958D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=421,430) /  &
!!$0.000130D-6,   -5888.449964932D0, 4.527681115D0,  &
!!$0.000121D-6,    -543.918059096D0, 6.109429504D0,  &
!!$0.000162D-6,    9683.594581116D0, 5.720092446D0,  &
!!$0.000141D-6,    6219.339951688D0, 0.679068671D0,  &
!!$0.000118D-6,   22743.409379516D0, 4.881123092D0,  &
!!$0.000129D-6,    1692.165669502D0, 0.351407289D0,  &
!!$0.000126D-6,    5657.405657679D0, 5.146592349D0,  &
!!$0.000114D-6,     728.762966531D0, 0.520791814D0,  &
!!$0.000120D-6,      52.596639600D0, 0.948516300D0,  &
!!$0.000115D-6,      65.220371012D0, 3.504914846D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=431,440) /  &
!!$0.000126D-6,    5881.403728234D0, 5.577502482D0,  &
!!$0.000158D-6,  163096.180360983D0, 2.957128968D0,  &
!!$0.000134D-6,   12341.806904281D0, 2.598576764D0,  &
!!$0.000151D-6,   16627.370915377D0, 3.985702050D0,  &
!!$0.000109D-6,    1368.660252845D0, 0.014730471D0,  &
!!$0.000131D-6,    6211.263196841D0, 0.085077024D0,  &
!!$0.000146D-6,    5792.741760812D0, 0.708426604D0,  &
!!$0.000146D-6,     -77.750543984D0, 3.121576600D0,  &
!!$0.000107D-6,    5341.013788022D0, 0.288231904D0,  &
!!$0.000138D-6,    6281.591377283D0, 2.797450317D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=441,450) /  &
!!$0.000113D-6,   -6277.552925684D0, 2.788904128D0,  &
!!$0.000115D-6,    -525.758811831D0, 5.895222200D0,  &
!!$0.000138D-6,    6016.468808270D0, 6.096188999D0,  &
!!$0.000139D-6,   23539.707386333D0, 2.028195445D0,  &
!!$0.000146D-6,   -4176.041342449D0, 4.660008502D0,  &
!!$0.000107D-6,   16062.184526117D0, 4.066520001D0,  &
!!$0.000142D-6,   83783.548222473D0, 2.936315115D0,  &
!!$0.000128D-6,    9380.959672717D0, 3.223844306D0,  &
!!$0.000135D-6,    6205.325306007D0, 1.638054048D0,  &
!!$0.000101D-6,    2699.734819318D0, 5.481603249D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=451,460) /  &
!!$0.000104D-6,    -568.821874027D0, 2.205734493D0,  &
!!$0.000103D-6,    6321.103522627D0, 2.440421099D0,  &
!!$0.000119D-6,    6321.208885629D0, 2.547496264D0,  &
!!$0.000138D-6,    1975.492545856D0, 2.314608466D0,  &
!!$0.000121D-6,     137.033024162D0, 4.539108237D0,  &
!!$0.000123D-6,   19402.796952817D0, 4.538074405D0,  &
!!$0.000119D-6,   22805.735565994D0, 2.869040566D0,  &
!!$0.000133D-6,   64471.991241142D0, 6.056405489D0,  &
!!$0.000129D-6,     -85.827298831D0, 2.540635083D0,  &
!!$0.000131D-6,   13613.804277336D0, 4.005732868D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=461,470) /  &
!!$0.000104D-6,    9814.604100291D0, 1.959967212D0,  &
!!$0.000112D-6,   16097.679950283D0, 3.589026260D0,  &
!!$0.000123D-6,    2107.034507542D0, 1.728627253D0,  &
!!$0.000121D-6,   36949.230808424D0, 6.072332087D0,  &
!!$0.000108D-6,  -12539.853380183D0, 3.716133846D0,  &
!!$0.000113D-6,   -7875.671863624D0, 2.725771122D0,  &
!!$0.000109D-6,    4171.425536614D0, 4.033338079D0,  &
!!$0.000101D-6,    6247.911759770D0, 3.441347021D0,  &
!!$0.000113D-6,    7330.728427345D0, 0.656372122D0,  &
!!$0.000113D-6,   51092.726050855D0, 2.791483066D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=471,480) /  &
!!$0.000106D-6,    5621.842923210D0, 1.815323326D0,  &
!!$0.000101D-6,     111.430161497D0, 5.711033677D0,  &
!!$0.000103D-6,     909.818733055D0, 2.812745443D0,  &
!!$0.000101D-6,    1790.642637886D0, 1.965746028D0,  &
!!$102.156724D-6,    6283.075849991D0, 4.249032005D0,  &
!!$1.706807D-6,   12566.151699983D0, 4.205904248D0,  &
!!$0.269668D-6,     213.299095438D0, 3.400290479D0,  &
!!$0.265919D-6,     529.690965095D0, 5.836047367D0,  &
!!$0.210568D-6,      -3.523118349D0, 6.262738348D0,  &
!!$0.077996D-6,    5223.693919802D0, 4.670344204D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=481,490) /  &
!!$0.054764D-6,    1577.343542448D0, 4.534800170D0,  &
!!$0.059146D-6,      26.298319800D0, 1.083044735D0,  &
!!$0.034420D-6,    -398.149003408D0, 5.980077351D0,  &
!!$0.032088D-6,   18849.227549974D0, 4.162913471D0,  &
!!$0.033595D-6,    5507.553238667D0, 5.980162321D0,  &
!!$0.029198D-6,    5856.477659115D0, 0.623811863D0,  &
!!$0.027764D-6,     155.420399434D0, 3.745318113D0,  &
!!$0.025190D-6,    5746.271337896D0, 2.980330535D0,  &
!!$0.022997D-6,    -796.298006816D0, 1.174411803D0,  &
!!$0.024976D-6,    5760.498431898D0, 2.467913690D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=491,500) /  &
!!$0.021774D-6,     206.185548437D0, 3.854787540D0,  &
!!$0.017925D-6,    -775.522611324D0, 1.092065955D0,  &
!!$0.013794D-6,     426.598190876D0, 2.699831988D0,  &
!!$0.013276D-6,    6062.663207553D0, 5.845801920D0,  &
!!$0.011774D-6,   12036.460734888D0, 2.292832062D0,  &
!!$0.012869D-6,    6076.890301554D0, 5.333425680D0,  &
!!$0.012152D-6,    1059.381930189D0, 6.222874454D0,  &
!!$0.011081D-6,      -7.113547001D0, 5.154724984D0,  &
!!$0.010143D-6,    4694.002954708D0, 4.044013795D0,  &
!!$0.009357D-6,    5486.777843175D0, 3.416081409D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=501,510) /  &
!!$0.010084D-6,     522.577418094D0, 0.749320262D0,  &
!!$0.008587D-6,   10977.078804699D0, 2.777152598D0,  &
!!$0.008628D-6,    6275.962302991D0, 4.562060226D0,  &
!!$0.008158D-6,    -220.412642439D0, 5.806891533D0,  &
!!$0.007746D-6,    2544.314419883D0, 1.603197066D0,  &
!!$0.007670D-6,    2146.165416475D0, 3.000200440D0,  &
!!$0.007098D-6,      74.781598567D0, 0.443725817D0,  &
!!$0.006180D-6,    -536.804512095D0, 1.302642751D0,  &
!!$0.005818D-6,    5088.628839767D0, 4.827723531D0,  &
!!$0.004945D-6,   -6286.598968340D0, 0.268305170D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=511,520) /  &
!!$0.004774D-6,    1349.867409659D0, 5.808636673D0,  &
!!$0.004687D-6,    -242.728603974D0, 5.154890570D0,  &
!!$0.006089D-6,    1748.016413067D0, 4.403765209D0,  &
!!$0.005975D-6,   -1194.447010225D0, 2.583472591D0,  &
!!$0.004229D-6,     951.718406251D0, 0.931172179D0,  &
!!$0.005264D-6,     553.569402842D0, 2.336107252D0,  &
!!$0.003049D-6,    5643.178563677D0, 1.362634430D0,  &
!!$0.002974D-6,    6812.766815086D0, 1.583012668D0,  &
!!$0.003403D-6,   -2352.866153772D0, 2.552189886D0,  &
!!$0.003030D-6,     419.484643875D0, 5.286473844D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=521,530) /  &
!!$0.003210D-6,      -7.046236698D0, 1.863796539D0,  &
!!$0.003058D-6,    9437.762934887D0, 4.226420633D0,  &
!!$0.002589D-6,   12352.852604545D0, 1.991935820D0,  &
!!$0.002927D-6,    5216.580372801D0, 2.319951253D0,  &
!!$0.002425D-6,    5230.807466803D0, 3.084752833D0,  &
!!$0.002656D-6,    3154.687084896D0, 2.487447866D0,  &
!!$0.002445D-6,   10447.387839604D0, 2.347139160D0,  &
!!$0.002990D-6,    4690.479836359D0, 6.235872050D0,  &
!!$0.002890D-6,    5863.591206116D0, 0.095197563D0,  &
!!$0.002498D-6,    6438.496249426D0, 2.994779800D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=531,540) /  &
!!$0.001889D-6,    8031.092263058D0, 3.569003717D0,  &
!!$0.002567D-6,     801.820931124D0, 3.425611498D0,  &
!!$0.001803D-6,  -71430.695617928D0, 2.192295512D0,  &
!!$0.001782D-6,       3.932153263D0, 5.180433689D0,  &
!!$0.001694D-6,   -4705.732307544D0, 4.641779174D0,  &
!!$0.001704D-6,   -1592.596013633D0, 3.997097652D0,  &
!!$0.001735D-6,    5849.364112115D0, 0.417558428D0,  &
!!$0.001643D-6,    8429.241266467D0, 2.180619584D0,  &
!!$0.001680D-6,      38.133035638D0, 4.164529426D0,  &
!!$0.002045D-6,    7084.896781115D0, 0.526323854D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=541,550) /  &
!!$0.001458D-6,    4292.330832950D0, 1.356098141D0,  &
!!$0.001437D-6,      20.355319399D0, 3.895439360D0,  &
!!$0.001738D-6,    6279.552731642D0, 0.087484036D0,  &
!!$0.001367D-6,   14143.495242431D0, 3.987576591D0,  &
!!$0.001344D-6,    7234.794256242D0, 0.090454338D0,  &
!!$0.001438D-6,   11499.656222793D0, 0.974387904D0,  &
!!$0.001257D-6,    6836.645252834D0, 1.509069366D0,  &
!!$0.001358D-6,   11513.883316794D0, 0.495572260D0,  &
!!$0.001628D-6,    7632.943259650D0, 4.968445721D0,  &
!!$0.001169D-6,     103.092774219D0, 2.838496795D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=551,560) /  &
!!$0.001162D-6,    4164.311989613D0, 3.408387778D0,  &
!!$0.001092D-6,    6069.776754553D0, 3.617942651D0,  &
!!$0.001008D-6,   17789.845619785D0, 0.286350174D0,  &
!!$0.001008D-6,     639.897286314D0, 1.610762073D0,  &
!!$0.000918D-6,   10213.285546211D0, 5.532798067D0,  &
!!$0.001011D-6,   -6256.777530192D0, 0.661826484D0,  &
!!$0.000753D-6,   16730.463689596D0, 3.905030235D0,  &
!!$0.000737D-6,   11926.254413669D0, 4.641956361D0,  &
!!$0.000694D-6,    3340.612426700D0, 2.111120332D0,  &
!!$0.000701D-6,    3894.181829542D0, 2.760823491D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=561,570) /  &
!!$0.000689D-6,    -135.065080035D0, 4.768800780D0,  &
!!$0.000700D-6,   13367.972631107D0, 5.760439898D0,  &
!!$0.000664D-6,    6040.347246017D0, 1.051215840D0,  &
!!$0.000654D-6,    5650.292110678D0, 4.911332503D0,  &
!!$0.000788D-6,    6681.224853400D0, 4.699648011D0,  &
!!$0.000628D-6,    5333.900241022D0, 5.024608847D0,  &
!!$0.000755D-6,    -110.206321219D0, 4.370971253D0,  &
!!$0.000628D-6,    6290.189396992D0, 3.660478857D0,  &
!!$0.000635D-6,   25132.303399966D0, 4.121051532D0,  &
!!$0.000534D-6,    5966.683980335D0, 1.173284524D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=571,580) /  &
!!$0.000543D-6,    -433.711737877D0, 0.345585464D0,  &
!!$0.000517D-6,   -1990.745017041D0, 5.414571768D0,  &
!!$0.000504D-6,    5767.611978898D0, 2.328281115D0,  &
!!$0.000485D-6,    5753.384884897D0, 1.685874771D0,  &
!!$0.000463D-6,    7860.419392439D0, 5.297703006D0,  &
!!$0.000604D-6,     515.463871093D0, 0.591998446D0,  &
!!$0.000443D-6,   12168.002696575D0, 4.830881244D0,  &
!!$0.000570D-6,     199.072001436D0, 3.899190272D0,  &
!!$0.000465D-6,   10969.965257698D0, 0.476681802D0,  &
!!$0.000424D-6,   -7079.373856808D0, 1.112242763D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=581,590) /  &
!!$0.000427D-6,     735.876513532D0, 1.994214480D0,  &
!!$0.000478D-6,   -6127.655450557D0, 3.778025483D0,  &
!!$0.000414D-6,   10973.555686350D0, 5.441088327D0,  &
!!$0.000512D-6,    1589.072895284D0, 0.107123853D0,  &
!!$0.000378D-6,   10984.192351700D0, 0.915087231D0,  &
!!$0.000402D-6,   11371.704689758D0, 4.107281715D0,  &
!!$0.000453D-6,    9917.696874510D0, 1.917490952D0,  &
!!$0.000395D-6,     149.563197135D0, 2.763124165D0,  &
!!$0.000371D-6,    5739.157790895D0, 3.112111866D0,  &
!!$0.000350D-6,   11790.629088659D0, 0.440639857D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=591,600) /  &
!!$0.000356D-6,    6133.512652857D0, 5.444568842D0,  &
!!$0.000344D-6,     412.371096874D0, 5.676832684D0,  &
!!$0.000383D-6,     955.599741609D0, 5.559734846D0,  &
!!$0.000333D-6,    6496.374945429D0, 0.261537984D0,  &
!!$0.000340D-6,    6055.549660552D0, 5.975534987D0,  &
!!$0.000334D-6,    1066.495477190D0, 2.335063907D0,  &
!!$0.000399D-6,   11506.769769794D0, 5.321230910D0,  &
!!$0.000314D-6,   18319.536584880D0, 2.313312404D0,  &
!!$0.000424D-6,    1052.268383188D0, 1.211961766D0,  &
!!$0.000307D-6,      63.735898303D0, 3.169551388D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=601,610) /  &
!!$0.000329D-6,      29.821438149D0, 6.106912080D0,  &
!!$0.000357D-6,    6309.374169791D0, 4.223760346D0,  &
!!$0.000312D-6,   -3738.761430108D0, 2.180556645D0,  &
!!$0.000301D-6,     309.278322656D0, 1.499984572D0,  &
!!$0.000268D-6,   12043.574281889D0, 2.447520648D0,  &
!!$0.000257D-6,   12491.370101415D0, 3.662331761D0,  &
!!$0.000290D-6,     625.670192312D0, 1.272834584D0,  &
!!$0.000256D-6,    5429.879468239D0, 1.913426912D0,  &
!!$0.000339D-6,    3496.032826134D0, 4.165930011D0,  &
!!$0.000283D-6,    3930.209696220D0, 4.325565754D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=611,620) /  &
!!$0.000241D-6,   12528.018664345D0, 3.832324536D0,  &
!!$0.000304D-6,    4686.889407707D0, 1.612348468D0,  &
!!$0.000259D-6,   16200.772724501D0, 3.470173146D0,  &
!!$0.000238D-6,   12139.553509107D0, 1.147977842D0,  &
!!$0.000236D-6,    6172.869528772D0, 3.776271728D0,  &
!!$0.000296D-6,   -7058.598461315D0, 0.460368852D0,  &
!!$0.000306D-6,   10575.406682942D0, 0.554749016D0,  &
!!$0.000251D-6,   17298.182327326D0, 0.834332510D0,  &
!!$0.000290D-6,    4732.030627343D0, 4.759564091D0,  &
!!$0.000261D-6,    5884.926846583D0, 0.298259862D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=621,630) /  &
!!$0.000249D-6,    5547.199336460D0, 3.749366406D0,  &
!!$0.000213D-6,   11712.955318231D0, 5.415666119D0,  &
!!$0.000223D-6,    4701.116501708D0, 2.703203558D0,  &
!!$0.000268D-6,    -640.877607382D0, 0.283670793D0,  &
!!$0.000209D-6,    5636.065016677D0, 1.238477199D0,  &
!!$0.000193D-6,   10177.257679534D0, 1.943251340D0,  &
!!$0.000182D-6,    6283.143160294D0, 2.456157599D0,  &
!!$0.000184D-6,    -227.526189440D0, 5.888038582D0,  &
!!$0.000182D-6,   -6283.008539689D0, 0.241332086D0,  &
!!$0.000228D-6,   -6284.056171060D0, 2.657323816D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=631,640) /  &
!!$0.000166D-6,    7238.675591600D0, 5.930629110D0,  &
!!$0.000167D-6,    3097.883822726D0, 5.570955333D0,  &
!!$0.000159D-6,    -323.505416657D0, 5.786670700D0,  &
!!$0.000154D-6,   -4136.910433516D0, 1.517805532D0,  &
!!$0.000176D-6,   12029.347187887D0, 3.139266834D0,  &
!!$0.000167D-6,   12132.439962106D0, 3.556352289D0,  &
!!$0.000153D-6,     202.253395174D0, 1.463313961D0,  &
!!$0.000157D-6,   17267.268201691D0, 1.586837396D0,  &
!!$0.000142D-6,   83996.847317911D0, 0.022670115D0,  &
!!$0.000152D-6,   17260.154654690D0, 0.708528947D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=641,650) /  &
!!$0.000144D-6,    6084.003848555D0, 5.187075177D0,  &
!!$0.000135D-6,    5756.566278634D0, 1.993229262D0,  &
!!$0.000134D-6,    5750.203491159D0, 3.457197134D0,  &
!!$0.000144D-6,    5326.786694021D0, 6.066193291D0,  &
!!$0.000160D-6,   11015.106477335D0, 1.710431974D0,  &
!!$0.000133D-6,    3634.621024518D0, 2.836451652D0,  &
!!$0.000134D-6,   18073.704938650D0, 5.453106665D0,  &
!!$0.000134D-6,    1162.474704408D0, 5.326898811D0,  &
!!$0.000128D-6,    5642.198242609D0, 2.511652591D0,  &
!!$0.000160D-6,     632.783739313D0, 5.628785365D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=651,660) /  &
!!$0.000132D-6,   13916.019109642D0, 0.819294053D0,  &
!!$0.000122D-6,   14314.168113050D0, 5.677408071D0,  &
!!$0.000125D-6,   12359.966151546D0, 5.251984735D0,  &
!!$0.000121D-6,    5749.452731634D0, 2.210924603D0,  &
!!$0.000136D-6,    -245.831646229D0, 1.646502367D0,  &
!!$0.000120D-6,    5757.317038160D0, 3.240883049D0,  &
!!$0.000134D-6,   12146.667056108D0, 3.059480037D0,  &
!!$0.000137D-6,    6206.809778716D0, 1.867105418D0,  &
!!$0.000141D-6,   17253.041107690D0, 2.069217456D0,  &
!!$0.000129D-6,   -7477.522860216D0, 2.781469314D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=661,670) /  &
!!$0.000116D-6,    5540.085789459D0, 4.281176991D0,  &
!!$0.000116D-6,    9779.108676125D0, 3.320925381D0,  &
!!$0.000129D-6,    5237.921013804D0, 3.497704076D0,  &
!!$0.000113D-6,    5959.570433334D0, 0.983210840D0,  &
!!$0.000122D-6,    6282.095528923D0, 2.674938860D0,  &
!!$0.000140D-6,     -11.045700264D0, 4.957936982D0,  &
!!$0.000108D-6,   23543.230504682D0, 1.390113589D0,  &
!!$0.000106D-6,  -12569.674818332D0, 0.429631317D0,  &
!!$0.000110D-6,    -266.607041722D0, 5.501340197D0,  &
!!$0.000115D-6,   12559.038152982D0, 4.691456618D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=671,680) /  &
!!$0.000134D-6,   -2388.894020449D0, 0.577313584D0,  &
!!$0.000109D-6,   10440.274292604D0, 6.218148717D0,  &
!!$0.000102D-6,    -543.918059096D0, 1.477842615D0,  &
!!$0.000108D-6,   21228.392023546D0, 2.237753948D0,  &
!!$0.000101D-6,   -4535.059436924D0, 3.100492232D0,  &
!!$0.000103D-6,      76.266071276D0, 5.594294322D0,  &
!!$0.000104D-6,     949.175608970D0, 5.674287810D0,  &
!!$0.000101D-6,   13517.870106233D0, 2.196632348D0,  &
!!$0.000100D-6,   11933.367960670D0, 4.056084160D0,  &
!!$4.322990D-6,    6283.075849991D0, 2.642893748D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=681,690) /  &
!!$0.406495D-6,       0.000000000D0, 4.712388980D0,  &
!!$0.122605D-6,   12566.151699983D0, 2.438140634D0,  &
!!$0.019476D-6,     213.299095438D0, 1.642186981D0,  &
!!$0.016916D-6,     529.690965095D0, 4.510959344D0,  &
!!$0.013374D-6,      -3.523118349D0, 1.502210314D0,  &
!!$0.008042D-6,      26.298319800D0, 0.478549024D0,  &
!!$0.007824D-6,     155.420399434D0, 5.254710405D0,  &
!!$0.004894D-6,    5746.271337896D0, 4.683210850D0,  &
!!$0.004875D-6,    5760.498431898D0, 0.759507698D0,  &
!!$0.004416D-6,    5223.693919802D0, 6.028853166D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=691,700) /  &
!!$0.004088D-6,      -7.113547001D0, 0.060926389D0,  &
!!$0.004433D-6,   77713.771467920D0, 3.627734103D0,  &
!!$0.003277D-6,   18849.227549974D0, 2.327912542D0,  &
!!$0.002703D-6,    6062.663207553D0, 1.271941729D0,  &
!!$0.003435D-6,    -775.522611324D0, 0.747446224D0,  &
!!$0.002618D-6,    6076.890301554D0, 3.633715689D0,  &
!!$0.003146D-6,     206.185548437D0, 5.647874613D0,  &
!!$0.002544D-6,    1577.343542448D0, 6.232904270D0,  &
!!$0.002218D-6,    -220.412642439D0, 1.309509946D0,  &
!!$0.002197D-6,    5856.477659115D0, 2.407212349D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=701,710) /  &
!!$0.002897D-6,    5753.384884897D0, 5.863842246D0,  &
!!$0.001766D-6,     426.598190876D0, 0.754113147D0,  &
!!$0.001738D-6,    -796.298006816D0, 2.714942671D0,  &
!!$0.001695D-6,     522.577418094D0, 2.629369842D0,  &
!!$0.001584D-6,    5507.553238667D0, 1.341138229D0,  &
!!$0.001503D-6,    -242.728603974D0, 0.377699736D0,  &
!!$0.001552D-6,    -536.804512095D0, 2.904684667D0,  &
!!$0.001370D-6,    -398.149003408D0, 1.265599125D0,  &
!!$0.001889D-6,   -5573.142801634D0, 4.413514859D0,  &
!!$0.001722D-6,    6069.776754553D0, 2.445966339D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=711,720) /  &
!!$0.001124D-6,    1059.381930189D0, 5.041799657D0,  &
!!$0.001258D-6,     553.569402842D0, 3.849557278D0,  &
!!$0.000831D-6,     951.718406251D0, 2.471094709D0,  &
!!$0.000767D-6,    4694.002954708D0, 5.363125422D0,  &
!!$0.000756D-6,    1349.867409659D0, 1.046195744D0,  &
!!$0.000775D-6,     -11.045700264D0, 0.245548001D0,  &
!!$0.000597D-6,    2146.165416475D0, 4.543268798D0,  &
!!$0.000568D-6,    5216.580372801D0, 4.178853144D0,  &
!!$0.000711D-6,    1748.016413067D0, 5.934271972D0,  &
!!$0.000499D-6,   12036.460734888D0, 0.624434410D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=721,730) /  &
!!$0.000671D-6,   -1194.447010225D0, 4.136047594D0,  &
!!$0.000488D-6,    5849.364112115D0, 2.209679987D0,  &
!!$0.000621D-6,    6438.496249426D0, 4.518860804D0,  &
!!$0.000495D-6,   -6286.598968340D0, 1.868201275D0,  &
!!$0.000456D-6,    5230.807466803D0, 1.271231591D0,  &
!!$0.000451D-6,    5088.628839767D0, 0.084060889D0,  &
!!$0.000435D-6,    5643.178563677D0, 3.324456609D0,  &
!!$0.000387D-6,   10977.078804699D0, 4.052488477D0,  &
!!$0.000547D-6,  161000.685737473D0, 2.841633844D0,  &
!!$0.000522D-6,    3154.687084896D0, 2.171979966D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=731,740) /  &
!!$0.000375D-6,    5486.777843175D0, 4.983027306D0,  &
!!$0.000421D-6,    5863.591206116D0, 4.546432249D0,  &
!!$0.000439D-6,    7084.896781115D0, 0.522967921D0,  &
!!$0.000309D-6,    2544.314419883D0, 3.172606705D0,  &
!!$0.000347D-6,    4690.479836359D0, 1.479586566D0,  &
!!$0.000317D-6,     801.820931124D0, 3.553088096D0,  &
!!$0.000262D-6,     419.484643875D0, 0.606635550D0,  &
!!$0.000248D-6,    6836.645252834D0, 3.014082064D0,  &
!!$0.000245D-6,   -1592.596013633D0, 5.519526220D0,  &
!!$0.000225D-6,    4292.330832950D0, 2.877956536D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=741,750) /  &
!!$0.000214D-6,    7234.794256242D0, 1.605227587D0,  &
!!$0.000205D-6,    5767.611978898D0, 0.625804796D0,  &
!!$0.000180D-6,   10447.387839604D0, 3.499954526D0,  &
!!$0.000229D-6,     199.072001436D0, 5.632304604D0,  &
!!$0.000214D-6,     639.897286314D0, 5.960227667D0,  &
!!$0.000175D-6,    -433.711737877D0, 2.162417992D0,  &
!!$0.000209D-6,     515.463871093D0, 2.322150893D0,  &
!!$0.000173D-6,    6040.347246017D0, 2.556183691D0,  &
!!$0.000184D-6,    6309.374169791D0, 4.732296790D0,  &
!!$0.000227D-6,  149854.400134205D0, 5.385812217D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=751,760) /  &
!!$0.000154D-6,    8031.092263058D0, 5.120720920D0,  &
!!$0.000151D-6,    5739.157790895D0, 4.815000443D0,  &
!!$0.000197D-6,    7632.943259650D0, 0.222827271D0,  &
!!$0.000197D-6,      74.781598567D0, 3.910456770D0,  &
!!$0.000138D-6,    6055.549660552D0, 1.397484253D0,  &
!!$0.000149D-6,   -6127.655450557D0, 5.333727496D0,  &
!!$0.000137D-6,    3894.181829542D0, 4.281749907D0,  &
!!$0.000135D-6,    9437.762934887D0, 5.979971885D0,  &
!!$0.000139D-6,   -2352.866153772D0, 4.715630782D0,  &
!!$0.000142D-6,    6812.766815086D0, 0.513330157D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=761,770) /  &
!!$0.000120D-6,   -4705.732307544D0, 0.194160689D0,  &
!!$0.000131D-6,  -71430.695617928D0, 0.000379226D0,  &
!!$0.000124D-6,    6279.552731642D0, 2.122264908D0,  &
!!$0.000108D-6,   -6256.777530192D0, 0.883445696D0,  &
!!$0.143388D-6,    6283.075849991D0, 1.131453581D0,  &
!!$0.006671D-6,   12566.151699983D0, 0.775148887D0,  &
!!$0.001480D-6,     155.420399434D0, 0.480016880D0,  &
!!$0.000934D-6,     213.299095438D0, 6.144453084D0,  &
!!$0.000795D-6,     529.690965095D0, 2.941595619D0,  &
!!$0.000673D-6,    5746.271337896D0, 0.120415406D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=771,780) /  &
!!$0.000672D-6,    5760.498431898D0, 5.317009738D0,  &
!!$0.000389D-6,    -220.412642439D0, 3.090323467D0,  &
!!$0.000373D-6,    6062.663207553D0, 3.003551964D0,  &
!!$0.000360D-6,    6076.890301554D0, 1.918913041D0,  &
!!$0.000316D-6,     -21.340641002D0, 5.545798121D0,  &
!!$0.000315D-6,    -242.728603974D0, 1.884932563D0,  &
!!$0.000278D-6,     206.185548437D0, 1.266254859D0,  &
!!$0.000238D-6,    -536.804512095D0, 4.532664830D0,  &
!!$0.000185D-6,     522.577418094D0, 4.578313856D0,  &
!!$0.000245D-6,   18849.227549974D0, 0.587467082D0 /
!!$DATA ((fairhd(i,j),i=1,3),j=781,787) /  &
!!$0.000180D-6,     426.598190876D0, 5.151178553D0,  &
!!$0.000200D-6,     553.569402842D0, 5.355983739D0,  &
!!$0.000141D-6,    5223.693919802D0, 1.336556009D0,  &
!!$0.000104D-6,    5856.477659115D0, 4.239842759D0,  &
!!$0.003826D-6,    6283.075849991D0, 5.705257275D0,  &
!!$0.000303D-6,   12566.151699983D0, 5.407132842D0,  &
!!$0.000209D-6,     155.420399434D0, 1.989815753D0 /
!!$! -----------------------------------------------------------------------
!!$
!!$
!!$
!!$!  Time since J2000.0 in Julian millennia
!!$t=(tdb-51544.5D0)/365250D0
!!$
!!$! -------------------- Topocentric terms -------------------------------
!!$
!!$!  Convert UT1 to local solar time in radians
!!$tsol = MOD(ut1,1D0)*d2pi - wl
!!$
!!$!  Planetary arguments:  taken or derived from Bretagnon 1982
!!$
!!$!  RLp = mean longitude, mean equinox of date
!!$!  RGp = mean anomaly
!!$!  RTp = mean longitude difference, EMB minus planet
!!$
!!$!  EMB
!!$rle = MOD (1.75347031435D0 + 6283.3196666635D0 * t, d2pi)
!!$rge = MOD (6.24005997418D0 + 6283.0195512850D0 * t, d2pi)
!!$!  Jupiter
!!$rtj = MOD (1.15392381696D0 + 5753.3848591241D0 * t, d2pi)
!!$!  Saturn
!!$rts = MOD (0.87945355785D0 + 6069.7767103654D0 * t, d2pi)
!!$
!!$!  Lunar mean elongation:  derived from Meeus 1984.
!!$rdmoon = MOD (5.198468D0 + 77713.77144D0 * t, d2pi)
!!$
!!$!  Topocentric terms (Moyer 1981 and Murray 1983)
!!$wt =  + 0.00029D-10 * u * SIN (tsol+rts)  &
!!$+ 0.00100D-10 * u * SIN (tsol-2D0*rge)+ 0.00133D-10 * u * SIN (tsol-rdmoon)  &
!!$+ 0.00133D-10 * u * SIN (tsol+rtj)- 0.00229D-10 * u * SIN (tsol+2D0*rle+rge)  &
!!$- 0.0220d-10 * v * SIN (rle+rge)+ 0.05312D-10 * u * SIN (tsol-rge)  &
!!$- 0.13677D-10 * u * SIN (tsol+2D0*rle)- 1.3184d-10 * v * COS (rle)  &
!!$+ 3.17679D-10 * u * SIN (tsol)
!!$
!!$! --------------- Fairhead model ---------------------------------------
!!$
!!$!  T**0
!!$w0=0D0
!!$DO i=474,1,-1
!!$  w0=w0+fairhd(1,i)*SIN(fairhd(2,i)*t+fairhd(3,i))
!!$END DO
!!$
!!$!  T**1
!!$w1=0D0
!!$DO i=679,475,-1
!!$  w1=w1+fairhd(1,i)*SIN(fairhd(2,i)*t+fairhd(3,i))
!!$END DO
!!$
!!$!  T**2
!!$w2=0D0
!!$DO i=764,680,-1
!!$  w2=w2+fairhd(1,i)*SIN(fairhd(2,i)*t+fairhd(3,i))
!!$END DO
!!$
!!$!  T**3
!!$w3=0D0
!!$DO i=784,765,-1
!!$  w3=w3+fairhd(1,i)*SIN(fairhd(2,i)*t+fairhd(3,i))
!!$END DO
!!$
!!$!  T**4
!!$w4=0D0
!!$DO i=787,785,-1
!!$  w4=w4+fairhd(1,i)*SIN(fairhd(2,i)*t+fairhd(3,i))
!!$END DO
!!$
!!$!  Multiply by powers of T and combine
!!$wf=t*(t*(t*(t*w4+w3)+w2)+w1)+w0
!!$
!!$!  Adjustments to use JPL planetary masses instead of IAU
!!$wj=     0.00065D-6  * SIN(   6069.776754D0   *t + 4.021194D0   ) +  &
!!$0.00033D-6  * SIN(    213.299095D0   *t + 5.543132D0   ) +  &
!!$(-0.00196D-6  * SIN(   6208.294251D0   *t + 5.696701D0   ))+  &
!!$(-0.00173D-6  * SIN(     74.781599D0   *t + 2.435900D0   ))+0.03638D-6*t*t
!!$
!!$! -----------------------------------------------------------------------
!!$
!!$!  Final result:  TDB-TT in seconds
!!$sla_rcc=wt+wf+wj
!!$
!!$END FUNCTION sla_rcc
SUBROUTINE sla_rdplan (date, np, elong, phi, ra, dec, diam)
!+
!     - - - - - - -
!      R D P L A N
!     - - - - - - -

!  Approximate topocentric apparent RA,Dec of a planet, and its
!  angular diameter.

!  Given:
!     DATE        d       MJD of observation (JD - 2400000.5)
!     NP          i       planet: 1 = Mercury
!                                 2 = Venus
!                                 3 = Moon
!                                 4 = Mars
!                                 5 = Jupiter
!                                 6 = Saturn
!                                 7 = Uranus
!                                 8 = Neptune
!                                 9 = Pluto
!                              else = Sun
!     ELONG,PHI   d       observer's east longitude and geodetic
!                                               latitude (radians)

!  Returned:
!     RA,DEC      d        RA, Dec (topocentric apparent, radians)
!     DIAM        d        angular diameter (equatorial, radians)

!  Notes:

!  1  The date is in a dynamical timescale (TDB, formerly ET) and is
!     in the form of a Modified Julian Date (JD-2400000.5).  For all
!     practical purposes, TT can be used instead of TDB, and for many
!     applications UT will do (except for the Moon).

!  2  The longitude and latitude allow correction for geocentric
!     parallax.  This is a major effect for the Moon, but in the
!     context of the limited accuracy of the present routine its
!     effect on planetary positions is small (negligible for the
!     outer planets).  Geocentric positions can be generated by
!     appropriate use of the routines sla_DMOON and sla_PLANET.

!  3  The direction accuracy (arcsec, 1000-3000AD) is of order:

!            Sun              5
!            Mercury          2
!            Venus           10
!            Moon            30
!            Mars            50
!            Jupiter         90
!            Saturn          90
!            Uranus          90
!            Neptune         10
!            Pluto            1   (1885-2099AD only)

!     The angular diameter accuracy is about 0.4% for the Moon,
!     and 0.01% or better for the Sun and planets.

!  See the sla_PLANET routine for references.

!  Called: sla_GMST, sla_DT, sla_EPJ, sla_DMOON, sla_PVOBS, sla_PRENUT,
!          sla_PLANET, sla_DMXV, sla_DCC2S, sla_DRANRM

!  P.T.Wallace   Starlink   26 May 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: date
INTEGER :: np
DOUBLE PRECISION :: elong,phi,ra,dec,diam

!  AU in km
DOUBLE PRECISION :: aukm
PARAMETER (aukm=1.49597870D8)

!  Light time for unit distance (sec)
DOUBLE PRECISION :: tau
PARAMETER (tau=499.004782D0)

INTEGER :: ip,j,i
DOUBLE PRECISION :: eqrau(0:9),stl,vgm(6),v(6),rmat(3,3),  &
vse(6),vsg(6),vsp(6),vgo(6),dx,dy,dz,r,tl

!  Equatorial radii (km)
DATA eqrau / 696000D0,2439.7D0,6051.9D0,1738D0,3397D0,71492D0,  &
60268D0,25559D0,24764D0,1151D0 /



!  Classify NP
ip=np
IF (ip < 0.OR.ip > 9) ip=0

!  Approximate local ST
stl=sla_gmst(date-sla_dt(sla_epj(date))/86400D0)+elong

!  Geocentre to Moon (mean of date)
CALL sla_dmoon(date,v)

!  Nutation to true of date
CALL sla_nut(date,rmat)
CALL sla_dmxv(rmat,v,vgm)
CALL sla_dmxv(rmat,v(4),vgm(4))

!  Moon?
IF (ip == 3) THEN
  
!     Yes: geocentre to Moon (true of date)
  DO i=1,6
    v(i)=vgm(i)
  END DO
ELSE
  
!     No: precession/nutation matrix, J2000 to date
  CALL sla_prenut(2000D0,date,rmat)
  
!     Sun to Earth-Moon Barycentre (J2000)
  CALL sla_planet(date,3,v,j)
  
!     Precession and nutation to date
  CALL sla_dmxv(rmat,v,vse)
  CALL sla_dmxv(rmat,v(4),vse(4))
  
!     Sun to geocentre (true of date)
  DO i=1,6
    vsg(i)=vse(i)-0.012150581D0*vgm(i)
  END DO
  
!     Sun?
  IF (ip == 0) THEN
    
!        Yes: geocentre to Sun
    DO i=1,6
      v(i)=-vsg(i)
    END DO
  ELSE
    
!        No: Sun to Planet (J2000)
    CALL sla_planet(date,ip,v,j)
    
!        Precession and nutation to date
    CALL sla_dmxv(rmat,v,vsp)
    CALL sla_dmxv(rmat,v(4),vsp(4))
    
!        Geocentre to planet
    DO i=1,6
      v(i)=vsp(i)-vsg(i)
    END DO
  END IF
END IF

!  Refer to origin at the observer
CALL sla_pvobs(phi,0D0,stl,vgo)
DO i=1,6
  v(i)=v(i)-vgo(i)
END DO

!  Geometric distance (AU)
dx=v(1)
dy=v(2)
dz=v(3)
r=SQRT(dx*dx+dy*dy+dz*dz)

!  Light time (sec)
tl=tau*r

!  Correct position for planetary aberration
DO i=1,3
  v(i)=v(i)-tl*v(i+3)
END DO

!  To RA,Dec
CALL sla_dcc2s(v,ra,dec)
ra=sla_dranrm(ra)

!  Angular diameter (radians)
diam=2D0*ASIN(eqrau(ip)/(r*aukm))

END SUBROUTINE sla_rdplan
SUBROUTINE sla_refco (hm, tdk, pmb, rh, wl, phi, tlr, eps,refa, refb)
!+
!     - - - - - -
!      R E F C O
!     - - - - - -

!  Determine the constants A and B in the atmospheric refraction
!  model dZ = A tan Z + B tan**3 Z.

!  Z is the "observed" zenith distance (i.e. affected by refraction)
!  and dZ is what to add to Z to give the "topocentric" (i.e. in vacuo)
!  zenith distance.

!  Given:
!    HM      d     height of the observer above sea level (metre)
!    TDK     d     ambient temperature at the observer (deg K)
!    PMB     d     pressure at the observer (millibar)
!    RH      d     relative humidity at the observer (range 0-1)
!    WL      d     effective wavelength of the source (micrometre)
!    PHI     d     latitude of the observer (radian, astronomical)
!    TLR     d     temperature lapse rate in the troposphere (degK/metre)
!    EPS     d     precision required to terminate iteration (radian)

!  Returned:
!    REFA    d     tan Z coefficient (radian)
!    REFB    d     tan**3 Z coefficient (radian)

!  Called:  sla_REFRO

!  Notes:

!  1  Typical values for the TLR and EPS arguments might be 0.0065D0 and
!     1D-10 respectively.

!  2  The radio refraction is chosen by specifying WL > 100 micrometres.

!  3  The routine is a slower but more accurate alternative to the
!     sla_REFCOQ routine.  The constants it produces give perfect
!     agreement with sla_REFRO at zenith distances arctan(1) (45 deg)
!     and arctan(4) (about 76 deg).  It achieves 0.5 arcsec accuracy
!     for ZD < 80 deg, 0.01 arcsec accuracy for ZD < 60 deg, and
!     0.001 arcsec accuracy for ZD < 45 deg.

!  P.T.Wallace   Starlink   3 June 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: hm,tdk,pmb,rh,wl,phi,tlr,eps,refa,refb

DOUBLE PRECISION :: atn1,atn4,r1,r2

!  Sample zenith distances: arctan(1) and arctan(4)
PARAMETER (atn1=0.7853981633974483D0,atn4=1.325817663668033D0)



!  Determine refraction for the two sample zenith distances
CALL sla_refro(atn1,hm,tdk,pmb,rh,wl,phi,tlr,eps,r1)
CALL sla_refro(atn4,hm,tdk,pmb,rh,wl,phi,tlr,eps,r2)

!  Solve for refraction constants
refa = (64D0*r1-r2)/60D0
refb = (r2-4D0*r1)/60D0

END SUBROUTINE sla_refco
SUBROUTINE sla_refcoq (tdk, pmb, rh, wl, refa, refb)
!+
!     - - - - - - -
!      R E F C O Q
!     - - - - - - -

!  Determine the constants A and B in the atmospheric refraction
!  model dZ = A tan Z + B tan**3 Z.  This is a fast alternative
!  to the sla_REFCO routine - see notes.

!  Z is the "observed" zenith distance (i.e. affected by refraction)
!  and dZ is what to add to Z to give the "topocentric" (i.e. in vacuo)
!  zenith distance.

!  Given:
!    TDK      d      ambient temperature at the observer (deg K)
!    PMB      d      pressure at the observer (millibar)
!    RH       d      relative humidity at the observer (range 0-1)
!    WL       d      effective wavelength of the source (micrometre)

!  Returned:
!    REFA     d      tan Z coefficient (radian)
!    REFB     d      tan**3 Z coefficient (radian)

!  The radio refraction is chosen by specifying WL > 100 micrometres.

!  Notes:

!  1  The model is an approximation, for moderate zenith distances,
!     to the predictions of the sla_REFRO routine.  The approximation
!     is maintained across a range of conditions, and applies to
!     both optical/IR and radio.

!  2  The algorithm is a fast alternative to the sla_REFCO routine.
!     The latter calls the sla_REFRO routine itself:  this involves
!     integrations through a model atmosphere, and is costly in
!     processor time.  However, the model which is produced is precisely
!     correct for two zenith distance (45 degrees and about 76 degrees)
!     and at other zenith distances is limited in accuracy only by the
!     A tan Z + B tan**3 Z formulation itself.  The present routine
!     is not as accurate, though it satisfies most practical
!     requirements.

!  3  The model omits the effects of (i) height above sea level (apart
!     from the reduced pressure itself), (ii) latitude (i.e. the
!     flattening of the Earth) and (iii) variations in tropospheric
!     lapse rate.

!     The model was tested using the following range of conditions:

!       lapse rates 0.0055, 0.0065, 0.0075 deg/metre
!       latitudes 0, 25, 50, 75 degrees
!       heights 0, 2500, 5000 metres ASL
!       pressures mean for height -10% to +5% in steps of 5%
!       temperatures -10 deg to +20 deg with respect to 280 deg at SL
!       relative humidity 0, 0.5, 1
!       wavelengths 0.4, 0.6, ... 2 micron, + radio
!       zenith distances 15, 45, 75 degrees

!     The accuracy with respect to direct use of the sla_REFRO routine
!     was as follows:

!                            worst         RMS

!       optical/IR           62 mas       8 mas
!       radio               319 mas      49 mas

!     For this particular set of conditions:

!       lapse rate 0.0065 degK/metre
!       latitude 50 degrees
!       sea level
!       pressure 1005 mB
!       temperature 280.15 degK
!       humidity 80%
!       wavelength 5740 Angstroms

!     the results were as follows:

!       ZD        sla_REFRO   sla_REFCOQ  Saastamoinen

!       10         10.27        10.27        10.27
!       20         21.19        21.20        21.19
!       30         33.61        33.61        33.60
!       40         48.82        48.83        48.81
!       45         58.16        58.18        58.16
!       50         69.28        69.30        69.27
!       55         82.97        82.99        82.95
!       60        100.51       100.54       100.50
!       65        124.23       124.26       124.20
!       70        158.63       158.68       158.61
!       72        177.32       177.37       177.31
!       74        200.35       200.38       200.32
!       76        229.45       229.43       229.42
!       78        267.44       267.29       267.41
!       80        319.13       318.55       319.10

!      deg        arcsec       arcsec       arcsec

!     The values for Saastamoinen's formula (which includes terms
!     up to tan^5) are taken from Hohenkerk and Sinclair (1985).

!     The results from the much slower but more accurate sla_REFCO
!     routine have not been included in the tabulation as they are
!     identical to those in the sla_REFRO column to the 0.01 arcsec
!     resolution used.

!  4  Outlandish input parameters are silently limited to mathematically
!     safe values.  Zero pressure is permissible, and causes zeroes to
!     be returned.

!  5  The algorithm draws on several sources, as follows:

!     a) The formula for the saturation vapour pressure of water as
!        a function of temperature and temperature is taken from
!        expressions A4.5-A4.7 of Gill (1982).

!     b) The formula for the water vapour pressure, given the
!        saturation pressure and the relative humidity, is from
!        Crane (1976), expression 2.5.5.

!     c) The refractivity of air is a function of temperature,
!        total pressure, water-vapour pressure and, in the case
!        of optical/IR but not radio, wavelength.  The formulae
!        for the two cases are developed from the Essen and Froome
!        expressions adopted in Resolution 1 of the 12th International
!        Geodesy Association General Assembly (1963).

!     The above three items are as used in the sla_REFRO routine.

!     d) The formula for beta, the ratio of the scale height of the
!        atmosphere to the geocentric distance of the observer, is
!        an adaption of expression 9 from Stone (1996).  The
!        adaptations, arrived at empirically, consist of (i) a
!        small adjustment to the coefficient and (ii) a humidity
!        term for the radio case only.

!     e) The formulae for the refraction constants as a function of
!        n-1 and beta are from Green (1987), expression 4.31.

!  References:

!     Crane, R.K., Meeks, M.L. (ed), "Refraction Effects in the Neutral
!     Atmosphere", Methods of Experimental Physics: Astrophysics 12B,
!     Academic Press, 1976.

!     Gill, Adrian E., "Atmosphere-Ocean Dynamics", Academic Press, 1982.

!     Hohenkerk, C.Y., & Sinclair, A.T., NAO Technical Note No. 63, 1985.

!     International Geodesy Association General Assembly, Bulletin
!     Geodesique 70 p390, 1963.

!     Stone, Ronald C., P.A.S.P. 108 1051-1058, 1996.

!     Green, R.M., "Spherical Astronomy", Cambridge University Press, 1987.

!  P.T.Wallace   Starlink   4 June 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: tdk,pmb,rh,wl,refa,refb

LOGICAL :: optic
DOUBLE PRECISION :: t,p,r,w,tdc,ps,pw,wlsq,gamma,beta



!  Decide whether optical/IR or radio case:  switch at 100 microns.
optic = wl <= 100D0

!  Restrict parameters to safe values.
t = MIN(MAX(tdk,100D0),500D0)
p = MIN(MAX(pmb,0D0),10000D0)
r = MIN(MAX(rh,0D0),1D0)
w = MIN(MAX(wl,0.1D0),1D6)

!  Water vapour pressure at the observer.
IF (p > 0D0) THEN
  tdc = t-273.15D0
  ps = 10D0**((0.7859D0+0.03477D0*tdc)/(1D0+0.00412D0*tdc))*  &
  (1D0+p*(4.5D-6+6D-10*tdc*tdc))
  pw = r*ps/(1D0-(1D0-r)*ps/p)
ELSE
  pw = 0D0
END IF

!  Refractive index minus 1 at the observer.
IF (optic) THEN
  wlsq = wl*wl
  gamma = ((77.532D-6+(4.391D-7+3.57D-9/wlsq)/wlsq)*p-11.2684D-6*pw)/t
ELSE
  gamma = (77.624D-6*p-(12.92D-6-0.371897D0/t)*pw)/t
END IF

!  Formula for beta adapted from Stone, with empirical adjustments.
beta=4.4474D-6*t
IF (.NOT.optic) beta=beta-0.0074D0*pw*beta

!  Refraction constants from Green.
refa = gamma*(1D0-beta)
refb = -gamma*(beta-gamma/2D0)

END SUBROUTINE sla_refcoq
SUBROUTINE sla_refro (zobs, hm, tdk, pmb, rh, wl, phi, tlr,eps, ref)
!+
!     - - - - - -
!      R E F R O
!     - - - - - -

!  Atmospheric refraction for radio and optical/IR wavelengths.

!  Given:
!    ZOBS    d  observed zenith distance of the source (radian)
!    HM      d  height of the observer above sea level (metre)
!    TDK     d  ambient temperature at the observer (deg K)
!    PMB     d  pressure at the observer (millibar)
!    RH      d  relative humidity at the observer (range 0-1)
!    WL      d  effective wavelength of the source (micrometre)
!    PHI     d  latitude of the observer (radian, astronomical)
!    TLR     d  temperature lapse rate in the troposphere (degK/metre)
!    EPS     d  precision required to terminate iteration (radian)

!  Returned:
!    REF     d  refraction: in vacuo ZD minus observed ZD (radian)

!  Notes:

!  1  A suggested value for the TLR argument is 0.0065D0.  The
!     refraction is significantly affected by TLR, and if studies
!     of the local atmosphere have been carried out a better TLR
!     value may be available.

!  2  A suggested value for the EPS argument is 1D-8.  The result is
!     usually at least two orders of magnitude more computationally
!     precise than the supplied EPS value.

!  3  The routine computes the refraction for zenith distances up
!     to and a little beyond 90 deg using the method of Hohenkerk
!     and Sinclair (NAO Technical Notes 59 and 63, subsequently adopted
!     in the Explanatory Supplement, 1992 edition - see section 3.281).

!  4  The code is a development of the optical/IR refraction subroutine
!     AREF of C.Hohenkerk (HMNAO, September 1984), with extensions to
!     support the radio case.  Apart from merely cosmetic changes, the
!     following modifications to the original HMNAO optical/IR refraction
!     code have been made:

!     .  The angle arguments have been changed to radians.

!     .  Any value of ZOBS is allowed (see note 6, below).

!     .  Other argument values have been limited to safe values.

!     .  Murray's values for the gas constants have been used
!        (Vectorial Astrometry, Adam Hilger, 1983).

!     .  The numerical integration phase has been rearranged for
!        extra clarity.

!     .  A better model for Ps(T) has been adopted (taken from
!        Gill, Atmosphere-Ocean Dynamics, Academic Press, 1982).

!     .  More accurate expressions for Pwo have been adopted
!        (again from Gill 1982).

!     .  Provision for radio wavelengths has been added using
!        expressions devised by A.T.Sinclair, RGO (private
!        communication 1989), based on the Essen & Froome
!        refractivity formula adopted in Resolution 1 of the
!        13th International Geodesy Association General Assembly
!        (Bulletin Geodesique 70 p390, 1963).

!     .  Various small changes have been made to gain speed.

!     None of the changes significantly affects the optical/IR results
!     with respect to the algorithm given in the 1992 Explanatory
!     Supplement.  For example, at 70 deg zenith distance the present
!     routine agrees with the ES algorithm to better than 0.05 arcsec
!     for any reasonable combination of parameters.  However, the
!     improved water-vapour expressions do make a significant difference
!     in the radio band, at 70 deg zenith distance reaching almost
!     4 arcsec for a hot, humid, low-altitude site during a period of
!     low pressure.

!  5  The radio refraction is chosen by specifying WL > 100 micrometres.
!     Because the algorithm takes no account of the ionosphere, the
!     accuracy deteriorates at low frequencies, below about 30 MHz.

!  6  Before use, the value of ZOBS is expressed in the range +/- pi.
!     If this ranged ZOBS is -ve, the result REF is computed from its
!     absolute value before being made -ve to match.  In addition, if
!     it has an absolute value greater than 93 deg, a fixed REF value
!     equal to the result for ZOBS = 93 deg is returned, appropriately
!     signed.

!  7  As in the original Hohenkerk and Sinclair algorithm, fixed values
!     of the water vapour polytrope exponent, the height of the
!     tropopause, and the height at which refraction is negligible are
!     used.

!  8  The radio refraction has been tested against work done by
!     Iain Coulson, JACH, (private communication 1995) for the
!     James Clerk Maxwell Telescope, Mauna Kea.  For typical conditions,
!     agreement at the 0.1 arcsec level is achieved for moderate ZD,
!     worsening to perhaps 0.5-1.0 arcsec at ZD 80 deg.  At hot and
!     humid sea-level sites the accuracy will not be as good.

!  9  It should be noted that the relative humidity RH is formally
!     defined in terms of "mixing ratio" rather than pressures or
!     densities as is often stated.  It is the mass of water per unit
!     mass of dry air divided by that for saturated air at the same
!     temperature and pressure (see Gill 1982).

!  Called:  sla_DRANGE, sla__ATMT, sla__ATMS

!  P.T.Wallace   Starlink   3 June 1997

!  Copyright (C) 1997 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: zobs,hm,tdk,pmb,rh,wl,phi,tlr,eps,ref


!  Fixed parameters

DOUBLE PRECISION :: d93,gcr,dmd,dmw,s,delta,ht,hs
!  93 degrees in radians
PARAMETER (d93=1.623156204D0)
!  Universal gas constant
PARAMETER (gcr=8314.32D0)
!  Molecular weight of dry air
PARAMETER (dmd=28.9644D0)
!  Molecular weight of water vapour
PARAMETER (dmw=18.0152D0)
!  Mean Earth radius (metre)
PARAMETER (s=6378120D0)
!  Exponent of temperature dependence of water vapour pressure
PARAMETER (delta=18.36D0)
!  Height of tropopause (metre)
PARAMETER (ht=11000D0)
!  Upper limit for refractive effects (metre)
PARAMETER (hs=80000D0)

INTEGER :: is,k,n,i,j
LOGICAL :: optic,loop
DOUBLE PRECISION :: zobs1,zobs2,hmok,tdkok,pmbok,rhok,wlok,alpha,  &
tol,wlsq,gb,a,gamal,gamma,gamm2,delm2,tdc,psat,pwo,w,  &
c1,c2,c3,c4,c5,c6,r0,tempo,dn0,rdndr0,sk0,f0,  &
rt,tt,dnt,rdndrt,sine,zt,ft,dnts,rdndrp,zts,fts,  &
rs,dns,rdndrs,zs,fs,refold,z0,zrange,fb,ff,fo,fe,  &
h,r,sz,rg,dr,tg,dn,rdndr,t,f,refp,reft


!  The refraction integrand
DOUBLE PRECISION :: refi
refi(r,dn,rdndr) = rdndr/(dn+rdndr)



!  Transform ZOBS into the normal range.
zobs1 = sla_drange(zobs)
zobs2 = MIN(ABS(zobs1),d93)

!  Keep other arguments within safe bounds.
hmok = MIN(MAX(hm,-1D3),10D3)
tdkok = MIN(MAX(tdk,100D0),500D0)
pmbok = MIN(MAX(pmb,0D0),10000D0)
rhok = MIN(MAX(rh,0D0),1D0)
wlok = MAX(wl,0.1D0)
alpha = MIN(MAX(ABS(tlr),0.001D0),0.01D0)

!  Tolerance for iteration.
tol = MIN(MAX(ABS(eps),1D-12),0.1D0)/2D0

!  Decide whether optical/IR or radio case - switch at 100 microns.
optic = wlok <= 100D0

!  Set up model atmosphere parameters defined at the observer.
wlsq = wlok*wlok
gb = 9.784D0*(1D0-0.0026D0*COS(phi+phi)-0.00000028D0*hmok)
IF (optic) THEN
  a = (287.604D0+(1.6288D0+0.0136D0/wlsq)/wlsq)*273.15D-6/1013.25D0
ELSE
  a = 77.624D-6
END IF
gamal = (gb*dmd)/gcr
gamma = gamal/alpha
gamm2 = gamma-2D0
delm2 = delta-2D0
tdc = tdkok-273.15D0
psat = 10D0**((0.7859D0+0.03477D0*tdc)/(1D0+0.00412D0*tdc))*  &
(1D0+pmbok*(4.5D-6+6D-10*tdc*tdc))
IF (pmbok > 0D0) THEN
  pwo = rhok*psat/(1D0-(1D0-rhok)*psat/pmbok)
ELSE
  pwo = 0D0
END IF
w = pwo*(1D0-dmw/dmd)*gamma/(delta-gamma)
c1 = a*(pmbok+w)/tdkok
IF (optic) THEN
  c2 = (a*w+11.2684D-6*pwo)/tdkok
ELSE
  c2 = (a*w+12.92D-6*pwo)/tdkok
END IF
c3 = (gamma-1D0)*alpha*c1/tdkok
c4 = (delta-1D0)*alpha*c2/tdkok
IF (optic) THEN
  c5 = 0D0
  c6 = 0D0
ELSE
  c5 = 371897D-6*pwo/tdkok
  c6 = c5*delm2*alpha/(tdkok*tdkok)
END IF

!  Conditions at the observer.
r0 = s+hmok
CALL sla__atmt(r0,tdkok,alpha,gamm2,delm2,c1,c2,c3,c4,c5,c6,  &
r0,tempo,dn0,rdndr0)
sk0 = dn0*r0*SIN(zobs2)
f0 = refi(r0,dn0,rdndr0)

!  Conditions in the troposphere at the tropopause.
rt = s+ht
CALL sla__atmt(r0,tdkok,alpha,gamm2,delm2,c1,c2,c3,c4,c5,c6,rt,tt,dnt,rdndrt)
sine = sk0/(rt*dnt)
zt = ATAN2(sine,SQRT(MAX(1D0-sine*sine,0D0)))
ft = refi(rt,dnt,rdndrt)

!  Conditions in the stratosphere at the tropopause.
CALL sla__atms(rt,tt,dnt,gamal,rt,dnts,rdndrp)
sine = sk0/(rt*dnts)
zts = ATAN2(sine,SQRT(MAX(1D0-sine*sine,0D0)))
fts = refi(rt,dnts,rdndrp)

!  Conditions at the stratosphere limit.
rs = s+hs
CALL sla__atms(rt,tt,dnt,gamal,rs,dns,rdndrs)
sine = sk0/(rs*dns)
zs = ATAN2(sine,SQRT(MAX(1D0-sine*sine,0D0)))
fs = refi(rs,dns,rdndrs)


!  Integrate the refraction integral in two parts;  first in the
!  troposphere (K=1), then in the stratosphere (K=2).


!  Initialize previous refraction to ensure at least two iterations.
refold = 1D6

!  Start off with 8 strips for the troposphere integration, and then
!  use the final troposphere value for the stratosphere integration,
!  which tends to need more strips.
is = 8

!  Troposphere then stratosphere.
DO k = 1,2
  
!     Start Z, Z range, and start and end values.
  IF (k == 1) THEN
    z0 = zobs2
    zrange = zt-z0
    fb = f0
    ff = ft
  ELSE
    z0 = zts
    zrange = zs-z0
    fb = fts
    ff = fs
  END IF
  
!     Sums of odd and even values.
  fo = 0D0
  fe = 0D0
  
!     First time through the loop we have to do every point.
  n = 1
  
!     Start of iteration loop (terminates at specified precision).
  loop = .true.
  DO WHILE (loop)
    
!        Strip width.
    h = zrange/DBLE(is)
    
!        Initialize distance from Earth centre for quadrature pass.
    IF (k == 1) THEN
      r = r0
    ELSE
      r = rt
    END IF
    
!        One pass (no need to compute evens after first time).
    DO i=1,is-1,n
      
!           Sine of observed zenith distance.
      sz = SIN(z0+h*DBLE(i))
      
!           Find R (to the nearest metre, maximum four iterations).
      IF (sz > 1D-20) THEN
        w = sk0/sz
        rg = r
        dr = 1D6
        j = 0
        DO WHILE (ABS(dr) > 1D0.AND.j < 4)
          j=j+1
          IF (k == 1) THEN
            CALL sla__atmt(r0,tdkok,alpha,gamm2,delm2,  &
            c1,c2,c3,c4,c5,c6,rg,tg,dn,rdndr)
          ELSE
            CALL sla__atms(rt,tt,dnt,gamal,rg,dn,rdndr)
          END IF
          dr = (rg*dn-w)/(dn+rdndr)
          rg = rg-dr
        END DO
        r = rg
      END IF
      
!           Find the refractive index and integrand at R.
      IF (k == 1) THEN
        CALL sla__atmt(r0,tdkok,alpha,gamm2,delm2,  &
        c1,c2,c3,c4,c5,c6,r,t,dn,rdndr)
      ELSE
        CALL sla__atms(rt,tt,dnt,gamal,r,dn,rdndr)
      END IF
      f = refi(r,dn,rdndr)
      
!           Accumulate odd and (first time only) even values.
      IF (n == 1.AND.MOD(i,2) == 0) THEN
        fe = fe+f
      ELSE
        fo = fo+f
      END IF
    END DO
    
!        Evaluate the integrand using Simpson's Rule.
    refp = h*(fb+4D0*fo+2D0*fe+ff)/3D0
    
!        Has the required precision been achieved?
    IF (ABS(refp-refold) > tol) THEN
      
!           No: prepare for next iteration.
      
!           Save current value for convergence test.
      refold = refp
      
!           Double the number of strips.
      is = is+is
      
!           Sum of all current values = sum of next pass's even values.
      fe = fe+fo
      
!           Prepare for new odd values.
      fo = 0D0
      
!           Skip even values next time.
      n = 2
    ELSE
      
!           Yes: save troposphere component and terminate the loop.
      IF (k == 1) reft = refp
      loop = .false.
    END IF
  END DO
END DO

!  Result.
ref = reft+refp
IF (zobs1 < 0D0) ref = -ref

END SUBROUTINE sla_refro
SUBROUTINE sla_refv (vu, refa, refb, vr)
!+
!     - - - - -
!      R E F V
!     - - - - -

!  Adjust an unrefracted Cartesian vector to include the effect of
!  atmospheric refraction, using the simple A tan Z + B tan**3 Z
!  model.

!  Given:
!    VU    dp    unrefracted position of the source (Az/El 3-vector)
!    REFA  dp    tan Z coefficient (radian)
!    REFB  dp    tan**3 Z coefficient (radian)

!  Returned:
!    VR    dp    refracted position of the source (Az/El 3-vector)

!  Notes:

!  1  This routine applies the adjustment for refraction in the
!     opposite sense to the usual one - it takes an unrefracted
!     (in vacuo) position and produces an observed (refracted)
!     position, whereas the A tan Z + B tan**3 Z model strictly
!     applies to the case where an observed position is to have the
!     refraction removed.  The unrefracted to refracted case is
!     harder, and requires an inverted form of the text-book
!     refraction models;  the algorithm used here is equivalent to
!     one iteration of the Newton-Raphson method applied to the above
!     formula.

!  2  Though optimized for speed rather than precision, the present
!     routine achieves consistency with the refracted-to-unrefracted
!     A tan Z + B tan**3 Z model at better than 1 microarcsecond within
!     30 degrees of the zenith and remains within 1 milliarcsecond to
!     beyond ZD 70 degrees.  The inherent accuracy of the model is, of
!     course, far worse than this - see the documentation for sla_REFCO
!     for more information.

!  3  At low elevations (below about 3 degrees) the refraction
!     correction is held back to prevent arithmetic problems and
!     wildly wrong results.  Over a wide range of observer heights
!     and corresponding temperatures and pressures, the following
!     levels of accuracy (arcsec) are achieved, relative to numerical
!     integration through a model atmosphere:

!              ZD    error

!              80      0.4
!              81      0.8
!              82      1.6
!              83      3
!              84      7
!              85     17
!              86     45
!              87    150
!              88    340
!              89    620
!              90   1100
!              91   1900         } relevant only to
!              92   3200         } high-elevation sites

!  4  See also the routine sla_REFZ, which performs the adjustment to
!     the zenith distance rather than in Cartesian Az/El coordinates.
!     The present routine is faster than sla_REFZ and, except very low down,
!     is equally accurate for all practical purposes.  However, beyond
!     about ZD 84 degrees sla_REFZ should be used, and for the utmost
!     accuracy iterative use of sla_REFRO should be considered.

!  P.T.Wallace   Starlink   26 December 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: vu(3),refa,refb,vr(3)

DOUBLE PRECISION :: x,y,z1,z,zsq,rsq,r,wb,wt,d,cd,f



!  Initial estimate = unrefracted vector
x = vu(1)
y = vu(2)
z1 = vu(3)

!  Keep correction approximately constant below about 3 deg elevation
z = MAX(z1,0.05D0)

!  One Newton-Raphson iteration
zsq = z*z
rsq = x*x+y*y
r = SQRT(rsq)
wb = refb*rsq/zsq
wt = (refa+wb)/(1D0+(refa+3D0*wb)*(zsq+rsq)/zsq)
d = wt*r/z
cd = 1D0-d*d/2D0
f = cd*(1D0-wt)

!  Post-refraction x,y,z
vr(1) = x*f
vr(2) = y*f
vr(3) = cd*(z+d*r)+(z1-z)

END SUBROUTINE sla_refv
SUBROUTINE sla_refz (zu, refa, refb, zr)
!+
!     - - - - -
!      R E F Z
!     - - - - -

!  Adjust an unrefracted zenith distance to include the effect of
!  atmospheric refraction, using the simple A tan Z + B tan**3 Z
!  model (plus special handling for large ZDs).

!  Given:
!    ZU    dp    unrefracted zenith distance of the source (radian)
!    REFA  dp    tan Z coefficient (radian)
!    REFB  dp    tan**3 Z coefficient (radian)

!  Returned:
!    ZR    dp    refracted zenith distance (radian)

!  Notes:

!  1  This routine applies the adjustment for refraction in the
!     opposite sense to the usual one - it takes an unrefracted
!     (in vacuo) position and produces an observed (refracted)
!     position, whereas the A tan Z + B tan**3 Z model strictly
!     applies to the case where an observed position is to have the
!     refraction removed.  The unrefracted to refracted case is
!     harder, and requires an inverted form of the text-book
!     refraction models;  the formula used here is based on the
!     Newton-Raphson method.  For the utmost numerical consistency
!     with the refracted to unrefracted model, two iterations are
!     carried out, achieving agreement at the 1D-11 arcseconds level
!     for a ZD of 80 degrees.  The inherent accuracy of the model
!     is, of course, far worse than this - see the documentation for
!     sla_REFCO for more information.

!  2  At ZD 83 degrees, the rapidly-worsening A tan Z + B tan**3 Z
!     model is abandoned and an empirical formula takes over.  Over a
!     wide range of observer heights and corresponding temperatures and
!     pressures, the following levels of accuracy (arcsec) are
!     typically achieved, relative to numerical integration through a
!     model atmosphere:

!              ZR    error

!              80      0.4
!              81      0.8
!              82      1.5
!              83      3.2
!              84      4.9
!              85      5.8
!              86      6.1
!              87      7.1
!              88     10
!              89     20
!              90     40
!              91    100         } relevant only to
!              92    200         } high-elevation sites

!     The high-ZD model is scaled to match the normal model at the
!     transition point;  there is no glitch.

!  3  Beyond 93 deg zenith distance, the refraction is held at its
!     93 deg value.

!  4  See also the routine sla_REFV, which performs the adjustment in
!     Cartesian Az/El coordinates, and with the emphasis on speed
!     rather than numerical accuracy.

!  P.T.Wallace   Starlink   19 September 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: zu,refa,refb,zr

!  Radians to degrees
DOUBLE PRECISION :: r2d
PARAMETER (r2d=57.29577951308232D0)

!  Largest usable ZD (deg)
DOUBLE PRECISION :: d93
PARAMETER (d93=93D0)

!  Coefficients for high ZD model (used beyond ZD 83 deg)
DOUBLE PRECISION :: c1,c2,c3,c4,c5
PARAMETER (c1=+0.55445D0,c2=-0.01133D0,  &
c3=+0.00202D0,c4=+0.28385D0,  &
c5=+0.02390D0)

!  ZD at which one model hands over to the other (radians)
DOUBLE PRECISION :: z83
PARAMETER (z83=83D0/r2d)

!  High-ZD-model prediction (deg) for that point
DOUBLE PRECISION :: ref83
PARAMETER (ref83=(c1+c2*7D0+c3*49D0)/(1D0+c4*7D0+c5*49D0))

DOUBLE PRECISION :: zu1,zl,s,c,t,tsq,tcu,ref,e,e2



!  Perform calculations for ZU or 83 deg, whichever is smaller
zu1 = MIN(zu,z83)

!  Functions of ZD
zl = zu1
s = SIN(zl)
c = COS(zl)
t = s/c
tsq = t*t
tcu = t*tsq

!  Refracted ZD (mathematically to better than 1 mas at 70 deg)
zl = zl-(refa*t+refb*tcu)/(1D0+(refa+3D0*refb*tsq)/(c*c))

!  Further iteration
s = SIN(zl)
c = COS(zl)
t = s/c
tsq = t*t
tcu = t*tsq
ref = zu1-zl+(zl-zu1+refa*t+refb*tcu)/(1D0+(refa+3D0*refb*tsq)/(c*c))

!  Special handling for large ZU
IF (zu > zu1) THEN
  e = 90D0-MIN(d93,zu*r2d)
  e2 = e*e
  ref = (ref/ref83)*(c1+c2*e+c3*e2)/(1D0+c4*e+c5*e2)
END IF

!  Return refracted ZD
zr = zu-ref

END SUBROUTINE sla_refz
REAL FUNCTION sla_rverot (phi, ra, da, st)
!+
!     - - - - - - -
!      R V E R O T
!     - - - - - - -

!  Velocity component in a given direction due to Earth rotation
!  (single precision)

!  Given:
!     PHI     real    latitude of observing station (geodetic)
!     RA,DA   real    apparent RA,DEC
!     ST      real    local apparent sidereal time

!  PHI, RA, DEC and ST are all in radians.

!  Result:
!     Component of Earth rotation in direction RA,DA (km/s)

!  Sign convention:
!     The result is +ve when the observatory is receding from the
!     given point on the sky.

!  Accuracy:
!     The simple algorithm used assumes a spherical Earth, of
!     a radius chosen to give results accurate to about 0.0005 km/s
!     for observing stations at typical latitudes and heights.  For
!     applications requiring greater precision, use the routine
!     sla_PVOBS.

!  P.T.Wallace   Starlink   20 July 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: phi,ra,da,st

!  Nominal mean sidereal speed of Earth equator in km/s (the actual
!  value is about 0.4651)
REAL :: espeed
PARAMETER (espeed=0.4655)


sla_rverot=espeed*COS(phi)*SIN(st-ra)*COS(da)

END FUNCTION sla_rverot
REAL FUNCTION sla_rvgalc (r2000, d2000)
!+
!     - - - - - - -
!      R V G A L C
!     - - - - - - -

!  Velocity component in a given direction due to the rotation
!  of the Galaxy (single precision)

!  Given:
!     R2000,D2000   real    J2000.0 mean RA,Dec (radians)

!  Result:
!     Component of dynamical LSR motion in direction R2000,D2000 (km/s)

!  Sign convention:
!     The result is +ve when the dynamical LSR is receding from the
!     given point on the sky.

!  Note:  The Local Standard of Rest used here is a point in the
!         vicinity of the Sun which is in a circular orbit around
!         the Galactic centre.  Sometimes called the "dynamical" LSR,
!         it is not to be confused with a "kinematical" LSR, which
!         is the mean standard of rest of star catalogues or stellar
!         populations.

!  Reference:  The orbital speed of 220 km/s used here comes from
!              Kerr & Lynden-Bell (1986), MNRAS, 221, p1023.

!  Called:
!     sla_CS2C, sla_VDV

!  P.T.Wallace   Starlink   23 March 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: r2000,d2000

REAL :: va(3), vb(3)



!  LSR velocity due to Galactic rotation

!  Speed = 220 km/s
!  Apex  = L2,B2  90deg, 0deg
!        = RA,Dec  21 12 01.1  +48 19 47  J2000.0

!  This is expressed in the form of a J2000.0 x,y,z vector:

!      VA(1) = X = -SPEED*COS(RA)*COS(DEC)
!      VA(2) = Y = -SPEED*SIN(RA)*COS(DEC)
!      VA(3) = Z = -SPEED*SIN(DEC)

DATA va / -108.70408, +97.86251, -164.33610 /



!  Convert given J2000 RA,Dec to x,y,z
CALL sla_cs2c(r2000,d2000,vb)

!  Compute dot product with LSR motion vector
sla_rvgalc=sla_vdv(va,vb)

END FUNCTION sla_rvgalc
REAL FUNCTION sla_rvlg (r2000, d2000)
!+
!     - - - - -
!      R V L G
!     - - - - -

!  Velocity component in a given direction due to the combination
!  of the rotation of the Galaxy and the motion of the Galaxy
!  relative to the mean motion of the local group (single precision)

!  Given:
!     R2000,D2000   real    J2000.0 mean RA,Dec (radians)

!  Result:
!     Component of SOLAR motion in direction R2000,D2000 (km/s)

!  Sign convention:
!     The result is +ve when the Sun is receding from the
!     given point on the sky.

!  Reference:
!     IAU Trans 1976, 168, p201.

!  Called:
!     sla_CS2C, sla_VDV

!  P.T.Wallace   Starlink   June 1985

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: r2000,d2000

REAL :: va(3), vb(3)



!  Solar velocity due to Galactic rotation and translation

!  Speed = 300 km/s

!  Apex  = L2,B2  90deg, 0deg
!        = RA,Dec  21 12 01.1  +48 19 47  J2000.0

!  This is expressed in the form of a J2000.0 x,y,z vector:

!      VA(1) = X = -SPEED*COS(RA)*COS(DEC)
!      VA(2) = Y = -SPEED*SIN(RA)*COS(DEC)
!      VA(3) = Z = -SPEED*SIN(DEC)

DATA va / -148.23284, +133.44888, -224.09467 /



!  Convert given J2000 RA,Dec to x,y,z
CALL sla_cs2c(r2000,d2000,vb)

!  Compute dot product with Solar motion vector
sla_rvlg=sla_vdv(va,vb)

END FUNCTION sla_rvlg
REAL FUNCTION sla_rvlsrd (r2000, d2000)
!+
!     - - - - - - -
!      R V L S R D
!     - - - - - - -

!  Velocity component in a given direction due to the Sun's motion
!  with respect to the dynamical Local Standard of Rest.

!  (single precision)

!  Given:
!     R2000,D2000   r    J2000.0 mean RA,Dec (radians)

!  Result:
!     Component of "peculiar" solar motion in direction R2000,D2000 (km/s)

!  Sign convention:
!     The result is +ve when the Sun is receding from the given point on
!     the sky.

!  Note:  The Local Standard of Rest used here is the "dynamical" LSR,
!         a point in the vicinity of the Sun which is in a circular
!         orbit around the Galactic centre.  The Sun's motion with
!         respect to the dynamical LSR is called the "peculiar" solar
!         motion.

!         There is another type of LSR, called a "kinematical" LSR.  A
!         kinematical LSR is the mean standard of rest of specified star
!         catalogues or stellar populations, and several slightly
!         different kinematical LSRs are in use.  The Sun's motion with
!         respect to an agreed kinematical LSR is known as the "standard"
!         solar motion.  To obtain a radial velocity correction with
!         respect to an adopted kinematical LSR use the routine sla_RVLSRK.

!  Reference:  Delhaye (1965), in "Stars and Stellar Systems", vol 5,
!              p73.

!  Called:
!     sla_CS2C, sla_VDV

!  P.T.Wallace   Starlink   9 March 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: r2000,d2000

REAL :: va(3), vb(3)



!  Peculiar solar motion from Delhaye 1965: in Galactic Cartesian
!  coordinates (+9,+12,+7) km/s.  This corresponds to about 16.6 km/s
!  towards Galactic coordinates L2 = 53 deg, B2 = +25 deg, or RA,Dec
!  17 49 58.7 +28 07 04 J2000.

!  The solar motion is expressed here in the form of a J2000.0
!  equatorial Cartesian vector:

!      VA(1) = X = -SPEED*COS(RA)*COS(DEC)
!      VA(2) = Y = -SPEED*SIN(RA)*COS(DEC)
!      VA(3) = Z = -SPEED*SIN(DEC)

DATA va / +0.63823, +14.58542, -7.80116 /



!  Convert given J2000 RA,Dec to x,y,z
CALL sla_cs2c(r2000,d2000,vb)

!  Compute dot product with solar motion vector
sla_rvlsrd=sla_vdv(va,vb)

END FUNCTION sla_rvlsrd
REAL FUNCTION sla_rvlsrk (r2000, d2000)
!+
!     - - - - - - -
!      R V L S R K
!     - - - - - - -

!  Velocity component in a given direction due to the Sun's motion
!  with respect to an adopted kinematic Local Standard of Rest.

!  (single precision)

!  Given:
!     R2000,D2000   r    J2000.0 mean RA,Dec (radians)

!  Result:
!     Component of "standard" solar motion in direction R2000,D2000 (km/s)

!  Sign convention:
!     The result is +ve when the Sun is receding from the given point on
!     the sky.

!  Note:  The Local Standard of Rest used here is one of several
!         "kinematical" LSRs in common use.  A kinematical LSR is the
!         mean standard of rest of specified star catalogues or stellar
!         populations.  The Sun's motion with respect to a kinematical
!         LSR is known as the "standard" solar motion.

!         There is another sort of LSR, the "dynamical" LSR, which is a
!         point in the vicinity of the Sun which is in a circular orbit
!         around the Galactic centre.  The Sun's motion with respect to
!         the dynamical LSR is called the "peculiar" solar motion.  To
!         obtain a radial velocity correction with respect to the
!         dynamical LSR use the routine sla_RVLSRD.

!  Reference:  Delhaye (1965), in "Stars and Stellar Systems", vol 5,
!              p73.

!  Called:
!     sla_CS2C, sla_VDV

!  P.T.Wallace   Starlink   11 March 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: r2000,d2000

REAL :: va(3), vb(3)



!  Standard solar motion (from Methods of Experimental Physics, ed Meeks,
!  vol 12, part C, sec 6.1.5.2, p281):

!  20 km/s towards RA 18h Dec +30d (1900).

!  The solar motion is expressed here in the form of a J2000.0
!  equatorial Cartesian vector:

!      VA(1) = X = -SPEED*COS(RA)*COS(DEC)
!      VA(2) = Y = -SPEED*SIN(RA)*COS(DEC)
!      VA(3) = Z = -SPEED*SIN(DEC)

DATA va / -0.29000, +17.31726, -10.00141 /



!  Convert given J2000 RA,Dec to x,y,z
CALL sla_cs2c(r2000,d2000,vb)

!  Compute dot product with solar motion vector
sla_rvlsrk=sla_vdv(va,vb)

END FUNCTION sla_rvlsrk
SUBROUTINE sla_s2tp (ra, dec, raz, decz, xi, eta, j)
!+
!     - - - - -
!      S 2 T P
!     - - - - -

!  Projection of spherical coordinates onto tangent plane:
!  "gnomonic" projection - "standard coordinates"
!  (single precision)

!  Given:
!     RA,DEC      real  spherical coordinates of point to be projected
!     RAZ,DECZ    real  spherical coordinates of tangent point

!  Returned:
!     XI,ETA      real  rectangular coordinates on tangent plane
!     J           int   status:   0 = OK, star on tangent plane
!                                 1 = error, star too far from axis
!                                 2 = error, antistar on tangent plane
!                                 3 = error, antistar too far from axis

!  P.T.Wallace   Starlink   18 July 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: ra,dec,raz,decz,xi,eta
INTEGER :: j

REAL :: sdecz,sdec,cdecz,cdec,radif,sradif,cradif,denom

REAL :: tiny
PARAMETER (tiny=1E-6)


!  Trig functions
sdecz=SIN(decz)
sdec=SIN(dec)
cdecz=COS(decz)
cdec=COS(dec)
radif=ra-raz
sradif=SIN(radif)
cradif=COS(radif)

!  Reciprocal of star vector length to tangent plane
denom=sdec*sdecz+cdec*cdecz*cradif

!  Handle vectors too far from axis
IF (denom > tiny) THEN
  j=0
ELSE IF (denom >= 0.0) THEN
  j=1
  denom=tiny
ELSE IF (denom > -tiny) THEN
  j=2
  denom=-tiny
ELSE
  j=3
END IF

!  Compute tangent plane coordinates (even in dubious cases)
xi=cdec*sradif/denom
eta=(sdec*cdecz-cdec*sdecz*cradif)/denom

END SUBROUTINE sla_s2tp
REAL FUNCTION sla_sep (a1, b1, a2, b2)
!+
!     - - - -
!      S E P
!     - - - -

!  Angle between two points on a sphere (single precision)

!  Given:
!     A1,B1    real    spherical coordinates of one point
!     A2,B2    real    spherical coordinates of the other point

!  (The spherical coordinates are RA,Dec, Long,Lat etc, in radians.)

!  The result is the angle, in radians, between the two points.  It
!  is always positive.

!  Called:  sla_CS2C

!  P.T.Wallace   Starlink   April 1985

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: a1,b1,a2,b2

INTEGER :: i
REAL :: v1(3),v2(3),w



!  Convert coordinates from spherical to Cartesian
CALL sla_cs2c(a1,b1,v1)
CALL sla_cs2c(a2,b2,v2)

!  Modulus squared of half the difference vector
w=0.0
DO i=1,3
  w=w+(v1(i)-v2(i))**2
END DO
w=w/4.0

!  Angle between the vectors
sla_sep=2.0*ATAN2(SQRT(w),SQRT(MAX(0.0,1.0-w)))

END FUNCTION sla_sep
SUBROUTINE sla_smat (n, a, y, d, jf, iw)
!+
!     - - - - -
!      S M A T
!     - - - - -

!  Matrix inversion & solution of simultaneous equations
!  (single precision)

!  For the set of n simultaneous equations in n unknowns:
!     A.Y = X

!  where:
!     A is a non-singular N x N matrix
!     Y is the vector of N unknowns
!     X is the known vector

!  SMATRX computes:
!     the inverse of matrix A
!     the determinant of matrix A
!     the vector of N unknowns

!  Arguments:

!     symbol  type dimension           before              after

!       N      int                 no. of unknowns       unchanged
!       A      real  (N,N)             matrix             inverse
!       Y      real   (N)              vector            solution
!       D      real                       -             determinant
!     * JF     int                        -           singularity flag
!       IW     int    (N)                 -              workspace

!  *  JF is the singularity flag.  If the matrix is non-singular,
!    JF=0 is returned.  If the matrix is singular, JF=-1 & D=0.0 are
!    returned.  In the latter case, the contents of array A on return
!    are undefined.

!  Algorithm:
!     Gaussian elimination with partial pivoting.

!  Speed:
!     Very fast.

!  Accuracy:
!     Fairly accurate - errors 1 to 4 times those of routines optimised
!     for accuracy.

!  Note:  replaces the obsolete sla_SMATRX routine.

!  P.T.Wallace   Starlink   10 September 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: n
REAL :: a(n,n),y(n),d
INTEGER :: jf
INTEGER :: iw(n)

REAL :: sfa
PARAMETER (sfa=1E-20)

INTEGER :: k,imx,i,j,np1mk,ki
REAL :: amx,t,akk,yk,aik


jf=0
d=1.0
DO k=1,n
  amx=ABS(a(k,k))
  imx=k
  IF (k /= n) THEN
    DO i=k+1,n
      t=ABS(a(i,k))
      IF (t > amx) THEN
        amx=t
        imx=i
      END IF
    END DO
  END IF
  IF (amx < sfa) THEN
    jf=-1
  ELSE
    IF (imx /= k) THEN
      DO j=1,n
        t=a(k,j)
        a(k,j)=a(imx,j)
        a(imx,j)=t
      END DO
      t=y(k)
      y(k)=y(imx)
      y(imx)=t
      d=-d
    END IF
    iw(k)=imx
    akk=a(k,k)
    d=d*akk
    IF (ABS(d) < sfa) THEN
      jf=-1
    ELSE
      akk=1.0/akk
      a(k,k)=akk
      DO j=1,n
        IF (j /= k) a(k,j)=a(k,j)*akk
      END DO
      yk=y(k)*akk
      y(k)=yk
      DO i=1,n
        aik=a(i,k)
        IF (i /= k) THEN
          DO j=1,n
            IF (j /= k) a(i,j)=a(i,j)-aik*a(k,j)
          END DO
          y(i)=y(i)-aik*yk
        END IF
      END DO
      DO i=1,n
        IF (i /= k) a(i,k)=-a(i,k)*akk
      END DO
    END IF
  END IF
END DO
IF (jf /= 0) THEN
  d=0.0
ELSE
  DO k=1,n
    np1mk=n+1-k
    ki=iw(np1mk)
    IF (np1mk /= ki) THEN
      DO i=1,n
        t=a(i,np1mk)
        a(i,np1mk)=a(i,ki)
        a(i,ki)=t
      END DO
    END IF
  END DO
END IF
END SUBROUTINE sla_smat
SUBROUTINE sla_subet (rc, dc, EQ, rm, dm)
!+
!     - - - - - -
!      S U B E T
!     - - - - - -

!  Remove the E-terms (elliptic component of annual aberration)
!  from a pre IAU 1976 catalogue RA,Dec to give a mean place
!  (double precision)

!  Given:
!     RC,DC     dp     RA,Dec (radians) with E-terms included
!     EQ        dp     Besselian epoch of mean equator and equinox

!  Returned:
!     RM,DM     dp     RA,Dec (radians) without E-terms

!  Called:
!     sla_ETRMS, sla_DCS2C, sla_,DVDV, sla_DCC2S, sla_DRANRM

!  Explanation:
!     Most star positions from pre-1984 optical catalogues (or
!     derived from astrometry using such stars) embody the
!     E-terms.  This routine converts such a position to a
!     formal mean place (allowing, for example, comparison with a
!     pulsar timing position).

!  Reference:
!     Explanatory Supplement to the Astronomical Ephemeris,
!     section 2D, page 48.

!  P.T.Wallace   Starlink   10 May 1990

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: rc,dc,EQ,rm,dm

DOUBLE PRECISION :: a(3),v(3),f

INTEGER :: i



!  E-terms
CALL sla_etrms(EQ,a)

!  Spherical to Cartesian
CALL sla_dcs2c(rc,dc,v)

!  Include the E-terms
f=1D0+sla_dvdv(v,a)
DO i=1,3
  v(i)=f*v(i)-a(i)
END DO

!  Cartesian to spherical
CALL sla_dcc2s(v,rm,dm)

!  Bring RA into conventional range
rm=sla_dranrm(rm)

END SUBROUTINE sla_subet
SUBROUTINE sla_supgal (dsl, dsb, dl, db)
!+
!     - - - - - - -
!      S U P G A L
!     - - - - - - -

!  Transformation from de Vaucouleurs supergalactic coordinates
!  to IAU 1958 galactic coordinates (double precision)

!  Given:
!     DSL,DSB     dp       supergalactic longitude and latitude

!  Returned:
!     DL,DB       dp       galactic longitude and latitude L2,B2

!  (all arguments are radians)

!  Called:
!     sla_DCS2C, sla_DIMXV, sla_DCC2S, sla_DRANRM, sla_DRANGE

!  References:

!     de Vaucouleurs, de Vaucouleurs, & Corwin, Second Reference
!     Catalogue of Bright Galaxies, U. Texas, page 8.

!     Systems & Applied Sciences Corp., Documentation for the
!     machine-readable version of the above catalogue,
!     Contract NAS 5-26490.

!    (These two references give different values for the galactic
!     longitude of the supergalactic origin.  Both are wrong;  the
!     correct value is L2=137.37.)

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: dsl,dsb,dl,db


DOUBLE PRECISION :: v1(3),v2(3)


!  System of supergalactic coordinates:

!    SGL   SGB        L2     B2      (deg)
!     -    +90      47.37  +6.32
!     0     0         -      0

!  Galactic to supergalactic rotation matrix:

DOUBLE PRECISION :: rmat(3,3)
DATA rmat(1,1),rmat(1,2),rmat(1,3),rmat(2,1),rmat(2,2),rmat(2,3),  &
rmat(3,1),rmat(3,2),rmat(3,3)/  &
-0.735742574804D0,+0.677261296414D0,+0.000000000000D0,  &
-0.074553778365D0,-0.080991471307D0,+0.993922590400D0,  &
+0.673145302109D0,+0.731271165817D0,+0.110081262225D0/



!  Spherical to Cartesian
CALL sla_dcs2c(dsl,dsb,v1)

!  Supergalactic to galactic
CALL sla_dimxv(rmat,v1,v2)

!  Cartesian to spherical
CALL sla_dcc2s(v2,dl,db)

!  Express in conventional ranges
dl=sla_dranrm(dl)
db=sla_drange(db)

END SUBROUTINE sla_supgal
SUBROUTINE sla_svd (m, n, mp, np, a, w, v, work, jstat)
!+
!     - - - -
!      S V D
!     - - - -

!  Singular value decomposition  (double precision)

!  This routine expresses a given matrix A as the product of
!  three matrices U, W, V:

!     A = U x W x VT

!  Where:

!     A   is any M (rows) x N (columns) matrix, where M.GE.N
!     U   is an M x N column-orthogonal matrix
!     W   is an N x N diagonal matrix with W(I,I).GE.0
!     VT  is the transpose of an N x N orthogonal matrix

!     Note that M and N, above, are the LOGICAL dimensions of the
!     matrices and vectors concerned, which can be located in
!     arrays of larger PHYSICAL dimensions, given by MP and NP.

!  Given:
!     M,N    i         numbers of rows and columns in matrix A
!     MP,NP  i         physical dimensions of array containing matrix A
!     A      d(MP,NP)  array containing MxN matrix A

!  Returned:
!     A      d(MP,NP)  array containing MxN column-orthogonal matrix U
!     W      d(N)      NxN diagonal matrix W (diagonal elements only)
!     V      d(NP,NP)  array containing NxN orthogonal matrix V
!     WORK   d(N)      workspace
!     JSTAT  i         0 = OK, -1 = A wrong shape, >0 = index of W
!                      for which convergence failed.  See note 2, below.

!   Notes:

!   1)  V contains matrix V, not the transpose of matrix V.

!   2)  If the status JSTAT is greater than zero, this need not
!       necessarily be treated as a failure.  It means that, due to
!       chance properties of the matrix A, the QR transformation
!       phase of the routine did not fully converge in a predefined
!       number of iterations, something that very seldom occurs.
!       When this condition does arise, it is possible that the
!       elements of the diagonal matrix W have not been correctly
!       found.  However, in practice the results are likely to
!       be trustworthy.  Applications should report the condition
!       as a warning, but then proceed normally.

!  References:
!     The algorithm is an adaptation of the routine SVD in the EISPACK
!     library (Garbow et al 1977, EISPACK Guide Extension, Springer
!     Verlag), which is a FORTRAN 66 implementation of the Algol
!     routine SVD of Wilkinson & Reinsch 1971 (Handbook for Automatic
!     Computation, vol 2, ed Bauer et al, Springer Verlag).  These
!     references give full details of the algorithm used here.  A good
!     account of the use of SVD in least squares problems is given in
!     Numerical Recipes (Press et al 1986, Cambridge University Press),
!     which includes another variant of the EISPACK code.

!  P.T.Wallace   Starlink   22 December 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: m,n,mp,np
DOUBLE PRECISION :: a(mp,np),w(n),v(np,np),work(n)
INTEGER :: jstat

!  Maximum number of iterations in QR phase
INTEGER :: itmax
PARAMETER (itmax=30)

INTEGER :: i,k,l,j,k1,its,l1,i1
LOGICAL :: cancel
DOUBLE PRECISION :: g,scale,an,s,x,f,h,c,y,z



!  Check that the matrix is the right shape
IF (m < n) THEN
  
!     No:  error status
  jstat = -1
  
ELSE
  
!     Yes:  preset the status to OK
  jstat = 0
  
  
!     Householder reduction to bidiagonal form
!     ----------------------------------------
  
  g = 0D0
  scale = 0D0
  an = 0D0
  DO i=1,n
    l = i+1
    work(i) = scale*g
    g = 0D0
    s = 0D0
    scale = 0D0
    IF (i <= m) THEN
      DO k=i,m
        scale = scale+ABS(a(k,i))
      END DO
      IF (scale /= 0D0) THEN
        DO k=i,m
          x = a(k,i)/scale
          a(k,i) = x
          s = s+x*x
        END DO
        f = a(i,i)
        g = -SIGN(SQRT(s),f)
        h = f*g-s
        a(i,i) = f-g
        IF (i /= n) THEN
          DO j=l,n
            s = 0D0
            DO k=i,m
              s = s+a(k,i)*a(k,j)
            END DO
            f = s/h
            DO k=i,m
              a(k,j) = a(k,j)+f*a(k,i)
            END DO
          END DO
        END IF
        DO k=i,m
          a(k,i) = scale*a(k,i)
        END DO
      END IF
    END IF
    w(i) = scale*g
    g = 0D0
    s = 0D0
    scale = 0D0
    IF (i <= m .AND. i /= n) THEN
      DO k=l,n
        scale = scale+ABS(a(i,k))
      END DO
      IF (scale /= 0D0) THEN
        DO k=l,n
          x = a(i,k)/scale
          a(i,k) = x
          s = s+x*x
        END DO
        f = a(i,l)
        g = -SIGN(SQRT(s),f)
        h = f*g-s
        a(i,l) = f-g
        DO k=l,n
          work(k) = a(i,k)/h
        END DO
        IF (i /= m) THEN
          DO j=l,m
            s = 0D0
            DO k=l,n
              s = s+a(j,k)*a(i,k)
            END DO
            DO k=l,n
              a(j,k) = a(j,k)+s*work(k)
            END DO
          END DO
        END IF
        DO k=l,n
          a(i,k) = scale*a(i,k)
        END DO
      END IF
    END IF
    
!        Overestimate of largest column norm for convergence test
    an = MAX(an,ABS(w(i))+ABS(work(i)))
    
  END DO
  
  
!     Accumulation of right-hand transformations
!     ------------------------------------------
  
  DO i=n,1,-1
    IF (i /= n) THEN
      IF (g /= 0D0) THEN
        DO j=l,n
          v(j,i) = (a(i,j)/a(i,l))/g
        END DO
        DO j=l,n
          s = 0D0
          DO k=l,n
            s = s+a(i,k)*v(k,j)
          END DO
          DO k=l,n
            v(k,j) = v(k,j)+s*v(k,i)
          END DO
        END DO
      END IF
      DO j=l,n
        v(i,j) = 0D0
        v(j,i) = 0D0
      END DO
    END IF
    v(i,i) = 1D0
    g = work(i)
    l = i
  END DO
  
  
!     Accumulation of left-hand transformations
!     -----------------------------------------
  
  DO i=n,1,-1
    l = i+1
    g = w(i)
    IF (i /= n) THEN
      DO j=l,n
        a(i,j) = 0D0
      END DO
    END IF
    IF (g /= 0D0) THEN
      IF (i /= n) THEN
        DO j=l,n
          s = 0D0
          DO k=l,m
            s = s+a(k,i)*a(k,j)
          END DO
          f = (s/a(i,i))/g
          DO k=i,m
            a(k,j) = a(k,j)+f*a(k,i)
          END DO
        END DO
      END IF
      DO j=i,m
        a(j,i) = a(j,i)/g
      END DO
    ELSE
      DO j=i,m
        a(j,i) = 0D0
      END DO
    END IF
    a(i,i) = a(i,i)+1D0
  END DO
  
  
!     Diagonalisation of the bidiagonal form
!     --------------------------------------
  
  DO k=n,1,-1
    k1 = k-1
    
!        Iterate until converged
    its = 0
    DO WHILE (its < itmax)
      its = its+1
      
!           Test for splitting into submatrices
      cancel = .true.
      DO l=k,1,-1
        l1 = l-1
        IF (an+ABS(work(l)) == an) THEN
          cancel = .false.
          GO TO 10
        END IF
!              (Following never attempted for L=1 because WORK(1) is zero)
        IF (an+ABS(w(l1)) == an) GO TO 10
      END DO
      10            CONTINUE
      
!           Cancellation of WORK(L) if L>1
      IF (cancel) THEN
        c = 0D0
        s = 1D0
        DO i=l,k
          f = s*work(i)
          IF (an+ABS(f) == an) GO TO 20
          g = w(i)
          h = SQRT(f*f+g*g)
          w(i) = h
          c = g/h
          s = -f/h
          DO j=1,m
            y = a(j,l1)
            z = a(j,i)
            a(j,l1) = y*c+z*s
            a(j,i) = -y*s+z*c
          END DO
        END DO
        20               CONTINUE
      END IF
      
!           Converged?
      z = w(k)
      IF (l == k) THEN
        
!              Yes:  stop iterating
        its = itmax
        
!              Ensure singular values non-negative
        IF (z < 0D0) THEN
          w(k) = -z
          DO j=1,n
            v(j,k) = -v(j,k)
          END DO
        END IF
      ELSE
        
!              Not converged yet:  set status if iteration limit reached
        IF (its == itmax) jstat = k
        
!              Shift from bottom 2x2 minor
        x = w(l)
        y = w(k1)
        g = work(k1)
        h = work(k)
        f = ((y-z)*(y+z)+(g-h)*(g+h))/(2D0*h*y)
        IF (ABS(f) <= 1D15) THEN
          g = SQRT(f*f+1D0)
        ELSE
          g = ABS(f)
        END IF
        f = ((x-z)*(x+z)+h*(y/(f+SIGN(g,f))-h))/x
        
!              Next QR transformation
        c = 1D0
        s = 1D0
        DO i1=l,k1
          i = i1+1
          g = work(i)
          y = w(i)
          h = s*g
          g = c*g
          z = SQRT(f*f+h*h)
          work(i1) = z
          IF (z /= 0D0) THEN
            c = f/z
            s = h/z
          ELSE
            c = 1D0
            s = 0D0
          END IF
          f = x*c+g*s
          g = -x*s+g*c
          h = y*s
          y = y*c
          DO j=1,n
            x = v(j,i1)
            z = v(j,i)
            v(j,i1) = x*c+z*s
            v(j,i) = -x*s+z*c
          END DO
          z = SQRT(f*f+h*h)
          w(i1) = z
          IF (z /= 0D0) THEN
            c = f/z
            s = h/z
          END IF
          f = c*g+s*y
          x = -s*g+c*y
          DO j=1,m
            y = a(j,i1)
            z = a(j,i)
            a(j,i1) = y*c+z*s
            a(j,i) = -y*s+z*c
          END DO
        END DO
        work(l) = 0D0
        work(k) = f
        w(k) = x
      END IF
    END DO
  END DO
END IF

END SUBROUTINE sla_svd
SUBROUTINE sla_svdcov (n, np, nc, w, v, work, cvm)
!+
!     - - - - - - -
!      S V D C O V
!     - - - - - - -

!  From the W and V matrices from the SVD factorisation of a matrix
!  (as obtained from the sla_SVD routine), obtain the covariance matrix.

!  (double precision)

!  Given:
!     N      i         number of rows and columns in matrices W and V
!     NP     i         first dimension of array containing matrix V
!     NC     i         first dimension of array to receive CVM
!     W      d(N)      NxN diagonal matrix W (diagonal elements only)
!     V      d(NP,NP)  array containing NxN orthogonal matrix V

!  Returned:
!     WORK   d(N)      workspace
!     CVM    d(NC,NC)  array to receive covariance matrix

!  Reference:
!     Numerical Recipes, section 14.3.

!  P.T.Wallace   Starlink   December 1988

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: n,np,nc
DOUBLE PRECISION :: w(n),v(np,np),work(n),cvm(nc,nc)

INTEGER :: i,j,k
DOUBLE PRECISION :: s



DO i=1,n
  s=w(i)
  IF (s /= 0D0) THEN
    work(i)=1D0/(s*s)
  ELSE
    work(i)=0D0
  END IF
END DO
DO i=1,n
  DO j=1,i
    s=0D0
    DO k=1,n
      s=s+v(i,k)*v(j,k)*work(k)
    END DO
    cvm(i,j)=s
    cvm(j,i)=s
  END DO
END DO

END SUBROUTINE sla_svdcov
SUBROUTINE sla_svdsol (m, n, mp, np, b, u, w, v, work, x)
!+
!     - - - - - - -
!      S V D S O L
!     - - - - - - -

!  From a given vector and the SVD of a matrix (as obtained from
!  the SVD routine), obtain the solution vector (double precision)

!  This routine solves the equation:

!     A . x = b

!  where:

!     A   is a given M (rows) x N (columns) matrix, where M.GE.N
!     x   is the N-vector we wish to find
!     b   is a given M-vector

!  by means of the Singular Value Decomposition method (SVD).  In
!  this method, the matrix A is first factorised (for example by
!  the routine sla_SVD) into the following components:

!     A = U x W x VT

!  where:

!     A   is the M (rows) x N (columns) matrix
!     U   is an M x N column-orthogonal matrix
!     W   is an N x N diagonal matrix with W(I,I).GE.0
!     VT  is the transpose of an NxN orthogonal matrix

!     Note that M and N, above, are the LOGICAL dimensions of the
!     matrices and vectors concerned, which can be located in
!     arrays of larger PHYSICAL dimensions MP and NP.

!  The solution is found from the expression:

!     x = V . [diag(1/Wj)] . (transpose(U) . b)

!  Notes:

!  1)  If matrix A is square, and if the diagonal matrix W is not
!      adjusted, the method is equivalent to conventional solution
!      of simultaneous equations.

!  2)  If M>N, the result is a least-squares fit.

!  3)  If the solution is poorly determined, this shows up in the
!      SVD factorisation as very small or zero Wj values.  Where
!      a Wj value is small but non-zero it can be set to zero to
!      avoid ill effects.  The present routine detects such zero
!      Wj values and produces a sensible solution, with highly
!      correlated terms kept under control rather than being allowed
!      to elope to infinity, and with meaningful values for the
!      other terms.

!  Given:
!     M,N    i         numbers of rows and columns in matrix A
!     MP,NP  i         physical dimensions of array containing matrix A
!     B      d(M)      known vector b
!     U      d(MP,NP)  array containing MxN matrix U
!     W      d(N)      NxN diagonal matrix W (diagonal elements only)
!     V      d(NP,NP)  array containing NxN orthogonal matrix V

!  Returned:
!     WORK   d(N)      workspace
!     X      d(N)      unknown vector x

!  Reference:
!     Numerical Recipes, section 2.9.

!  P.T.Wallace   Starlink   29 October 1993

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

INTEGER :: m,n,mp,np
DOUBLE PRECISION :: b(m),u(mp,np),w(n),v(np,np),work(n),x(n)

INTEGER :: j,i,jj
DOUBLE PRECISION :: s



!  Calculate [diag(1/Wj)] . transpose(U) . b (or zero for zero Wj)
DO j=1,n
  s=0D0
  IF (w(j) /= 0D0) THEN
    DO i=1,m
      s=s+u(i,j)*b(i)
    END DO
    s=s/w(j)
  END IF
  work(j)=s
END DO

!  Multiply by matrix V to get result
DO j=1,n
  s=0D0
  DO jj=1,n
    s=s+v(j,jj)*work(jj)
  END DO
  x(j)=s
END DO

END SUBROUTINE sla_svdsol
SUBROUTINE sla_tp2s (xi, eta, raz, decz, ra, dec)
!+
!     - - - - -
!      T P 2 S
!     - - - - -

!  Transform tangent plane coordinates into spherical
!  (single precision)

!  Given:
!     XI,ETA      real  tangent plane rectangular coordinates
!     RAZ,DECZ    real  spherical coordinates of tangent point

!  Returned:
!     RA,DEC      real  spherical coordinates (0-2pi,+/-pi/2)

!  Called:        sla_RANORM

!  P.T.Wallace   Starlink   24 July 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: xi,eta,raz,decz,ra,dec


REAL :: sdecz,cdecz,denom



sdecz=SIN(decz)
cdecz=COS(decz)

denom=cdecz-eta*sdecz

ra=sla_ranorm(ATAN2(xi,denom)+raz)
dec=ATAN2(sdecz+eta*cdecz,SQRT(xi*xi+denom*denom))

END SUBROUTINE sla_tp2s
SUBROUTINE sla_tp2v (xi, eta, v0, v)
!+
!     - - - - -
!      T P 2 V
!     - - - - -

!  Given the tangent-plane coordinates of a star and the direction
!  cosines of the tangent point, determine the direction cosines
!  of the star.

!  (single precision)

!  Given:
!     XI,ETA    r       tangent plane coordinates of star
!     V0        r(3)    direction cosines of tangent point

!  Returned:
!     V         r(3)    direction cosines of star

!  Notes:

!  1  If vector V0 is not of unit length, the returned vector V will
!     be wrong.

!  2  If vector V0 points at a pole, the returned vector V will be
!     based on the arbitrary assumption that the RA of the tangent
!     point is zero.

!  3  This routine is the Cartesian equivalent of the routine sla_TP2S.

!  P.T.Wallace   Starlink   11 February 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: xi,eta,v0(3),v(3)

REAL :: x,y,z,f,r


x=v0(1)
y=v0(2)
z=v0(3)
f=SQRT(1.0+xi*xi+eta*eta)
r=SQRT(x*x+y*y)
IF (r == 0.0) THEN
  r=1E-20
  x=r
END IF
v(1)=(x-(xi*y+eta*x*z)/r)/f
v(2)=(y+(xi*x-eta*y*z)/r)/f
v(3)=(z+eta*r)/f

END SUBROUTINE sla_tp2v
SUBROUTINE sla_tps2c (xi, eta, ra, dec, raz1, decz1,raz2, decz2, n)
!+
!     - - - - - -
!      T P S 2 C
!     - - - - - -

!  From the tangent plane coordinates of a star of known RA,Dec,
!  determine the RA,Dec of the tangent point.

!  (single precision)

!  Given:
!     XI,ETA      r    tangent plane rectangular coordinates
!     RA,DEC      r    spherical coordinates

!  Returned:
!     RAZ1,DECZ1  r    spherical coordinates of tangent point, solution 1
!     RAZ2,DECZ2  r    spherical coordinates of tangent point, solution 2
!     N           i    number of solutions:
!                        0 = no solutions returned (note 2)
!                        1 = only the first solution is useful (note 3)
!                        2 = both solutions are useful (note 3)

!  Notes:

!  1  The RAZ1 and RAZ2 values are returned in the range 0-2pi.

!  2  Cases where there is no solution can only arise near the poles.
!     For example, it is clearly impossible for a star at the pole
!     itself to have a non-zero XI value, and hence it is
!     meaningless to ask where the tangent point would have to be
!     to bring about this combination of XI and DEC.

!  3  Also near the poles, cases can arise where there are two useful
!     solutions.  The argument N indicates whether the second of the
!     two solutions returned is useful.  N=1 indicates only one useful
!     solution, the usual case;  under these circumstances, the second
!     solution corresponds to the "over-the-pole" case, and this is
!     reflected in the values of RAZ2 and DECZ2 which are returned.

!  4  The DECZ1 and DECZ2 values are returned in the range +/-pi, but
!     in the usual, non-pole-crossing, case, the range is +/-pi/2.

!  5  This routine is the spherical equivalent of the routine sla_DTPV2C.

!  Called:  sla_RANORM

!  P.T.Wallace   Starlink   5 June 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: xi,eta,ra,dec,raz1,decz1,raz2,decz2
INTEGER :: n

REAL :: x2,y2,sd,cd,sdf,r2,r,s,c



x2=xi*xi
y2=eta*eta
sd=SIN(dec)
cd=COS(dec)
sdf=sd*SQRT(1.0+x2+y2)
r2=cd*cd*(1.0+y2)-sd*sd*x2
IF (r2 >= 0.0) THEN
  r=SQRT(r2)
  s=sdf-eta*r
  c=sdf*eta+r
  IF (xi == 0.0.AND.r == 0.0) r=1.0
  raz1=sla_ranorm(ra-ATAN2(xi,r))
  decz1=ATAN2(s,c)
  r=-r
  s=sdf-eta*r
  c=sdf*eta+r
  raz2=sla_ranorm(ra-ATAN2(xi,r))
  decz2=ATAN2(s,c)
  IF (ABS(sdf) < 1.0) THEN
    n=1
  ELSE
    n=2
  END IF
ELSE
  n=0
END IF

END SUBROUTINE sla_tps2c
SUBROUTINE sla_tpv2c (xi, eta, v, v01, v02, n)
!+
!     - - - - - -
!      T P V 2 C
!     - - - - - -

!  Given the tangent-plane coordinates of a star and its direction
!  cosines, determine the direction cosines of the tangent-point.

!  (single precision)

!  Given:
!     XI,ETA    r       tangent plane coordinates of star
!     V         r(3)    direction cosines of star

!  Returned:
!     V01       r(3)    direction cosines of tangent point, solution 1
!     V02       r(3)    direction cosines of tangent point, solution 2
!     N         i       number of solutions:
!                         0 = no solutions returned (note 2)
!                         1 = only the first solution is useful (note 3)
!                         2 = both solutions are useful (note 3)

!  Notes:

!  1  The vector V must be of unit length or the result will be wrong.

!  2  Cases where there is no solution can only arise near the poles.
!     For example, it is clearly impossible for a star at the pole
!     itself to have a non-zero XI value, and hence it is meaningless
!     to ask where the tangent point would have to be.

!  3  Also near the poles, cases can arise where there are two useful
!     solutions.  The argument N indicates whether the second of the
!     two solutions returned is useful.  N=1 indicates only one useful
!     solution, the usual case;  under these circumstances, the second
!     solution can be regarded as valid if the vector V02 is interpreted
!     as the "over-the-pole" case.

!  4  This routine is the Cartesian equivalent of the routine sla_TPS2C.

!  P.T.Wallace   Starlink   5 June 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: xi,eta,v(3),v01(3),v02(3)
INTEGER :: n

REAL :: x,y,z,rxy2,xi2,eta2p1,sdf,r2,r,c


x=v(1)
y=v(2)
z=v(3)
rxy2=x*x+y*y
xi2=xi*xi
eta2p1=eta*eta+1.0
sdf=z*SQRT(xi2+eta2p1)
r2=rxy2*eta2p1-z*z*xi2
IF (r2 > 0.0) THEN
  r=SQRT(r2)
  c=(sdf*eta+r)/(eta2p1*SQRT(rxy2*(r2+xi2)))
  v01(1)=c*(x*r+y*xi)
  v01(2)=c*(y*r-x*xi)
  v01(3)=(sdf-eta*r)/eta2p1
  r=-r
  c=(sdf*eta+r)/(eta2p1*SQRT(rxy2*(r2+xi2)))
  v02(1)=c*(x*r+y*xi)
  v02(2)=c*(y*r-x*xi)
  v02(3)=(sdf-eta*r)/eta2p1
  IF (ABS(sdf) < 1.0) THEN
    n=1
  ELSE
    n=2
  END IF
ELSE
  n=0
END IF

END SUBROUTINE sla_tpv2c
SUBROUTINE sla_unpcd (disco,x,y)
!+
!     - - - - - -
!      U N P C D
!     - - - - - -

!  Remove pincushion/barrel distortion from a distorted [x,y]
!  to give tangent-plane [x,y].

!  Given:
!     DISCO    d      pincushion/barrel distortion coefficient
!     X,Y      d      distorted coordinates

!  Returned:
!     X,Y      d      tangent-plane coordinates

!  Notes:

!  1)  The distortion is of the form RP = R*(1 + C*R**2), where R is
!      the radial distance from the tangent point, C is the DISCO
!      argument, and RP is the radial distance in the presence of
!      the distortion.

!  2)  For pincushion distortion, C is +ve;  for barrel distortion,
!      C is -ve.

!  3)  For X,Y in "radians" - units of one projection radius,
!      which in the case of a photograph is the focal length of
!      the camera - the following DISCO values apply:

!          Geometry          DISCO

!          astrograph         0.0
!          Schmidt           -0.3333
!          AAT PF doublet  +147.069
!          AAT PF triplet  +178.585
!          AAT f/8          +21.20
!          JKT f/8          +13.32

!  4)  The present routine is an approximate inverse to the
!      companion routine sla_PCD, obtained from two iterations
!      of Newton's method.  The mismatch between the sla_PCD and
!      sla_UNPCD routines is negligible for astrometric applications;
!      to reach 1 milliarcsec at the edge of the AAT triplet or
!      Schmidt field would require field diameters of 2.4 degrees
!      and 42 degrees respectively.

!  P.T.Wallace   Starlink   1 August 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: disco,x,y

DOUBLE PRECISION :: cr2,a,cr2a2,f



cr2=disco*(x*x+y*y)
a=(2D0*cr2+1D0)/(3D0*cr2+1D0)
cr2a2=cr2*a*a
f=(2D0*cr2a2*a+1D0)/(3D0*cr2a2+1D0)
x=x*f
y=y*f

END SUBROUTINE sla_unpcd
SUBROUTINE sla_v2tp (v, v0, xi, eta, j)
!+
!     - - - - -
!      V 2 T P
!     - - - - -

!  Given the direction cosines of a star and of the tangent point,
!  determine the star's tangent-plane coordinates.

!  (single precision)

!  Given:
!     V         r(3)    direction cosines of star
!     V0        r(3)    direction cosines of tangent point

!  Returned:
!     XI,ETA    r       tangent plane coordinates of star
!     J         i       status:   0 = OK
!                                 1 = error, star too far from axis
!                                 2 = error, antistar on tangent plane
!                                 3 = error, antistar too far from axis

!  Notes:

!  1  If vector V0 is not of unit length, or if vector V is of zero
!     length, the results will be wrong.

!  2  If V0 points at a pole, the returned XI,ETA will be based on the
!     arbitrary assumption that the RA of the tangent point is zero.

!  3  This routine is the Cartesian equivalent of the routine sla_S2TP.

!  P.T.Wallace   Starlink   27 November 1996

!  Copyright (C) 1996 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: v(3),v0(3),xi,eta
INTEGER :: j

REAL :: x,y,z,x0,y0,z0,r2,r,w,d

REAL :: tiny
PARAMETER (tiny=1E-6)


x=v(1)
y=v(2)
z=v(3)
x0=v0(1)
y0=v0(2)
z0=v0(3)
r2=x0*x0+y0*y0
r=SQRT(r2)
IF (r == 0.0) THEN
  r=1E-20
  x0=r
END IF
w=x*x0+y*y0
d=w+z*z0
IF (d > tiny) THEN
  j=0
ELSE IF (d >= 0.0) THEN
  j=1
  d=tiny
ELSE IF (d > -tiny) THEN
  j=2
  d=-tiny
ELSE
  j=3
END IF
d=d*r
xi=(y*x0-x*y0)/d
eta=(z*r2-z0*w)/d

END SUBROUTINE sla_v2tp
REAL FUNCTION sla_vdv (va, vb)
!+
!     - - - -
!      V D V
!     - - - -

!  Scalar product of two 3-vectors  (single precision)

!  Given:
!      VA      real(3)     first vector
!      VB      real(3)     second vector

!  The result is the scalar product VA.VB (single precision)

!  P.T.Wallace   Starlink   November 1984

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: va(3),vb(3)


sla_vdv=va(1)*vb(1)+va(2)*vb(2)+va(3)*vb(3)

END FUNCTION sla_vdv
SUBROUTINE sla_vn (v, uv, vm)
!+
!     - - -
!      V N
!     - - -

!  Normalizes a 3-vector also giving the modulus (single precision)

!  Given:
!     V       real(3)      vector

!  Returned:
!     UV      real(3)      unit vector in direction of V
!     VM      real         modulus of V

!  If the modulus of V is zero, UV is set to zero as well

!  P.T.Wallace   Starlink   23 November 1995

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: v(3),uv(3),vm

INTEGER :: i
REAL :: w1,w2


!  Modulus
w1=0.0
DO i=1,3
  w2=v(i)
  w1=w1+w2*w2
END DO
w1=SQRT(w1)
vm=w1

!  Normalize the vector
IF (w1 <= 0.0) w1=1.0
DO i=1,3
  uv(i)=v(i)/w1
END DO

END SUBROUTINE sla_vn
SUBROUTINE sla_vxv (va, vb, vc)
!+
!     - - - -
!      V X V
!     - - - -

!  Vector product of two 3-vectors (single precision)

!  Given:
!      VA      real(3)     first vector
!      VB      real(3)     second vector

!  Returned:
!      VC      real(3)     vector result

!  P.T.Wallace   Starlink   March 1986

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

REAL :: va(3),vb(3),vc(3)

REAL :: vw(3)
INTEGER :: i


!  Form the vector product VA cross VB
vw(1)=va(2)*vb(3)-va(3)*vb(2)
vw(2)=va(3)*vb(1)-va(1)*vb(3)
vw(3)=va(1)*vb(2)-va(2)*vb(1)

!  Return the result
DO i=1,3
  vc(i)=vw(i)
END DO

END SUBROUTINE sla_vxv
SUBROUTINE sla_xy2xy (x1,y1,coeffs,x2,y2)
!+
!     - - - - - -
!      X Y 2 X Y
!     - - - - - -

!  Transform one [X,Y] into another using a linear model of the type
!  produced by the sla_FITXY routine.

!  Given:
!     X1       d        x-coordinate
!     Y1       d        y-coordinate
!     COEFFS  d(6)      transformation coefficients (see note)

!  Returned:
!     X2       d        x-coordinate
!     Y2       d        y-coordinate

!  The model relates two sets of [X,Y] coordinates as follows.
!  Naming the elements of COEFFS:

!     COEFFS(1) = A
!     COEFFS(2) = B
!     COEFFS(3) = C
!     COEFFS(4) = D
!     COEFFS(5) = E
!     COEFFS(6) = F

!  the present routine performs the transformation:

!     X2 = A + B*X1 + C*Y1
!     Y2 = D + E*X1 + F*Y1

!  See also sla_FITXY, sla_PXY, sla_INVF, sla_DCMPF

!  P.T.Wallace   Starlink   5 December 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: x1,y1,coeffs(6),x2,y2


x2=coeffs(1)+coeffs(2)*x1+coeffs(3)*y1
y2=coeffs(4)+coeffs(5)*x1+coeffs(6)*y1

END SUBROUTINE sla_xy2xy
DOUBLE PRECISION FUNCTION sla_zd (ha, dec, phi)
!+
!     - - -
!      Z D
!     - - -

!  HA, Dec to Zenith Distance (double precision)

!  Given:
!     HA     d     Hour Angle in radians
!     DEC    d     declination in radians
!     PHI    d     observatory latitude in radians

!  The result is in the range 0 to pi.

!  Notes:

!  1)  The latitude must be geodetic.  In critical applications,
!      corrections for polar motion should be applied.

!  2)  In some applications it will be important to specify the
!      correct type of hour angle and declination in order to
!      produce the required type of zenith distance.  In particular,
!      it may be important to distinguish between the zenith distance
!      as affected by refraction, which would require the "observed"
!      HA,Dec, and the zenith distance in vacuo, which would require
!      the "topocentric" HA,Dec.  If the effects of diurnal aberration
!      can be neglected, the "apparent" HA,Dec may be used instead of
!      the topocentric HA,Dec.

!  3)  No range checking of arguments is done.

!  4)  In applications which involve many zenith distance calculations,
!      rather than calling the present routine it will be more efficient
!      to use inline code, having previously computed fixed terms such
!      as sine and cosine of latitude, and perhaps sine and cosine of
!      declination.

!  P.T.Wallace   Starlink   3 April 1994

!  Copyright (C) 1995 Rutherford Appleton Laboratory
!-

IMPLICIT NONE

DOUBLE PRECISION :: ha,dec,phi

DOUBLE PRECISION :: sh,ch,sd,cd,sp,cp,x,y,z


sh=SIN(ha)
ch=COS(ha)
sd=SIN(dec)
cd=COS(dec)
sp=SIN(phi)
cp=COS(phi)
x=ch*cd*sp-sd*cp
y=sh*cd
z=ch*cd*cp+sd*sp
sla_zd=ATAN2(SQRT(x*x+y*y),z)

END FUNCTION sla_zd
END MODULE sla
